#!/usr/bin/perl
#
# $Id: paraloop.pl 247 2014-02-11 08:19:15Z manu $
#

=head1 Copyright Notice

 Copyright I.N.R.A. - C.N.R.S. 

 emmanuel.courcelle@toulouse.inra.fr
 jerome.gouzy@toulouse.inra.fr

 This software is a computer program whose purpose is to help you run 
 your scripts on parallel computers.

 This software is governed by the CeCILL  license under French law and
 abiding by the rules of distribution of free software.  You can  use, 
 modify and/ or redistribute the software under the terms of the CeCILL
 license as circulated by CEA, CNRS and INRIA at the following URL
 http://www.cecill.info" 

 As a counterpart to the access to the source code and  rights to copy,
 modify and redistribute granted by the license, users are provided only
 with a limited warranty  and the software's author,  the holder of the
 economic rights,  and the successive licensors  have only  limited
 liability. 

 In this respect, the user's attention is drawn to the risks associated
 with loading,  using,  modifying and/or developing or reproducing the
 software by the user in light of its specific status of free software,
 that may mean  that it is complicated to manipulate,  and  that  also
 therefore means  that it is reserved for developers  and  experienced
 professionals having in-depth computer knowledge. Users are therefore
 encouraged to load and test the software's suitability as regards their
 requirements in conditions enabling the security of their systems and/or 
 data to be ensured and,  more generally, to use and operate it in the 
 same conditions as regards security. 

 The fact that you are presently reading this means that you have had
 knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

paraloop.pl - Run something in a PARAllel LOOP

=head1 SYNOPSIS

paraloop.pl [--switches | --parameters | --plugins]

   --switches    ==> to know everything about available switches
   --parameters  ==> to know everything about parameters
   --plugins     ==> to know the list of installed plugins
   --schedulers  ==> to know the list of installed schedulers

=head1 DESCRIPTION

Execute a loop in parallel on several processors or several nodes of a cluster.

=head1 SUBROUTINES

=cut

################################################################################
#
# THE GLOBAL BEGIN BLOCK
#
# We try to initialize @INC from some environment variables if they exist, of from
# the paraloop directory if they not.

# The best solution is to defined the environment variable $PARALOOP
#
################################################################################

BEGIN
{
    if (exists $ENV{'PARALOOP'})
    {
        unshift(@INC, $ENV{'PARALOOP'}.'/lib');
    }
    elsif (exists $ENV{'LIPMLIB'})
    {
        unshift(@INC, $ENV{'LIPMLIB'});
    }
    else
    {
        my ($dirprg, $nameprg) = $0 =~ /(.+)\/(.+)/;
        unshift(@INC, "$dirprg/lib", "$dirprg/../lib");
    }
}

################################################################################
#
# THE PACKAGE ParaloopTimer
#
# This tiny object is a timer. The timer is started when the object is allocated,
# and the elapsed time, system and user times are computed and written to the
# lock file when the object is destroyed.
#
# The object is allocated from inside MainLoop, so that only the MASTER Tasks
# are not counted, which should be correct as they only monitor their slaves.
#
################################################################################
#
# 14 oct 2013 - paraloop deleted from sourcesup, release 1.3.1 created
#

package ParaloopTimer;

use strict;
use warnings;
require "UNIVERSAL.pl";
use ParamParser;

sub _Init
{
    my $self    = shift;
    my $o_param = shift;
    $self->{'__o_param'} = $o_param;
    $self->{'__timer'}   = time()+0;
    $o_param->SetUnlessDefined('time_start',time());
    $o_param->SetUnlessDefined('time_elapsed',0);
    $o_param->SetUnlessDefined('time_user',0);
    $o_param->SetUnlessDefined('time_system',0);
    $o_param->SetUnlessDefined('time_children_user',0);
    $o_param->SetUnlessDefined('time_children_system',0);
}

sub UpdateElapsed
{
    my $self = shift;
    my $o_param = $self->{'__o_param'};
    my $new_timer = time();
    my $elapsed = $new_timer - $self->{'__timer'};
    $o_param->Set('time_elapsed',$o_param->Get('time_elapsed') + $elapsed);
    $self->{'__timer'} = $new_timer;
}
    
sub DESTROY
{
    my $self = shift;
    my $o_param = $self->{'__o_param'};
    my $elapsed = time() - $self->{'__timer'};
    my ($user,$system,$cuser,$csystem) = times();
    $o_param->Set('time_elapsed',$o_param->Get('time_elapsed') + $elapsed);
    $o_param->Set('time_user',$o_param->Get('time_user') + $user);
    $o_param->Set('time_system',$o_param->Get('time_system') + $system);
    $o_param->Set('time_children_user',$o_param->Get('time_children_user') + $cuser);
    $o_param->Set('time_children_system',$o_param->Get('time_children_system') + $csystem);
    $o_param->Dump($o_param->Get('lock_file'));
};

################################################################################
#
# THE PACKAGE MAIN
#
################################################################################

package main;

use strict;
use warnings;
no warnings qw(once);
use FileHandle;
use Getopt::Long;
use IO::File;
use Cwd;
use File::Basename;
use ParamParser;
use Logger;
use Runner;

#
# some const
#

use constant MAXFILESIZE       => 1000;
use constant NAMESPACE         => 'PARALOOP_';
use constant LOCK_DIR          => NAMESPACE . 'lock';
use constant ERROR_DIR         => NAMESPACE . 'error';
use constant LOG_LEVEL_DEFAULT => '01';
use constant WAIT_TIME         => 20;
#use constant WAIT_TIME         => 5;
use constant DEF_LOAD_BALANCING_THRESHOLD => 1;

# The "normal" behaviour is locking "lock files" before modifying them, thus enabling load_balancing mode
# But some file systems do not work when thrying to lock files, if you are using such systems put the constant to 1

use constant UNSET_LOCK_FILE => 0;

&ParamParser::SetDefaultBehaviour('exit_on_getopt_error');

#
# M A I N   P R O G R A M
#

my $VERSION = do {my @r = (q$Rev: 247 $ =~ /\d+/g); $r[0]};

#
# Was paraloop.pl ran by itself ? we test the existence of the environment variable subcount
# This variable is SET INTERNALLY BY THIS SCRIPT - DO NOT SET THE VARIABLE MANUALLY !!!
#

my $subcount_variable = NAMESPACE . 'subcount';
if (!exists $ENV{$subcount_variable})
{
    &MainCommandLineThenExit();    # Was called by user
}
else
{
    my $current_mode_variable = NAMESPACE . 'current_mode';
    if (!exists $ENV{$current_mode_variable} or $ENV{$current_mode_variable} eq 'SLAVE' or $ENV{$current_mode_variable} eq 'AUTONOMOUS')
    {
        # to debug env variables - if(1) to activate debug
        if (0)
        {
            foreach my $k (sort keys %ENV)
            {
                warn "ENV ".length($ENV{$k})." - $k - " . $ENV{$k}."\n";
            }
        }
        &MainSubmitThenExit();         # Was called by paraloop through the Submit() procedure.
    }
    else
    {
        &MainSubmitMasterThenExit();
    }
}

#
# Title: MainCommandLineThenExit
# Usage: &MainCommandLineThenExit()
# procedure: Called when paraloop is entered through a command line
#            The caller may be a human, or another instance of paraloop
# Args:      None
# Return:    NEVER RETURN
# Globals:   none
#

sub MainCommandLineThenExit
{
    my $o_param = New ParamParser();    # The object handling parameters
    $o_param->SetUsage(\&Usage);
    $o_param->SetBehaviour('exit_on_getopt_error');
    $o_param->SelectNameSpace(NAMESPACE);
    $o_param->Update(
                      'GETOPTLONG', 'I',              "resume:s",     "check:s",      "cfg=s",
                      "db=s",       "queue=s",        "local",        "output=s",     "mail=s",
                      "input=s",    "slice_offset=i", "slice_size=i", "slice_nb=i",   "restart:s",
                      "start=i",    "end=i",          "step=i",       "program=s",    "interrupt:s",
                      "kill:s",     "account=s",      "ncpus=i",      "monoprocessor","interleaved", "no_slicing",
                      "verbose",    "quiet",          "help",         "switches",     "parameters",
                      "cluster=s",  "plugins",        "schedulers",   "wait",         "plugin_debug", 
                      "show_parameters","clean",      "html",         "load_balancing_enabled"
                      );

    unless ($o_param->IsDefined('quiet') or $o_param->IsDefined('html'))
    {
        print "paraloop.pl - VERSION $VERSION - Using ParamParser v$ParamParser::VERSION, Logger v$Logger::VERSION\n\n";
    }
    
    my $plugin_debug = ($o_param->IsDefined('plugin_debug')) ? 1 : 0;
    &Switches($plugin_debug)   if ($o_param->IsDefined('switches'));
    &Parameters($plugin_debug) if ($o_param->IsDefined('parameters'));
    &Plugins($plugin_debug)    if ($o_param->IsDefined('plugins'));
    &Schedulers($plugin_debug) if ($o_param->IsDefined('schedulers'));

    my $verbose = ($o_param->IsDefined('verbose')) ? 1 : 0;
    my $quiet   = ($o_param->IsDefined('quiet'))   ? 1 : 0;
    my $monoprocessor = ($o_param->IsDefined('monoprocessor')) ? 1 : 0;
    if ($verbose == 1 and $quiet == 1)
    {
        die "ERROR - do NOT specify --verbose together with --quiet\n";
        #&Usage();
    }

    #
    # Better to manage later
    #
    
    $o_param->Set('local','0') unless ($o_param->IsDefined('local'));
    
    #
    # show_parameters specified: if resume, call the Resume function, else display the parameters and leave
    #
    
    if ($o_param->IsDefined('show_parameters') and !$o_param->IsDefined('resume'))
    {
        &ReadCfgFiles($o_param);
        print "\nSHOW PARAMETERS SWITCH\n";
        print "======================";
        $o_param->Print();
        print("\n");
        exit(1);
    };
    
    #
    # monoprocessor specified: start one job, with (start,end,step) specified or with (slice_offset,slice_size,slice_nb) specified
    #
    
    if ($monoprocessor==1)
    {
        MainCommandLineOneLoopThenExit($o_param);
    }
    
    #
    # resume specified: resume from an interrupted job, or from a monoprocessor interrupted job
    #
    
    elsif ($o_param->IsDefined('resume'))
    {
        MainCommandLineResumeThenExit($o_param);
    }

    #
    # restart specified: restart from an interrupted job, or from a monoprocessor interrupted job
    #
    
    elsif ($o_param->IsDefined('restart'))
    {
        MainCommandLineRestartThenExit($o_param);
    }

    #
    # interrupt/kill specified: interrupt/kill a job, or all the jobs
    #
    
    elsif ($o_param->IsDefined('interrupt') or $o_param->IsDefined('kill'))
    {
        MainCommandLineInterruptThenExit($o_param);
    }

    #
    # check specified: we just display some information about the current jobs
    #
    
    elsif ($o_param->IsDefined('check'))
    {
        MainCommandLineCheckThenExit($o_param);
    }
    
    #
    # The parameter --program should be specified, OR --cfg
    #
    
    elsif (!$o_param->IsDefined('program') and !$o_param->IsDefined('cfg'))
    {
        MainCommandLineWaitOrCleanThenExit($o_param);
    }

    #
    # ncpus mode: call paraloop from a high level loop (ncpus iterations)
    #
    
    else
    {
        MainCommandLineNcpusThenExit($o_param);
    }
}

#
# Title: ValidateParameters
# Usage: &CheckParam($o_param)
# function:  Called by MainCommandLineResumeThenExit, RestartThenExit
#            Validate the parameters: --restart or --resume, never both
#            The lock directory may be specified with any option: resume, restart,
#            wait, clean) but the values must be coherent                                                                  
# Args:      $o_param
# Return:    The lock directory or file
# Globals:   none
#

sub ValidateParameters
{
    my $o_param = shift;
    if ($o_param->IsDefined('resume'))
    {
        $o_param->Update('GETOPTLONG', 'I', "resume:s", "local", "cluster=s", "wait:s", "clean:s");
    }
    elsif ($o_param->IsDefined('restart'))
    {
        $o_param->Update('GETOPTLONG', 'I', "restart:s", "local", "cluster=s", "wait:s", "clean:s");
    }
    elsif ($o_param->IsDefined('interrupt'))
    {
        $o_param->Update('GETOPTLONG', 'I', "interrupt:s");
    }
    elsif ($o_param->IsDefined('kill'))
    {
        $o_param->Update('GETOPTLONG', 'I', "kill:s");
    }
    elsif ($o_param->IsDefined('check'))
    {
        $o_param->Update('GETOPTLONG', 'I', "check:s", "quiet", "html");
    }
    else
    {
        $o_param->Update('GETOPTLONG', 'I', "wait:s", "clean:s");
    };
    $o_param->SelectNameSpace(NAMESPACE);
    $o_param->Set('local',0) unless ($o_param->IsDefined('local'));

    #
    # Extract the value of the lock directory
    #
    
    my @prms = ($o_param->Get('resume'),
                $o_param->Get('restart'),
                $o_param->Get('check'),
                $o_param->Get('interrupt'),
                $o_param->Get('kill'),
                $o_param->Get('clean'));
    
    my $lock_dirfile='';
    for my $p (@prms)
    {
        if ($p ne '' and $lock_dirfile eq '')
        {
            $lock_dirfile = $p;
            next;
        };
        if ($p ne '' and $p ne $lock_dirfile)
        {
            die "ERROR - Incoherent values for lock directory: $p or $lock_dirfile ?";
        };
    };
    $lock_dirfile = &ArgToLockFileName($lock_dirfile);
    
    #
    # If --clean specified, force --wait
    #
    
    $o_param->Set('wait',1) if ($o_param->IsDefined('clean'));
    
    return $lock_dirfile;
}

        
#
# Title: MainCommandLineNcpusThenExit
# Usage: &MainCommandLineNcpusThenExit($o_param)
# procedure: Called by MainCommandLineThenExit when paraloop is entered
#            through a command line with the ncpus switch
# Args:      $o_param
# Return:    NEVER RETURN
# Globals:   none
#

# MODE = multi/init

sub MainCommandLineNcpusThenExit
{
    
    my $o_param = shift;
    #
    # Update $o_param ONLY from command line: Only some parameters allowed
    #

    $o_param->Init();               # Workaround a bug in ParamParser
    $o_param->SetUsage(\&Usage);    # kifkif
    $o_param->SelectNameSpace(NAMESPACE);
    $o_param->SetBehaviour('exit_on_getopt_error');
    $o_param->Update(
                      'GETOPTLONG', 'I',                "cfg=s",   "db=s",      "queue=s",                "local",
                      "output=s",   "input=s",          "start=i", "end=i",     "program=s",              "account=s",
                      "ncpus=i",    "interleaved",      "wait",    "clean",     "verbose",                "quiet",     
                      "plugin_debug","show_parameters", "mail=s",  "cluster=s", "load_balancing_enabled", "help"
                      );
    &ReadCfgFiles($o_param);
    $o_param->AssertDefined('program','input');
    if ($o_param->IsDefined('output'))
    {
        $o_param->Set('output_prefix',$o_param->Get('output'));
    }
    else
    {
        $o_param->Set('output_prefix','NULL');
    }
    $o_param->Set('local',0) unless ($o_param->IsDefined('local'));
    

    # Refuse to start if the parameter set is incoherent
    if ($o_param->IsDefined('load_balancing_enabled') and $o_param->Get('load_balancing_enabled') eq '1'
        and
        $o_param->IsDefined('interleaved') and $o_param->Get('interleaved') eq '1')
    {
        die "ERROR - The interleave switch and the load_balancing_enabled parameters are not compatible\n";
    };
    if ($o_param->IsDefined('load_balancing_enabled') and $o_param->Get('load_balancing_enabled') eq '1'
        and
        &UNSET_LOCK_FILE==1)
	{
		die "ERROR - Load balancing mode CANNOT BE USED in this installation - See the constant UNSET_LOCK_FILE";
	}

    # Some substitutions to make, and some parameters to set and print
    &MakeSubstitutions($o_param);
    &InitFileParametersNcpus($o_param);

    # Verify that end higher than start
    if ($o_param->IsDefined('start') and $o_param->IsDefined('end'))
    {
        if ($o_param->Get('start') >= $o_param->Get('end'))
        {
            die "ERROR - If specified, start should be lower than stop\n";
        }
    }
    
    # Init some local variables from those parameters
    my $verbose = ($o_param->IsDefined('verbose')) ? 1 : 0;
    my $quiet   = ($o_param->IsDefined('quiet'))   ? 1 : 0;

    # Create a log object, to log to $output.log (WITHOUT the numeric fields)
    
    my $output_prefix = $o_param->Get('output_prefix');
    my $log_file = $o_param->Get('log_file');
    my $log_level= $o_param->Get('log_level');

    my $o_log = Logger->New(">>$log_file", $log_level);

    # Init the $scheduler and the $ncpus variable:
    #      If the ncpus switch was specified, we take its value.
    #      Else, we look for the parameter PARALOOP_<Scheduler>_ncpus, where <Scheduler> is the name of the chosen scheduler.
    #      Else, we die with an error message (NO default value !)
    #

    my $ncpus = '';
    my $scheduler='';
    if ($o_param->Get('local')!=0)
    {
        $scheduler = 'System';
        $o_log->Trace("Scheduler $scheduler selected (--local)\n",".","1");
    }
    elsif ($o_param->IsDefined('Scheduler')) 
    {
        $scheduler = $o_param->Get('Scheduler');
        $o_log->Trace("Scheduler $scheduler selected\n",".","1");
    }
    else
    {
        $o_log->Trace("ERROR - The Scheduler parameter should be defined\n",".","1");
        die "\nERROR - The Scheduler parameter should be defined\n";
    };
    
    if ($o_param->IsDefined('ncpus'))
    {
        $ncpus = $o_param->Get('ncpus');
    }
    elsif ($o_param->Get('local')!=0 and $o_param->IsDefined('local_ncpus'))
    {
        $ncpus = $o_param->Get('local_ncpus');
        
    }
    else
    {
        my $def_ncpus_param = $scheduler . '_ncpus';
        if ($o_param->IsDefined($def_ncpus_param))
        {
            $ncpus = $o_param->Get($def_ncpus_param);
        }
        else
        {
            my $msg = "ERROR - The ncpus switch";
            $msg   .= " OR the local_ncpus parameter" if ($o_param->Get('local')!=0);
            $msg   .= " OR the $def_ncpus_param parameter should be defined";
            $o_log->Trace("msg\n",".","1");
            die "\n$msg";
        }
    }
    $o_log->Trace("ncpus = $ncpus\n",".","1");
   
    # The plugin is created to check everything is OK, to print a welcome message and to know the number of records
    # The scheduler is created ONLY to display a message

    my $plugin = &CreatePlugin($o_param,$o_log);

    if ($quiet==0)
    {
        my $o_sched = &CreateScheduler($o_param);

        print "Found " . $plugin->WhatParaloopPlugin() . "\n";
        print "Found " . $o_sched->WhatParaloopScheduler() . " \n";

        print "LOCK   directory = ".$o_param->Get('lock_directory')."\n";
        print "ERROR  directory = ".$o_param->Get('error_directory')."\n";
        
        my $output_prefix = $o_param->Get('output_prefix');
        if ($output_prefix eq 'NULL')
        {
            print "WARNING          - NO OUTPUT FILE or LOG FILE WILL BE GENERATED - SEE THE --output SWITCH\n";
        }
        else
        {
            print "OUTPUT prefix    = ".$o_param->Get('output_prefix')."\n";
        }
    }


    # Delete the wait switch for the children
    my $wait = ($o_param->IsDefined('wait')) ? 1 : 0;
    $o_param->Delete('wait');

    # Delete the clean switch for the children
    my $clean = ($o_param->IsDefined('clean')) ? 1 : 0;
    $o_param->Delete('clean');

    # Please wait before cleaning...
    $wait=1 if ($clean != 0);
    
    # Set the monoprocessor switch for the children
    $o_param->Set('monoprocessor',1);
    
    # Compute start,end values (may be nothing specified here)
    my ($start, $end);
    if ($o_param->IsDefined('start'))
    {
        $start = $o_param->Get('start');
    }
    else
    {
        $start = 0;
    }
    print "start            = $start\n" unless ($quiet == 1);

    if ($o_param->IsDefined('end'))
    {
        $end = $o_param->Get('end');
    }
    else
    {
        print "Counting the number of records in the input files\n" unless ($quiet == 1);
        my $nb_seq = $plugin->TellLength();
        $end = $nb_seq - 1;
        print "Found $nb_seq records.\n" if ($verbose == 1);
        $o_log->Trace("Found $nb_seq records.\n",".","1");
    }
    print "end              = $end\n\n" unless ($quiet == 1);
    
    if ($end-$start+1 < $ncpus) {		# Could happen, specially with the Shell plugin
        my $old_ncpus = $ncpus; 
        $ncpus = $end-$start + 1;
        print "WARNING - ncpus was changed from $old_ncpus to $ncpus\n";
        $o_log->Trace("WARNING - ncpus was changed from $old_ncpus to $ncpus.\n",".","1");
    };
    
    if ($o_param->IsDefined('plugin_debug'))
    {
        &PrintVar($o_param);
        print "Now leaving - Please resume without the plugin_debug switch\n";
        exit(0);
    }
    
    #
    #  The mode: default mode is AUTONOMOUS
    #    
    if (! $o_param->IsDefined('mode'))
    {
        $o_param->Set('mode','AUTONOMOUS');
    }
    
    #
    #  MASTER/SLAVE mode:
    #    -select current_mode to MASTER
    #    -compute slave_ncpus and master_ncpus
    #    -call CpuLoop, passing $master_ncpus
    #    
    if (($o_param->Get('mode') ne 'MASTER/SLAVE') and ($o_param->Get('mode') ne 'AUTONOMOUS'))
    {
        warn '#'.$o_param->Get('mode').'#';
        die "ERROR - MASTER/SLAVE or AUTONOMOUS are the ONLY authorized values for mode\n";
    }
    elsif ($o_param->Get('mode') eq 'MASTER/SLAVE')
    {
        $o_param->Set('current_mode','MASTER');
        my ($slave_ncpus,$master_ncpus);
        if (($o_param->IsDefined('slave_ncpus') and $o_param->IsDefined('master_ncpus')) or (!$o_param->IsDefined('slave_ncpus') and !$o_param->IsDefined('master_ncpus')))
        {
            die "ERROR - MASTER/SLAVE mode: 1 and ONLY 1 of slave/ncpus, master_ncpus parameters must be specified\n";
        }
        if ($o_param->IsDefined('slave_ncpus'))
        {
            $slave_ncpus = $o_param->Get('slave_ncpus');
            die "ERROR - If defined, slave_ncpus should be > 0\n" if ($slave_ncpus<=0);
            if ($ncpus % $slave_ncpus == 0)
            {
                $master_ncpus = $ncpus / $slave_ncpus;
            }
            else
            {
                $master_ncpus = int($ncpus / $slave_ncpus) + 1;
            }
            $o_param->Set('master_ncpus',$master_ncpus);
        }
        else
        {
            $master_ncpus = $o_param->Get('master_ncpus');
            die "ERROR - If defined, master_ncpus should be > 0\n" if ($master_ncpus<=0);
            if ($ncpus % $master_ncpus == 0)
            {
                $slave_ncpus = $ncpus / $master_ncpus;
            }
            else
            {
                $slave_ncpus = int($ncpus / $master_ncpus) + 1;
            }
            $o_param->Set('slave_ncpus',$slave_ncpus);
        }
        
        # status = STARTING
        $o_param->Set('status','STARTING');
        
        # Enter the CpuLoop         
        &CpuLoop($start,$end,$master_ncpus,$o_param,$o_log);
    }
    
    #
    #  AUTONOMOUS mode:
    #     -call CpuLoop, passing $ncpus
    #    
    elsif ($o_param->Get('mode') eq 'AUTONOMOUS')
    {
        # status = STARTING
        $o_param->Set('status','STARTING');
        
        # Enter the CpuLoop         
        &CpuLoop($start,$end,$ncpus,$o_param,$o_log);
    }
    
    #
    #  Only 2 modes supported
    #
    else
    {
        warn '#'.$o_param->Get('mode').'#';
        die "ERROR - MASTER/SLAVE or AUTONOMOUS are the ONLY authorized values for mode\n";
    } 
    
    my $lock_directory = $o_param->Get('lock_directory');
    if ( &UNSET_LOCK_FILE==0 )
    {
	    &ParamParser::SetDefaultBehaviour('lock_file');     # From now on, we lock param files before reading them
	}
	else
	{
		&ParamParser::UnSetDefaultBehaviour('lock_file');
	}
    &WaitParaloop($lock_directory)  if ($wait == 1);
    &CleanParaloop($lock_directory) if ($clean == 1);

    exit 0;

}

#
# Title: MainCommandLineResumeThenExit
# Usage: &MainCommandLineResumeThenExit($o_param)
# procedure: Called by MainCommandLineThenExit when paraloop is entered
#            through a command line with the resume switch
# Args:      $o_param
# Return:    NEVER RETURN
# Globals:   None
#

sub MainCommandLineResumeThenExit
{
    my $o_param = shift;
    my $resume_file = &ValidateParameters($o_param);
    #$o_param->Update('GETOPTLONG', 'I', "resume:s", "local", "cluster=s", "wait", "clean"); 
    #$o_param->SelectNameSpace(NAMESPACE);
    my $wait = ($o_param->IsDefined('wait')) ? 1 : 0;
    my $clean = ($o_param->IsDefined('clean')) ? 1 : 0;
    #$o_param->Set('local',0) unless ($o_param->IsDefined('local'));
    $o_param->Delete('wait');
    $o_param->Delete('clean');
    $o_param->Delete('resume');
    
    if (!-e $resume_file)
    {
        die "ERROR - The resume file $resume_file should exist";
    }

    #
    # Directory specified: look for every lock file here, and call paraloop with the correct switch
    #
    if ( &UNSET_LOCK_FILE==0 )
    {
	    &ParamParser::SetDefaultBehaviour('lock_file');     # From now on, we lock param files before reading them
	}
	else
	{
		&ParamParser::UnSetDefaultBehaviour('lock_file');
	}

    if (-d $resume_file)
    {
        MainCommandLineResumeIsDir($o_param,$resume_file);
    }
    else
    {
        MainCommandLineResumeIsFile($o_param,$resume_file);
    }

    &WaitParaloop($resume_file)  if ($wait == 1);
    &CleanParaloop($resume_file) if ($clean == 1);
    exit (0);
}

#
# Title: MainCommandLineResumeIsDir
# Usage: &MainCommandLineResumeIsDir($o_param,$resume_dir)
# procedure: Called by MainCommandLineResumeThenExit when paraloop is entered
#            through a command line with the resume switch AND the argument of the
#            switch is a directory
# Args:      $o_param    = The parameters
#            $resume_dir = The lock directory
# Globals:   None
#

# MODE = multi/resume

sub MainCommandLineResumeIsDir
{
    my $o_param = shift;
    my $resume_dir = shift;
    
    my @resume_files = glob("$resume_dir/lock.*.lck");
    foreach my $f (@resume_files)
    {
        my $o_param_resume = New ParamParser('PARAMPARSER',$o_param);
        $o_param_resume->Update($f,'A');
        
        $o_param_resume->SelectNameSpace(NAMESPACE); 
        my $plp_id = $o_param_resume->Get('plp_id');
        next if ($o_param_resume->Get('status') eq 'TERMINATED');
        
        # We do not resume the slaves: they will be resumed by their masters
        next if ($o_param_resume->IsDefined('current_mode') and $o_param_resume->Get('current_mode') eq 'SLAVE');
        
        # We do not resume the terminated jobs nor the MASTER MONITORING jobs, nor RUNNING jobs (even hanged)
        # You can resume them with specifying the plp_id    
        my $status = $o_param_resume->Get('status');
        if ($status eq 'MONITORING')
        {
            print "Paraloop job $plp_id not resumed ($status) - Please try specifying --resume $plp_id\n";
            next;
        };

        # OK - Let's try to resume the job
        unlink($o_param_resume->Get('stop_all'), $o_param_resume->Get('stop_this'));    # Remove stop files, if any
        
        $o_param_resume->Delete('subcount');    # Delete subcount, because we are resumed by the user

        my $log_file  = $o_param_resume->Get('log_file');
        my $log_level = $o_param_resume->IsDefined("log_level") ? $o_param_resume->Get("log_level") : LOG_LEVEL_DEFAULT;
        my $o_log_resume = Logger->New(">>$log_file", $log_level);

        my $o_sched_resume = &CreateScheduler($o_param_resume,$o_log_resume);
        
        # the status was running: is it true ? If it's true, we don't resume it
        my $job_id;
        if ($o_param_resume->IsDefined('job_id'))
        {
            $job_id = $o_param_resume->Get('job_id');
            if ($o_sched_resume->IsLaunched($job_id)!=0)
            {
                print "Paraloop job $plp_id not resumed: it is already running\n";
                next;
            };
        }
        
        $o_log_resume->Trace("RESUMING PARALOOP FROM THE DIRECTORY $resume_dir\n", '*', '1');
        $o_log_resume->Trace("PARAMETERS WERE SET FROM THE FILE $f\n",    '.', '2');

        $o_param_resume->Set('status', 'RESUMING');

        if ($o_param->IsDefined('show_parameters'))
        {
            print "\nSHOW PARAMETERS SWITCH\n";
            print "======================";
            $o_param_resume->Print();
            print("\n");
        }
        else
        {
            $job_id = $o_sched_resume->Submit();
            if ($job_id eq '0')
            {
                print "Paraloop job $plp_id could not be resumed (the nodes are too loaded) - please try again later\n";
            }
            else
            {
                print "Paraloop job $plp_id     resumed - new job ID = $job_id\n";
            }
        }
    }
}

#
# Title: MainCommandLineResumeIsFile
# Usage: &MainCommandLineResumeIsFile($o_param,$resume_file)
# procedure: Called by MainCommandLineResumeThenExit when paraloop is entered
#            through a command line with the resume switch AND the argument of the
#            switch is a lock file
# Args:      $o_param = The parameters object      
#$resume_file = The lock file to resume from
# Return:    NEVER RETURN
# Globals:   None
#

# MODE = mono/resume

sub MainCommandLineResumeIsFile
{
    my $o_param = shift;
    my $resume_file = shift;

    #
    # File specified: resume paraloop reading the parameters from this file
    #

    $o_param->SelectNameSpace();    # Cancel the namespace before reading the files
    $o_param->Update($resume_file, 'A');    # Read the parameters from the file that was saved then
    $o_param->SelectNameSpace(NAMESPACE);    # The conf files are NOT used
    exit(0) if ($o_param->Get('status') eq 'TERMINATED'); #Nothing to resume
    my $plp_id = $o_param->Get('plp_id');

    my $lock_file = $o_param->Get('lock_file');
    if (!-e $lock_file)
    {
        die "ERROR - The lock file $lock_file should exist.";
    }

    # We do not resume the slaves: they should be resumed by their masters
    if ($o_param->IsDefined('current_mode') and $o_param->Get('current_mode') eq 'SLAVE')
    {
        die "ERROR - You cannot resume a slave alone. Please resume the master";
    }

    # We resume the MASTER MONITORING jobs, RUNNING jobs
    # But with a warning
    my $status = $o_param->Get('status');
    if ($status eq 'MONITORING' or $status eq 'RUNNING')
    {
        print "Paraloop job $plp_id has a status $status - I'll try to resume it anyway.\n";
    };
    
    unlink($o_param->Get('stop_all'), $o_param->Get('stop_this'));    # Remove stop files, if any
    
    my $job_id;
    $o_param->Delete('subcount');    # Delete subcount, because we are resumed by the user

    my $log_file  = $o_param->Get('log_file');
    my $log_level = $o_param->Get("log_level");
    my $o_log = Logger->New(">>$log_file", $log_level);

    my $sched = &CreateScheduler($o_param,$o_log);
    if ($o_param->IsDefined('job_id'))
    {
        $job_id = $o_param->Get('job_id');
        if ($sched->IsLaunched($job_id) != 0)
        {
            die "ERROR - The job $job_id (see $lock_file) is already launched";
        }
    };
    
    $o_log->Trace("RESUMING PARALOOP FROM THE FILE $resume_file\n", '*', '1');
    $o_log->Trace("PARAMETERS WERE SET FROM THE FILE $lock_file\n",    '.', '2');

    $o_param->Set('status', 'RESUMING');

    if ($o_param->IsDefined('show_parameters'))
    {
        print "\nSHOW PARAMETERS SWITCH\n";
        print "======================";
        $o_param->Print();
        print("\n");
    }
    else
    {
        $job_id = $sched->Submit();
        if ($job_id eq '0')
        {
            print "Paraloop job $plp_id could not be resumed (the nodes are too loaded) - please try again later\n";
        }
        else
        {
            print "Paraloop job $plp_id     resumed - new job ID = $job_id\n";
        }
    };
}

#
# Title: MainCommandLineRestartThenExit
# Usage: &MainCommandLineRestartThenExit($o_param)
# procedure: Called by MainCommandLineThenExit when paraloop is entered
#            through a command line with the restart switch
# Args:      $o_param
# Return:    NEVER RETURN
# Globals:   None
#

sub MainCommandLineRestartThenExit
{
    my $o_param = shift;
    my $restart_file = &ValidateParameters($o_param);
    #$o_param->Update('GETOPTLONG', 'I', "restart:s", "local", "cluster=s", "wait", "clean"); 
    #$o_param->SelectNameSpace(NAMESPACE);
    my $wait = ($o_param->IsDefined('wait')) ? 1 : 0;
    my $clean = ($o_param->IsDefined('clean')) ? 1 : 0;
    #my $restart_file = $o_param->Get('restart');
    #$o_param->Set('local',0) unless ($o_param->IsDefined('local'));
    $o_param->Delete('wait');
    $o_param->Delete('clean');
    $o_param->Delete('restart');
    
    if (!-e $restart_file)
    {
        die "ERROR - The restart file $restart_file should exist";
    }

    #
    # Directory specified: look for every lock file here, and call paraloop with the correct switch
    #
    if ( &UNSET_LOCK_FILE==0 )
    {
	    &ParamParser::SetDefaultBehaviour('lock_file');     # From now on, we lock param files before reading them
	}
	else
	{
		&ParamParser::UnSetDefaultBehaviour('lock_file');
	}
    if (-d $restart_file)
    {
        MainCommandLineRestartIsDir($o_param,$restart_file);
    }
    else
    {
        MainCommandLineRestartIsFile($o_param,$restart_file);
    }

    &WaitParaloop($restart_file)  if ($wait == 1);
    &CleanParaloop($restart_file) if ($clean == 1);
    exit (0);
}

#
# Title: MainCommandLineRestartIsDir
# Usage: &MainCommandLineRestartIsDir($o_param,$restart_dir)
# procedure: Called by MainCommandLineRestartThenExit when paraloop is entered
#            through a command line with the restart switch AND the argument of the
#            switch is a directory
# Args:      $o_param    = The parameters
#            $restart_dir = The lock directory
# Return:    NEVER RETURN
# Globals:   None
#

# MODE = multi/restart

sub MainCommandLineRestartIsDir
{
    my $o_param = shift;
    my $restart_dir = shift;

    my @restart_files = glob("$restart_dir/lock.*.lck");
    foreach my $f (@restart_files)
    {
        my $o_param_restart = New ParamParser('PARAMPARSER',$o_param);
        $o_param_restart->Update($f,'A');
        $o_param_restart->SelectNameSpace(NAMESPACE); 
        my $plp_id = $o_param_restart->Get('plp_id');

        # We do not restart the slaves: they will be restarted by their masters
        next if ($o_param_restart->IsDefined('current_mode') and $o_param_restart->Get('current_mode') eq 'SLAVE');
        
        # We do not restart the terminated jobs nor the MASTER MONITORING jobs, nor RUNNING jobs (even hanged)
        # You can restart them with specifying the plp_id    
        my $status = $o_param_restart->Get('status');
        if ($status eq 'TERMINATED' or $status eq 'MONITORING' or $status eq 'RUNNING')
        {
            print "Paraloop job $plp_id not restarted ($status) - Please try specifying --restart $plp_id\n";
            next;
        };

        # We do not restart a job if it has done old segments
        if ($o_param_restart->IsDefined('done_segments'))
        { 
            print "Paraloop job $plp_id not restarted because it processed some segments\n";
            next;
        };
        
        # OK - Let's try to restart the job

        unlink($o_param_restart->Get('stop_all'), $o_param_restart->Get('stop_this'));    # Remove stop files, if any
        
        $o_param_restart->Delete('subcount');    # Delete subcount, because we are restarted by the user

        #
        # We increment the log file and the output file, AND rename the current log/output files to XXX.bkp
        # It is better than deleting the file, but it will not be seen by the cleaning procedures
        #
        
        my $log_file  = $o_param_restart->Get('log_file');
        link($log_file,$log_file.'.bkp');
        unlink $log_file;
        $log_file     = &IncrementName($log_file);
        $o_param_restart->Set('log_file',$log_file);
        
        my $output = $o_param_restart->Get('output');
        link($output, $output.'.bkp');
        unlink($output);
        $output    = &IncrementName($output);
        $o_param_restart->Set('output',$output);
        
        my $log_level = $o_param_restart->IsDefined("log_level") ? $o_param_restart->Get("log_level") : LOG_LEVEL_DEFAULT;
        my $o_log_restart = Logger->New(">>$log_file", $log_level);

        my $o_sched_restart = &CreateScheduler($o_param_restart,$o_log_restart);
        my $job_id;
        if($o_param_restart->IsDefined('job_id'))
        {
            $job_id = $o_param_restart->Get('job_id');
            next if ($o_sched_restart->IsLaunched($job_id)!=0);     # Already started
        };
        
        #
        # RESTART means that we set current to start
        #
        
        my $start = $o_param_restart->Get('start');
        $o_param_restart->Set('current',$start);

        $o_log_restart->Trace("RESTARTING PARALOOP FROM THE DIRECTORY $restart_dir\n", '*', '1');
        $o_log_restart->Trace("PARAMETERS WERE SET FROM THE FILE $f\n",    '.', '2');
        $o_log_restart->Trace("RESETTING current TO $start\n",'.','2'); 

        $o_param_restart->Set('status', 'RESTARTING');

        if ($o_param->IsDefined('show_parameters'))
        {
            print "\nSHOW PARAMETERS SWITCH\n";
            print "======================";
            $o_param_restart->Print();
            print("\n");
        }
        else
        {
            $job_id = $o_sched_restart->Submit();
            if ($job_id eq '0')
            {
                print "Paraloop job $plp_id could not be restarted (the nodes are too loaded) - please try again later\n";
            }
            else
            {
                print "Paraloop job $plp_id     restarted - new job ID = $job_id\n";
            }
        }
    }
}

#
# Title: MainCommandLineRestartIsFile
# Usage: &MainCommandLineRestartIsFile($o_param,$restart_file)
# procedure: Called by MainCommandLineRestartThenExit when paraloop is entered
#            through a command line with the restart switch AND the argument of the
#            switch is a lock file
# Args:      $o_param = The parameters object      
# $restart_file = The lock file to restart from
# Return:    NEVER RETURN
# Globals:   None
#

# MODE = mono/restart

sub MainCommandLineRestartIsFile
{
    my $o_param = shift;
    my $restart_file = shift;

    #
    # File specified: restart paraloop reading the parameters from this file
    #

    $o_param->SelectNameSpace();    # Cancel the namespace before reading the files
    $o_param->Update($restart_file, 'O');    # Read the parameters from the file that was saved then
    $o_param->SelectNameSpace(NAMESPACE);    # The conf files are NOT used
    my $plp_id = $o_param->Get('plp_id');
    my $lock_file = $o_param->Get('lock_file');
    if (!-e $lock_file)
    {
        die "ERROR - The lock file $lock_file should exist.";
    }
    
    # We do not restart the slaves: they should be restarted by their masters
    if ($o_param->IsDefined('current_mode') and $o_param->Get('current_mode') eq 'SLAVE')
    {
        die "ERROR - You cannot restart a slave alone. Please restart the master";
    }

    # We do not restart a job if it has done old segments
    if ($o_param->IsDefined('done_segments'))
    { 
        die "Paraloop job $plp_id not restarted because segments were processed by this job\n";
    };
    
    # We restart the terminated jobs, MASTER MONITORING jobs, RUNNING jobs
    # But with a warning
    my $status = $o_param->Get('status');
    if ($status eq 'TERMINATED' or $status eq 'MONITORING' or $status eq 'RUNNING')
    {
        print "Paraloop job $plp_id has a status $status - I'll try to restart it anyway.\n";
    };
    
    # Create a new Logger object, then a new Scheduler object
    
    $o_param->Delete('subcount');    # Delete subcount, because we are restartd by the user
    my $log_file  = $o_param->Get('log_file');
    my $log_level = $o_param->Get("log_level");
    my $o_log     = Logger->New(">>$log_file", $log_level,'NOCOMPRESSION NEWQUIET');
    my $o_sched = &CreateScheduler($o_param,$o_log);
    my $job_id;
    if ($o_param->IsDefined('job_id'))
    {
        $job_id = $o_param->Get('job_id');
        if ($o_sched->IsLaunched($job_id) != 0)
        {
            die "ERROR - The job $job_id (see $lock_file) is already launched";
        }
    };

    # OK - The job is not running, we'll restart it
    
    unlink($o_param->Get('stop_all'), $o_param->Get('stop_this'));    # Remove stop files, if any
    
    #
    # We increment the log file and the output file, AND rename the current log/output files to XXX.bkp
    # It is better than deleting the file, but it will not be seen by the cleaning procedures
    #
    
    my $new_log_file     = &IncrementName($log_file);
    $o_log->SetLogFile($new_log_file);
    link($log_file,$log_file.'.bkp');
    unlink $log_file;
    $o_param->Set('log_file',$new_log_file);
    
    my $output = $o_param->Get('output');
    link($output, $output.'.bkp');
    unlink($output);
    $output    = &IncrementName($output);
    $o_param->Set('output',$output);
    
    #
    # RESTART means that we set current to start
    #
    
    my $start = $o_param->Get('start');
    $o_param->Set('current',$start);

    $o_log->Trace("RESTARTING PARALOOP FROM THE FILE $restart_file\n", '*', '1');
    $o_log->Trace("PARAMETERS WERE SET FROM THE FILE $lock_file\n",    '.', '2');
    $o_log->Trace("RESETTING current TO $start\n",'.','2'); 

    $o_param->Set('status', 'RESTARTING');

    if ($o_param->IsDefined('show_parameters'))
    {
        print "\nSHOW PARAMETERS SWITCH\n";
        print "======================";
        $o_param->Print();
        print("\n");
    }
    else
    {
        $job_id = $o_sched->Submit();
        if ($job_id eq '0')
        {
            print "Paraloop job $plp_id could not be restarted (the nodes are too loaded) - please try again later\n";
        }
        else
        {
            print "Paraloop job $plp_id     restarted - new job ID = $job_id\n";
        }
    };
}

#
# Title: MainCommandLineInterruptThenExit
# Usage: &MainCommandLineInterruptThenExit($o_param)
# procedure: Called by MainCommandLineThenExit when paraloop is entered
#            through a command line with the interrupt switch
# Args:      $o_param
# Return:    NEVER RETURN
# Globals:   None
#

sub MainCommandLineInterruptThenExit
{
    my $o_param = shift;
    my $interrupt_file = &ValidateParameters($o_param);
    my $kill_it = $o_param->IsDefined('kill');
    
    if (!-e $interrupt_file)
    {
        die "ERROR - The interrupt file $interrupt_file should exist";
    }

    if ( &UNSET_LOCK_FILE==0 )
    {
	    &ParamParser::SetDefaultBehaviour('lock_file');     # From now on, we lock param files before reading them
	}
	else
	{
		&ParamParser::UnSetDefaultBehaviour('lock_file');
	}
    if (-d $interrupt_file)
    {
        #MainCommandLineInterruptIsDirThenExit($o_param,$interrupt_file,$kill_it);
        MainCommandLineInterruptIsDirThenExit($interrupt_file,$kill_it);
   }
    else
    {
        #MainCommandLineInterruptIsFileThenExit($o_param,$interrupt_file,$kill_it);
        MainCommandLineInterruptIsFileThenExit($interrupt_file,$kill_it);
   }
}

#
# Title: MainCommandLineInterruptIsDirThenExit
# Usage: &MainCommandLineInterruptIsDirThenExit($o_param,$interrupt_dir)
# procedure: Called by MainCommandLineInterruptThenExit when paraloop is entered
#            through a command line with the interrupt switch AND the argument of the
#            switch is a directory
# Args:      $interrupt_dir = The lock directory
#            $kill_it (0/1, default=0)= If 1, kill the job
# Return:    NEVER RETURN
# Globals:   None
#

sub MainCommandLineInterruptIsDirThenExit
{
    my ($interrupt_dir,$kill_it) = (@_);
    $kill_it = 0 unless (defined $kill_it);

#    my $o_param = shift;
#    my $interrupt_dir = shift;

    foreach my $f (glob("$interrupt_dir/lock.*.lck"))
    {    
        # For each lock file call Interruptfile
        next if (-d $f);    # Should not happen !!!
        &InterruptFile($f,$kill_it);
    }
    exit 0;
}

#
# Title: MainCommandLineInterruptIsFileThenExit
# Usage: &MainCommandLineInterruptIsFileThenExit($o_param,$interrupt_file)
# procedure: Called by MainCommandLineInterruptThenExit when paraloop is entered
#            through a command line with the interrupt switch AND the argument of the
#            switch is a file
# Args:      $interrupt_file = The lock file
#            $kill_it (0/1, default=0)= If 1, kill the job
# Return:    NEVER RETURN
# Globals:   None
#

sub MainCommandLineInterruptIsFileThenExit
{
    my ($interrupt_file,$kill_it) = (@_);
    $kill_it = 0 unless (defined $kill_it);

#    my $o_param = shift;
#    my $interrupt_file = shift;
    
    &InterruptFile($interrupt_file,$kill_it);
    exit 0;
}

#
# Title: InterruptFile
# Usage: &InterruptFile($Interrupt_file)
# procedure:  Interrupt a job
# Args:      $interrupt_file          = The lock file
#            $kill_it (0/1, default=0)= If 1, kill the job
# Return:    Nothing    
# Globals:   None
#

sub InterruptFile
{
    my ($interrupt_file,$kill_it) = (@_);
    $kill_it = 0 unless (defined $kill_it);
    
    my $o_param = New ParamParser($interrupt_file);
    $o_param->SetBehaviour("ignore_space");
    $o_param->SelectNameSpace(NAMESPACE);    # The conf files are NOT used
    
    my $job_id = $o_param->Get('job_id');
   
    my $status = $o_param->Get('status');
    if ($status eq 'TERMINATED' or $status eq 'INTERRUPTED') { # Nothing to interrupt
        print "Job $job_id ($interrupt_file) is already stopped\n";
        return;
    };

    my $stop_file = $o_param->Get('stop_this');

    # Create a "stop" file, when the main loop will see this file it will stop
    open(STOP,">$stop_file") or die "ERROR - Cannot create $stop_file (lock file = $interrupt_file)";
    close(STOP);

    # If we are in status RUNNING, there is nothing to do, just to wait
    # If $kill_it was specified, it's time to kill the job 
    if ($kill_it!=0 or ($status ne 'RUNNING' and $status ne 'MONITORING'))
    {
        my $log_file  = $o_param->Get('log_file');
        my $log_level = $o_param->Get("log_level");
        my $o_log = Logger->New(">>$log_file", $log_level);
        
        my $o_sched = &CreateScheduler($o_param,$o_log);
        
        my $state = $o_sched->IsLaunched($job_id);

        # The job is still running: just kill it !
        if ($state != 0)
        {
            $o_sched->KillJob($job_id);
            sleep 1;
            $o_log->Trace("TRYING TO KILL JOB $job_id\n",'*','1');
            $o_param->Set('status','KILLED');
            &DumpToFile($o_param);
            print ("trying to kill $job_id\n");
        }
        
        # the job is already killed: nothing to do, just to print a warning
        else
        {
            $o_log->Trace("THE JOB $job_id IS ALREADY KILLED\n",'*','1');
            print ("The job $job_id is already killed\n");
        }
    }
}

#
# Title: MainCommandLineCheckThenExit
# Usage: &MainCommandLineCheckThenExit($o_param)
# procedure: Called by MainCommandLineThenExit when paraloop is entered
#            through a command line with the check switch
# Args:      $o_param
# Return:    NEVER RETURN
# Globals:   None
#

sub MainCommandLineCheckThenExit
{
    my $o_param = shift;
    my $html = $o_param->IsDefined('html') ? 1 : 0;
    
    my $check_file = &ValidateParameters($o_param);
    my $quiet = $o_param->Get('quiet');

    if (!-e $check_file)
    {
        die "ERROR - The file/dir $check_file should exist";
    }

    #
    # Print header
    #
    my $fmt='';
    
    if ($html)
    {
        #print '<html><head></head><body>';
        $fmt='<table class="plp"><tr><th>%8s</th><th>%10s</th><th>%10s</th><th>%10s</th><th>%10s</th><th>%10s</th><th>%5s</th><th>%22s</th><th>%12s</th><th>%9s</th><th>%9s</th><th>%9s</th></tr>';
    }
    else
    {
        $fmt="%8s %10s %10s %10s %10s %10s %5s %22s %12s  %9s %9s %9s\n";
    };
    printf($fmt,"Id","start","end","step","size","current","adv","job Id","status","elps time","rem time","job ctime");
    print ('-' x 136 . "\n") unless ($html);

    #
    # Directory specified
    #
    if ( &UNSET_LOCK_FILE==0 )
    {
	    &ParamParser::SetDefaultBehaviour('lock_file');     # From now on, we lock param files before reading them
	}
	else
	{
		&ParamParser::UnSetDefaultBehaviour('lock_file');
	}
    if (-d $check_file)
    {
        MainCommandLineCheckIsDirThenExit($check_file,$html,$quiet);
    }
    else
    {
        MainCommandLineCheckIsFileThenExit($check_file,$html,$quiet);
    }
}

#
# Title: MainCommandLineCheckIsDirThenExit
# Usage: &MainCommandLineCheckIsDirThenExit($check_dir,$html)
# procedure: Called by MainCommandLineCheckThenExit when paraloop is entered
#            through a command line with the check switch AND the argument of the
#            switch is a directory
# Args:      $check_dir = The lock directory
#            $html      = If 1, print inside an html table
#            $quiet     = If 1, print quiet, ie removing the 'DONE' lines, if any
# Return:    NEVER RETURN
# Globals:   None
#

# MODE = check

sub MainCommandLineCheckIsDirThenExit
{
    my $check_dir = shift;
    my $html = shift;
    my $quiet = shift;
    $quiet=0 unless(defined $quiet);
    $quiet += 0;
    
    my %output_lines=();
    my $time_start=0;
    my $cumulated_user_time = 0.0;
    my $cumulated_system_time = 0.0;
    my $cumulated_elapsed_time = 0;
    my $cumulated_atoms_done = 0;
    my $cumulated_atoms_todo = 0;
    my $running_jobs = 0;
    my @a_estimated = ();
    my $load_balancing;
    
    # read every lock file to check the jobs, fill some variables
    my $ncpu=0;
    foreach my $f (glob("$check_dir/lock.*.lck"))
    {    # For each lock file...
        next if (-d $f);    # Should not happen !!!
        my $estimated = '';
        my $local_time_start=0;
        $running_jobs += &CheckFile($f,\%output_lines,\$local_time_start,\$cumulated_elapsed_time,\$cumulated_system_time,\$cumulated_user_time,
        \$cumulated_atoms_done,\$cumulated_atoms_todo,\$estimated,\$load_balancing,$quiet);
        push (@a_estimated,$estimated);
        
        # Keep the oldest start time (but NOT 0)
        if ($time_start==0 or ($local_time_start>0 and $local_time_start<$time_start))
        {
            $time_start = $local_time_start;
        };     
        $ncpu++;
    };

    # @mkeys = the MASTER jobs (Mxxx)
    # @nkeys = the SLAVE jobs (in MASTER/SLAVE mode) or the jobs in normal mode    
    my @mkeys = grep (/\AM/,keys %output_lines);
    my @nkeys = grep (/\A\d/, keys %output_lines);

    # Print first the master jobs
    if (@mkeys > 0)
    {
        foreach my $k (sort {my ($a1)=($a=~/M(\d+)/);my ($b1)=($b=~/M(\d+)/);$a1<=>$b1} @mkeys) 
        {
            my $line = $output_lines{$k};
            if ($html)
            {
                $line =~ s!\A +!<tr><td>!;
                $line =~ s! +!</td><td>!;
                $line =~ s! +!</td><td>!;
                $line =~ s! +!</td><td>!;
                $line =~ s! +!</td><td>-</td><td>-</td><td>-</td><td>!;
                $line =~ s! +!</td><td>!g;
                $line .= "</td></tr>";
            };
            print $line . "\n";
        };

        if ($html)
        {
            print '<tr><td colspan="11"><hr /></td></tr>' . "\n";
        }
        else
        {
            print ('-' x 136 . "\n");
        };
    };

    # then the slave jobs
    foreach my $k (sort {my ($a1)=($a=~/\d+_(\d+)/);my ($b1)=($b=~/\d+_(\d+)/);$a1<=>$b1} @nkeys)
    {
        my $line = $output_lines{$k};
        if ($html)
        {
            $line =~ s!\A +!<tr><td>!;
            $line =~ s! +!</td><td>!g;
            $line .= "</td></tr>";
        };
        print $line . "\n";
    };

    # Print the summary if everything is terminated
    if ($running_jobs == 0)
    {
        if ($html)
        {
            print '<tr><td colspan="11"><hr /></td></tr>' . "\n";
            print ('<tr><td colspan="5">NUMBER OF JOBS</td><td colspan=5>' . $ncpu,1 .'</td></tr>');
            print ('<tr><td colspan="5">CUMULATED ELAPSED TIME</td><td colspan=5>' . &Time2Hour($cumulated_elapsed_time,1) .'</td></tr>');
            print ('<tr><td colspan="5">CUMULATED     CPU TIME</td><td colspan=5>' . &Time2Hour($cumulated_system_time + $cumulated_user_time,0) .'</td></tr>');
            print ('<tr><td colspan="5">SYSTEM</td><td colspan=5>'. &Time2Hour($cumulated_system_time,0) .'</td></tr>');
            print ('<tr><td colspan="5">USER</td><td colspan=5>' .  &Time2Hour($cumulated_user_time,0) ."\n");
        }
        else
        {
            print ('-' x 136 . "\n");
            print ("        NUMBER OF JOBS = " . $ncpu . "\n");
            print ("CUMULATED ELAPSED TIME = " . &Time2Hour($cumulated_elapsed_time,1) ."\n");
            print ("CUMULATED     CPU TIME = " . &Time2Hour($cumulated_system_time + $cumulated_user_time,0) ."\n");
            print ("                SYSTEM = " . &Time2Hour($cumulated_system_time,0) ."\n");
            print ("                  USER = " . &Time2Hour($cumulated_user_time,0) ."\n");
        };
    }
    
    # Print the remaining estimated time if still running
    else
    {
        # Try to estimate the global remaining time
        my ($max_estimated,$ttl_estimated)=(0,0);
        for my $e (@a_estimated)
        {
            $max_estimated = $e if ($e > $max_estimated);
            $ttl_estimated += $e;
        }
        my $estimated = ($load_balancing==0) ? $max_estimated : int($ttl_estimated/$running_jobs);
        
        # Try to estimate the global remaining time, taking into account the interruptions
        my $large_estimated = $estimated;
        if ($time_start != 0 and $cumulated_atoms_done != 0)
        {
            $large_estimated = (time()-$time_start)*($cumulated_atoms_todo/$cumulated_atoms_done-1);
        };

        # Compute the predicted end date
        my $end = time() + $large_estimated;
        my @end_date = localtime($end);
        my $end_date = ($end_date[5]+1900).'-'.($end_date[4]+1).'-'.$end_date[3].' '.$end_date[2].':'.$end_date[1].':'.$end_date[0];
        
        if ($html)
        {
            print '<tr><td colspan="11"><hr /></td></tr>' . "\n";
            print '<tr><td>LOAD BALANCING ENABLED</td></tr>' . "\n" if ($load_balancing!=0); 
            if ($cumulated_atoms_done != 0)
            {
                printf("<tr><td colspan=\"5\">Total advancement (%%)</td><td colspan=\"5\">%2i</td></tr>",100*$cumulated_atoms_done/$cumulated_atoms_todo);
                print ('<tr><td colspan="5">Remaining estimated cpu time</td><td colspan="5">'   . &Time2Hour($estimated,1)       . '</td></tr>');
                print ('<tr><td colspan="5">Remaining estimated elapsed time</td><td colspan="5">' . &Time2Hour($large_estimated,1) . '</td></tr>');
                print ('<tr><td colspan="5">Estimated end date=</td><td colspan="5">' . $end_date . '</td></tr>');
            }
         }
        else
        {
            print ('-' x 136 . "\n");
            print 'LOAD BALANCING ENABLED' . "\n" if ($load_balancing!=0); 
            if ($cumulated_atoms_done != 0)
            {
                printf ("Total advancement (%%)            = %2i\n",100*$cumulated_atoms_done/$cumulated_atoms_todo);
                print ("Remaining estimated cpu time     = " . &Time2Hour($estimated,1)."\n");
                print ("Remaining estimated elapsed time = " . &Time2Hour($large_estimated,1) ."\n");
                print ("Estimated end date               = " . $end_date ."\n");
            }
        }
    }
    if ($html)
    {
        print ("</table>\n");
    }
    else
    {
            print ('-' x 126 . "\n"); 
    };
	if ( $load_balancing!=0)
	{
		print $html?'<p>':'';
		if (!$quiet)
		{
			print "TRY THE --quiet SWITCH TO DISPLAY LESS STUFF";
		}
		else
		{
			print "REMOVE THE --quiet SWITCH TO DISPLAY MORE INFORMATION";
		}
		print $html ? "</p>\n":"\n";
	}
    exit 0;
}

#
# Title: MainCommandLineCheckIsFileThenExit
# Usage: &MainCommandLineCheckIsFileThenExit($check_file,$html)
# procedure: Called by MainCommandLineCheckThenExit when paraloop is entered
#            through a command line with the check switch AND the argument of the
#            switch is a file
# Args:      $check_file = The lock file
#            $html      = If 1, print inside an html table
# Return:    NEVER RETURN
# Globals:   None
#

# MODE = check

sub MainCommandLineCheckIsFileThenExit
{
    my $check_file = shift;
    my $html = shift;
    my %output_lines=();
    &CheckFile($check_file,\%output_lines);
    foreach my $k (keys %output_lines) 
    {
        my $line = $output_lines{$k};
        if ($html)
        {
            $line =~ s! +!</td><td>!g;
            $line = "<th><td>".$line."</td></th>";
        };
        print $line . "\n";
    };
    exit 0;
}

#
# Title: CheckFile
# Usage: &CheckFile($check_file,\%result, \$elapsed, \$system, \$user, $quiet)
# function:  Check a file, add some data to %result
# Args[in]:  $check_file[in] = The lock file
#            \%result[out]   = A hash for storing the result
#            \$elapsed[out], \$system[out], \$user[out] = 3 scalars for storing the accumulated elapsed, user, system times
#            \$atoms_done[out],\$atoms_todo[out] = 2 scalars for storing the accumulated nb of atom jobs already done, and the total number of atom jobs to do
#            \$estimated[out] = The estimated remaining time for this job
#            \$load_balancing[out] = the load_balancing parameter, as read from check_file
#            $quiet                = If set, do not display the lines marked as "DONE"
# Return:    1 if terminated, 0 if not
# Globals:   None
#

# MODE = check

sub CheckFile
{
    my $f = shift;
    my $rh_result = shift;
    my ($r_time_start,$r_elapsed,$r_system,$r_user,$r_atoms_done,$r_atoms_todo,$r_estimated,$r_load_balancing,$quiet) = (@_);
    $quiet=0 unless(defined $quiet);
    
    my $o_param;
    my $i=0;
    while($i<4)
    {
        $o_param = New ParamParser($f);
        $o_param->Update($f.'_tmp','O') if (-e $f.'_tmp');
        $o_param->SelectNameSpace(NAMESPACE);
        
        if (!$o_param->IsDefined('stop_all'))
        {
            warn "$f empty - THE LOCKS SHOULD BE ENABLED ON YOUR nfs INSTALLATION";
            $i++;
            sleep 1;
        }
        else {
            last;
        };
    };
    
    my $plp_id  = $o_param->Get('plp_id');
    my $job_id  = $o_param->Get('job_id');
    my $status  = $o_param->Get('status');
    my $run_flg = ($status eq 'TERMINATED') ? 0 : 1;
    my $start   = $o_param->Get('start')+0;
    my $end     = $o_param->Get('end')+0;
    my $step    = $o_param->Get('step')+0;
    my $current = ($run_flg==0)?'*':$o_param->Get('current')+0;
    my $elapsed = ($o_param->IsDefined('time_elapsed')) ? $o_param->Get('time_elapsed')+0 : 0;
    my $size    = 1 + int(($end - $start) / $step);
    my $adv;
    my $line;
    my $estimated;
    my $mode    = ($o_param->IsDefined('current_mode')) ? $o_param->Get('current_mode') : 'AUTONOMOUS';
    if ($mode eq 'MASTER')
    {
        $adv = ' ';
        $size= ' ';
        $current=' ';
        $line = sprintf("%8s %10i %10i %10i %10s %10s %5s %22s %12s",$plp_id,$start,$end,$step,$size,$current,$adv,$job_id,$status);
        $rh_result->{'M'.$start} = $line;
    } 
    else
    {
        if ($run_flg==0)
        {
            $adv = 100;
            $estimated = 0;
        }
        elsif ($end==$start)
        {
            $adv = '*';
            $estimated = 0;
        } 
        else
        {
            my $adv1 = 100.0 * ($current-$start) / (1+$end-$start);
            $adv     = int($adv1);
            $estimated = ($current==$start)? 0 : $elapsed * (100.0/$adv1 - 1);
        };

        my $cpu_user = 0;
        $cpu_user = $o_param->Get('time_user') + $o_param->Get('time_children_user') if ($o_param->IsDefined('time_user'));

        $line = sprintf("%8s %10i %10i %10i %10i %10s %5s %22s %12s  %9s %9s %9s",$plp_id,$start,$end,$step,$size,$current,$adv,$job_id,$status,&Time2Hour($elapsed,1),&Time2Hour($estimated,1),&Time2Hour($cpu_user,1));
        $rh_result->{$plp_id.'_'.$start} = $line;
        
        if ($o_param->IsDefined('done_segments'))
        {
            my @a_done_segments     = split(',',$o_param->Get('done_segments'));
            my @a_done_time_elapsed = split(',',$o_param->Get('done_time_elapsed'));
            foreach my $s (@a_done_segments) 
            {
                my ($start,$end,$step) = split ('-',$s);
                $size                  = 1 + int(($end - $start) / $step);
                my $done_elapsed       = shift @a_done_time_elapsed;
                if ( $quiet==0 )
                {
	                my $line = sprintf("%8s %10i %10i %10i %10i %10s %5s %22s %12s  %9s",$plp_id,$start,$end,$step,$size,'*',100,$job_id,'DONE',&Time2Hour($done_elapsed,1));
	                $rh_result->{$plp_id.'_'.$start} = $line;
				}
                $$r_atoms_done += $size;
                $$r_atoms_todo += $size;
                $$r_elapsed    += $done_elapsed;
            };
        };
    }
    
    $$r_time_start = $o_param->Get('time_start') if ($o_param->IsDefined('time_start')); 
    $$r_elapsed += $o_param->Get('time_elapsed') if ($o_param->IsDefined('time_elapsed'));
    $$r_user    += $o_param->Get('time_user') + $o_param->Get('time_children_user') if ($o_param->IsDefined('time_user'));
    $$r_system  += $o_param->Get('time_system') + $o_param->Get('time_children_system') if ($o_param->IsDefined('time_system'));
    $$r_atoms_done += ($current eq '*') ? $size : ($current-$start)/$step;
    $$r_atoms_todo += $size;
    $$r_estimated = $estimated;
    $$r_load_balancing = ($o_param->IsDefined('load_balancing_enabled')) ? $o_param->Get('load_balancing_enabled') : 0;
    return $run_flg;
};

#
# Title: MainCommandLineWaitOrCleanThenExit
# Usage: &MainCommandLineWaitOrCleanThenExit($o_param)
# procedure: Called by MainCommandLineThenExit when paraloop is entered
#            through a command line with ONLY --wait or --clean
# Args:      $o_param
# Return:    NEVER RETURN
# Globals:   None
#

# MODE = waiting


sub MainCommandLineWaitOrCleanThenExit
{
    my $o_param = shift;
    my $wait_dir = &ValidateParameters($o_param);
    my $wait = ($o_param->IsDefined('wait')) ? 1 : 0;
    my $clean = ($o_param->IsDefined('clean')) ? 1 : 0;
    if ( &UNSET_LOCK_FILE==0 )
    {
	    &ParamParser::SetDefaultBehaviour('lock_file');     # From now on, we lock param files before reading them
	}
	else
	{
		&ParamParser::UnSetDefaultBehaviour('lock_file');
	}
    &WaitParaloop($wait_dir) if ($wait);
    &CleanParaloop($wait_dir) if ($clean);
}

#
# Title: MainCommandLineOneLoopThenExit
# Usage: &MainCommandLineOneLoopThenExit($o_param)
# procedure: Called by MainCommandLineThenExit when paraloop is entered
#            through a command line for executing a monoprocessor loop
# Args:      $o_param
# Return:    NEVER RETURN
# Globals:   None
#

# MODE = mono/init

sub MainCommandLineOneLoopThenExit
{    # STARTING A NEW JOB

    my $o_param = shift;
    my $ra_cfg_files = &ReadCfgFiles($o_param);
    $o_param->AssertDefined("input", "program");
    if ($o_param->IsDefined('output'))
    {
        $o_param->Set('output_prefix',$o_param->Get('output'));
    }
    else
    {
        $o_param->Set('output_prefix','NULL');
    }
   
    # Some substitutions to make

    &MakeSubstitutions($o_param);
    
    my $command_line = $0 . ' ' . join(' ',@ARGV);   # Keep track of the complete command line
    $o_param->Set('command_line', $command_line);

    $o_param->Set('workdir', cwd);    # Keep track of the current directory for the qsub calls

    &InitBoundaries($o_param);
    &InitFileParameters($o_param);

    # Create lock directory and lock file, die if something wrong
    
    my $lock_file = $o_param->Get('lock_file');
    if (-e $lock_file)
    {
        die "ERROR - LOCKFILE $lock_file FOUND";
    }
    my $lck = new IO::File(">$lock_file") or die "Cannot create $lock_file";
    print "created $lock_file\n";

    # Create objects

    my $log_file = $o_param->Get('log_file');
    my $log_level= $o_param->Get('log_level');
    my $o_log = Logger->New(">>$log_file", $log_level);

    my $plugin = &CreatePlugin($o_param,$o_log);
    
    my $step   = $o_param->Get("step") + 0;
    if ($step < 0)
    {
        $plugin->SkipRecords(-1);    # Just to check that this object is able to skip a negative number of records
    }

    my $sched = &CreateScheduler($o_param,$o_log);

    $o_log->Trace("STARTING PARALOOP\n",                         '*', '1');
    $o_log->Trace("PARAMETERS WERE SET FROM THE COMMAND LINE\n", '.', '2');
    for my $f (@$ra_cfg_files)
    {
        $o_log->Trace("PARAMETERS WERE SET FROM THE FILE $f\n", '.', '2');
    }
    
    &InitStealStatus($o_param,$o_log);      # For load balancing 

    $o_param->Set('status', 'STARTING');

    #
    # debug mode: Start MainLoop without forking a new process:
    #             This is convenient only for debugging a new pligin
    #

    # NOTE - We are now leaving the mono/init mode: in the normal operation we submit
    #        a new paraloop job, but in debug mode we just switch to mono/running and we call MainLoop
    
    if ($o_param->IsDefined('plugin_debug'))
    {                                     # Start MainLoop without forking a new process
        &MainLoop($o_param,$o_log,$plugin,$sched);
        &PrintVar($o_param);
    }
    else
    {
        my $job_id = $sched->Submit();
        if ($job_id eq '0')
        {
            print "No job could be launched (the nodes are too much loaded)\n";
        }
        else
        {        
            print "launched job, job ID = $job_id\n";
        }
    }
    exit(0);
}

#
# Title: MainSubmitThenExit
# Usage: &MainSubmitThenExit()
# procedure: Called when paraloop is entered after a call to the Submit sub
#            of a previous instance
#            Called for submits in current_mode SLAVE or AUTONOMOUS
# Args:      None
# Return:    NEVER RETURN
# Globals:   None
#

# MODE = mono/running

sub MainSubmitThenExit
{
    if ( &UNSET_LOCK_FILE==0 )
    {
	    &ParamParser::SetDefaultBehaviour('lock_file');     # From now on, we lock param files before reading them
	}
	else
	{
		&ParamParser::UnSetDefaultBehaviour('lock_file');
	}
    my $o_param = New ParamParser();    # The object handling parameters
    $o_param->SetUsage(\&Usage);
    $o_param->SetBehaviour('exit_on_getopt_error');
    $o_param->Update('ENV', 'I');    # Read parameters from ENV. Some of them start with a prefix
    #print "================================\n";
    #$o_param->Print();
    #print "\n================================\n";
    $o_param->SelectNameSpace(NAMESPACE);    # Consider this prefix as a space name

    # Set the quiet option avoiding a lot of redundant messages
    $o_param->Delete('verbose');
    $o_param->Set('quiet', 1);

    my $lock_file = $o_param->Get('lock_file');    # Read last parameters (namely job_id) from the lock file
    #&PrintVar($o_param);
    
    # Make the substitutions which are specific to each job
    &MakeSubstitutionsSubmitTime($o_param);
    
    chdir($o_param->Get('workdir'));

    my $log_file  = $o_param->Get('log_file');
    my $log_level = $o_param->Get("log_level");
    my $o_log = Logger->New(">>$log_file", $log_level);
    $o_log->Trace("PARAMETERS WERE SET FROM ENV\n", '.', '2');

    # Create the Scheduler object
    my $o_sched = &CreateScheduler($o_param,$o_log);

    # Ask the scheduler for the job Id, and store this ID in the parameters
    my $job_id = $o_sched->JobId();
    $o_param->Set('job_id', $job_id);

    # Create the Plugin object
    my $o_plugin = &CreatePlugin($o_param,$o_log);

    &MainLoop($o_param,$o_log,$o_plugin,$o_sched);

    # We arrive here ONLY if the main loop is terminated
    # Change the status of the job to 'TERMINATED'
    $o_param->Set('status', 'TERMINATED');
    &DumpToFile($o_param);

    #
    # --mail switch: send a mail if ALL the jobs are terminated
    #
    
    if ($o_param->IsDefined('mail')) {
        my $hostname = `$o_param->Get('hostname')`;
        my $mail_agent = $o_param->Get('mail_agent');
        my $echo = $o_param->Get('echo');
        my $lock_dir = dirname($lock_file);
        if (IsInterrupted($lock_dir)==2) {          # Every job is marked as TERMINATED
            my $mail_addr = $o_param->Get('mail');
            my $subject   = "PARALOOP terminated";
            my $body      = "PARALOOP terminated all his jobs.\nTry paraloop --check $lock_dir for the details\n\n";
            $body .= "Regards, \n" . $hostname;
            `$echo \"$body\" | $mail_agent -s \"$subject\" $mail_addr`;
        };
    };
    exit(0);

}

#
# Title: MainSubmitMasterThenExit
# Usage: &MainSubmitMasterThenExit()
# procedure: Called when paraloop is entered after a call to the Submit sub
#            of a previous instance, AND the current_mode parameter is set to master
#            meaning we are treating a master job in MASTER/SLAVE mode
# Args:      None
# Return:    NEVER RETURN
# Globals:   None
#

# MODE = mono/running

sub MainSubmitMasterThenExit
{
    if ( &UNSET_LOCK_FILE==0 )
    {
	    &ParamParser::SetDefaultBehaviour('lock_file');     # From now on, we lock param files before reading them
	}
	else
	{
		&ParamParser::UnSetDefaultBehaviour('lock_file');
	}
    my $o_param = New ParamParser('ENV');    # The object handling parameters - Init from ENV
    $o_param->SelectNameSpace(NAMESPACE);    # Consider this prefix as a space name

    # Set the quiet option avoiding a lot of redundant messages
    $o_param->Delete('verbose');
    $o_param->Set('quiet', 1);

    #$o_param->Print();
    my $lock_file = $o_param->Get('lock_file');    # Read last parameters (namely job_id) from the lock file
    $o_param->SelectNameSpace('');
    $o_param->Update($lock_file,'A');
    $o_param->SelectNameSpace(NAMESPACE);    #&PrintVar($o_param);
    chdir($o_param->Get('workdir'));

    my $log_file  = $o_param->Get('log_file');
    my $log_level = $o_param->Get("log_level");
    my $o_log = Logger->New(">>$log_file", $log_level);
    $o_log->Trace("ENTERING MASTER MODE\n", '.','1');
    $o_log->Trace("PARAMETERS WERE SET FROM ENV\n", '.', '2');

    # Create the Scheduler object (will be used later)
    my $o_sched = &CreateScheduler($o_param,$o_log);

    # Ask the scheduler for the job Id, and store this ID in the parameters
    my $job_id = $o_sched->JobId();
    $o_param->Set('job_id', $job_id);

    #
    # 3 different modes, driven by the status: RESUMING, RESTARTING, STARTING
    #
    
    my $status = $o_param->Get('status');
    #$o_log->Trace("KOUKOU $status\n",'.','2');
    my $ra_slave_files;
    if ($status eq 'STARTING')
    {
        #
        # We are now switching to the SLAVE MODE
        # 
        # NB - stop_all and stop_this files: those parameters are copied from $o_param, like the others
        #      BUT they are set by a SetUnlessDefined function in InitFileParameters
        #      Thus every slave job belonging to this master will be stopped by the SAME file
        #      Thus all the slaves belonging to the same slave will be stopped together
        #
        
        $o_param->Set('master_lock_file',$lock_file);
        my $o_param_slave = New ParamParser('PARAMPARSER', $o_param);
        $o_param_slave->SelectNameSpace(NAMESPACE);
        $o_param_slave->Set('current_mode','SLAVE');
        if (! $o_param_slave->IsDefined('slave_Scheduler')) {
            die "ERROR - The slave_Scheduler parameter should be defined\n";
        };
        $o_param_slave->Set('Scheduler',$o_param_slave->Get('slave_Scheduler'));
        if (! $o_param_slave->IsDefined('slave_ncpus')) {
            die "ERROR - The slave_ncpus parameter should be defined\n";
        };
        my $slave_ncpus = $o_param_slave->Get('slave_ncpus');
        
        my $global_start = $o_param_slave->Get('start');
        my $global_end   = $o_param_slave->Get('end');
        
        #$o_log->Trace("koukou je lance $slave_ncpus jobs\n","","0");
        
        $ra_slave_files = &CpuLoop($global_start,$global_end,$slave_ncpus,$o_param_slave,$o_log);
        
        $o_param->Set('slave_lock_files',join(',',@$ra_slave_files));
    }
    elsif ($status eq 'RESTARTING')
    {
        my @slaves = split(',',$o_param->Get('slave_lock_files'));
        $ra_slave_files = \@slaves;
        foreach my $s (@$ra_slave_files)
        {
            my $o_param_slave = New ParamParser($s);
            $o_param_slave->SelectNameSpace(NAMESPACE);

            $o_param_slave->Set('current',$o_param_slave->Get('start'));
            
            my $output = $o_param_slave->Get('output');
            link($output, $output.'.bkp');
            unlink($output);
            $o_param_slave->Set('output',&IncrementName($output));
            
            my $log_file = $o_param_slave->Get('log_file');
            link($log_file,$log_file.'.bkp');
            unlink($log_file);
            $log_file = &IncrementName($log_file);
            $o_param_slave->Set('log_file',$log_file);

            my $log_level= $o_param_slave->Get('log_level');
            my $o_log_slave = Logger->New(">>$log_file", $log_level);
            my $o_sched = &CreateScheduler($o_param_slave,$o_log_slave); 
            $o_param_slave->Set('subcount',0);
            $o_log->Trace("RESTARTING PARALOOP FOR SLAVE $s\n",'.','1');
            my $job_id = $o_sched->Submit();
            if ($job_id eq '0')
            {
                print "No job could be launched (the nodes are too much loaded)\n";
            }
            else
            {        
                print "launched job, job ID = $job_id\n";
            }
        }
    }
    elsif ($status eq 'RESUMING' or $status eq 'SUBMIT')
    {
        my @slaves = split(',',$o_param->Get('slave_lock_files'));
        $ra_slave_files = \@slaves;
        foreach my $s (@$ra_slave_files)
        {
            my $o_param_slave = New ParamParser($s);
            $o_param_slave->SelectNameSpace(NAMESPACE);

            my $log_file = $o_param_slave->Get('log_file');
            my $log_level= $o_param_slave->Get('log_level');
            my $o_log_slave = Logger->New(">>$log_file", $log_level);
            my $o_sched = &CreateScheduler($o_param_slave,$o_log_slave); 
            $o_param_slave->Set('subcount',0);
            $o_log->Trace("RESUMING PARALOOP FOR SLAVE $s\n",'.','1');
            my $job_id = $o_sched->Submit();
            if ($job_id eq '0')
            {
                print "No job could be launched (the nodes are too much loaded)\n";
            }
            else
            {        
                print "launched job, job ID = $job_id\n";
            }
        }
    }       
    else        # At this point, the master's status should be RESUMING,RESTARTING or STARTING. Nothing else.
    {
        die "\nERROR - Internal Error, status = $status";
    }
    
    $o_param->Set('status', 'MONITORING');
    &DumpToFile($o_param);
    my $monitor_status = &MonitorJobs($ra_slave_files);
    if ($monitor_status == 2)
    {
        $o_param->Set('status','SUBMIT');
        $o_sched->Submit();
    };
    if ($monitor_status == 1)
    {
        $o_param->Set('status','INTERRUPTED');
        $o_log->Trace("ALL THE SLAVES ARE INTERRUPTED. PLEASE RESUME WITH --resume $lock_file\n",'*','1');
    };
    if ($monitor_status == 0)
    {
        $o_param->Set('status','TERMINATED');
        $o_log->Trace("ALL THE SLAVES ARE TERMINATED.\n",'*','1');
    };
    &DumpToFile($o_param);
    exit(0);
}

#
# Title: MainLoop
# Usage: &MainLoop($o_param,$o_log,$o_plugin,$o_scheduler)
# procedure: Enter the main loop of the program
#            1/ Retrieve 4 objects: param, log, plugin, scheduler
#            2/ Skip the beginning of the file if necessary, reaching the correct start record
#            3/ Call preproc script if necessary
#            4/ Loop from current to end. At each iteration:
#                     call preatom script if necessary
#                     call $plugin->Exec(), doing what must be done
#                     call postatom script if necessary
#                     If the output file is too big, change output and log file
#                     If time spent in the queue is too high, call postproc script, call submit and never return
#                     If SIGINT was received, OR if paraloop.stp was created, OR if paraloop.START.stp was created, call StopJob and never return
#            5/ If END was reached, call postproc script if necessary then return
#
# Return:    Nothing
# Args:      4 objects
# Globals:   none
#

sub MainLoop
{
    my $o_param        = shift;
    my $o_log           = shift;
    my $o_plugin        = shift;
    my $o_sched         = shift;
    my $lock_file       = $o_param->Get('lock_file');
    my $fair_time_limit = 0 + $o_param->IsDefined('fair_time_limit') ? $o_param->Get('fair_time_limit') : 0;
    
    # Cancel fair_time_limit if the scheduler does not support this
    if ($o_sched->IsSelfPossible()==0 and $fair_time_limit != 0)
    {
        $fair_time_limit = 0;
        $o_log->Trace("FAIR TIME LIMIT CANCELLED BECAUSE THE SCHEDULER DOES NOT SUPPORT SELF-SUBMIT\n",'.','1');
    };

    #
    # We start a "timer" - This timer will be cleared when the exit from MainLoop
    #                      See the code of this tiny object at the beginning of this file
    #
    
    my $o_timer = New ParaloopTimer($o_param);

    #
    # Calling the preprocessing script
    # Initializing the start timestamp and the current timestamp
    # start_timestamp is garanteed not to be modified during this run (ie up to the next Submit)
    # current_timestamp is computed at each iteration
    #
    
    my $start_timestamp = $o_log->CalcStamp();
    my $current_timestamp = $start_timestamp;
    $o_param->Set('start_timestamp',$start_timestamp);
    $o_param->Set('current_timestamp',$current_timestamp);
    &DumpToFile($o_param);
    $o_plugin->CallExternalScript('preproc_script',$o_param->Get('input'),'>>'.$o_param->Get('output'));
    
    while(1)
    {    
        #$o_log->Trace("koukou ici $$, my father is ".getppid()."\n","","0");
        my $start           = $o_param->Get('start') + 0;
        my $current         = $o_param->Get('current') + 0;
        my $step            = $o_param->Get('step') + 0;
        my $end             = $o_param->Get('end') + 0;
        my $db              = '';
        $db = $o_param->Get('db') if ($o_param->IsDefined("db"));
        my $input       = $o_param->Get('input');
        my $output      = $o_param->Get('output');
        my $output_size = (-s $output);
        
        #
        # skip the first n sequences
        #
        
        if ($current >= 0)
        {    # Nothing to do if $current < 0
            
            $o_param->Set('status', 'RUNNING');
            &DumpToFile($o_param);
            
            if ($current > 0)
            {   # Nothing to skip if $current == 0
                my $last_skipped = $current - 1;
                $o_log->Trace("SKIPPING SEQUENCES FROM 0 TO $last_skipped\n",'.', '1');
                $o_plugin->SkipRecords($current);
            }
            my $i_end           = 1 + int(($end - $current) / $step);
            my $first_iteration = 1;
            my $start_time      = time();
            my $elapsed_time    = 0;
            
            #
            # Here is the loop
            # We do not use a for loop, because $i_end is computed at every loop
            # Not sure this should work with a for loop
            #
            
            my $i = 0;
            while ($i < $i_end)
            {
                
                last if ($step>0 and $o_plugin->Tell()>$end);
                last if ($step<0 and $o_plugin->Tell()<$end);
                
                
                #
                # Do we still have time ?
                #
                
                my $timeout = $o_sched->RemTime();
                if ($timeout == 0)
                {
                    $o_plugin->CallExternalScript('postproc_script',$o_param->Get('input'),'>>'.$o_param->Get('output'));
                    $o_param->Set('status', 'SUBMIT');
                    $o_sched->Submit();
                    exit;                      # DO NOT RETURN FROM MAINLOOP
                }
                if ($timeout > 0)
                {
                    $o_log->Trace("THIS JOB WILL BE INTERRUPTED BY THE SCHEDULER IN $timeout s\n", '.', '2');
                }
                if ($fair_time_limit != 0)
                {
                    my $fair_remaining_time = $fair_time_limit - $elapsed_time;
                    $o_log->Trace("fair_time_limit=$fair_time_limit s; TIME REMAINING=$fair_remaining_time s\n", '.', '2');
                }
                
                $o_log->Trace("==========\n",$current_timestamp,'1');
                my ($incr_size, $sts, $kld) = $o_plugin->Exec($timeout);    # EXECUTE THE PROG
                
                ############################### my ($incr_size,$sts,$kld) = $o_plugin->Exec(4);	# EXECUTE THE PROG
                $sts = 0 if (!defined($sts));                             # Some objects do not care of return status
                $kld = 0 if (!defined($kld));                             # Some objects do not care of timeout
                if ($sts != 0)
                {
                    $o_log->Trace("THE EXEC FUNCTION RETURNED $sts\n", ".", '2');
                    last;
                }
                if (($kld == 0) or ($kld == -1 and $first_iteration == 1))
                {    # If all OK, or if timeout detected at 1st iteration:
                    if ($kld == -1)
                    {    # start next step in either case
                        $o_log->Trace("THE COMMAND TAKES TOO MUCH TIME FOR THIS QUEUE\n", ".", '1');
                    }
                    
                    #
                    # Keep the lock file synchronized with the program
                    #
                    
                    $o_plugin->SkipRecords($step - 1) if ($step != 1);    # skip some records, unless step=1
                    $current = $o_plugin->Tell();
                    $current_timestamp = $o_log->CalcStamp();
                    $o_param->Set('current', $current);
                    $o_param->Set('current_timestamp',$current_timestamp);
                    $o_param->Set('status',  'RUNNING');
                    $first_iteration = 0;
                }
                
                if ($kld == -1)
                {                                                       # Timeout reached: resubmit
                    $o_log->Trace("NO MORE TIME IN QUEUE\n", ".", '1');
                    $o_plugin->CallExternalScript('postproc_script',$o_param->Get('input'),'>>'.$o_param->Get('output'));
                    $o_param->Set('status', 'SUBMIT');
                    $o_sched->Submit();
                    exit;                      # DO NOT RETURN FROM MAINLOOP
                }
                
                $elapsed_time = time() - $start_time;
                if ($fair_time_limit > 0 and $elapsed_time >= $fair_time_limit)
                {
                    $o_log->Trace(
                    "TOO MUCH TIME SPENT IN THIS QUEUE (elapsed $elapsed_time, fair_time_limit $fair_time_limit)\n",
                    ".", '1');
                    $o_plugin->CallExternalScript('postproc_script',$o_param->Get('input'),'>>'.$o_param->Get('output'));
                    $o_param->Set('status', 'SUBMIT');
                    $o_sched->Submit();
                    exit;                      # DO NOT RETURN FROM MAINLOOP
                }
                
                &DumpToFile($o_param);
                
                #
                # Size matters: we don't want the output file to become too large
                #
                
                $output_size += $incr_size;
                my $max_file_size = $o_param->Get('max_file_size') + 0;
                if ($output_size / 1000000 >= $max_file_size)
                {
                    my $log_file      = $o_param->Get('log_file');
                    my $next_log_file = &IncrementName($log_file);
                    my $next_output   = &IncrementName($output);
                    $o_log->Trace("NOW WRITING OUTPUT TO $next_output\n", ".", '1');
                    $o_param->Set('log_file', $next_log_file);
                    $o_log->SetLogFile($next_log_file);
                    $o_param->Set('output', $next_output);
                    $output      = $next_output;
                    $output_size = 0;
                    $o_plugin->SetOutputFile($output);
                }
                
                #
                # Do we have to interrupt our job ?
                #
                
                my $stop_all  = $o_param->Get('stop_all');
                my $stop_this = $o_param->Get('stop_this');
                if (-r $stop_all or -r $stop_this or $kld == -2)
                {    # Some .stp file found, or term signal received (qdel)
                    if ($i >= $i_end)
                    {
                        last;    # The job is finished
                    }
                    $o_plugin->CallExternalScript('postproc_script',$o_param->Get('input'),'>>'.$o_param->Get('output'));
                    $o_log->Trace("JOB INTERRUPTED BY THE USER (found $stop_all)\n", '*', '1') if (-r $stop_all);
                    $o_log->Trace("JOB INTERRUPTED BY THE USER (found $stop_this)\n",'*', '1') if (-r $stop_this);
                    $o_log->Trace("JOB INTERRUPTED BY THE USER (received SIGINT)\n", '*', '1') if ($kld==-2);
                    $o_param->Set('status', 'INTERRUPTED');
                    &DumpToFile($o_param);
                    &StopJob($o_param,$o_log);
                    exit(0);
                };
                
                #
                # Read the temporary lock file, if it exists
                # Update $o_param in Overwrite mode, then remove the file
                # Adjust $end and $i_end
                #
                
                my $tmp_lock_file = $lock_file.'_tmp';
                if (-r $tmp_lock_file)
                {
                    $o_param->SelectNameSpace();
                    $o_param->Update($tmp_lock_file,'O');
                    $o_param->SelectNameSpace(NAMESPACE);
                    unlink ($tmp_lock_file);
                    $o_param->Dump($lock_file);
                    $end = $o_param->Get('end');
                    $i_end = 1 + int(($end - $start) / $step);
                    $o_log->Trace("JOB DYNAMICALLY RECONFIGURED - END = $end\n",$current_timestamp,'1');
                };
                $i++;
                $o_timer->UpdateElapsed();
            }
        }

        #
        # Load balancing is not enabled: exit from the loop
        #
        
        last unless ($o_param->IsDefined('load_balancing_enabled') and $o_param->Get('load_balancing_enabled') != 0);

        #
        # Load balancing enabled: call StealWork, trying to steal some work to another job
        #
        
        last unless ($o_plugin->StealWork($o_sched)==1);
    
    };
    
    $o_plugin->CallExternalScript('postproc_script',$o_param->Get('input'),'>>'.$o_param->Get('output'));
    $o_log->Trace("JOB FINISHED - BYE BYE\n", '*', '1');
    
    return;

}

#
# Title: MakeSubstitutions
# Usage: &MakeSubstitutions($o_param)
# procedure: Make the substitutions in $o_param
# Return:   Nothing
# Args:     $o_param
# Globals:  none
#

sub MakeSubstitutions
{
    my $o_param = shift;
    my ($second, $minute, $hour, $day,$mon,$year) = localtime();
    $second = "0$second" if ($second < 10);
    $minute = "0$minute" if ($minute < 10);
    $hour   = "0$hour" if ($hour < 10);
    $day    = "0$day" if ($day < 10);
    $mon += 1;      # 01..12
    $mon    = "0$mon" if ($mon < 10);
    $year -= 100;   # 2005 ==> 05
    $year   = "0$year" if ($year < 10);

    my $ncpus       = ($o_param->IsDefined('ncpus'))       ?  $o_param->Get('ncpus') : 1;
    my $slave_ncpus = ($o_param->IsDefined('slave_ncpus')) ?  $o_param->Get('slave_ncpus') : 1;
    my $master_ncpus= ($o_param->IsDefined('master_ncpus'))?  $o_param->Get('master_ncpus') : 1;
    my $local_ncpus = ($o_param->IsDefined('local_ncpus')) ?  $o_param->Get('local_ncpus') : 1;

    $o_param->SetSubstitution('%s', \$second);   
    $o_param->SetSubstitution('%m', \$minute);
    $o_param->SetSubstitution('%h', \$hour);
    $o_param->SetSubstitution('%D', \$day);
    $o_param->SetSubstitution('%M', \$mon);
    $o_param->SetSubstitution('%Y', \$year);
    $o_param->SetSubstitution('%p', \$ncpus);
    $o_param->SetSubstitution('%v', \$slave_ncpus);
    $o_param->SetSubstitution('%t', \$master_ncpus);
    $o_param->SetSubstitution('%l', \$local_ncpus);
    $o_param->SetBehaviour('use_substitution_table');           # The substitution are done now
};

#
# Title: MakeSubstitutionsSubmitTime
# Usage: &MakeSubstitutionsSubmitTime($o_param)
# procedure: Make the substitutions in $o_param at submit time 
# Return:   Nothing
# Args:     $o_param
# Globals:  none
#

sub MakeSubstitutionsSubmitTime
{
    my $o_param = shift;
    
    my $id = $o_param->Get('plp_id');
    
    $o_param->SetSubstitution('%i', \$id);   
    $o_param->SetBehaviour('use_substitution_table');           # The substitution are done now
};

#
# Title: CpuLoop
# Usage: &CpuLoop($global_start, $global_end, $ncpus,$o_param,$o_log)
# function: Submit several jobs (1 per cpu), computing the slices
# Return:   an array initialized with the created lock file names
# Args:     $o_param,$o_log = The needed objects
#           $global_start, $global_end = The global boundaries used for computing the limits
#           $ncpus = The number of cpus
# Globals:  none
#

sub CpuLoop {
    my $global_start = shift;
    my $global_end = shift;
    my $ncpus = shift;
    my $o_param = shift;
    my $o_log = shift;
    my @result = ();
    
    my $mode = ($o_param->IsDefined('current_mode')) ? $o_param->Get('current_mode') : 'AUTONOMOUS';
    my $plp_id_prefix='';
    if ($mode eq 'SLAVE')       # We are entering CpuLoop in SLAVE mode: if there is a plp_id, this is taken as a prefix 
    {
        if ($o_param->IsDefined('plp_id'))
        {
            $plp_id_prefix = $o_param->Get('plp_id') . '-';
        }
        else
        {
            $plp_id_prefix = '1-';
        };
    };
    my $plp_id = 0;
    
    $o_param->Set('workdir', cwd);    # Keep track of the current directory for the qsub calls
    $o_param->Set('subcount',0);

    #
    # Interleaved mode: the load is distributed as follows (for ncpus=4):
    #                   CPU0 => 0,4,8,...
    #                   CPU1 => 1,5,9,...
    #                   CPU2 => 2,6,10,...
    #                   CPU3 => 3,7,11,...
    # WARNING: If the end specified was 10, the REAL LAST read record will be 11
    #
    # However, the MASTER jobs distribute the data through a SLICE mode
    #          We go to interleaved mode ONLY for SLAVE jobs
    #

    if ($o_param->IsDefined('interleaved') and $mode ne 'MASTER')
    {
        $o_param->Delete('interleaved');
        $o_param->Set('end', $global_end);
        $o_param->Set('step', ($global_end > $global_start) ? $ncpus : -$ncpus);
        for (my $cpu = 0; $cpu < $ncpus; $cpu++)
        {
            # Work on a copy of $o_param
            
            my $o_work_param = New ParamParser('PARAMPARSER',$o_param);
            $o_work_param->SelectNameSpace(NAMESPACE);
            
            # Set start
            $o_work_param->Set('start', ($global_end > $global_start) ? $global_start + $cpu : $global_start - $cpu);
            
            # Set plp_id
            $plp_id++;
            $o_work_param->Set('plp_id',$plp_id_prefix.$plp_id);
            
            &InitBoundaries($o_work_param);
            &InitFileParameters($o_work_param);
            my $log_file = $o_work_param->Get('log_file');
            my $log_level= $o_work_param->Get('log_level');
            my $o_work_log = Logger->New(">>$log_file", $log_level);
            
            my $lock_file = $o_work_param->Get('lock_file');
            push (@result, $lock_file);
            #$o_work_param->Dump('ENV');
            #$o_work_param->Dump($lock_file);
            $o_work_log->Trace("$mode INTERLEAVED MODE - STARTING PARALOOP FOR CPU $cpu\n",'.','1');
            if ($mode eq 'SLAVE')
            {
                $o_work_log->Trace('MASTER LOCK FILE = '.$o_work_param->Get('master_lock_file')."\n",'.','1');
            }
            
            my $o_work_sched = &CreateScheduler($o_work_param,$o_work_log);
            my $job_id = $o_work_sched->Submit();
            if ($job_id eq '0')
            {
                print "No job could be launched (the nodes are too much loaded)\n";
                next;
            };
            my $msg = "start = ".$o_work_param->Get('start')." - end = ".$o_work_param->Get('end');
            $msg .= " - step = ".$o_work_param->Get('step')." - job Id = $job_id\n";
            $o_work_log->Trace($msg,'.','1');
            print "launched job, job ID = $job_id\n";
            $o_log->Trace("LAUNCHED JOB $job_id\n",'.','1');
        }
    }

    #
    # Slice mode: the input file is splitted in slices
    #

    else
    {
        my $no_slicing = $o_param->IsDefined('no_slicing');
        my $total_nb = $global_end - $global_start + 1;
        my $slice_size;
        my $slice_offset;
        if ($no_slicing)
        {
            $slice_size = $total_nb;
        }
        else
        {
            if ($total_nb % $ncpus == 0)
            {
                $slice_size = $total_nb / $ncpus;
            }
            else
            {
                $slice_size = int($total_nb / $ncpus) + 1;
            }
        }
        
        $o_param->Delete('start');
        $o_param->Delete('end');
        $o_param->Delete('step');
        $o_param->Set('slice_offset', $global_start);
        $o_param->Set('slice_size',   $slice_size);
        for (my $cpu = 0; $cpu < $ncpus; $cpu++)
        {
            # Work on a copy of $o_param

            my $o_work_param = New ParamParser('PARAMPARSER',$o_param);
            $o_work_param->SelectNameSpace(NAMESPACE);

            # Set slice nb            
            if ($no_slicing)
            {
                $o_work_param->Set('slice_nb', 0);
            }
            else
            {
                $o_work_param->Set('slice_nb', $cpu);
            };
            
            # Set plp_id
            $plp_id++;
            $o_work_param->Set('plp_id',$plp_id_prefix.$plp_id);
            
            &InitBoundaries($o_work_param,$global_end) or last;         # Set start,end versus slice information
            &InitFileParameters($o_work_param);
           
            my $log_file = $o_work_param->Get('log_file');
            my $log_level= $o_work_param->Get('log_level');
            my $o_work_log = Logger->New(">>$log_file", $log_level);
            my $lock_file = $o_work_param->Get('lock_file');
            push(@result, $lock_file);
            $o_work_log->Trace("ENTERING $mode SLICE MODE - STARTING PARALOOP FOR SLICE $cpu\n",'.','1');
            if ($mode eq 'SLAVE')
            {
                $o_work_log->Trace('MASTER LOCK FILE = '.$o_work_param->Get('master_lock_file')."\n",'.','1');
            }
            my $o_work_sched = &CreateScheduler($o_work_param,$o_work_log);
            my $job_id = $o_work_sched->Submit();
            #$o_work_log->Trace("koukou ici $$ - called Submit - returned $job_id\n","","0"); 
            if ($job_id eq '0')
            {
                print "No job could be launched (the nodes are too loaded)\n";
                next;
            };
                    
            my $msg = "start = ".$o_work_param->Get('start')." - end = ".$o_work_param->Get('end');
            $msg .= " - job Id = $job_id\n";
            $o_work_log->Trace($msg,'.','1');
            print "launched job, job ID = $job_id\n";
            $o_log->Trace("LAUNCHED JOB $job_id\n",'.','1');
       }
    }
    return \@result;
}

#
# Title: InitFileParameters
# Usage: &InitFileParameters($o_param)
# procedure: Modify $o_param, computing several file-related parameters
# precondition: start, output parameters must be set
# Return:   nothing returned if OK, die if errors
# Args:     $o_param
# Globals:  none
#

sub InitFileParameters {
    my $o_param = shift;
    
    # lock directory
    
    my $lock_directory;
    if ($o_param->IsDefined('lock_directory'))
    {
        $lock_directory = $o_param->Get('lock_directory');
    }
    else
    {
        $lock_directory = LOCK_DIR;
        $o_param->Set('lock_directory', $lock_directory);
    }
    unless (-d $lock_directory)
    {
        mkdir $lock_directory;
    }
    
    # output, log and lock files
    
    my $start  = $o_param->Get("start");
    my $plp_id = ($o_param->IsDefined('plp_id')) ? $o_param->Get('plp_id'):'';
    my $tag    = ($plp_id eq '') ? $start : $plp_id;
    
    my ($log_file,$lock_file,$output);

    my $output_prefix = $o_param->Get("output_prefix");
    my $flg_master = ($o_param->IsDefined('current_mode') and ($o_param->Get('current_mode') eq 'MASTER'));
    if ($flg_master==1)
    {
        $lock_file = $lock_directory . "/lock.$tag.master.lck";
    }
    else
    {
        $lock_file = $lock_directory . "/lock.$tag.lck";
   };
   if ($output_prefix ne "NULL")
   {
       if ($flg_master==1)
       {
           $log_file = $output_prefix . '.' . $tag . '.0.master.log';
           $output   = $output_prefix . '.' . $tag . '.0.master.txt';
       }
       else
       {
           $log_file = $output_prefix . '.' . $tag . '.0.log';
           $output   = $output_prefix . '.' . $tag . '.0.txt';
       }
   }
   else
   {
       $log_file = '/dev/null';
       $output   = '/dev/null';
   }       
   $o_param->Set('log_file', $log_file);
   $o_param->Set('output', $output);
   $o_param->Set('lock_file', $lock_file);

   # log level
   
   my $log_level;
   if ($o_param->IsDefined("log_level"))
   {
       $log_level = $o_param->Get("log_level");
   }
   else
   {
       $log_level = LOG_LEVEL_DEFAULT;
       $o_param->Set("log_level", $log_level);
   }
   
   # max file size
   
   my $max_file_size;
   if ($o_param->IsDefined('max_file_size'))
   {
       $max_file_size = $o_param->Get('max_file_size');
   }
   else
   {
       $max_file_size = MAXFILESIZE;
       $o_param->Set('max_file_size', $max_file_size);
   }
   
   # error directory
   
   my $error_directory;
   if ($o_param->IsDefined('error_directory'))
   {
       $error_directory = $o_param->Get('error_directory');
   }
   else
   {
       $error_directory = ERROR_DIR;
       $o_param->Set('error_directory', $error_directory);
   }
   
   unless (-d $error_directory)
   {
       mkdir $error_directory;
   }
   
   # stop files
   
   my $stop_all  = $lock_directory . '/paraloop.stp';
   my $stop_this = $lock_directory . "/paraloop.$plp_id.stp";
   $o_param->SetUnlessDefined('stop_all',  $stop_all);
   $o_param->SetUnlessDefined('stop_this', $stop_this);
   
};

#
# Title: InitFileParametersNcpus
# Usage: &InitFileParametersNcpus($o_param)
# procedure: Modify $o_param, computing several file-related parameters - 
#            Called by MainCommandLineNcpusThenExit
# precondition: output parameter must be set, but may be NOT start
# Return:   nothing returned if OK, die if errors
# Args:     $o_param
# Globals:  none
#

sub InitFileParametersNcpus {
    my $o_param = shift;
    
    # lock directory
    $o_param->SetUnlessDefined('lock_directory',&LOCK_DIR);
    my $lock_directory = $o_param->Get('lock_directory');
    my $standard_lock_directory = &LOCK_DIR;
    if ( -e $standard_lock_directory and !-l $standard_lock_directory)
    {
        if (rmdir $standard_lock_directory == 0)
        {
            die "PARALOOP ERROR !!! - Please Remove $standard_lock_directory, then try again\n";
        }
    }
    else
    {
        unlink $standard_lock_directory;
    }
    if (-d $lock_directory)
    {
        if (rmdir $lock_directory == 0)
        {
            die "PARALOOP ERROR !!! - Please Remove $lock_directory, then try again\n";
        }
    }
    else
    {
        mkdir $lock_directory;
        symlink($lock_directory, $standard_lock_directory);
    }
    
    # output, log and lock files
    my $output_prefix = $o_param->Get("output_prefix");
    my $out_dir       = dirname($output_prefix);
    
    if ($out_dir ne '' and (!-d $out_dir))
    {
        system($o_param->Get('mkdir')." -p $out_dir") and die "Cannot make $out_dir";
    };
    
    my ($log_file,$output);
    $log_file = $output_prefix . '.log';
    $output   = $output_prefix . '.txt';
    
    # If output_prefix eq NULL, we log and we output to /dev/null
    if ($output_prefix eq "NULL")
    {
        $o_param->Set('log_file', "/dev/null");
        $o_param->Set('output', "/dev/null");
    }
    
    #  else, we log and we output to previously computed file names
    else
    {
        $o_param->Set('log_file', $log_file);
        $o_param->Set('output', $output);
    };
    
    # log level
    
    my $log_level;
    if ($o_param->IsDefined("log_level"))
    {
        $log_level = $o_param->Get("log_level");
    }
    else
    {
        $log_level = LOG_LEVEL_DEFAULT;
        $o_param->Set("log_level", $log_level);
    }
    
    # error directory
    
    my $error_directory;
    if ($o_param->IsDefined('error_directory'))
    {
        $error_directory = $o_param->Get('error_directory');
        my $standard_error_directory = &ERROR_DIR;
        if ( -e $standard_error_directory and ! -l $standard_error_directory )
        {
            die "PARALOOP ERROR !!! - Please Remove $standard_error_directory, then try again\n";
        }
        else
        {
            unlink $standard_error_directory;
        }
        symlink($error_directory, $standard_error_directory);
    }
    else
    {
        $error_directory = &ERROR_DIR;
        $o_param->Set('error_directory', $error_directory);
    }
    
    unless (-d $error_directory)
    {
        mkdir $error_directory;
    }
    
    # stop files
    
    my $stop_all  = $lock_directory . '/paraloop.stp';
    $o_param->SetUnlessDefined('stop_all',  $stop_all);
    
};

#
# Title: ArgToLockFileName
# Usage: &ArgToLockFileName($name)
# function: Transform an argument to --check, --interrupt, etc. to a real file or directory name
#           If the argument does not exist, return PARALOOP_lock
#           If the argument exists as a file or a directory, it is returned as-is
#           If not, we try to resolve it to a lock file name
# Return:   The file name found, if found
# Args:     $o_param
# Globals:  none
#

sub ArgToLockFileName
{
    my $arg = shift;
    return 'PARALOOP_lock' if (!defined($arg) or $arg eq '');
    return $arg if (-e $arg);
    if ($arg =~ /(.+)[,:\/]([0-9\-]+)/)
    {
        my $file = $1.'/lock.';
        my $id = $2;
        if ($id =~ /-/)                        # It is a slave
        {
            return $file.$id.'.lck';
        }
        else
        {
            my $f = $file.$id.'.master.lck';    # Try a master
            return (-e $f) ? $f : $file.$id.'.lck';
        };
    }
    else
    {
        my $f = 'PARALOOP_lock/lock.'.$arg.'.master.lck';
        return (-e $f) ? $f : 'PARALOOP_lock/lock.'.$arg.'.lck';
    };
};

#
# Title: InitStealStatus
# Usage: &InitStealStatus($o_param,$o_log)
# procedure: Steal or not steal job to the others ?
#            That is the question - Resolved question, and answer logged
# Return:   nothing returned
# Args:     $o_param, $o_log
# Globals:  none
#

sub InitStealStatus {
    my $o_param = shift;
    my $o_log = shift;
    
    my $msg = "LOAD BALANCING ".($o_param->IsDefined('load_balancing_enabled') and $o_param->Get('load_balancing_enabled') != 0) ? " ENABLED" : " DISABLED";
    my $steal_threshold =  ($o_param->IsDefined('load_balancing_threshold')) ? $o_param->Get('load_balancing_threshold') : DEF_LOAD_BALANCING_THRESHOLD;
    $msg   .= ($msg =~ /ENABLED/) ? " - THRESHOLD = $steal_threshold\n" : "\n"; 
    $o_log->Trace($msg, '.','1');
};

#
# Title: InitBoundaries
# Usage: &InitBoundaries($o_param, $global_end)
# function: Modify $o_param, computing start, end from slices information, if necessary
#           Init step (defaults to 1) and current
# Return:   1 if OK
#           0 if the boundaries cannot be computed ($start, $end higher than $global_end
#           die if errors
# Args:     $o_param
#           $global_end (facultatif) - If specified, we force $end to be <= $global_end
# Globals:  none
#

sub InitBoundaries {
    my $o_param=shift;
    my $global_end = shift;
    my ($start, $end, $slice_offset, $slice_size, $slice_nb) = (-1, -1, 0, -1, -1);
    $start        = $o_param->Get("start") + 0 if ($o_param->IsDefined("start"));
    $end          = $o_param->Get("end") + 0   if ($o_param->IsDefined("end"));
    $slice_offset = $o_param->Get("slice_offset") + 0 if ($o_param->IsDefined("slice_offset"));
    $slice_size   = $o_param->Get("slice_size") + 0   if ($o_param->IsDefined("slice_size"));
    $slice_nb     = $o_param->Get("slice_nb") + 0     if ($o_param->IsDefined("slice_nb"));
    
    # compute start, end
    
    if ($start == -1 or $end == -1)
    {
        if ($o_param->IsDefined('step'))
        {
            die "ERROR - The switch --step must be used with the switches --start and --end\n";
        };
        if ($slice_size == 0 or $slice_nb == -1)
        {
            die "ERROR - Please specify --start,--end OR --slice_size,--slice_nb\n";
        };
        
        $start = $slice_offset + $slice_size * $slice_nb;
        $end   = $start + $slice_size - 1;  # The job is done in the interval [$start,$end]
        if (defined ($global_end) and $end > $global_end)
        {
            $end = $global_end;
            return 0 if ($end < $start);
        };
        $o_param->Set('start',   $start);
        $o_param->Set('end',     $end);
    }
    
    # (start,end) specified: delete any slice specification, just for coherence
    
    else
    {
        $o_param->Delete('slice_nb');
        $o_param->Delete('slice_size');
        $o_param->Delete('slice_offset');
    };

    # Give a default value to step, unless already defined
    
    $o_param->SetUnlessDefined('step',1);
    my $step = $o_param->Get('step');
    
    # Init current
    
    $o_param->Set('current', $start);

    die "step=$step must be >0 or <0"                             if ($step == 0);
    die "start=$start and end=$end must be > 0"                   if ($start < 0 or $end < 0);
    die "end=$end must be > start=$start, because step=$step > 0" if ($end < $start and $step > 0);
    die "end=$end must be < start=$start, because step=$step < 0" if ($end > $start and $step < 0);
    return 1;
};

#
# Title: CreatePlugin
# Usage: my $o_plugin = &CreatePlugin($o_param,$o_log)
# function: Using the parameter as a switch, create a 'program' object
# precondition: $o_param and $o_log must exist and be initialized
# Return:   the new object
# Args:     $o_param, $o_log
# Globals:  none
#

sub CreatePlugin
{
    my $o_param = shift;
    my $o_log    = shift;
    
    my $program = $o_param->Get('program');
    my $module  = 'PARALOOP::PLUGIN::' . $program;
    eval("use $module");
    if ($@ ne "")
    {    # Something wrong in eval (may be the object does not work)
        die $@;
    }
    my $o_plugin;
    my $to_eval = "PARALOOP::PLUGIN::$program->New" . '(-params=>$o_param, -log=>$o_log)';
    $o_plugin = eval "$to_eval";
    if ($@ ne "")
    {    # Something wrong in eval (missing object of bad parameter passed
        die $@;
    }
    return $o_plugin;
}

#
# Title: CreateScheduler
# Usage: my $o_sched = &CreateScheduler($o_param,$o_log)
# function: Using the parameters --local and Scheduler, create a Scheduler object
# precondition: $o_param and $o_log must exist and be initialized
# Return:   the new scheduler object
# Args:     $o_param,$o_log
# Globals:  none
#

sub CreateScheduler
{
    my $o_param = shift;
    my $o_log    = shift;
    my $module    = "";
    my $scheduler = "";

    if ($o_param->IsDefined('Scheduler'))
    {
        $scheduler = $o_param->Get('Scheduler');
    }
    else
    {
        die "The parameter PARALOOP_Scheduler should be defined\n";
    }

    if ($scheduler eq 'System' or $o_param->Get('local')!=0)
    {
        $module = 'PARALOOP::SCHEDULER::System';
    }
    else
    {
        $module = "PARALOOP::SCHEDULER::$scheduler";
    }

    eval("use $module");
    if ($@ ne "")
    {    # Something wrong in eval (may be the object does not work)
        die $@;
    }

    my $cmd;
    my $to_eval = "$module->New" . '(-params=>$o_param, -log=>$o_log)';
    $cmd = eval "$to_eval";
    if ($@ ne "")
    {    # Something wrong in eval (missing object of bad parameter passed
        die $@;
    }

    return $cmd;
}

#
# Title: StopJob
# Usage: &StopJob($o_param,$o_log)
# procedure: Log some message and exit.
# Return:   NEVER RETURN - CALL exit
# Args:     $o_param,$o_log
# Globals:  None
#

sub StopJob
{
    my $o_param = shift;
    my $o_log    = shift;
    my $lock_file = ($o_param->IsDefined('master_lock_file')) ? $o_param->Get('master_lock_file') : $o_param->Get('lock_file');
    $o_log->Trace("INTERRUPTION REQUEST RECEIVED - PLEASE RESUME WITH --resume $lock_file\n", '.', '1');
    exit(0);
}

#
# Title: PrintVar
# Usage: &PrintVar($o_param);
# Procedure: print (to the standard output) the environment
# Return:   none
# Args:     $o_param
# Globals:  %ENV
#

sub PrintVar
{
    my $o_param = shift;
    
    print "\nPARALOOP IMPORTANT ENVIRONMENT VARIABLES\n";
    print "==========================================\n";
    foreach my $v (sort keys %ENV)
    {
        if ($v =~ /^PLP_/ or $v =~ /PBS_/)
        {
            print "$v => $ENV{$v}\n";
        }
    }
    print "\nPARALOOP PARAMETERS\n";
    print "=======================\n";
    my @keys = $o_param->GetKeys('.*');
    foreach my $v (@keys)
    {
        print "$v => " . $o_param->Get($v) . "\n";
    }
}

#
# Title: IncrementName
# Usage: $new_name = &IncrementName($old_name)
# Function: search and increment the last number in the name passed by parameter
#           build and return a new name, incrementing this number
#           If no number found, do nothing
# Return:   the new name
# Args:     the name to increment
# globals:  none
#

sub IncrementName
{
    my $name = shift;
    if ($name =~ /(\d+)([^ \d\/]+)\Z/)
    {
        my $nb = $1;
        $nb++;
        $name =~ s/(\d+)([^ \d\/]+)\Z/$nb$2/;
    }
    return $name;
}

#
# Title: Time2Hour
# Usage: my $time = &Time2Hour($elapsed,$is_integer)
# Function: Converts an integer or floating number of seconds in format:
#           HH:mm:ss ($is_integer=1) or HH:mm:ss.xx ($is_integer=0)
# Return:   The time in the wanted format (string)
# Args:     $elapsed: a number of seconds. 
#           $is_integer = 1 for first format, 0 for second format
# globals:  none
#

sub Time2Hour {
    my $time = shift;
    my $is_integer = shift;
    my $rvl = '';
    if ($is_integer==1)
    {
        my $s = $time % 60;
        my $m = int($time / 60);
        my $h = $m / 60;
        $m = $m % 60;
        $rvl = sprintf("%03d:%02d:%02d",$h,$m,$s);
    }
    else
    {
        my $time = 100*$time;
        my $s = int ($time) % 6000;
        my $m = int ($time / 6000);
        my $h = $m / 60;
        $m = $m % 60;
        $rvl = sprintf("%03d:%02d:%05.2f",$h,$m,$s/100.0);
    }
    return $rvl;
};

#
# Title: DumpToFile
# Usage: &DumpToFile($o_param)
# Procedure: Search in $o_param the name of the lock file and dump to this file.
#            Dump also to the environment, this can be useful for external scripts
# Return:    nothing
# Args:      the ParamPArser object to dump
# globals:  none
#

sub DumpToFile
{
    my $o_param = shift;
    my $lock_file = $o_param->Get('lock_file');
    $o_param->Dump($lock_file);
    $o_param->Dump('ENV');
    return;
};

# Title: MonitorJobs
# Usage: &MonitorJobs(\@files)
# Function: Sleep some time (5 s), then read the lock files passed by parameter
#           Go on sleeping/reading while there is at least one running job
# Return:   0 if every job is terminated
#           1 if some job is only interrupted
#           2 if some job is Submitting
# Args:     The list of jobs (lock files) to monitor
#

sub MonitorJobs {
    my $ra_files = shift;
    #`echo HOUHOU >> TOTO`;
    my ($running,$interrupted,$submitting);
    sleep 30;
    #my $cpt=0;
    while(1)
    {
        $running=0;
        $interrupted=0;
        $submitting=0;
        foreach my $f (@$ra_files) 
        {
            my $o_param = New ParamParser($f);
            $o_param->SelectNameSpace(NAMESPACE);
            my $status = $o_param->Get('status');
            if ($status eq 'RUNNING')
            {
                &ResubmitIfNecessary($o_param);
                $running++;
            }
            $interrupted++ if ($status eq 'INTERRUPTED');
            $submitting++ if ($status =~ /SUBMIT/);
        };
        last if ($running==0);
        #`echo "REKOUKOU $running" >>TOTO`;
        # cf. comment above same instruction in the WaitParaloop sub
        
        #$cpt++;
        #`echo KOUKOU $cpt >> /tmp/TOTO`;
        #if (wait()==-1)
        #{
            #`echo $running $interrupted $submitting - sleeping >>/tmp/TOTO`;
        #    sleep WAIT_TIME;
        #}
        #else
        #{
            #`echo $running $interrupted $submitting - waiting >>/tmp/TOTO`;
        #    while (wait() != -1) {};
        #    sleep 5;
        #};
        sleep WAIT_TIME; 
    };
    return 0 if ($interrupted==0 and $submitting==0);
    return 2 if ($submitting>0);
    return 1;
}

# Title: WaitParaloop
# Usage: &WaitParaloop($directory)
# Procedure: Sleep some time (100 s), then read the lock files under $directory
#            Return only when all the jobs are interrupted, terminated or killed
# Return:   none
# Args:     $directory = The directory to check
# globals:  none
#

sub WaitParaloop
{
    my $lock = shift;
    my $int_status = 0;
    sleep 5;
    
    #my $cpt;
    if (-d $lock)
    {
        print "\nNow Waiting until all subprocesses are finished\n";
        do
        {
            $int_status = &IsInterrupted($lock);
            if ($int_status == 1)
            {
                print "THE JOBS ARE INTERRUPTED OR TERMINATED\n";
                print "You can resume the jobs with paraloop.pl --resume $lock --wait\n";
                return;
            };
            if ($int_status == 2)
            {
                return;
            };
            
            sleep WAIT_TIME;
            
            # If this process has a child, IT IS a paraloop running paraloop process.
            # This could happen with the System or Rsystem scheduler when using the --wait switch
            # In this case, the best thing to do is to wait for any child to terminate, THEN to check again the lock files
            # BUT when using other schedulers, like PBS, OR when using the --wait switch alone,
            # we should not have any child. In this case, just wait for some time before checking the lock files.
            
            #$cpt++;
            #if (wait()==-1)
            #{
            #    #`echo $cpt - sleeping >>/tmp/TOTO`;
            #    sleep WAIT_TIME;
            #}
            #else
            #{
                #`echo $cpt - waiting >>/tmp/TOTO`;
                # Wait until every child is finished.
            #    while (wait() != -1) {};
            #    # Sleep a little, for the file synchronization
            #    sleep 5;
            #};
        } while (1);
    }
    elsif (-f $lock)
    {
        print "\nNow Waiting until one subprocess is finished (see $lock)\n";
        do
        {
            $int_status = &IsInterrupted($lock);
            if ($int_status == 1)
            {
                print "JOB INTERRUPTED\n";
                print "You can resume the job with paraloop.pl --resume $lock --wait\n";
                return;
            };
            if ($int_status == 2)
            {
                return;
            };
            
            # Same comment as above
            
            if (wait()==-1)
            {
                sleep WAIT_TIME;
            }
            else
            {
                while (wait() != -1) {};
                sleep 5;
            };

        } while (1);
    }
    else
    {
        die "\nERROR - $lock does not exist";
    }
}

#
# Title: CleanParaloop
# Usage: &CleanParaloop($directory)
# Procedure: Read the lock files under $lock
#            Do not do anything if all jobs are not terminated
#            cat every output file to one big output file
#            remove every log file
#            remove every lock file and error file
#            remove error directory and lock directory
# Return:   none
# Args:     $directory = The directory to use for cleaning
# globals:  none
#

sub CleanParaloop
{
    my $lock = shift;
    die "\nERROR - $lock is NOT a directory\n" unless (-d $lock);

    my @params = ();
    my $output_prefix = '';
    
    # Check for the coherence of all the lock files
    my $cat="";
    foreach my $f (glob("$lock/lock.*.lck"))
    {
        my $o_param = New ParamParser($f);
        $o_param->SelectNameSpace(NAMESPACE);
        die "\nERROR - Some jobs ($f) are NOT terminated\n" unless ($o_param->Get('status') eq 'TERMINATED');
        if ($output_prefix eq '')
        {
            $output_prefix = $o_param->Get('output_prefix');
        }
        else
        {
            if ($output_prefix ne $o_param->Get('output_prefix'))
            {
                die "ERROR - File $f: value of output_prefix is " . $o_param->Get('output_prefix') . ", should be $output_prefix\n";
            };
        };
        if ($cat eq "")
        {
            $cat = $o_param->Get('cat');
        }            
        push(@params, $o_param);
    };
    
    if ($output_prefix ne "NULL")
    {
        # Remove the dummy and empty file xxx.txt
        my $dummy_txt = $output_prefix.'.txt';
        if (-z $dummy_txt)
        {
            unlink $dummy_txt;
        }
        # Remove the last log file
        unlink($output_prefix.'.log');
    }    
    
    # Clean the files referred to by the lock files                                                               
    print "\nNow cleaning the output and lock files\n";
    foreach my $o_param (@params)
    {
        if ($output_prefix ne "NULL")
        {
            my $output = $o_param->Get('output');
            my ($prefix,$serial) = ($output=~/(.+)\.(\d+)\.txt\Z/);
            if ($o_param->IsDefined('current_mode')==0 or $o_param->Get('current_mode') ne 'MASTER')
            {
                
                # Cat and remove all the files xxx.n.txt
                # Remove all the files xxx.n.log
                
                for (my $s=0; $s <= $serial; $s++)
                {
                    my $out = $prefix.'.'.$s.'.txt';
                    my $log = $prefix.'.'.$s.'.log';
                    if (! -e $out)
                    {
                        warn "WARNING - $out does not exist: did you make a restart ?\n";
                        next;
                    };
                    if (! -e $log)
                    {
                        warn "WARNING - $log does not exist: I just hope there is no problem ?\n";
                    };
                    
                    my $cmd = "$cat $out";
                    my ($status,$kld) = &RunExt(-cmd=>$cmd, -stdout=>">> $output_prefix");
                    #warn "KOUKOU $cmd $status $kld";
                    die "ERROR - Cannot execute the command $cmd >> $output_prefix - Paraloop not cleaned" if ($status != 0);
                    unlink($out,$log);
                }
                
            }
            
            # For a 'master', there is ONLY the log file to remove - NO output file produced
            else
            {
                my $log_file = $o_param->Get('log_file');
                unlink($log_file);
            }
        }
        
        # Remove the lock file
        my $lock_file = $o_param->Get('lock_file');
        unlink($lock_file, $lock_file.'.lock');
    }
    
    # Remove the lock directory
    
    rmdir $lock;
}


#
# Title: IsInterrupted
# Usage: $sts = &IsInterrupted()
# Function:  Return 1 if all the jobs are interrupted or terminated (if parameter is a directory)
#                     or if the job is interrupted (if parameter is a file)
#            Return 2 if all the jobs are terminated (if parameter is a directory)
#                     or if the job is terminated (if parameter is a file)
#            Return 0 in the other cases
# Return:    0/1/2
# Args:      The lock directory or a lock file
# globals:   none
#

sub IsInterrupted
{
    my $lock = shift;
    if (-d $lock)
    {
        my $all_interrupted = 1;
        my $all_terminated= 1;
        foreach my $f (glob("$lock/lock.*.lck"))
        {
            my $o_param = New ParamParser($f);
            $o_param->SelectNameSpace(NAMESPACE);
            next if ($o_param->IsDefined('current_mode') and $o_param->Get('current_mode') eq 'SLAVE');
            my $status = $o_param->Get('status');
            if ($status ne 'INTERRUPTED' and $status ne 'TERMINATED' and $status ne 'KILLED')
            {
                &ResubmitIfNecessary($o_param);
                $all_interrupted = 0;
            };
            if ($status ne 'TERMINATED')
            {
                $all_terminated = 0;
            }
        };
        return 2 if ($all_terminated==1);   # All jobs terminated
        return 1 if ($all_interrupted==1);  # All jobs interrupted OR terminated;
        return 0;                           # Some jobs still running
    }
    elsif (-f $lock)
    {
        my $o_param = New ParamParser($lock);
        $o_param->SelectNameSpace(NAMESPACE);
        my $status = $o_param->Get('status');
        return 2 if ($status eq 'TERMINATED');
        return 1 if ($status eq 'INTERRUPTED');
        return 0;
    } else
    {
        die "\nERROR - $lock DOES NOT EXIST ANY MORE";
    }
}

#
# Title: ResubmitIfNecessary
# Usage: &ResubmitIfNecessary($o_param)
# Function:  Create a Scheduler and a Loggger objects from the info found in $o_param
#            Call the IsLaunched function to know if the job is still alive
#            If not alive, resubmit the job
#
# Return:    0 ==> The job was alive
#            1 ==> The job was submitted again
# Args:      A ParamParser object
# globals:   none
#

sub ResubmitIfNecessary
{
    my $o_param  = shift;
    my $log_file = $o_param->Get('log_file');
    my $log_level= $o_param->Get('log_level');
    my $job_id;
    $o_param->Delete('subcount');    # Delete subcount, because we are not working on a subprocess
    my $o_log = Logger->New(">>$log_file", $log_level, 'NOCOMPRESSION NEWQUIET');
    my $o_sched  = &CreateScheduler($o_param, $o_log);
    if ($o_param->IsDefined('job_id'))
    {
        $job_id = $o_param->Get('job_id');
        #$o_log->Trace("koukou - ici $$ - is $job_id still alive ?\n","","0");
        if ($o_sched->IsLaunched($job_id)!=0)   # Job launched
        {
            return 0;
        };
    }
    else
    {
        $job_id = '0';
    };
    my $new_job_id = $o_sched->Submit();
    #$o_log->Trace("koukou, ici $$ - resubmitting job L3361 - new job = $new_job_id\n","","1");
    if ($new_job_id eq '0')
    {
        print "job $job_id could not be resubmitted (the nodes are too loaded) - please try again later\n";
    }
    else
    {
        print "resubmitted job, job ID was $job_id, now $new_job_id\n";
    };
    return 1;
};

#
# Title: ReadCfgFiles
# Usage: &ReadCfgFiles($object)
# Function:  Using the cfg parameter if defined, create a list of configuration files
#            If we are in verbose mode, test the existence of the file and print a message for each file found
#            Update the object passed by reference
#			 Correct some parameters if necessary
# Return:    The list of found configuration files
# Args:      a ParamParser object
# globals:   $0
#

sub ReadCfgFiles
{
    my $o_param = shift;
    my ($dirprg, $nameprg) = $0 =~ /(.+)\/(.+)/;

    # First create an array with the config files
    my @config_files = ();

    # Read the config /etc/paraloop.root.cfg and /etc/unix_cmds.cfg files
    if (exists $ENV{'PARALOOP'})
    {
        push(@config_files,$ENV{'PARALOOP'}.'/etc/unix_cmds.cfg',$ENV{'PARALOOP'}.'/etc/paraloop.root.cfg');
    }
    elsif (exists $ENV{'LIPMLIB'})
    {
        push(@config_files, $ENV{'LIPMLIB'} . '/../etc/unix_cmds.cfg', $ENV{'LIPMLIB'} . '/../etc/paraloop.root.cfg');
    }
    else
    {
        push(@config_files, "$dirprg/etc/unix_cmds.cfg", "$dirprg/../etc/unix_cmds.cfg", "$dirprg/etc/paraloop.root.cfg", "$dirprg/../etc/paraloop.root.cfg");
    }

    # Read the config files defined by the cfg switch
    if ($o_param->IsDefined('cfg'))
    {
        push(@config_files, split(',', $o_param->Get('cfg')));
    }

    # Read .paralooprc
    push(@config_files, $ENV{'HOME'} . '/.paralooprc');

    # Then, read the etc/paraloop.cfg file

    if (exists $ENV{'PARALOOP'})
    {
        push(@config_files,$ENV{'PARALOOP'}.'/etc/paraloop.cfg');
    }
    elsif (exists $ENV{'LIPMLIB'})
    {
        push(@config_files, $ENV{'LIPMLIB'} . '/../etc/paraloop.cfg');
    }
    else
    {
        push(@config_files, "$dirprg/etc/paraloop.cfg", "$dirprg/../etc/paraloop.cfg");
    }

    # Build the list of really found configuration files
    # If verbose, display which file exists and which file does not exist
    my @found_config_files;
    foreach my $f (@config_files)
    {
        if ($o_param->IsDefined('verbose'))
        {

            print "Configuration file $f";
            (-r $f) ? print " OK\n" : print " DOES NOT EXIST !\n";
        }
        push(@found_config_files, $f) if (-r $f);
    }

    # Update the ParamParser object from those files
    $o_param->SetBehaviour("ignore_space");
    $o_param->SelectNameSpace();    # Cancel the namespace before reading the conf files

    foreach my $f (@found_config_files)
    {
        $o_param->Update($f, 'A');
        #$o_param->Print();                # DEBUG ONLY
        #print("\n======\n";
    }
    $o_param->SelectNameSpace(NAMESPACE);

    # Corrrect some parameters

    if ($o_param->IsDefined('no_local_mode'))    # May be set by the administrator
    {
        if ($o_param->IsDefined('local') and $o_param->Get('local') ne '0')
        {
            print "SORRY, the local mode is disabled on this site - Terminating\n";
            exit(0);
        }
    }
    if ($o_param->IsDefined('local') and $o_param->Get('local') ne '0')
    {
        $o_param->Delete('queue');                 # local has the precedence
        $o_param->Delete('fair_time_limit');       # No fair_time in local mode
    }

    return \@found_config_files;
}

#
# Title: Usage,Switches,Parameters,Plugins
# Usage: &Usage(),Switches(),Parameters(),Plugins()
# Function: print some documentation...
# Return:   none
# Args:     $plugin_debug: print more things if 1
# globals:  none
#

sub Usage
{
    print "\nIf you are in a hurry, you could start with:\n\n";
    print "\tparaloop.pl --program=SomePlugin [--db=database] --input=infile [--output=outputfile] --start=0 --end=100 --load_balancing_enabled\n\n";
    print "\nTo read the complete online documentation:\n";
    print "\tparaloop.pl --switches\t\tThe command line switches\n";
    print "\tparaloop.pl --parameters\tThe parameters you may write into the configuration files\n";
    print "\tparaloop.pl --plugins\t\tThe available plugins\n";
    print "\tparaloop.pl --schedulers\tThe available schedulers\n";
}

sub Switches
{
    print "\nThe switches of general use\n";
    print "---------------------------\n";
    print "--cfg=f1,f2,f3   \tThe list of configuration files\n";
    print "--mail=mail_address\tSend a mail after completion of the LAST job\n";
    print "--verbose       \tsome verbosity added\n";
    print "--quiet         \tless messages\n";
    print "--show_parameters\tUsed with other switches and parameters, shows the value of each parameter or switch,\n";
    print "\t\t\twithout submitting the job\n";
    print "--switches      \tThis screen\n";
    print "--parameters    \tThe parameters you can put in configuration files\n";
    print "--plugins       \tThe list of installed plugins\n";
    print "--plugins --debug=D \tMay be useful to debug a new plugin\n";
    print "--plugin_debug  \tThis special mode is only useful for debugging the plugins\n";
    print "--help          \tA very quick help screen\n";
    print "\nSpecifying start - end\n";
    print "-----------------------------\n";
    print "--start=integer \tRecord number (1st = 0) to begin the loop with, including this number\n";
    print "--end=integer   \tRecord number to end the loop with, including this number\n";
    print "\nThe switches related to the chosen plugin\n";
    print "------------------------------------------\n";
    print "--program       \tThe name of a Perl object to run inside the loop\n";
    print "                \tExamples: Dummy, Shell (try perldoc -I \$PARALOOP PARALOOP::Shell)\n";
    print "--input         \tThe input file/database/directory name\n";
    print "--output=OUT    \tThe output prefix-filename. Real name will be:\n";
    print "                \tOUT.X.Y.txt (output files) and OUT.X.Y.log (log files)\n";
    print "--db=string     \tOne more paramter used by some plugins (Blastp, Blastpgp,..)\n";
    print "\nSpecifying cpus\n";
    print "----------------\n";
#    print "--monoprocessor\t\tRun the job ONLY in one processor\n";
    print "--ncpus=integer\t\tThe number of paraloop instances to start, generally less than the number of cpus\n";
    print "--start=integer\t\tThe global start record\n";
    print "--end=integer\t\tThe global end record\n";
    print "--interleaved\t\tIf specified, The records are distributed on a round-robin basis, NOT by slices\n";
    print "--wait\t\t\tWait until every child process returns\n";
    print "--clean\t\t\tWhen every child returns, clean everything, concatening every per-job output file to a single file\n";
    print "\nLoad balancing\n";
    print "---------------\n";
    print "load_balancing_enabled\tMake fastest jobs to steal some work to slowest jobs, balancing the load between the processors\n";
    print "\nMonitoring the job\n";
    print "------------------\n";
    print "--check=lock\t\tMonitor every job whose lock file lives in this directory\n";
    print "--check=lock/lock.x.lck\tMonitor only the job whose lock file is specified\n";
    print "--Check --html\t\tPrint result of check inside an html table\n";
    print "\nInterrupting the jobs\n";
    print "---------------------\n";
    print "--interrupt\t\tInterrupt every job whose lock file lives in the PARALOOP_lock directory\n";
    print "--interrupt=lock\tInterrupt every job whose lock file lives in this directory\n";
    print "--interrupt=lock/2\tInterrupt only the job whose directory and Id are specified\n";
    print "---------------------\n";
    print "--kill\t\tKill every job whose lock file lives in the PARALOOP_lock directory\n";
    print "--kill=lock\tKill every job whose lock file lives in this directory\n";
    print "--kill=lock/2\tKill only the job whose directory and Id are specified\n";
    print "--------------------------------\n";
    print "--interrupt means = The processing will be stopped AT THE END OF THE CURRENT ATOMIC JOB - this is a safe process\n";
    print "--kill means      = The processing will be stopped AT ONCE - this is a very bad trip !\n";
    print "--------------------------------\n";
    print "\nResuming after an interruption\n";
    print "--resume\t\tResume every job whose lock file lives in the PARALOOP_lock directory\n";
    print "--resume=lock\t\tResume every job whose lock file lives in this directory\n";
    print "--resume=lock/2\t\tResume Only the job whose directory and Id are specified\n";
    print "\nRestarting after an interruption\n";
    print "--------------------------------\n";
    print "--restart\t\tRestart (start again from start) every interrupted job whose lock file lives in PARALOOP_lock\n";
    print "--restart=lock\t\tdon't touch the terminated or running jobs\n";
    print "--restart=lock/5\tRestart only job whose directory and Id are specified, EVEN IF IT IS TERMINATED\n";
    print "\nThe switches related to PBS\n";
    print "---------------------------\n";
    print "--queue=string  \tThe queue the jobs will be submitted to\n";
    print "--account=string\tAccount name, passed to qsub\n";
    print "--local         \tExecute the jobs on the local machine (But see also no_local_mode)\n";
    print "\nThe switches related to the cluster (scheduler=Rsystem)\n";
    print "---------------------------------------------------------\n";
    print "--cluster=string  \tThe name of a cluster to use (Rsystem scheduler only).\n";
    print "                  \tIf you specify --cluster=toto, you must have a parameter Rsystem_cluster_toto describing the nodes\n";
    print "--local         \tExecute the jobs on the local machine (But see also no_local_mode)\n";
   
    print "\nSpecifying slices instead of Start/End parameters:\n"; 
    print "----------------------------------------------------\n";
    print "the input file is divided into equal-sized slices, each instance works on a specific slice.\n";
    print "The Start and End parameters are computed from the slice_offset, slice_size and slice_nb parameters.\n";
    print "The step parameter is set to 1\n\n";
    print "--slice_offset=integer\tThe offset to take into account for computing the start\n";
    print "--slice_size=integer\tThe size of each slice (number of records)\n";
    print "--slice_nb=interger\tThe slice number to use\n";
    print "\n";
    exit(0);
}

sub Parameters
{
    my $plugin_debug = shift;
    my ($dirprg, $nameprg) = $0 =~ /(.+)\/(.+)/;

    print "\nThe PARALOOP parameters\n";
    print "------------------------\n\n";
    print "The parameter name is ALWAYS preceded by PARALOOP_\n";
    print "\nNearly all switches can be specified as a parameter, in a configuration file\n";
    print "Example: --queue workq can be specified as PARALOOP_queue=workq\n";
    print "\nThe parameters files may live in several standard locations, and you may specify additional files\n";
    print "\nEx.: paraloop.pl ... --cfg f1.cfg,f2.cfg,f3.cfg gives following priorities (highest to lowest):\n";
    print "\t1. \tThe switch on the command line\n";
    print "\t2. \t\$PARALOOP/etc/paraloop.root.cfg\n";
    print "\t3. \tf1.cfg\n";
    print "\t4. \tf2.cfg\n";
    print "\t5. \tf3.cfg\n";
    print "\t6. \t\$HOME/.paralooprc\n";
    print "\t7. \t\$PARALOOP/paraloop.cfg\n";
    print "\nSUBSTITUTED STRINGS\n";
    print "-------------------\n";
    print "Some strings are substituted when paraloop is started:\n";
    print "%h\tThe hour part of time (11 for 11:30:05)\n";
    print "%m\tThe minute part of time (30 for 11:30:05)\n";
    print "%s\tThe seconds part of time (05 for 11:30:05)\n";
    print "%Y\tThe year part of date (05 for Sept 6th 2005)\n";
    print "%M\tThe month part of date (09 for Sept 6th 2005)\n";
    print "%D\tThe day part of date (06 for Sept 6th 2005)\n";
    print "%p\tThe ncpus parameter\n";
    print "%l\tThe local_ncpus parameter\n";
    print "%v\tThe slave_ncpus parameter\n";
    print "%t\tThe master_ncpus parameter\n";
    print "%i\ttThe paraloop job number (1..10 for 10 ncpus)\n"; 

    print "\nGENERAL PARAMETERS:\n";
    print "-------------------\n";
    print "PARALOOP_log_level0\t-> do not log anything\n";
    print "                    01\t-> log more things\n";
    print "                    012\t-> log a lot of things\n\n";
    print "PARALOOP_no_local_mode\t\tWhen specified, the switch --local cannot be used.\n";
    print "PARALOOP_fair_time_limit\tWhen the time elapsed since the beginning of the loop reachs this value (in s),\n";
    print
      "\t\t\t\tthe loop is interrupted and another qsub is submitted. However, this NOT managed as a timeout\n";
    print "\t\t\t\tso that your running job should not be inexpectedly interrupted\n";
    print "\t\t\t\tThus, it is fair specifying this parameter\n";
    print "PARALOOP_max_file_size\t\tThe output file is NEVER bigger thant this size in Mb (default 1000)\n";
    print "PARALOOP_error_directory\tThe directory for the storage of output/error files\n";
    print "PARALOOP_lock_directory\t\tThe lock directory\n";
    print "PARALOOP_tmp\t\t\tThe temporary directory (defaults to /tmp)\n";
    print "\n";
    print "PARALOOP_load_balancing_enabled=1\tWhen set, this parameter permits fastest jobs to steal some work to slowest jobs,\n";
    print "\t\t\t\t\tthus balancing the load between the processors (default=disabled)\n";
    print "PARALOOP_load_balancing_threshold\tIf a job has less than load_balancing_threshold records to compute, his work will not be stolen.\n";
    print "\t\t\t\t\tDefault = 1\n";
    print "\n";
    print "PARALOOP_mode\tAUTONOMOUS\t(default)Every job is launched from the command line\n";
    print "\t\tMASTER/SLAVE\tOnly master jobs are launched from the command line, slave jobs are launched by the master modes\n";
    print "PARALOOP_slave_Scheduler\tThe scheduler used to launch the slave jobs\n";
    print "PARALOOP_slave_ncpus\t\tThe number of instances launched by each master job\n";
    print "PARALOOP_master_ncpus\t\tThe number of instances launched by the command line\n";
    print "\t\t\t\tWARNING - You should specify PARALOOP_slave_ncpus OR PARALOOP_master_ncpus - NOT both\n";
    print "PARALOOP_local_ncpus\t\tThe default value for ncpus when the --local switch is specified\n\n";
    print "PARALOOP_preproc_script\t\tExternal script, will be executed at begin or resume of each job\n";
    print "PARALOOP_postproc_script\tExternal script, will be executed at end or interruption of each job\n";
    print "PARALOOP_preatom_script\t\tExternal script, will be executed before each atomic job\n";
    print "PARALOOP_postatom_script\tExternal script, will be executed after each atomic job\n";
    print "\nPARAMETERS RELATED TO THE AVAILABLE PLUGINS AND SCHEDULERS:\n";
    print "-----------------------------------------------------------\n";
    my $lib = '';
    foreach my $d (@INC)
    {
        $lib = $d . '/' . 'PARALOOP/PLUGIN';
        if (-d $lib)
        {
            foreach my $f (glob("$lib/*.pm"))
            {    # For file
                chomp($f);
                if ($f =~ /$lib\/(.+)\.pm\Z/)
                {
                    my $module = 'PARALOOP::PLUGIN::' . $1;
                    eval("use $module");
                    print $@ if ($plugin_debug != 0);
                    next if ($@ ne "");
                    my $msg;
                    my $plugin = $module . '::Parameters' . '()';
                    $msg = eval($plugin);
                    print $@ if ($plugin_debug != 0);
                    next if ($@ ne "");
                    print "$msg\n";
                }
            }
        }

        $lib = $d . '/' . 'PARALOOP/SCHEDULER';
        if (-d $lib)
        {
            foreach my $f (glob("$lib/*.pm"))
            {    # For file
                chomp($f);
                if ($f =~ /$lib\/(.+)\.pm\Z/)
                {
                    my $module = 'PARALOOP::SCHEDULER::' . $1;
                    eval("use $module");
                    print $@ if ($plugin_debug != 0);
                    next if ($@ ne "");
                    my $msg;
                    my $plugin = $module . '::Parameters' . '()';
                    $msg = eval($plugin);
                    print $@ if ($plugin_debug != 0);
                    next if ($@ ne "");
                    print "$msg\n";
                }
            }
        }
    }
    exit(0);
}

sub Plugins
{
    my $plugin_debug = shift;
    print "\nCurrently installed Plugins\n\n";
    foreach my $d (@INC)
    {
        my $lib = $d . '/' . 'PARALOOP/PLUGIN';
        if (-d $lib)
        {
            foreach my $f (glob("$lib/*.pm"))
            {    # For file
                chomp($f);
                if ($f =~ /$lib\/(.+)\.pm\Z/)
                {
                    my $module = 'PARALOOP::PLUGIN::' . $1;
                    eval("use $module");
                    print $@ if ($plugin_debug != 0);
                    next if ($@ ne "");
                    my $msg;
                    my $plugin = $module . '::WhatParaloopPlugin' . '()';
                    $msg = eval($plugin);
                    print $@ if ($plugin_debug != 0);
                    next if ($@ ne "");
                    print "$msg\n";
                }
            }
        }
    }
    exit(0);
}

sub Schedulers
{
    my $plugin_debug = shift;
    print "\nCurrently installed Schedulers\n\n";
    foreach my $d (@INC)
    {
        my $lib = $d . '/' . 'PARALOOP/SCHEDULER/';
        if (-d $lib)
        {
            foreach my $f (glob("$lib/*.pm"))
            {    # For file
                chomp($f);
                if ($f =~ /$lib\/(.+)\.pm\Z/)
                {
                    my $module = 'PARALOOP::SCHEDULER::' . $1;
                    eval("use $module");
                    print $@ if ($plugin_debug != 0);
                    next if ($@ ne "");
                    my $msg;
                    my $plugin = $module . '::WhatParaloopScheduler' . '()';
                    $msg = eval($plugin);
                    print $@ if ($plugin_debug != 0);
                    next if ($@ ne "");
                    print "$msg\n";
                }
            }
        }
    }
    exit(0);
}


