
# $Id: Bioperl.pm 187 2007-03-16 14:35:00Z manu $

package PARALOOP::PLUGIN::Bioperl;

=pod

=head1 Copyright Notice

 Copyright I.N.R.A. - C.N.R.S. 

 emmanuel.courcelle@toulouse.inra.fr
 jerome.gouzy@toulouse.inra.fr

 This software is a computer program whose purpose is to help you run 
 your scripts on parallel computers.

 This software is governed by the CeCILL  license under French law and
 abiding by the rules of distribution of free software.  You can  use, 
 modify and/ or redistribute the software under the terms of the CeCILL
 license as circulated by CEA, CNRS and INRIA at the following URL
 http://www.cecill.info" 

 As a counterpart to the access to the source code and  rights to copy,
 modify and redistribute granted by the license, users are provided only
 with a limited warranty  and the software's author,  the holder of the
 economic rights,  and the successive licensors  have only  limited
 liability. 

 In this respect, the user's attention is drawn to the risks associated
 with loading,  using,  modifying and/or developing or reproducing the
 software by the user in light of its specific status of free software,
 that may mean  that it is complicated to manipulate,  and  that  also
 therefore means  that it is reserved for developers  and  experienced
 professionals having in-depth computer knowledge. Users are therefore
 encouraged to load and test the software's suitability as regards their
 requirements in conditions enabling the security of their systems and/or 
 data to be ensured and,  more generally, to use and operate it in the 
 same conditions as regards security. 

 The fact that you are presently reading this means that you have had
 knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

PARALOOP::PLUGIN::Bioperl - Executing something with files read using bioperl

=head1 Derives from...

PARALOOP::PLUGIN::BpInput

=head1 DESCRIPTION

This plugin reads the next record from the input file, whose format must be recognized by bioperl,
calls an external program or script, then writes the output of this program

The external script is passed as: script_name  param1 param2 ... input output, ie the last two parameters are the input and the output file,
the other parameters are declared by PARAMETER_Bioperl_params

Please note that when the script is ran, its current directory is a temporary directory, located in the same directory 
as the input file. You may securely create files here.

=head1 CONFIGURATION FILE PARAMETERS

 PARALOOP_Bioperl_path
     path to the external script, ran at each iteration. No default
 PARALOOP_Bioperl_params
     parameters passed to this script. Default 0
 PARALOOP_Bioperl_input_format
     Format for the input file (default to fasta)
     WARNING - This is the format to the file which is read by THE EXTERNAL script
               BUT the input file from paraloop may be of any format automatically recognized by bioperl

=head1 A very short example

The following is a very stupid script, which makes nothing which could be executed through this plugin:

 #!/bin/sh

 sleep $1;
 cat $2 > $3
 exit 0;

If the script is called stupid.sh and lives in the current paraloop directory, there should be two parameters defined:

 PARALOOP_Bioperl_path  = stupid.sh
 PARALOOP_Bioperl_params = 3 

The script is called by paraloop as:

 stupid.sh 3 input output

=head1 PUBLIC METHODS

 WhatParaloopPlugin 
     my $msg=$obj->WhatParaloopPlugin()

 Parameters
     my $msg=$obj->Parameters()

 New
     my $obj = PARALOOP::Dummy->New(-params => $params, -log => $o_log);

 Exec
     my ($incr_size,$error,$killed) = $obj->Exec();

 SetOutputFile
     $obj->SetOutputFile('filename');

=head1 METHODS AND SUBROUTINES - DETAILED DESCRIPTION

=cut

use strict;
use warnings;
use Cwd;
use File::Basename;
use Carp;

# The NAMESPACE constant are also set in paraloop
use constant NAMESPACE => 'PARALOOP_';

use Logger;
use PARALOOP::PLUGIN::BpInput;
use Runner;

our @ISA = qw( PARALOOP::PLUGIN::BpInput );

use constant BIOPERL_PARAMS_DEFAULT => "0";
use constant BIOPERL_FORMAT_DEFAULT => "fasta";

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 187 $ =~ /\d+/g); $r[0]};
}

=pod

=head2 Function WhatParaloopPlugin

 Usage:    my $msg = PARALOOP::Dummy::WhatParaloopPlugin()
        or my $msg = $obj->WhatParaloopPlugin()
 Function: Return some information (2 lines max) on the plugin
           An object is NOT considered as a plugin if this function is not implemented
 Arg:      none ($self is ignored)
 Return:   A message describing the plugin (1 or 2 lines)

=cut

sub WhatParaloopPlugin
{
    my $rvl = "";
    $rvl =
      "PARALOOP::PLUGIN::Bioperl v$PARALOOP::PLUGIN::Bioperl::VERSION - Using Runner v$Runner::VERSION - Derivating from PARALOOP::PLUGIN::BpInput v$PARALOOP::PLUGIN::BpInput::VERSION\n";
    $rvl .= "This plugin calls an external program for each record read from the input, using bioperl.\n";
    return $rvl;
}

=pod

=head2 Function Parameters

 Usage:    my $msg = PARALOOP::Dummy::Parameters()
        or my $msg = $obj->Parameters()
 Function: Return the list of allowed parameters for this plugin
           If your plugin defines some parameters, it is good pratice adding this function,
	   to display those parameters when the --parameters switch of paraloop.pl is used.
	   However, this is not mandatory
 Arg:      none ($self is ignored)
 Return:   A message describing the plugin (1 or 2 lines)

=cut

sub Parameters
{
    my $rvl = "";
    $rvl = "\nParameters for the Bioperl plugin\n=================\n";
    $rvl .= "PARALOOP_Bioperl_path\t\tpath to the external program (NO DEFAULT)\n";
    $rvl .= "PARALOOP_Bioperl_params\t\tParameters for this program (default \"0\")\n";
    $rvl .= "PARALOOP_Bioperl_input_format\tDefault format for the input file (default fasta)\n";
    return $rvl;
}

=head2 procedure _Init

 Usage:  CALLED BY UNIVERSAL::New
         my obj = PARALOOP::Dummy->New(-params => $params, -log => $o_log);
 Access: private (use New instead)
 procedure: Init the object
 Return:   nothing
 Args:     -params = an object of class ParamParser (already initialized)
           -log    = an object of class Logger (already initialized)
 globals:  none

=cut

sub _Init
{
    my $self    = shift;
    my %args    = @_;
    my $o_param = $args{'-params'};
    my $o_log   = $args{'-log'};

    my $current_dir = cwd;

    my ($inputfile, $outputfile);

    if ($o_param->IsDefined('input'))
    {    # check if inputfile is correctly defined
        $inputfile = $o_param->Get('input');
        if ($inputfile !~ /^\//)
        {    # Relative name
            $inputfile = $current_dir . '/' . $inputfile;
        }
        -r $inputfile or croak("ERROR - $inputfile cannot be opened in read mode");
    }
    else
    {
        croak("ERROR - Bioperl plugin: the input file is not defined");
        $o_param->Usage();
    }

    if ($o_param->IsDefined('output'))
    {        # check if outputfile is correctly defined
        $outputfile = $o_param->Get('output');
        if ($outputfile !~ /^\//)
        {    # Relative name
            $outputfile = $current_dir . '/' . $outputfile;
        }

        open(TEST, ">>$outputfile") or croak("ERROR - $outputfile cannot be opened in write mode");
        close(TEST);
    }
    else
    {
        croak("ERROR - Bioperl plugin: the output file is not defined");
        &$o_param->Usage();
    }

    # Build the __bin variable

    my $bin = "";
    if ($o_param->IsDefined('Bioperl_path'))
    {
        $bin = $o_param->Get('Bioperl_path');
        if ($bin !~ /^\//)
        {    # Relative name
            $bin = $current_dir . '/' . $bin;
        }

        system("which $bin  >/dev/null 2>&1")==0 or croak "ERROR - $bin cannot be found or executed";

    }
    else
    {
        croak("ERROR - Bioperl plugin: The Bioperl_path parameter should be defined");
        exit(1);
    }

    # Build the $parameters and $format variables

    my $parameters = $o_param->IsDefined('Bioperl_params') ? $o_param->Get('Bioperl_params') : BIOPERL_PARAMS_DEFAULT;
    my $format =
      $o_param->IsDefined('Bioperl_input_format') ? $o_param->Get('Bioperl_input_format') : BIOPERL_FORMAT_DEFAULT;

    $self->SUPER::_Init(-inputfile => $inputfile, -params => $o_param, -log => $o_log);    # Init the superclass

    # Compute job_id
    my $job_id = $self->_JobId($o_param);

    $self->{'__params'}     = $o_param;
    $self->{'__outputfile'}  = $outputfile;
    $self->{'__job_id'}      = $job_id;
    $self->{'__bin'}         = $bin;
    $self->{'__parameters'}  = $parameters;
    $self->{'__format'}      = $format;
    $self->{'__current_dir'} = $current_dir;

    # create a temporary directory
    my $tmp    = $self->{'__tmp'};
    my $stamp  = $o_log->CalcStamp();
    my $tmpdir = $tmp . '/' . &basename($inputfile) . '_' . $job_id . '_' . $stamp;
    mkdir($tmpdir) or croak "ERROR - Cannot create directory $tmpdir: $!";
    $self->{'__tmpdir'} = $tmpdir;

}

=pod

=head2 Procedure SetOutputFile

 Usage:     &$opt->SetOutputFile('filename');
 Procedure: change the name of the output file
 Arg :      the output file name

=cut

sub SetOutputFile
{
    my $self        = shift;
    my $outputfile  = shift;
    my $current_dir = $self->{'__current_dir'};
    if ($outputfile !~ /^\//)
    {    # Relative name
        $outputfile = $current_dir . '/' . $outputfile;
    }

    $self->{'__outputfile'} = $outputfile;
}

=pod

=head2 procedure DESTROY

 Usage:     The destructor of the object
            undef($obj);
 Access:    public
 procedure: Destroyes the resources allocated by _Init
 Return:    nothing
 Args:      nothing
 globals:   none

=cut

sub DESTROY
{
    my $self   = shift;
    my $o_param=$self->{'__params'};
    my $rm = $o_param->Get('rm');

    my $tmpdir = $self->{'__tmpdir'};
    `$rm -rf $tmpdir`;
}

=pod

=head2 Function Exec

 Usage:    my ($incr_size, $error,$killed) = &$opt->Exec($timeout);
 Function: Execute the external script
 Arg:      $timeout   = The number of seconds after which a timeout is sent
 Return:   $incr_size = The number of bytes added to the output file
           $error     = An error code. 
	                If != 0, paraloop WILL BE TERMINATED - Generally marks the end of file
           $killed    = A kill code
		        0: everything OK 
                       -1: killed by timeout
                       -2: killed by TERM signal (kill, qdel, ...)

=cut

sub Exec
{
    my $self        = shift;
    my $timeout     = shift;
    my $bin         = $self->{'__bin'};
    my $parameters  = $self->{'__parameters'};
    my $format      = $self->{'__format'};
    my $input_file  = $self->{'__inputfile'};
    my $output_file = $self->{'__outputfile'};
    my $o_log       = $self->{'__log'};
    my $job_id      = $self->{'__job_id'};
    my $tmpdir      = $self->{'__tmpdir'};
    my $rewrite_input = $self->{'__rewrite_input'};
    
    my $o_param=$self->{'__params'};
    my $wc = $o_param->Get('wc');
    my $cat = $o_param->Get('cat');

    my $stamp = $o_log->CalcStamp();
    my $itmp  = $tmpdir . '/input';
    my $otmp  = $tmpdir . '/output';

    my ($nb, $nl) = (0, 0);

    my $seq_nb = $self->Tell();
    my $seq    = $self->NextRecord();
    if (!defined($seq))
    {
        my $msg = "NO MORE SEQUENCES TO READ FROM $input_file\n";
        $o_log->Trace("$msg", $stamp, '1');
        return (0, -1, 0);
    }

    # Remember where we are and change to temporary directory

    my $savdir = cwd;    # Go to temporary directory
    chdir($tmpdir) or die "Cannot change to temporary directory";

    # Extract next sequence and save it to $itmp

    my $id = $seq->id();
    my $itmpfh = Bio::SeqIO->new(-file => ">$itmp", -format => $format) or die "Cannot open $itmp for writing";
    $itmpfh->write_seq($seq);
    undef($itmpfh);

    my $cmd = $bin;
    $cmd .= " $parameters $itmp $otmp ";
    $o_log->Trace("$cmd\n", '.', 2);

    $o_log->StartTimer('Exec_timer');
    my ($sts, $killed) = RunExt(-cmd => $cmd);

    if ($sts != 0)
    {
        $o_log->Trace("BIOPERL STATUS $sts - (REQUEST $seq_nb)\n", 'Exec_timer', '1');
    }
    elsif (!-r $otmp or -z $otmp)
    {
        $o_log->Trace("BIOPERL NO OUTPUT - (REQUEST $seq_nb)\n", 'Exec_timer', '1');
    }
    else
    {

        # Append the output to the main output file

        chdir($savdir);    # Back to work directory

        my $rvl_wc = `$wc -c -l $otmp`;
        chomp($rvl_wc);
        ($nl, $nb) = ($rvl_wc =~ /\A\s*(\d+)\s+(\d+) /);
        $o_log->Trace("BIOPERL OK - (REQUEST $seq_nb - $nl OUTPUT LINES)\n", 'Exec_timer', '1');
        my $out = new IO::File(">>$output_file") or die "Cannot open $output_file in append mode";
        my $in  = new IO::File($otmp)            or die "Cannot open $otmp for reading";
        while (my $line = <$in>)
        {
            print $out $line;
        }
    }

    # If status != 0 and if __rewrite_input on, write the input data to some file
    if ($sts != 0 and $rewrite_input != 0)
    {
        my $new_input = $output_file.'.in';
        `$cat $itmp >> $new_input`;
        my $msg = "Input data copied back to " . basename($new_input);
        $o_log->Trace("$msg\n", '.', 'R');
    };
        
    unlink($itmp, $otmp);
    return ($nb, 0, $killed);
}

