
# $Id: Dummy.pm 188 2007-04-13 10:38:51Z manu $

package PARALOOP::PLUGIN::Dummy;

=pod

=head1 Copyright Notice

 Copyright I.N.R.A. - C.N.R.S. 

 emmanuel.courcelle@toulouse.inra.fr
 jerome.gouzy@toulouse.inra.fr

 This software is a computer program whose purpose is to help you run 
 your scripts on parallel computers.

 This software is governed by the CeCILL  license under French law and
 abiding by the rules of distribution of free software.  You can  use, 
 modify and/ or redistribute the software under the terms of the CeCILL
 license as circulated by CEA, CNRS and INRIA at the following URL
 http://www.cecill.info" 

 As a counterpart to the access to the source code and  rights to copy,
 modify and redistribute granted by the license, users are provided only
 with a limited warranty  and the software's author,  the holder of the
 economic rights,  and the successive licensors  have only  limited
 liability. 

 In this respect, the user's attention is drawn to the risks associated
 with loading,  using,  modifying and/or developing or reproducing the
 software by the user in light of its specific status of free software,
 that may mean  that it is complicated to manipulate,  and  that  also
 therefore means  that it is reserved for developers  and  experienced
 professionals having in-depth computer knowledge. Users are therefore
 encouraged to load and test the software's suitability as regards their
 requirements in conditions enabling the security of their systems and/or 
 data to be ensured and,  more generally, to use and operate it in the 
 same conditions as regards security. 

 The fact that you are presently reading this means that you have had
 knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

PARALOOP::PLUGIN::Dummy - Template object

=head1 Derives from...

PARALOOP::PLUGIN::BpInput

=head1 DESCRIPTION

This plugin reads the next record from the input file, copies this record 
to the output file, then sleeps a few seconds: this is dummy, but this is just a demo, and a template 
to help you write your plugins.

=head1 CONFIGURATION FILE PARAMETERS

 PARALOOP_Dummy_sleep_path
     path to the sleep program (default sleep)
 PARALOOP_Dummy_sleep_time  
     time to sleep (default 6)

=head1 PUBLIC METHODS

 WhatParaloopPlugin
     my $msg=$obj->WhatParaloopPlugin()

 Parameters
     my $msg=$obj->Parameters()

 New 
     my $obj = PARALOOP::Dummy->New(-params => $params, -log => $o_log);

 Exec
     my ($incr_size,$error,$killed) = $obj->Exec();

 SetOutputFile
     $obj->SetOutputFile('filename');


=head1 METHODS INHERITED FROM BpInput

 NextRecord
     my $seq = $obj->NextRecord();
     
 SkipRecords
     $obj->SkipRecords($n);
     
 Tell()
     my $position = $obj->Tell();
     
 TellLength
     my $length = $obj->TellLength();
     
 SetOutputFile 
     $obj->SetOutputFile('filename');

=head1 METHODS AND SUBROUTINES - DETAILED DESCRIPTION

=cut

use strict;
use warnings;
use Cwd;
use File::Basename;
use Carp;

use Logger;
use PARALOOP::PLUGIN::BpInput;
use Runner;

our @ISA = qw( PARALOOP::PLUGIN::BpInput );

use constant SLEEP_DEFAULT      => 'sleep';    # sleep is already in the path
use constant SLEEP_TIME_DEFAULT => '6';        # sleep 6 seconds

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 188 $ =~ /\d+/g); $r[0]};
}

=pod

=head2 Function WhatParaloopPlugin

 Usage:    my $msg = PARALOOP::Dummy::WhatParaloopPlugin()
        or my $msg = $obj->WhatParaloopPlugin()
 Function: Return some information (2 lines max) on the plugin
           An object is NOT considered as a plugin if this function is not implemented
 Arg:      none ($self is ignored)
 Return:   A message describing the plugin (1 or 2 lines)

=cut

sub WhatParaloopPlugin
{
    my $rvl = "";
    $rvl =
      "PARALOOP::PLUGIN::Dummy v$PARALOOP::PLUGIN::Dummy::VERSION - Using Runner v$Runner::VERSION - Derivating from PARALOOP::PLUGIN::BpInput v$PARALOOP::PLUGIN::BpInput::VERSION\n";
    $rvl .= "This dummy plugin does not do anything - Consider it as a template for writing really useful plugins\n";
    return $rvl;
}

=pod

=head2 Function Parameters

 Usage:    my $msg = PARALOOP::Dummy::Parameters()
        or my $msg = $obj->Parameters()
 Function: Return the list of allowed parameters for this plugin
           If your plugin defines some parameters, it is good pratice adding this function,
           to display those parameters when the --parameters switch of paraloop.pl is used.
           However, this is not mandatory
 Arg:      none ($self is ignored)
 Return:   A message describing the plugin (1 or 2 lines)

=cut

sub Parameters
{
    my $rvl = "";
    $rvl = "\nParameters for the Dummy plugin\n=================\n";
    $rvl .= "PARALOOP_Dummy_sleep_path\tpath to the sleep program (default sleep)\n";
    $rvl .= "PARALOOP_Dummy_sleep_time\ttime to sleep (default 6)\n";
    return $rvl;
}

=pod

=head2 procedure _Init

 Usage:     CALLED BY UNIVERSAL::New
            my obj = PARALOOP::Dummy->New(-params => $params, -log => $o_log);
 Access:    private (use New instead)
 procedure: Init the object
 Return:    nothing
 Args:      -params = an object of class ParamParser (already initialized)
            -log    = an object of class Log (already initialized)
 globals:   none
 
 The plugin is created by paraloop at initialization, BEFORE calling qsub; 
 this lets you the opportunity to check everything carefully,
 so you can hope not to have any error message when your job will be submitted later
 If a resource is created (file, temporary directory, etc) it must be deleted in DESTROY

=cut

sub _Init
{
    my $self    = shift;
    my %args    = @_;
    my $o_param = $args{'-params'};
    my $o_log   = $args{'-log'};

    # Check inputfile, outputfile
    my ($inputfile, $outputfile);

    if ($o_param->IsDefined('input'))
    {    # check if inputfile is correctly defined
        $inputfile = $o_param->Get('input');
        -r $inputfile or croak("ERROR - $inputfile cannot be opened in read mode");
    }
    else
    {
        croak("ERROR - the input file is not defined");
        $o_param->Usage();
    }

    if ($o_param->IsDefined('output'))
    {    # check if outputfile is correctly defined
        $outputfile = $o_param->Get('output');
        open(TEST, ">>$outputfile") or croak("ERROR - $outputfile cannot be opened in write mode");
        close(TEST);
    }
    else
    {
        croak("ERROR - the output file is not defined");
        &$o_param->Usage();
    }

    my $bin = $o_param->IsDefined('Dummy_sleep_path') ? $o_param->Get('Dummy_sleep_path') : SLEEP_DEFAULT;

    system("which $bin  >/dev/null 2>&1")==0 or croak "ERROR - $bin cannot be found or executed";

    my $sleep_time = $o_param->IsDefined('Dummy_sleep_time') ? $o_param->Get('Dummy_sleep_time') : SLEEP_TIME_DEFAULT;

    $self->{'__params'}     = $o_param;
    $self->{'__outputfile'} = $outputfile;
    $self->{'__bin'}        = $bin;
    $self->{'__sleep_time'} = $sleep_time;

    # Initialize the superclass
    $self->SUPER::_Init(-params => $o_param, -inputfile => $inputfile, -log => $o_log);    # Init the superclass

    # Compute job_id
    my $job_id = $self->_JobId($o_param);
    $self->{'__job_id'} = $job_id;
    
    # create a temporary directory
    my $tmp    = $self->{'__tmp'};
    my $stamp  = $o_log->CalcStamp();
    my $tmpdir = $tmp . '/' . &basename($inputfile) . '_' . $job_id . '_' . $stamp;
    mkdir($tmpdir) or croak "ERROR - Cannot create directory $tmpdir: $!";
    $self->{'__tmpdir'} = $tmpdir;
}

=pod

=head2 procedure DESTROY

 Usage:     The destructor of the object
            undef($obj);
 Access:    public
 procedure: Destroyes the resources allocated by _Init
 Return:    nothing
 Args:      nothing
 globals:   none

=cut

sub DESTROY
{
    my $self   = shift;

    my $o_param=$self->{'__params'};
    my $rm = $o_param->Get('rm');
    
    my $tmpdir = $self->{'__tmpdir'};
    `$rm -rf $tmpdir`;
}

=pod

=head2 Function Exec

 Usage:    my ($incr_size, $error,$killed) = &$opt->Exec($timeout);
 Function: Copy the input record to the output
           Sleep a few seconds
 Arg:      $timeout   = The number of seconds after which a timeout is sent
 Return:   $incr_size = The number of bytes added to the output file
           $error     = An error code. 
	                If != 0, paraloop WILL BE TERMINATED - Generally marks the end of file
           $killed    = A kill code
		        0: everything OK 
                       -1: killed by timeout
                        2: killed by TERM signal (kill, qdel, ...)

=cut

sub Exec
{
    my $self    = shift;
    my $timeout = shift;

    my $input_file  = $self->{'__inputfile'};
    my $output_file = $self->{'__outputfile'};
    my $o_log       = $self->{'__log'};
    my $sleep_bin   = $self->{'__bin'};
    my $sleep_time  = $self->{'__sleep_time'};
    my $job_id      = $self->{'__job_id'};
    my $tmpdir      = $self->{'__tmp'};
    my $rewrite_input =  $self->{'__rewrite_input'};

    my $o_param=$self->{'__params'};
    my $wc = $o_param->Get('wc');
    my $cat = $o_param->Get('cat');
    my $cp = $o_param->Get('cp');

    my ($nb, $nl) = (0, 0);

    # Read the next record
    my $seq_nb = $self->Tell();
    my $seq    = $self->NextRecord();

    if (!defined($seq))
    {
        my $msg = "NO MORE SEQUENCES TO READ FROM $input_file\n";
        $o_log->Trace("$msg", '*', '1');
        return (0, -1);
    }

    my $stamp = $o_log->CalcStamp();
    my $itmp  = $tmpdir . '/' . &basename($input_file) . '_' . $job_id . '_' . $stamp;
    my $otmp  = $tmpdir . '/' . &basename($output_file) . '_' . $job_id . '_' . $stamp;

    # Write a copy of the sequence to the temporary directory
    my $id     = $seq->id();
    my $itmpfh = Bio::SeqIO->new(-file => ">$itmp", -format => 'fasta') or die "Cannot open $itmp for writing";
    $itmpfh->write_seq($seq);
    undef($itmpfh);

    # Call the preatom script (this has to be done AFTER data copy to the temporary directory, but BEFORE changing to the temporary directory)
    # The output file is truncated to 0
    $self->CallExternalScript('preatom_script',$itmp,">$otmp");
    
    # Remember where we are and change to temporary directory
    my $savdir = cwd;    # Go to temporary directory
    chdir($tmpdir) or die "Cannot change to temporary directory";

    # Start a timer
    $o_log->StartTimer('Exec_timer');

    # Sleep some time, simulating an external program:
    # 1/ Prepare the command
    my $cmd = "$sleep_bin $sleep_time";
    $o_log->Trace("$cmd\n",             '.', 2);
    $o_log->Trace("TIMEOUT $timeout\n", '.', 2);

    # 2/ Run the external program
    my ($sts, $killed) = RunExt(-cmd => $cmd, -timeout => $timeout);

    # 3/ Log the result
    $o_log->Trace("HELLO - Sequence $seq_nb - Return $sts - I did nothing during $sleep_time s\n", 'Exec_timer', '1');

    # In a real object, the output file would have been created by $cmd
    # Here, we just copy $itmp to $otmp. We APPEND data, thus preserving data written by preatom script, if any

    if ($sts==0)
    {
        `$cat $itmp >>$otmp`;
    };
    #warn ("cp $itmp $otmp");

    # Compute the number of bytes that will be added to the output file
    my $rvl_wc = `$wc -c $otmp`;
    chomp($rvl_wc);
    ($nb) = ($rvl_wc =~ /\A\s*(\d+)\s*/);

    # Back to work directory
    chdir($savdir);    

    # Call the postatom script (this has to be done AFTER retrieving the temporary directory)
    $self->CallExternalScript('postatom_script',$itmp,">>$otmp");

    # Append the output to the main output file
    `$cat $otmp >>$output_file`;

    # If status != 0 and if __rewrite_input on, write the input line to some file
    # Note - to get a status != 0, just specify Dummy_sleep_time to a negative number
    #        You should get .txt files and .txt.in files identical (only with log_level R)
    if ($sts != 0 and $rewrite_input != 0)
    {
        my $new_input = $output_file.'.in';
        `$cat $itmp >> $new_input`;
        my $msg = "Input data copied back to " . basename($new_input);
        $o_log->Trace("$msg\n", '.', 'R');
    };
        
    # Remove the temporary input file
    unlink($itmp);

    return ($nb, 0, $killed);
}

