#!/usr/bin/perl

#
# $Id$
#

# Copyright INRA

# Jerome.Gouzy@toulouse.inra.fr
# Erika.Sallet@toulouse.inra.fr
# Thomas.Schiex@toulouse.inra.fr

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
 
#
# 	Jerome.Gouzy@toulouse.inra.fr
#	Last updated: Nov 2011
#

=head1 NAME



=head1 SYNOPSIS

egn-prok_train.pl --cfg filename --infile fastafile --blastxfile seq.blastx.m8 --outdir $EUGENEDIRPROK/test [--verbose]

=head1 DESCRIPTION



=cut


my $DEBUG = 0;

BEGIN
{
	use Cwd 'abs_path';
	use File::Basename;
	my ($dirprg) =  dirname(abs_path($0));
	unshift(@INC,"$dirprg/../../lib");
	unshift(@INC,"$dirprg/../../corelib");
	our $VERSION = do {my @r = (q$Rev$ =~ /\d+/g); $r[0]};
}

use Getopt::Long;
use ParamParser;
use Logger;
use Carp;
use Data::Dumper;
use File::Path;
use File::Copy;
use File::Basename;
use General;
use GeneralBioinfo;

use strict;
use warnings;

MAIN:
{
	my $o_param =  New ParamParser('GETOPTLONG',\&Usage,("help:s", "cfg=s", "infile=s",
														 "blastxfile=s", 
														 "workingdir=s", 
														 "matrixfile=s", "verbose", 
														 "debug"));

	$o_param->SetBehaviour('use_substitution_table');
	$o_param->Update($o_param->Get('cfg'),'U') if (	$o_param->IsDefined('cfg') && 
													$o_param->AssertFileExists('cfg'));
	$o_param->Update('ENV','U');
	$o_param->AssertDefined('infile', 'matrixfile', 'EUGENEDIR');
	$o_param->AssertFileExists('blastxfile', 'infile');
	$o_param->SetUnlessDefined('train_imm', $o_param->Get('EUGENEDIR')."/bin/egn_TrainIMM");

	$o_param->AssertFullPath('infile', 'matrixfile', 'EUGENEDIR', 'train_imm', 
							 'prg_shuffleseq', 'blastxfile');

	$o_param->AssertInteger('prg_egn-prok_train_hsp_training_length', 
							'prg_egn-prok_train_hsp_training_pci');
	$o_param->AssertDefined('prg_egn-prok_train_hsp_training_evalue', 
							'prg_egn-prok_train_hsp_training_length', 
							'prg_egn-prok_train_hsp_training_pci');

	my $matrixfile = $o_param->Get('matrixfile');
	my ($outdir, $outfilename)  = $matrixfile =~ /(.+)\/(.+)/;
	$o_param->SetUnlessDefined('workingdir', $outdir);
	$o_param->AssertDirExists('workingdir');
	my $workingdir =  $o_param->Get('workingdir');
	chdir($workingdir);

	my $fasta_training = $workingdir."/egn-prok-training.".time.".$$.fa";
	# EXTRACT: extract the regions above the blast thresholds in multifasta format
	&_BuildFastaFromSelectedRegions($o_param, 
									$o_param->Get('infile'), 
									$fasta_training, 
									$o_param->Get('blastxfile'),  
									$o_param->Get('prg_egn-prok_train_hsp_training_pci'), 
									$o_param->Get('prg_egn-prok_train_hsp_training_evalue'), 
									$o_param->Get('prg_egn-prok_train_hsp_training_length'));

	if (!-s $fasta_training)
	{
		Carp::croak "ERROR during egn-prok_train.pl: not enough hsp to build training dataset. (Empty file >$fasta_training<)\n";
	}
	
	my $tmp_matrice = &BuildModels($o_param, $fasta_training);
	
	system("mv $tmp_matrice $matrixfile");
}
		 

=head2 function BuildModels

 Title        : BuildModels
 Usage        : &BuildModels($o_param,$fastafile);
 Prerequisite : none
 Fonction     : build a egn matrix from a multifasta file
 Returns      : the matrix filename

=cut

sub  BuildModels
{
	my ($o_param, $fastafile) = @_;

	my $shuffleseq = $o_param->Get('prg_shuffleseq');
	my $trainIMM   = $o_param->Get("train_imm");
	my $debug      = $o_param->Get("debug");

	my $intergenic = "$fastafile.shuffled";
	my $cde = "$shuffleseq --sequence $fastafile --outseq $intergenic 2> /dev/null";
	print STDERR $cde if ($DEBUG);
	system($cde);

	&RefFasta("$fastafile","$fastafile.train.cds");
	&RefFasta("$intergenic","$fastafile.train.pseudo-ig");
	my $f_intron = &GetStreamOut("$fastafile.train.intron");

	$cde = "$trainIMM $fastafile.train.cds $fastafile.train.intron -I $fastafile.train.pseudo-ig > $fastafile.train.log";
	system($cde);
	
	if (! $debug )
	{
		system("rm $fastafile $fastafile.train.intron $fastafile.train.pseudo-ig $fastafile.train.log");
	}

	return "$fastafile.train.cds.bin";
}



sub RefFasta
{
my($in,$out) = @_;

	my %h_seq = ();
	&FastaToHash($in,\%h_seq);
	my $f_out=&GetStreamOut("$out");
	my $cpt = 1;
	foreach my $id (keys(%h_seq))
	{
		printf $f_out "%d $h_seq{$id}{sequence}\n",$cpt++; 
	}
	$f_out->close;
}


=head2 procedure _BuildFastaFromSelectedRegions

 Title        : _BuildFastaFromSelectedRegions
 Usage        : &_BuildFastaFromSelectedRegions($infile,$selfasta,$blastresultfile,$hsp_pci,$hsp_evalue,$hsp_length)
 Prerequisite : none
 Fonction     : build a multifasta file based on blastx hits
 Returns      : none
 Args         :	$infile          multifasta file of sequences
                $selfasta        multifasta file to be generated
                $blastresultfile ncbi-blast result file (m8/m9 formatted)
 Globals      : none

=cut

sub _BuildFastaFromSelectedRegions
{
	my ($o_param, $infile, $selfasta, $blastresultfile,
		$hsp_pci, $hsp_evalue, $hsp_length) = @_;

	my %h_hit=();
	my %h_com=();
	my %h_seq=();
	my $ok = 0;
	my $id = '';
	my $debug      = $o_param->Get('debug');

	# Load sequences
	&FastaToHash($infile,\%h_seq);

	my $f_blast = &GetStreamIn($blastresultfile);
	my $f_fasta = &GetStreamOut($selfasta);
	#expected format ncbi-blast m8 or m9
	#Query id, Subject id, % identity, alignment length, mismatches, gap openings, q. start, q. end, s. start, s. end, e-value, bit score
	#AT1G50920.1	AT1G36960.1	40.00	20	12	0	625	644	161	180	3.1	18.5
	my %h_deja=();
	while(my $line=<$f_blast>)
	{
		next if ( $line =~ /^#/ );

		&ParseM8Line(\$line,1,\%h_hit);
		next if ( ! ( ($h_hit{pci} >= $hsp_pci) && ($h_hit{"e-value"} <= $hsp_evalue) && $h_hit{len} >= $hsp_length)  );	# for blastx the length is in aa
		next if ( defined($h_deja{$h_hit{target}}) );
		$h_deja{$h_hit{target}}=1;
		my($start,$end)=&MinMax($h_hit{qstart},$h_hit{qend});
		if ( ($end-$start+1) % 3 != 0 )
		{
			#print STDERR "no mod 3: $line\n";
			next;
		}
		my ($query) = $h_hit{query};
		my $sseq = &ExtractSeq(\$h_seq{$h_hit{query}}{sequence},$start,$end,$h_hit{strand});
		next if ( $sseq =~ /N/ );
		$sseq =~ s/\w\w\w$//; # j'enleve le dernier codon
		my $ok = &TRUE;
		while ( $sseq =~ /(...)/g )
		{	
			my $codon = $1;
			if ( $codon =~ /TAA|TGA|TAG/ )
			{
				$ok = &FALSE;
				last;
			}
		}
		next if  ( ! $ok );
		substr($h_seq{$h_hit{query}}{sequence},$start-1,length($sseq),'N' x length($sseq));
		$h_hit{query} =~ s/\|/-/g;
		my $header = "$h_hit{query}-$start,$end,$h_hit{strand}";
		&PrintSeq($f_fasta, $header, $sseq, 0);
	
		
	}
	$f_blast->close;
	$f_fasta->close;


	#unlink $blastresultfile;
}


sub PrintSeq
{
	my ($f_fasta, $header, $sseq, $cpt) = @_;

	my $current_header = $header;
	$current_header.="-$cpt" if ($cpt > 0);
	
	my $maxlen = 9000;
	if (length($sseq) <= $maxlen)
	{
		print $f_fasta ">$current_header\n$sseq\n";
		return;
	}
	else
	{
		my $subseq = substr($sseq, 0, $maxlen);
		my $endseq = substr($sseq, $maxlen);
		$cpt++;
		print $f_fasta ">$current_header\n$subseq\n";
		&PrintSeq($f_fasta, $header, $endseq, $cpt); 
		return;
	}
	
}
		
=head2 procedure Usage

 Title        : Usage
 Usage        : &Usage()
 Prerequisite : none
 Fonction     : print the program usage
 Returns      : none
 Args         :	none
 Globals      : none

=cut

sub Usage
{
print STDERR<<END
$0
[Mandatory]
	--help
	--cfg        filename  		configuration file
	--infile     filename  		multifasta file (full path)
	--blastxfile filename  		m8 blastx formatted file (full path)
    --matrixfile filename       Path of the resulting matrix file (full path)
	--outdir     dirname 		directory where are stored the framed matrices (full path)
[Optional]
    --workingdir dirname        [outdir] Directory where create the temporary files

[Required external software]
   \$EUGENEDIR/bin/egn_TrainIMM
END
;
}
