#!/usr/bin/perl

#
# $Id: get_BlastX.pl 1360 2019-03-01 15:15:57Z sallet $
#

# Copyright INRA

# Erika.Sallet@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Schiex@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based system for detecting, clustering and annotating non protein
# coding RNA genes.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod

=head1 NAME

get_BlastX.pl

=head1 SYNOPSIS 

get_BlastX.pl --sequence seq.fna --db database.fna --outfile  seq.blastx.gff3 --cfg eugene.par

=head1 DESCRIPTION

run blastx and convert output in GFF3

=cut


BEGIN
{
    $ENV{LANG} = "C";
}

use FindBin;
use lib "$FindBin::Bin/../../lib/";
use lib "$FindBin::Bin/../../corelib/"; 
use strict;
use warnings;
use ParamParser;
use General;
use GeneralBioinfo;
use Carp;

my $DEBUG;



MAIN:
{
	my $o_param = New ParamParser('GETOPTLONG', \&Usage, "cfg=s", "sequence=s", "db=s", "outfile=s", 
								  "workingdir=s", "verbose", "noclean", "protein_similarity_search_method=s");

	$o_param->AssertFileExists('cfg');
	$o_param->Update($o_param->Get('cfg'),'A');
	$o_param->SetSubstitution('%i', \$ENV{EGNPP});
	$o_param->SetBehaviour('use_substitution_table');

    # Test exist sequence and db
    $o_param->AssertFileExists('sequence');
	$o_param->AssertDefined('prg_mapwithblast');
	$DEBUG = ($o_param->IsDefined('verbose')) ? 1 : 0;

	my $db = $o_param->Get('db');
	Carp::croak "ERROR: $db does not exist or is not indexed\n" if (!-f $db.".pin" && !-f "$db.pal");  

	$o_param->AssertDefined('outfile', 'prg_convert_m8_to_gff3_param', 'prg_paraloop');
	$o_param->SetUnlessDefined('protein_similarity_search_method', "diamond_blastx");
	$o_param->AssertAllowedValue('protein_similarity_search_method', "blastx", "ublast_blastx", "diamond_blastx");
    my $sequence = $o_param->Get('sequence');
	my $outfile  = $o_param->Get('outfile');
	if (defined($outfile) && $outfile !~ /^\//)
    {
		my $pwd = `pwd`;
		chomp($pwd);
        $outfile = $pwd . "/" . $outfile;
        $o_param->Set('outfile', $outfile);
    }
	$o_param->AssertFullPath('outfile');
	$outfile = $o_param->Get('outfile');
	
	my ($outdir, $outfilename)  = $outfile  =~ /(.+)\/(.+)/;
	my ($dbdir, $dbname)        = $db       =~ /(.+)\/(.+)/;
	my ($seqdir, $seqname)      = $sequence =~ /(.+)\/(.+)/;
	$o_param->SetUnlessDefined('workingdir', $outdir);
	$o_param->AssertDirExists('workingdir');

	$o_param->AssertFullPath('prg_convert_m8_to_gff3');


	my $gff3file = $outfile .".gff3";
	&ParaloopBlastX($sequence, $db, $outfile, $o_param);

	&Convert2GFF3($outfile, $gff3file, $o_param);

}

	
=head2 procedure Convert2GFF3

 Title        : Convert2GFF3
 Usage        : Convert2GFF3($8file, $out, $param)
 Prerequisite : prg_convert_m8_to_gff3 and prg_convert_m8_to_gff3_param has to be defined
 Function     : Convert m8 file to gff3 file
 Args         : $m8file   Path of a m8 file
                $outfile  Resulting gff3 path
                $o_param  ParamParser object
 Globals      : none

=cut

sub Convert2GFF3
{
	my ($m8file, $outfile, $o_param) = @_;
	
	my $cmd = $o_param->Get('prg_convert_m8_to_gff3') . " " . $o_param->Get('prg_convert_m8_to_gff3_param') . 
		" --infile $m8file --outfile $outfile 2> /dev/null";
	print $cmd."\n" if ($DEBUG);
	system $cmd;

	return;
}	

   

=head2 procedure ParaloopBlastX

 Title        : ParaloopBlastX
 Usage        : ParaloopBlastX($seq, $db, $out, $param)
 Prerequisite : $o_param has to defined : workingdir
 Function     : Launch Blastx
 Args         : $sequence Input sequence file
                $db       Database
	            $out      Result m8 file
                $o_param  ParamParser object
 Globals      : none

=cut
sub ParaloopBlastX
{
	my ($sequence, $db, $out, $o_param) = @_;

	my $cmd = $o_param->Get('prg_mapwithblast') . " --sequence $sequence --db $db --output $out --workingdir ". 
		$o_param->Get('workingdir') . " --cfg " . $o_param->Get('cfg');
	$cmd .= " --protein_similarity_search_method " . $o_param->Get('protein_similarity_search_method');
	$cmd .= " --verbose" if ($DEBUG);
	$cmd .= " --noclean " if ($o_param->IsDefined('noclean'));
	print $cmd."\n" if $DEBUG;
	system($cmd) && Carp::croak "Error >$cmd<\n" ;

	return;
}

=item Usage

	$O_PARAM->Usage();
	Print the usage of the program

=cut

sub Usage
{
	print STDERR<<END
$0     
Launch blastx and create a m8 file (result) and a GFF3 file (result.gff3)
PREREQUISITE: environment variable EGNPP has to be defined.

[Mandatory]
 --sequence             filename	fasta sequence file
 --db                   filename    database 
 --outfile              filename    result file.
 --cfg                  filename    Configuration file
[Optional]
 --protein_similarity_search_method 	[blastx,ublast_blastx,diamond_blastx] Default: diamond_blastx
                               ublast_blastx and diamond_blastx methods perform a protein database reduction step with respectively 
                               ublast or diamond program, then launch blastx on this reduce database	
  --workingdir          dirname     [Default: outfile directory]  directory where create temporary files
  --noclean
  --verbose
  --help

END
}
