#
# $Id: GFF3.pm 1437 2015-04-20 12:36:30Z carrere $
#

package AnnotatedSeqFileReader::GFF3;
 
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=head1 NAME

AnnotatedSeqReader::GFF3 - parse a GFF3 formatted file and load an AnnotatedSeq object
                           See the GFF3 description at http://www.sequenceontology.org/gff3.shtml


=head1 SYNOPSIS
	See AnnotatedSeq

    Warnings: the current version of the parser does not manage 
        - embedded fasta sequence => you must provide both a gff3 file (__in_ann_file) and a sequence file (__in_seq_file)
        - exon features
    see the description of other public methods in the SYNOPSIS of AnnotatedSeq.pm

=head1 DESCRIPTION

=cut

use strict;
use warnings;
use Carp;
use General;

use Error qw(:try);
use Bio::SeqIO;
use Bio::Seq;
use AnnotatedSeq;
use Data::Dumper;
use File::Spec;
use URI::Escape;
use SeqOntology;
use AssemblyError;
use base qw(AnnotatedSeqFileReader);


BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1437 $ =~ /\d+/g); $r[0]};
}


=head2 procedure _Init

 Title        : _Init
 Usage        : my $o_gff3 = New AnnotatedSeqFile::GFF3(...);
 Function     : Constructor

=cut

sub _Init
{
	my $self = shift;
	$self->SUPER::_Init(@_);
	my $o_seq_ontoloy = New SeqOntology;
	$self -> { __o_seq_ontology } = $o_seq_ontoloy;
	#$self->{__seq_offset} = -1;
}
sub GetFormat
{
	return 'GFF3';
}
=head2 procedure DetectContig

 Title        : DetectContig
 Usage        : my $ctg = $o_gff3->DetectContig();
 Function     : Open the gff3 file until a tabulated line is found, the 1st col is the contig label
 Args         : none
 Globals      : none

=cut
sub DetectContig
{
	my $self    = shift;
	my $o_annot = $self -> { _o_annot };
	my $f_gff3 = &GetStreamIn($o_annot->{__in_ann_file});
	my $ctg = "";
	#SC20130709: on garde la trace de la ligne de separateur des contigs pour gérer dans la boucle suivante le cas des contigs non annotés
	my $sequence_region_line = "";
	while( my $line = <$f_gff3>)
	{
		chomp($line);
		$sequence_region_line = $line if ($line =~ /^##sequence-region/);
		next if ( $line =~ /^\#/ || $line eq '');
		($ctg) = split(/\t/,$line,2);
		last;
	}
	
	#SC20130709: dans le cas de contigs non annotés, $ctg est vide (pas de ligne)
	#            donc on récupere l'info du contig dans la derniere ligne ##sequence-region rencontrée
	if (!defined ($ctg) or ($ctg eq ""))
	{
		if ($sequence_region_line =~ /^##sequence-region\s+(\S+)/)
		{
			$ctg = $1;
		}
	}
	
	$o_annot -> { __contig } = $ctg;
}		

=head2 procedure LoadGenomicSequence

 Title        : LoadGenomicSequence
 Usage        : $o_gff3->LoadGenomicSequence();
 Function     : throw an exception, because the loader of the genomic seuqnece from a gff3 file is not yet implemented
 Returns      : none
 Args         : none
 Globals      : none

=cut
sub LoadGenomicSequence
{
	my ($self) = @_;
	my $o_annot = $self -> { _o_annot };
	my $genomic_seq = '';
	my $seq_flg= &FALSE;
	my $fh_gff3 = &GetStreamIn($o_annot->{__in_ann_file});
	while ( my $line = <$fh_gff3> )
	{
		chomp $line;
		if ( $line =~ /^##FASTA/ )
		{
			my $line1 = <$fh_gff3>;
			my $ctg = $o_annot->GetContigLabel();
			chomp $line1;
			if ( $line1 =~ />$ctg$/ )
			{
				$seq_flg = &TRUE;
				next;
			}
		}
		if ($seq_flg==&TRUE && $line =~ /^[CAGTNcagtn ]+$/ )
		{
			$genomic_seq .= $line;
		}
		elsif ($seq_flg==&TRUE)
		{
			last;
		}
	}
	return \$genomic_seq;
}

=head2 procedure Load

 Title        : Load
 Usage        : $o_gff3->Load();
 Prerequisite : the infile must be in GFF3 format and should contain unique IDs
 Function     : Load the protein-coding genes, the Ncrna and the repeat contained in a GFF3 formatted file
                Warning: only the features gene|RNA|repeat_region|CDS are loaded
 Returns      : none
 Args         : none
 Globals      : none

=cut

sub Load
{
	my $self = shift;
	my $o_annot = $self -> { _o_annot };
	my $o_so    = $self -> { __o_seq_ontology };

	# Keep some loci (genes, mrnas) in this hash, the key is the ID
	my %h_loci=();
	
	my $f_gff3 = &GetStreamIn($o_annot->{__in_ann_file});
	
	# To ensure that the gene is defined before the corresponding mRNA and mRNA before CDS, etc.
	# we split the gff3 file in temporary file: 1 file / subclass of Locus::
	my (%h_f_gff3,%h_fh_gff3);
	my @a_classes = qw( Gene Mrna Ncrna Exon CdsRegion Repeat UtrRegion MiscRegion );
	
	# Open the files
	my $tmp_dir = $o_annot -> { __tmpdir };
	foreach my $class ( @a_classes )
	{
		$h_f_gff3  { $class } = File::Spec->catfile( $tmp_dir, "import." . $class . ".gff3" );
		$h_fh_gff3 { $class }  = &GetStreamOut($h_f_gff3{$class});
	}
	
	# Read the gff3 file, set the contig label if necessary, and copy the line to the corresponding tmp file
	my $contig_label = $o_annot -> GetContigLabel();
	while( my $line = <$f_gff3>)
	{
		chomp($line);

		# Update sequence-region if specified
		# NOTE - The Locus object will update the sequence-region anyway, may be enlarging it
		if ( $line =~ /^##sequence-region/ )
		{
			my ($ctg,$start,$end) = ( $line=~/##sequence-region ([^ ]+) (\d+) (\d+)/ );
			if ( $ctg eq $contig_label )
			{
				$o_annot -> UpdateSeqRegion($start,$end);
			}
			next;
		}
		
		# Update header data if the line starts with #bfc contig
		if ( $line =~ /^#bfc / )
		{
			my ($key,$value) = ($line =~ / ([^ ]+)=(.+)$/ );
			$o_annot->SetHeaderData($key,$value);
			next;
		}
		
		# The FASTA sequence:
		# #FASTA
		# >contigs
		# GAGAGA
		#if ( $line =~ /^#FASTA/ )
		#{
		#	# skipping the contig name
		#	my $l = <$f_gff3>;
		#	$self->{__seq_offset} = $f_gff3->tell();
		#}
		
		# Ignored lines
		next if ( $line =~ /^\#/ || $line eq '');
		
		# data lines
		my ( $ctg,$dummy,$so) = split(/\t/,$line);
		next if ( $ctg ne $contig_label );
		my $class = $o_so -> GetLocusClass( $so );
		if ( $class eq "" )
		{
			# We do not use GetWarningHeader here, because we want $so to be part of the warning message, not only data
			warn ( "WARNING $so $so NOT in Sequence Ontology - line skipped !\n" );
		}
		else
		{
			$class =~ s/Locus:://;
			
			# should not happen
			if ( !exists $h_fh_gff3{$class} )
			{
				throw Error::Simple ("ERROR Line $line - $class unknown\n");
			}
			else
			{
				my $fh = $h_fh_gff3{$class};
				print $fh "$line\n";
			}
		}
	}
	
	# Close the temporary files
	foreach my $class ( @a_classes )
	{
		$h_fh_gff3 { $class }->close();
	}
	$f_gff3->close;
	print STDERR "Filtering GFF3 files finished\n" if ( $o_annot -> GetDebug() );

	# Read the files we just wrote, in the appropriate order
	foreach my $class ( @a_classes )
	{
		#my $class = 'Locus::'.$class;
		my $fh_gff3 = GetStreamIn($h_f_gff3{$class});
		while( my $line = <$fh_gff3>)
		{
			try
			{
				# This is a fake while loop, avoiding a warning when executing next
				# Only ONE iteration will be performed
				while ( &TRUE )
				{
					chomp($line);
					my $rh_param = $self -> __SplitGff3Line($line);
					my %h_param=%$rh_param;
					
					# CDS/exon/Utr do not need an ID, other objects do
					if ( $class !~ /Exon|CdsRegion|UtrRegion/ )
					{
						if (! defined $h_param{ID})
						{
							warn ($self -> GetWarningHeader() . "anonymous $class, no ID value in the feature field, skipped line $line\n");
							last;
						}
					}
	
					# Gene/Ncrna/Repeat/MiscRegion do not need a Parent, other objects do
					if ( $class !~ /Gene|Ncrna|Repeat|MiscRegion/ )
					{
						if (! defined $h_param{Parent})
						{
							throw Error::Simple ("orphan $class (no Parent specified), skipped line $line\n");
						}
					}
					
					# The remaining is dependant of $class
					if ( $class eq 'Gene' )
					{
						# Will be undef if gene skipped (error in constructor)
						$h_loci{$h_param{ID}} = undef;
						my $o_gene = New Locus::Gene(%h_param);
						$o_annot->AddGene($o_gene);
						$h_loci{$h_param{ID}} = $o_gene;
						last;
					}
					
					if ( $class eq 'Mrna' )
					{
						if ( !exists $h_loci{$h_param{Parent}} )
						{
							throw Error::Simple ( "No gene with ID=$h_param{Parent} - The gff3 file is invalid !" );
						}
						my $o_gene = $h_loci{$h_param{Parent}};
						
						# Will be undef  if Mrna skipped
						$h_loci{$h_param{ID}} = undef;
						
						# If the gene was skipped, skip the Mrna too
						if ( defined ( $o_gene ) )
						{
							delete $h_param{Parent};
							my $o_mrna = New Locus::Mrna(%h_param);
							$o_gene -> AddmRNA ( $o_mrna );
							$h_loci{$h_param{ID}} = $o_mrna;
							last;
						}
					}
	
					if ( $class eq 'Ncrna' )
					{
						# Will be undef  if Ncrna skipped
						$h_loci{$h_param{ID}} = undef;

						if ( defined($h_param{Parent}) )
						{
							if ( !exists $h_loci{$h_param{Parent}}  )
							{
								throw Error::Simple ( "No gene with ID=$h_param{Parent} - The gff3 file is invalid !" );
							}
							if ( defined $h_loci{$h_param{Parent}} )
							{
								my $o_gene = $h_loci{$h_param{Parent}};
								delete $h_param{Parent};
								my $o_ncrna = New Locus::Ncrna(%h_param);
								$o_gene -> AddNcrna($o_ncrna);
								$h_loci{$h_param{ID}} = $o_ncrna;
							}
						}
						else
						{
							my $o_ncrna = New Locus::Ncrna(%h_param);
							$o_annot->AddNcrna($o_ncrna);
							$h_loci{$h_param{ID}} = $o_ncrna;
						}
					}

					if ( $class eq 'Exon' )
					{
						my $ra_parents = $h_param{Parent};
						if ( ref($ra_parents) eq '' )
						{
							$ra_parents = [ $ra_parents ];
						}
						delete $h_param{Parent};
						my $i=0;
						my $id = $h_param{ID};
						foreach my $parent (@$ra_parents)
						{
							if ( !exists $h_loci{$parent} )
							{
								throw Error::Simple ( "No locus with ID=$parent - The gff3 file is invalid !" );
							}
							if ( exists $h_param{ID} && @$ra_parents>1 )  # Avoiding an invalid gff3 !
							{
								$i++;
								$h_param{ID} = $id.'.'.$i;
							}
							
							if ( defined $h_loci{$parent} )
							{
								my $o_exon  = New Locus::Exon(%h_param);
								my $o_locus = $h_loci{$parent};
								my $o_rna;
								if ( $o_locus -> isa ( 'Locus::Gene' ))
								{
									$o_rna = $o_locus -> GetMrna(0);
									if ( !defined $o_rna )
									{
										# POUR AJOUTER UN MRNA 
										#$o_rna = New Locus::Mrna ( $o_locus );
										#$o_locus -> AddmRNA($o_rna);
										# POUR LAISSER TEL QUEL
										#$o_rna = $o_locus;
										$o_rna = New Locus::Mrna ( $o_locus );
										$o_locus -> AddmRNA($o_rna);
									}
								}
								elsif ( $o_locus -> isa ( 'Locus::Rna' ) )
								{
									$o_rna = $o_locus;
								}
								else
								{
									my $ref = ref $o_locus;
									throw Error::Simple ( "The parent of an exon ($parent) should be an Rna, may be a gene, NOT a $ref\n" );
								}
								
								$o_rna -> AddExon ( $o_exon );
							}
						}
						last;
					}
		
					if ( $class eq 'CdsRegion' )
					{
						my $ra_parents = $h_param{Parent};
						if ( ref($ra_parents) eq '' )
						{
							$ra_parents = [ $ra_parents ];
						}
						delete $h_param{Parent};
						my $i  = 0;
						my $id = $h_param{ID};
						foreach my $parent (@$ra_parents)
						{
							if ( !exists $h_loci{$parent}  )
							{
								throw Error::Simple ( "No mRNA with ID=$parent - The gff3 file is invalid !" );
							}
							if ( exists $h_param{ID} && @$ra_parents>1 )  # Avoiding an invalid gff3 !
							{
								$i++;
								$h_param{ID} = $id.'.'.$i;
							}
							if ( defined $h_loci{$parent} )
							{
								my $o_cds   = New Locus::CdsRegion(%h_param);
								my $o_locus = $h_loci{$parent};

								# Handling incomplete gff3 files, with gene but no mRNA lines
								if ( $o_locus -> isa ( 'Locus::Gene' ) )
								{
									my $o_mrna = New Locus::Mrna ( $o_locus );
									$o_locus -> AddmRNA ( $o_mrna );
									$h_loci{$parent} = $o_mrna;
									$o_locus = $o_mrna;
								}
								$o_locus -> AddCDS ( $o_cds );
							}
						}
						last;
					}
					
					if ( $class eq 'Repeat' )
					{
						my $o_repeat = New Locus::Repeat(%h_param);
						$o_annot->AddRepeat($o_repeat);
						last;
					}
					
					if ( $class eq 'UtrRegion' )
					{
						#my @a_parent = split(',',$h_param{parent});
						my @a_parents = ref($h_param{Parent} eq '')?( $h_param{Parent} ) : $h_param{Parent};
						delete $h_param{Parent};
						foreach my $parent (@a_parents)
						{
							if ( !exists $h_loci{$parent}  )
							{
								throw Error::Simple ( "No gene/mRNA with ID=$parent - The gff3 file is invalid !" );
							}
							if ( defined $h_loci{$parent} )
							{
								my $o_utr = New Locus::UtrRegion(%h_param);
								my $o_locus = $h_loci{$parent};
								$o_locus -> AddUtr ( $o_utr );
							}
						}
						last;
					}
						
					if ( $class eq 'MiscRegion' )
					{
						if ( defined($h_param{Parent}) )
						{
							if ( !exists $h_loci{$h_param{Parent}}  )
							{
								throw Error::Simple ( "No region with ID=$h_param{Parent} - The gff3 file is invalid !" );
							}
							if ( defined $h_loci{$h_param{Parent}} )
							{
								my $o_parent_locus = $h_loci{$h_param{Parent}};
								delete $h_param{Parent};
								my $o_region = New Locus::MiscRegion (%h_param);
								$o_parent_locus -> AddUnsortedRegion( $o_region );
							}
						}
						else
						{
							my $o_region = New Locus::MiscRegion (%h_param);
							$o_annot->AddMiscRegion($o_region);
							$h_loci{$h_param{ID}} = $o_region;
						}
						last;
					}
					last;
				}
			}
			catch AssemblyError with
			{
				# SKIP Silently the locus !
				# warn ( "WARNING - Assembly error - Locus skipped\n" );
			}
			catch Error with
			{
				my ($o_error) = @_;
				my $msg = $o_error -> text();
				warn ( "ERROR $msg\n" );
				$o_annot -> SetFatalError();
			};
		}
		undef $f_gff3;
	}

	# check the strand if correctly assigned for each Locus (this is not obvious, as in (invalid) gff3 files  you may have a '.' for the strand
	if ( $self -> __CheckStrands() == &FALSE )
	{
		$o_annot -> SetFatalError();
	}
	
	# See the switch --load-desc-file in BioFileConverter
	if ( $o_annot->GetDescFile() ne "" )
	{
		$self->__LoadDescriptionFile($o_annot->GetDescFile());
	}
}

=head2 procedure __CheckStrands

 Title     : __CheckStrands
 Usage     : $self->__SplitGff3Line();
 function  : Checks every locus has a correct strand, and warn something for each pb
 Returns   : 1 if check OK, 0 if at least 1 locus has a pb
 Args      : $line The line to split

=cut
sub __CheckStrands
{
	my $self = shift;
	my $o_annot = $self -> { _o_annot };
	
	my $ra_loci = $o_annot -> GetSortedLoci();
	my $error = 0;
	foreach my $o_locus ( @$ra_loci)
	{
		if ( $o_locus -> GetStrand() == 0 )
		{
			my $pk = $o_locus -> GetpkExtended();
			warn ( "ERROR Locus $pk - bad strand - your gff3 file is NOT valid - this cannot be recovered\n" );
			$error = 1;
		}
	}
	return ! $error;
}
			
=head2 function __SplitGff3Line

 Title        : __SplitGff3Line
 Usage        : my $rh_param = $self->__SplitGff3Line($line);
 Function     : split the gff3 line, the results is put to a hash, this hash is returned
 Returns      : $rh_param, a hash with key/values ready to create a new Locus
 Args         : $line The line to split

=cut
sub __SplitGff3Line
{
	my $self = shift;
	my ($line) = @_;
	my %h_param;
	my ($ctg,$src,$type,$start,$end,$score,$strand,$phase,$comment) = split(/\t/,$line);
				
	$h_param{start}  = $start;
	$h_param{end}    = $end;
	$h_param{strand} = $strand = 1 if ( $strand eq '+' );
	$h_param{strand} = $strand =-1 if ( $strand eq '-' );
	$h_param{strand} = $strand = 0 if ( $strand eq '.' );

	$h_param{sub_type}= $type;
	$h_param{score}  = $score;

	# Do not read the phase, it will be computed again if necessary
	# $h_param{phase}  = $phase;
				
	my @a_com=split(';',$comment);
	for (my $i=0;$i<=$#a_com;$i++)
	{	
		my $pair = $a_com[$i];
		$pair =~ s/^\s+|\s+$//g ; # MANU Oct 2009 - Remove heading and trailing blanks
		next if ($pair eq "");
		if ( $pair !~ /=/ )
		{
			warn ($self->GetWarningHeader() . "Tags should be in the form key=value - $pair ignored\n");
			next;
		}
		my ($key,$value) = split('=',$pair,2);
			
		# If $key starts with a UC but is NOT a reserved tag, silently convert it to a lc-beginning tag
		if ( $key !~ /^Name$|^Note$|^Parent$|^ID$|^Dbxref$|^Alias$|^Target$|^Gap$|^Derives_from$|^Ontology_term$/ )
		{
			$key = "\l$key";
		}
		if ( $key =~ /ec_number/i )
		{
			$key = 'eC_number';
		}
		if ( $key ne 'Name' && $value =~ /,/ )
		{
			my @a_values   = split ( /,/,$value);
			foreach my $v (@a_values)
			{
				$v = uri_unescape $v;
			}
			$h_param{$key} = \@a_values;
		}
		elsif ( $value eq '""' )
		{
			$h_param{$key} = '_no_value';
		}
		else
		{
			$value = uri_unescape ( $value );
			$h_param{$key} = $value;
		}
	}
	return \%h_param;
}

sub __LoadDescriptionFile
{
	my $self = shift;
	my ($desc_file) = @_;
	my $o_annot = $self -> { _o_annot };

	my $o_splitter_class = "Splitter::Desc";
	eval "use $o_splitter_class";
	throw Error::Simple( $@ ) if ( $@ ne "" );

	my $o_splitter       = New $o_splitter_class( $desc_file );
	throw Error::Simple( $@ ) if ( $@ ne "" );
	my $ra_genes = $o_annot -> GetGenes();
	foreach my $o_gene ( @$ra_genes )
	{
		my $error = &FALSE;
		#my $id = $o_gene->GetId();
		my $id = $o_gene->GetGeneName();
		$id =~ s/gene://;
		my $r_desc;
		try
		{
			$r_desc = $o_splitter->GetContig($id);
		}
		catch Error with
		{
			print STDERR "NOT FOUND Desc FOR " . $o_gene->GetpkExtended() . " (looking for $id)\n";
			$error = &TRUE;
		};
		
		if ( ! $error )
		{
			my @a_children;
			if ( $o_gene -> IsProteinCoding() == &FALSE )
			{
				foreach my $o_rna (@{$o_gene->GetNcrnas()})
				{
					push @a_children,$o_rna;
				}
			}
			else
			{
				foreach my $o_rna (@{$o_gene->GetMrnas()})
				{
					push @a_children,$o_rna;
				}
			}

			my %h_ipr;
			my @a_desc = split /\n/,$$r_desc;
			foreach my $l (@a_desc)
			{
				if ( $l =~ /^ID   (.+)$/ ) 
				{ 
					$o_gene -> Delete( 'gene' );
					$o_gene -> Set('gene',$1);
					next;
				}
				if ( $l =~ /^LT   (.+)$/ ) 
				{ 
					$o_gene -> Delete( 'locus_tag' );
					$o_gene -> Set('locus_tag',$1);
					map { $_->Delete('locus_tag') } @a_children;
					map { $_->Set('locus_tag',$1) } @a_children;
					next;
				}
				if ( $l =~ /^DE   (.+)$/ ) 
				{ 
					map { $_->Delete('product') } @a_children;
					map { $_->Set('product',$1) } @a_children;
					next;
				}
				if ( $l =~ /^CL   (.+)$/ ) 
				{
					map { $_->Delete('function') } @a_children;
					map { $_->Set('function',$1) } @a_children;
					next;
				}
				if ( $l =~ /^SY   (.+)$/ ) 
				{ 
					my $sy = $1;
					my ($sy_sp,$sy_lt) = ( $sy =~ /(.+):.+:(.+)/ );
					map { $_->AddValue('note',"corresponds to $sy_lt in $sy_sp") } @a_children;
					next;
				}
				
				# keep the ipr inside a hash, removing the redundancy
				if ( $l =~ /^CC   AutoIPR_.+: *(IPR.+)/ ) 
				{ 
					my $val = $1;
					my @a_ipr = ($val =~ /(IPR\d+):/g);
					foreach my $ipr (@a_ipr) 
					{ 
						$h_ipr{$ipr} = 1;
					}
					next;
				}
				#SC20130711: correction regexp sinon ne recupere que le dernier GO (apres les dernier ":")
				#elsif ( $l =~ /^CC   AutoGO_.+: (.+)/ )
				elsif ( $l =~ /^CC   AutoGO_[^:]+: (.+)/ )
				{ 
					
					my $val = $1;
					my @a_go = ($val =~ /GO:(\d+)/g);
					foreach my $go (@a_go) 
					{ 
						map { $_->AddValue('Dbxref',"GO:$go") } @a_children;
					}
					next;
				}
				elsif ( $l =~ /^CC   (.+)/ )
				{
					my $val = $1;
					map { $_->AddValue('note',$1) } @a_children;
				}
				if ( $l =~ /^IP   (.+)/ )
				{
					my $val = $1;
					map { $_->AddValue('note',$1) } @a_children;
				}
				if ( $l =~ /^EC/ ) 
				{
					my @a_f = split ';', $l;
					map { $_->Delete('EC_number') } @a_children;
					foreach my $f (@a_f)
					{
						next if ( $f eq '');
						my ($ec_number) = ($f =~ /EC([0-9.\-]+):?/g);
						$ec_number = $Locus::O_ENZYMVAL->Check($ec_number) if (defined $Locus::O_ENZYMVAL);
						#~ map { $_->Delete('EC_number') } @a_children;
						#~ map { $_->Delete('eC_number') } @a_children;
						#~ map { $_->Delete('ec_number') } @a_children;
						#~ map { $_->Set('EC_number',$ec_number) } @a_children;
						map { $_->AddValue('EC_number',$ec_number) } @a_children;
						#map { $_->Set('ec_number',$ec_number) } @a_children;
					}
					next;
				}
			}
			
			# Add the ipr, if any
			foreach my $ipr ( keys %h_ipr )
			{
				map { $_->AddValue('Dbxref',"InterPro:$ipr") } @a_children;
			}
		}
	}
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
