#
# $Id: Fasta.pm 1309 2013-11-18 09:03:08Z manu $
#

package AnnotatedSeqFileWriter::FastaLike;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=head1 NAME

AnnotatedSeqFileWriter::Fasta - Using annotations stored in a companion AnnotatedSeq object, using the genomic fasta file
                                dump fasta files corresponding to the features

=head1 SYNOPSIS

 my $o_dumper = New AnnotatedSeqFileWriter::Narcisse( o_annot => $o_annot );
 $o_dumper -> Dump();

=head1 DESCRIPTION

 this class extends the AnnotatedSeqFileWriter, it dumps data from AnnotatedSeqFile to Narcisse files

=cut

use strict;
use warnings;
use Carp;

use Error qw(:try);
use LipmError::IOException;
use Bio::Seq;
use Data::Dumper;
use General;

use Locus;
use base qw( AnnotatedSeqFileWriter );

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1309 $ =~ /\d+/g); $r[0]};
}

sub _Init
{
	my $self = shift;

	$self->SUPER::_Init(@_);
}

=head2 procedure Dump

 Title     : Dump
 Usage     : $o_annot_writer -> Dump ()
 Procedure : Dump the genes objects to 4 fasta-formatted files
 Args      : -append 1/0 (optional, def 0)
 Error     : Throw an Error::Simple is there is no sequence loaded

=cut

sub Dump
{
	my $self = shift;
	my (%h_params) = (-append => 0, -output_utr => 0, -output_intron => 0,  @_);
	my $open_mode  = ($h_params{ -append }!=0)? 'a' : 'w';
	
	my $o_annot = $self -> { _o_annot };
	
	if ( $o_annot -> GetLength() == 0 )
	{
		throw Error::Simple( "Cannot generate " . $self->GetFormat() . " files, no sequence is loaded" );
	}
	
	my $ra_genes = $o_annot -> GetGenes();
	my $ra_ncrnas= $o_annot -> GetNcrnas();

	# Get and parse the file name, because the *REAL* file names is rebuilt
	my $seq_filename = $o_annot -> GetOutputFile();
	
	my ($filename,$ext,$comp_ext) = ("","","");

	if ( &GetCompressPath ( $seq_filename ) ne "" )
	{
		($filename, $comp_ext) = ( $seq_filename=~/^(.+)(\.\w+)$/ )
	}
	else
	{
		$filename = $seq_filename;
	}
	#print "\nPATH->#$filename##$comp_ext#\n";
	if ( $filename =~ /\.(\w+)$/)
	{
		($filename,$ext) = ( $filename=~/^(.+)(\.\w+)$/ )
	}
	#print "PATH->#$filename#$ext#$comp_ext#\n";

	my @a_seqtypes = ('prot','cds','gene','mrna','ncrna');
	push (@a_seqtypes,'intron') if ($o_annot->GetOutputIntron()==&TRUE);
	push (@a_seqtypes,'3utr','5utr') if ($o_annot->GetOutputUtr()==&TRUE);
	foreach my $seqtype (@a_seqtypes)
	{
		#
		# NOTE - Using FileCompressedAppend here, if you keep a ref to the objects after my death, you'll avoid
		#        Compression/decompression steps
		my $f_out    = $filename . '_' . $seqtype . $ext . $comp_ext;
		#print "KOUKOU $f_out\n";
		my $o_comp   = New FileCompressedAppend ( $f_out, $open_mode );
		$f_out       = $o_comp->GetFileName();
		unlink ($f_out) if ( $open_mode eq 'w' );
		my $fh_out   = new IO::File( $f_out, $open_mode ) || throw LipmError::IOException($f_out,'f');
		print STDERR $self -> GetWarningHeader() . " - Dumping $seqtype to $f_out\n" if ( $o_annot -> GetDebug());

		my @a_loci;
		push @a_loci, @$ra_genes;
		if ( $seqtype eq 'ncrna' )
		{
			push @a_loci, @$ra_ncrnas;
		}
		foreach my $o_locus (@a_loci)
		{
			$self -> __DumpFeature ( $seqtype,$fh_out,$o_locus );
		}
		$fh_out -> close();
		$o_annot -> AddDumped( $o_comp );
	}
}

sub _ParseFileName
{
	my $self = shift;

	my $o_annot = $self -> { _o_annot };
	
	# Get and parse the file name, because the *REAL* file names is rebuilt
	my $seq_filename = $o_annot -> GetOutputFile();
	
	my ($filename,$ext,$comp_ext) = ("","","");

	if ( &GetCompressPath ( $seq_filename ) ne "" )
	{
		($filename, $comp_ext) = ( $seq_filename=~/^(.+)(\.\w+)$/ )
	}
	else
	{
		$filename = $seq_filename;
	}
	#print "\nPATH->#$filename##$comp_ext#\n";
	if ( $filename =~ /\.(\w+)$/)
	{
		($filename,$ext) = ( $filename=~/^(.+)(\.\w+)$/ )
	}
	#print "PATH->#$filename#$ext#$comp_ext#\n";
	return ($filename,$ext,$comp_ext);
}

=head2 Function __DumpFeature, __DumpProt,__DumpCds, __DumpMrna, __DumpGene, __DumpIntro

 Title        : __DumpFeature
 Function     : Print on the $fh_out filehandle the fasta sequences corresponding to the CDS, the protein, the mrna or the gene
				For $seqtype = prot (dumping proteins), we call the GetTranslation function from Locus::Cds, MAY BE MODIFYING THE CONSISTENCY FLAG
                Warning: only protein coding genes are considered
 Returns      : The number of printed sequences
 Args         : $seqtype   : prot|mrna|cds|gene|ncrna|intron
                $fh_out    : An already opened file
				$o_locus   : A Locus::Gene object
                $protein_id: CDS|MRNA - REMOVED !!! AUTOMATICALLY SELECTING MRNA !!!
 Globals      : none

=cut

sub __DumpFeature
{
	my $self    = shift;
	my $seqtype = shift;

	if ( $seqtype =~ /prot|cds/ )
	{
		return $self -> __DumpProt($seqtype,@_);
	}
	elsif ( $seqtype eq 'mrna' )
	{
		return $self -> __DumpMrna(@_);
	}
	elsif ( $seqtype eq 'intron' )
	{
		return $self -> __DumpIntron(@_);
	}
	elsif ( $seqtype eq '5utr' )
	{
		return $self -> __DumpUtr('five_prime_UTR',@_);
	}
	elsif ( $seqtype eq '3utr' )
	{
		return $self -> __DumpUtr('three_prime_UTR',@_);
	}
	elsif ( $seqtype eq 'gene' )
	{
		return $self -> __DumpGene(@_);
	}
	elsif ( $seqtype eq 'ncrna' )
	{
		return $self -> __DumpNcrna(@_);
	}
}

=head2 Function __DumpProt

 Title        : __DumpProt
 Function     : Called by __DumpFeature, dumps the proteic translation of each mrna (proteic sequence), or the cds of each mrna (nucleic sequence)
				For $seqtype = prot (dumping proteins), we call the GetTranslation function from Locus::Cds,
				MAY BE MODIFYING THE CONSISTENCY FLAG
                Warning: only protein coding genes are considered
 Returns      : The number of printed sequences
 Args         : $seqtype   : prot|cds
                $fh_out    : An already opened file
				$o_gene    : A Locus::Gene object
 Globals      : none

=cut

sub __DumpProt
{
	my $self             = shift;
	my($seqtype,$fh_out, $o_gene) = @_;
	return 0 unless ( $o_gene -> IsSelected() );
	return 0 unless ( $o_gene->IsProteinCoding() );

	my $o_annot = $o_gene -> GetAncestor();
	my $ctg     = $o_annot -> GetChromosomeLabel();
	my $sp      = $o_annot -> GetSpecies();
	my $kingdom = $o_annot -> GetKingdom();
	my $debug   = $o_annot -> GetDebug();
	#my $create_cds = $o_annot -> GetCreateCds();
	
	my $nb_obj = 0;
	
	my $ra_mrna = $o_gene -> GetMrnas();
	foreach my $o_mrna (@$ra_mrna)
	{
		next unless $o_mrna -> IsSelected();
		print STDERR $self -> GetWarningHeader() . " - Writing a $seqtype " . $o_mrna->GetpkExtended()  ."\n" if ( $debug );
		#my $o_cds       = $o_mrna -> GetCDS($create_cds);
		my $o_cds       = $o_mrna -> GetCDS();
		if ( !defined ( $o_cds ) )
		{
			warn $self -> GetWarningHeader() . "No cds for " . $o_mrna->GetpkExtended() . "\n";
			next;
		}
		next unless ( $o_cds->IsSelected() );
		
		# TODO - REVOIR TOUT CA AVEC __GetMrnaInfo
		
		# JG/API defini seulement dans le parseur Ensembl => je ne pense pas qu'il y ait eu des tests de non regression, dont je garde egalement l'ancien code
		# TODO M'ETONNERAIT QUE CA MARCHE
		#my $data_source_ensembl = ( defined($o_mrna) && $o_mrna->Get('cds_start') ne '' ) ? 1 : 0; 
		my $data_source_ensembl = &FALSE;

		# next unless defined $o_cds;
		if ( $o_cds->GetRegions() eq '')
		{
			my ($s,$e) = $o_cds->GetEnds();
			throw Error::Simple("ERROR - No region for cds [$s..$e]\n");
		}
		#my %h_param     = $o_cds->GetAll('locus_tag','product','gene_name','accession');
		#my $locus_tag = $o_cds -> GetLocusTag();
		
		# NOTE - The Locus tag found in the fasta file is the GENE locus tag !
		my $locus_tag = $o_gene ->GetLocusTag();
		my $gene_name = $o_cds -> GetGeneName(&TRUE);
		$locus_tag = $gene_name if ( $locus_tag eq "");
		
		# l'accession de la proteine est le numero d'accession du mrna, mais c'est pas toujours le cas chez narcisse - A voir ce qu'on fait de ca
		# NOTE - protein_id est un parametre de nc_get_data.pl
		#my $accession = ($protein_id eq 'MRNA') ? $o_mrna -> GetAccession() : $o_cds -> GetAccession();
		my ($dum1,$dum2,$accession,$dum3) = $self->__GetMrnaInfo($o_gene,$o_mrna,$o_cds);
		$accession = $o_mrna -> GetLocusTag()     if ( $accession eq "" );
		$accession = $o_gene -> GetLocusTag()     if ( $accession eq "" );
		$accession = $o_gene -> GetGeneName()     if ( $accession eq "" );
		$accession =~ s/^mRNA://;

		my $note  = $o_cds -> GetProduct(&TRUE);
		$note     = $o_cds -> GetNote(&TRUE)    if ( $note eq '');

		# JG changement suite au merge API my ($len,$sseq) = $o_gene->GetSequence($r_seq,$h_param{strand},$exoncds);
		my ($len,$sseq);
		#if ( $data_source_ensembl )
		#{
		#	$sseq = $o_mrna -> GetSequence;
		#	$len  = $o_mrna -> GetSequenceLength();
		#}
		#else
		#{
			#($len,$sseq) = $o_gene->GetSequence($r_seq,$h_param{strand},$exoncds);
			$sseq = $o_cds -> GetSequence;
			$len  = $o_cds -> GetSequenceLength();
		#}
		next if ( $len == 0 );
		my $strand = $o_cds -> GetStrand();
		my ($begin,$end) = $o_cds -> GetEnds();

		#my $acc = $h_param{accession}; # $r_cds->GetAccession();
		#$acc = $h_param{locus_tag} if ( $acc eq '');
		
		#TF patch to provide an accesion number to the potato cds and prot
		$accession    = $o_gene -> GetId()       if ( $accession eq '');
		$locus_tag	  = $o_gene -> GetId()       if ( $locus_tag eq '');
		
		if ( $seqtype eq 'cds' )
		{
			my $gn = ( $gene_name ne '')  ? " gn=$gene_name" : "";
			$accession = "cds:$accession";# if ( $accession !~ /^cds:/i );
			if (!$o_annot -> GetNoCleanFastaHeader())
			{	
				$accession =~ s/^cds://i while ($accession =~ /^cds:/i);
			}
			my $header = ">$accession locus_tag=$locus_tag$gn begin=$begin end=$end len=$len chr=$ctg strand=$strand sp=$sp $note";
			$self -> _PrintFasta ($fh_out, $header, \$sseq);
			$nb_obj++;
		}
		else
		{
			my $seq_prot = $o_cds -> GetTranslation ( $debug );
			my $len_prot = length $seq_prot;
			
			if ( $len_prot > 0 )
			{
                # The translation is correct: format the proteic sequence for a fasta file
                my $gn = ( $gene_name ne '')  ? " gn=$gene_name" : "";
                $accession = "prot:$accession";# if ( $accession !~ /^prot:/i );
                if (!$o_annot -> GetNoCleanFastaHeader())
				{	
					$accession =~ s/^prot://i while ($accession =~ /^prot:/i);
				}
                my $product = $o_mrna -> GetConsolidatedProduct();
                $product    = $o_mrna -> GetConsolidatedNote() if ( $product eq '');
				
                my $header = ">$accession locus_tag=$locus_tag$gn len=$len_prot chr=$ctg strand=$strand sp=$sp def=$product";
                $header    .= " note=$note" if ($note ne '');
                $self -> _PrintFasta ($fh_out, $header, \$seq_prot);
				$nb_obj++;
			}
		}
	}
	return $nb_obj;
}

=head2 Function __DumpMrna

 Title        : __DumpMrna
 Function     : Called by __DumpFeature, dumps the mrna genomic sequence (exons only)
                Warning: only protein coding genes are considered
 Returns      : The number of printed sequences
 Args         : $fh_out    : An already opened file
				$o_gene    : A Locus::Gene object
 Globals      : none

=cut

sub __DumpMrna
{
	my $self             = shift;
	my($fh_out, $o_gene) = @_;
	return 0 unless ( $o_gene -> IsSelected() );
	return 0 unless ( $o_gene -> IsProteinCoding() );

	my $o_annot = $o_gene -> GetAncestor();
	my $ctg     = $o_annot -> GetChromosomeLabel();
	my $sp      = $o_annot -> GetSpecies();
	
	my $nb_obj = 0;
	my $ra_mrnas = $o_gene -> { __a_mrnas };
	foreach my $o_mrna ( @$ra_mrnas )
	{
		next unless ($o_mrna -> IsSelected() );
		print STDERR $self -> GetWarningHeader() . " -       Dumping " . $o_mrna->GetpkExtended() ."\n" if ( $o_annot -> GetDebug());
		my $o_cds     = $o_mrna -> GetCDS();
		if ( !defined ( $o_cds ) )
		{
			warn $self -> GetWarningHeader() . "No cds for " . $o_mrna->GetpkExtended() ." \n";
		}
		my ($locus_tag,$gene_name,$accession,$note) = $self->__GetMrnaInfo($o_gene,$o_mrna,$o_cds);

		my $o_exon_set= $o_mrna -> GetExonSet();
		if ( ! defined $o_exon_set )
		{
			my ($s,$e) = $self -> GetEnds();
			my $locus_tag = $self -> GetLocusTag();
			throw Error::Simple ( "ERROR - No exons for this mrna: $locus_tag [$s..$e]" );
		}
		
		my $sseq      = $o_exon_set -> GetSequence();
		my $len       = $o_mrna -> GetSequenceLength();
		my $strand    = $o_mrna -> GetStrand();
		my ($begin,$end) = $o_mrna -> GetEnds();
		next if ( $len == 0 );
		
		my $gn = ( $gene_name ne '' )  ? " gn=$gene_name" : "";
		$accession = "mrna:$accession";# if ( $accession !~ /^mrna:/i );
		if (!$o_annot -> GetNoCleanFastaHeader())
		{
			$accession =~ s/^mrna://i while ($accession =~ /^mrna:/i);
		}
		my $header = ">$accession locus_tag=$locus_tag$gn begin=$begin end=$end len=$len chr=$ctg strand=$strand sp=$sp $note";
		$self -> _PrintFasta ( $fh_out, $header, \$sseq );
		$nb_obj++;
	}
	return $nb_obj;
}

=head2 Function __DumpIntron

 Title        : __DumpIntron
 Function     : Called by __DumpFeature, dumps the introns of the mrna genomic sequence
                Warning: only protein coding genes are considered
                Note: If not defined, the utrs are created (the method CreateUtr is called)

 Returns      : The number of printed sequences
 Args         : $fh_out    : An already opened file
				$o_gene    : A Locus::Gene object
 Globals      : none

=cut

sub __DumpIntron
{
	my $self             = shift;
	my($fh_out, $o_gene) = @_;
	return 0 unless ( $o_gene -> IsSelected() );
	return 0 unless ( $o_gene -> IsProteinCoding() );

	my $o_annot = $o_gene -> GetAncestor();
	my $ctg     = $o_annot -> GetChromosomeLabel();
	my $sp      = $o_annot -> GetSpecies();
	
	my $nb_obj = 0;
	my $ra_mrnas = $o_gene -> { __a_mrnas };
	foreach my $o_mrna ( @$ra_mrnas )
	{
		next unless ($o_mrna -> IsSelected() );
		print STDERR $self -> GetWarningHeader() . " -       Dumping the introns of " . $o_mrna->GetpkExtended() ."\n" if ( $o_annot -> GetDebug());
		my $o_cds     = $o_mrna -> GetCDS();
		if ( !defined ( $o_cds ) )
		{
			warn $self -> GetWarningHeader() . "No cds for " . $o_mrna->GetpkExtended() ." \n";
			return;
		}
		
		my $strand    = $o_mrna -> GetStrand();
		my ($locus_tag,$gene_name,$accession,$note) = $self->__GetMrnaInfo($o_gene,$o_mrna,$o_cds);
		# clean prefixes 'mrna:' in mRNA accession
		if (!$o_annot -> GetNoCleanFastaHeader())
		{
			$accession =~ s/^mrna://i while ($accession =~ /^mrna:/i);
		}
		my $gn = ( $gene_name ne '' )  ? " gn=$gene_name" : "";
		my $tmpl_header = ">ACCESSION locus_tag=$locus_tag$gn begin=BEGIN end=END len=LEN chr=$ctg strand=$strand sp=$sp $note";

		my @a_introns = $o_mrna -> GetIntrons();
		my $cpt=0;
		foreach my $o_intron (@a_introns)
		{
			my $sseq      = $o_intron -> GetSequence();
			my $len       = $o_intron -> GetSequenceLength();
			next if ( $len == 0);
			$cpt++;
			my ($begin,$end) = $o_intron -> GetEnds();
			my $header = $tmpl_header;
			$header =~ s/>ACCESSION />${accession}i${cpt} /;
			$header =~ s/begin=BEGIN /begin=${begin} /;
			$header =~ s/end=END /end=${end} /;
			$self -> _PrintFasta ( $fh_out, $header, \$sseq );
			$nb_obj++;
		}
	}
	return $nb_obj;
}

=head2 Function __DumpUtr

 Title        : __DumpUtr
 Function     : Called by __DumpFeature, dumps the utr of the mrna genomic sequence
                Warning: only protein coding genes are considered
                Note: If not defined, the utrs are created (the method CreateUtr is called)
 Returns      : The number of printed sequences
 Args         : $type = five_prime_utr OR three_prime_utr
                $fh_out    : An already opened file
				$o_gene    : A Locus::Gene object
 Globals      : none

=cut

sub __DumpUtr
{
	my $self             = shift;
	my($utr_type,$fh_out, $o_gene) = @_;
	return 0 unless ( $o_gene -> IsSelected() );
	return 0 unless ( $o_gene -> IsProteinCoding() );

	my $o_annot = $o_gene -> GetAncestor();
	my $ctg     = $o_annot -> GetChromosomeLabel();
	my $sp      = $o_annot -> GetSpecies();
	
	my $nb_obj = 0;
	my $ra_mrnas = $o_gene -> { __a_mrnas };
	foreach my $o_mrna ( @$ra_mrnas )
	{
		next unless ($o_mrna -> IsSelected() );
		print STDERR $self -> GetWarningHeader() . " -       Dumping the utr of " . $o_mrna->GetpkExtended() ."\n" if ( $o_annot -> GetDebug());
		my $o_cds     = $o_mrna -> GetCDS();
		my $strand    = $o_mrna -> GetStrand();
		my ($locus_tag,$gene_name,$accession,$note) = $self->__GetMrnaInfo($o_gene,$o_mrna,$o_cds);
		# clean prefixes 'mrna:' in mRNA accession
		if (!$o_annot -> GetNoCleanFastaHeader())
		{
			$accession =~ s/^mrna://i while ($accession =~ /^mrna:/i);
		}
		my $gn = ( $gene_name ne '' )  ? " gn=$gene_name" : "";
		my $tmpl_header = ">ACCESSION type=TYPE locus_tag=$locus_tag$gn begin=BEGIN end=END len=LEN chr=$ctg strand=$strand sp=$sp $note";

		my $cpt=0;
		my $o_utr = $o_mrna -> GetUtr($utr_type);
		next if ( !defined ( $o_utr) );
		
		foreach my $o_utr_region ($o_utr->GetRegions())
		{
			my $sseq = $o_utr_region -> GetSequence();
			my $len  = $o_utr_region -> GetSequenceLength();
			next if ($len == 0);
			$cpt++;
			my ($begin,$end) = $o_utr_region -> GetEnds();
			my $header = $tmpl_header;
			$header =~ s/>ACCESSION />${accession}u${cpt} /;
			$header =~ s/begin=BEGIN /begin=${begin} /;
			$header =~ s/end=END /end=${end} /;
			$header =~ s/type=TYPE /type=${utr_type} /;
			$self -> _PrintFasta ( $fh_out, $header, \$sseq );
			$nb_obj++;
		}
	}
	return $nb_obj;
}

=head2 __GetMrnaInfo

 This little procedure is used by __DumpMrna and __DumpIntron
 
=cut
sub __GetMrnaInfo
{
	my $self = shift;
	my ($o_gene,$o_mrna,$o_cds) = @_;

	my ($locus_tag,$gene_name,$accession,$note);
	$locus_tag = $o_mrna -> GetLocusTag();
	$locus_tag = $o_gene -> GetLocusTag   if ( $locus_tag eq "" );
	$locus_tag = $o_mrna -> GetGeneName() if ( $locus_tag eq "" );
	$locus_tag = $o_gene -> GetNarcisseId() if ( $locus_tag eq "" ); 
	$gene_name = $o_mrna -> GetGeneName(&TRUE);
	# TODO - Remettre cette chose-la !
	#my $accession = ($protein_id eq 'MRNA') ? $o_mrna -> GetAccession() : $o_cds -> GetAccession();
	$accession  = "";
	if (defined $o_cds )
	{
		$accession = $o_cds -> GetProteinId();
	}
	$accession    = $o_mrna-> GetTranscriptId() if ( $accession eq '');
	$accession    = $o_mrna-> GetId() if ( $accession eq "");
	
	my $note = $o_mrna -> GetConsolidatedProduct();
	$note    = $o_mrna -> GetConsolidatedNote() if ( $note eq '');
	return ($locus_tag,$gene_name,$accession,$note);
}

=head2 Function __DumpNcrna

 Title        : __DumpNcrna
 Function     : Called by __DumpFeature, dumps the ncrna genomic sequence
                ncRna declared at top level as well as Ncrna declared inside a gene are considered
 Returns      : The number of printed sequences
 Args         : $fh_out    : An already opened file
				$o_locus   : A Locus::Gene object OR a Locus::Ncrna object
 Globals      : none

=cut

sub __DumpNcrna
{
	my $self             = shift;
	my($fh_out, $o_locus) = @_;
	return 0 unless ( $o_locus -> IsSelected() );

	my $o_annot = $o_locus -> GetAncestor();
	my $ctg     = $o_annot -> GetChromosomeLabel();
	my $sp      = $o_annot -> GetSpecies();
	#my $kingdom = $o_annot -> GetKingdom();

#	my $create_cds = $o_annot -> GetCreateCds();
	
	my $nb_obj = 0;
	if ( $o_locus -> isa ( "Locus::Gene" ) )
	{
		my $ra_ncrnas = $o_locus -> GetNcrnas();
		foreach my $o_ncrna ( @$ra_ncrnas )
		{
			$nb_obj += $self -> __DumpNcrna1( $fh_out, $o_ncrna, $o_annot -> GetDebug() );
		}
	}
	else
	{
		$nb_obj += $self -> __DumpNcrna1( $fh_out, $o_locus, $o_annot -> GetDebug() );
	}
	return $nb_obj;
}

sub __DumpNcrna1
{
	my $self             = shift;
	my($fh_out, $o_ncrna, $debug) = @_;
	my $o_annot   = $o_ncrna -> GetAncestor();

	return 0 unless ($o_ncrna -> IsSelected() );
	print STDERR $self -> GetWarningHeader() . " -       Dumping " . $o_ncrna->GetpkExtended() ."\n" if ( $o_annot -> GetDebug() );
		
	my $locus_tag = $o_ncrna -> GetLocusTag();
	$locus_tag = $o_ncrna -> GetGeneName() if ( $locus_tag eq "" );
	my $gene_name = $o_ncrna -> GetGeneName(&TRUE);

	my $accession  = $o_ncrna -> GetTranscriptId();  # oui ca existe meme pour un ncrna
	$accession    = $o_ncrna-> GetId() if ( $accession eq "");
		
	my $note  = $o_ncrna -> GetProduct(&TRUE);
	$note     = $o_ncrna -> GetNote(&TRUE) if ( $note eq '');

	my $sseq      = $o_ncrna -> GetSequence();
	my $len       = $o_ncrna -> GetSequenceLength();
	my $strand    = $o_ncrna -> GetStrand();
	my ($begin,$end) = $o_ncrna -> GetEnds();
	my $ctg     = $o_annot -> GetChromosomeLabel();
	my $sp      = $o_annot -> GetSpecies();
	
	next if ( $len == 0 );
	my $gn = ( $gene_name ne '' )  ? " gn=$gene_name" : "";
	$accession = "ncrna:$accession"; # if ( $accession !~ /^ncrna:/i );
	# Clean prefixes "ncrna:"
	if (!$o_annot -> GetNoCleanFastaHeader())
	{
		$accession =~ s/^ncrna://i while ($accession =~ /^ncrna:/i);
	}
		
	my $header = ">$accession locus_tag=$locus_tag$gn begin=$begin end=$end len=$len chr=$ctg strand=$strand sp=$sp $note";
	$self -> _PrintFasta ( $fh_out, $header, \$sseq );
	return 1;
}
	
=head2 Function __DumpGene

 Title        : __DumpMrna
 Function     : Called by __DumpFeature, dumps the mrna genomic sequence
                Warning: ALL genes are considered, protein coding or not
 Returns      : The number of printed sequences (0 or 1)
 Args         : $fh_out    : An already opened file
				$o_gene    : A Locus::Gene object
 Globals      : none

=cut
	
sub __DumpGene
{
	my $self             = shift;
	my($fh_out, $o_gene) = @_;
	return 0 unless ( $o_gene -> IsSelected() );
	#return 0 unless ( $o_gene->IsProteinCoding() );

	my $o_annot = $self -> { _o_annot };
	print STDERR $self -> GetWarningHeader() . " -       Dumping " . $o_gene->GetpkExtended() . "\n" if ( $o_annot -> GetDebug());
	my $ctg     = $o_annot -> GetChromosomeLabel();
	my $sp      = $o_annot -> GetSpecies();
	#my $kingdom = $o_annot -> GetKingdom();
	
	my $gene_name = $o_gene -> GetGeneName(&TRUE);
	my $locus_tag = $o_gene -> GetLocusTag();
	$locus_tag    = $o_gene -> GetGeneName() if ( $locus_tag eq '' );
	
	# Try first the accession if it exists, else the gene name, else the locus tag or the NarcisseId
	#my $accession = $o_gene -> GetAccession();
	my $accession    = $o_gene -> GetGeneName();
	$accession    = $o_gene -> GetId()       if ( $accession eq '');
	
	my $note  = $o_gene -> GetProduct(&TRUE);
	$note     = $o_gene -> GetNote(&TRUE) if ( $note eq '');

	my $sseq      = $o_gene -> GetSequence();
	my $len       = $o_gene -> GetSequenceLength();
	my $strand    = $o_gene -> GetStrand();
	my ($begin,$end) = $o_gene -> GetEnds();
	return 0 if ( $len == 0);
	my $gn     = ( $gene_name ne '' )  ? " gn=$gene_name" : "";
	$accession = "gene:$accession";# if ( $accession !~ /^gene:/i );
	
	# Clean prefixes "gene:"
	if (!$o_annot -> GetNoCleanFastaHeader())
	{
		$accession =~ s/^gene://i while ($accession =~ /^gene:/i);
	}

	my $header = ">$accession locus_tag=$locus_tag$gn begin=$begin end=$end len=$len chr=$ctg strand=$strand sp=$sp $note";
	$self -> _PrintFasta( $fh_out, $header, \$sseq);
	return 1;
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;

