#
# $Id: Narcisse.pm 1288 2013-10-21 08:39:57Z manu $
#

package AnnotatedSeqFileWriter::Narcisse;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=head1 NAME

AnnotatedSeqFileWriter::Narcisse - Dumping AnnotatedSeqFile objects for Narcisse

=head1 SYNOPSIS

 my $o_dumper = New AnnotatedSeqFileWriter::Narcisse( o_annot => $o_annot );

=head1 DESCRIPTION

 this class extends the AnnotatedSeqFileWriter, it dumps data from AnnotatedSeqFile to Narcisse files

=cut

use strict;
use warnings;
use Carp;

use Error qw(:try);
use Bio::Seq;
use Data::Dumper;
use General;

use Locus;
use base qw( AnnotatedSeqFileWriter );

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1288 $ =~ /\d+/g); $r[0]};
}

sub _Init
{
	my $self = shift;

	$self->SUPER::_Init(@_);
	$self -> {_rh_fasta_sequences} = {};
}
sub GetFormat
{
	return 'NARCISSE';
}
=head2 procedure Dump

 Title        : Dump
 Usage        : $o_annot_writer -> Dump( %h_params );
 Procedure    : Dump data to Narcisse-formatted files
 
 Args         : %h_params : a hash with named parameters
                Mandatory: 'repeat'    => "repeat_filename"
                Mandatory: 'rna'       => "rna_filename"
                Mandatory: 'exon'      => "exon_filename"
                Mandatory: 'gene'      => "gene_filename"
                Mandatory: 'transcript'=> "transcript_filename"
                Mandatory: 'protein_id'=> "MRNA|CDS"

=cut
sub Dump
{
	my $self = shift;
	my (%h_params) = (-append => 0, @_);
	my $open_mode  = ($h_params{ -append }!=0)? 'a' : 'w';
	
	my $o_annot = $self -> { _o_annot };
	#if ( $o_annot -> GetLength() == 0 )
	#{
		#throw Error::Simple( "Cannot generate NARCISSE files, there is no sequence" );
	#}
	
	my $o_annot = $self -> { _o_annot };
	
	my $ra_sorted_loci = $o_annot -> GetSortedLoci(&TRUE);

	# Get and parse the file name, because the *REAL* file names is rebuilt
	my $seq_filename = $o_annot -> GetOutputFile();
	
	my ($filename,$ext,$comp_ext) = ("","","");
	if ( &GetCompressPath ( $seq_filename ) ne "" )
	{
		($filename, $comp_ext) = ( $seq_filename=~/^(.+)(\.\w+)$/ )
	}
	else
	{
		$filename = $seq_filename;
	}

	if ( $filename =~ /\./ )
	{
		($filename,$ext) = ( $filename=~/^(.+)(\.\w+)$/ )
	}

	my %h_headers = (
		repeat     => [ qw( chrom element symbol start end strand sub_type ) ],
		ncrna      => [ qw( chrom element symbol start end strand sub_type ) ],
		gene       => [ qw( chrom element locus_tag symbol start end strand ) ],
		transcript => [ qw( chrom element parent locus_tag symbol accession start end strand nb_exons start_translation stop_translation protein_id regions consistency )],
		exon       => [ qw( chrom element parent symbol start end strand ) ],
		protein_id => [ qw( chrom element locus_tag gene_name protein_id len strand consistency product iprscan ) ]
	);
	
	foreach my $type ('repeat','ncrna','gene','transcript','protein_id','exon')
	{

		#
		# NOTE - Using FileCompressedAppend here, if you keep a ref to the objects after my death, you'll avoid
		#        Compression/decompression steps
		my $f_out    = $filename . '_' . $type . $ext . $comp_ext;
		my $o_comp   = New FileCompressedAppend ( $f_out, $open_mode );
		$f_out       = $o_comp->GetFileName();
		unlink ($f_out) if ( $open_mode eq 'w' );
		my $fh_out   = new IO::File( $f_out, $open_mode ) || throw(LipmError::IOException($f_out,'f'));
		print STDERR $self -> GetWarningHeader() . " - Dumping $type to $f_out\n" if ( $o_annot -> GetDebug());
		if ( $open_mode eq 'w' )
		{
			my $ra_headers = $h_headers { $type };
			print $fh_out join ( "\t", @$ra_headers ) . "\n";
		}

		foreach my $o_locus (@$ra_sorted_loci)
		{
			next unless ( $o_locus -> IsSelected() );
			
			if ( $type eq 'repeat' )
			{
				next unless ($o_locus -> isa ( 'Locus::Repeat' ));
				print $fh_out join( "\t", $self -> __GetLocus ( $o_locus ) ) . "\n";
			}
			
			elsif ( $type eq 'ncrna' )
			{
				if ( $o_locus->isa('Locus::Ncrna') )
				{
					print $fh_out join( "\t", $self -> __GetLocus ( $o_locus ) ) . "\n";
				}
				elsif ( $o_locus->isa('Locus::Gene') )
				{
					my $ra_ncrnas = $o_locus->GetNcrnas();
					foreach my $o_ncrna ( @$ra_ncrnas )
					{
						print $fh_out join( "\t", $self -> __GetLocus ( $o_locus ) ) . "\n";
					}
				}
			}
			elsif ( $type eq 'gene' )
			{
				next unless ( $o_locus->GetSubType() eq 'protein_coding_gene' );
				print $fh_out join( "\t", $self -> __GetGene ( $o_locus ) ) . "\n";
			}
			elsif ( $type eq 'transcript' )
			{
				next unless ( $o_locus->GetSubType() eq 'protein_coding_gene' );
				my $ra_mrnas = $o_locus -> GetMrnas();
				foreach my $o_mrna ( @$ra_mrnas )
				{
					my $o_cds = $o_mrna -> GetCDS();
					# Skip mRNAS without CDS (should not exist)
					if ( !defined ( $o_cds ) )
					{
						my ($s,$e) = $o_mrna->GetEnds();
						warn $self -> GetWarningHeader() . "No cds for this mrna ".$o_mrna->GetLocusTag()." [$s..$e]\n";
					}
					else
					{
						print $fh_out join ( "\t", $self -> __GetMrna ( $o_mrna ) ) . "\n";
					}
				}
			}
			elsif ( $type eq 'protein_id' )
			{
				next unless ( $o_locus->GetSubType() eq 'protein_coding_gene' );
				my $ra_mrnas = $o_locus -> GetMrnas();
				foreach my $o_mrna ( @$ra_mrnas )
				{
					my $o_cds = $o_mrna -> GetCDS();
					# Skip mRNAS without CDS (should not exist) but don't warn (already done)
					if ( defined $o_cds )
					{
						if ( $o_cds -> GetConsistency() ne 'I' )
						{
							print $fh_out join ( "\t", $self -> __GetMrna ( $o_mrna, &TRUE ) ) . "\n";
						}
					}
				}
			}
			elsif ( $type eq 'exon' )
			{
				next unless ( $o_locus->GetSubType() eq 'exon' );
				print $fh_out join ( "\t", $self -> __GetExon ( $o_locus ) ) . "\n";
			}
		}
	}
}

=head2 procedure __GetLocus

 Title    : __GetLocus
 Usage    : my @a_data = $self -> __GetLocus( $o_locus )
 Function : Retrieve data corresponding to a generic locus
 
 Args     : $o_locus A Locus object
 Return   : An array (the columns)

=cut
sub __GetLocus
{
	my $self     = shift;
	my ($o_locus) = @_;
	
	my $chr          = $o_locus -> GetChromosomeLabel();
	my $uid          = $o_locus -> GetNarcisseId();
	my $locus_tag    = $o_locus -> GetLocusTag();
	$locus_tag       = $o_locus -> GetNarcisseId() if ( $locus_tag eq '');
	my ($start,$end) = $o_locus -> GetEnds();
	my $strand       = $o_locus -> GetStrand();
	my $sub_type      = $o_locus -> GetSubType();

	return ( $chr, $uid, $locus_tag, $start, $end, $strand, $sub_type );
}

=head2 procedure __GetGene

 Title    : __GetGene
 Usage    : my @a_data = $self -> __GetGene( $o_gene )
 Function : Retrieve data corresponding to a gene
 
 Args     : $o_locus A Locus::Gene object
 Return   : An array (the columns)

=cut			
sub __GetGene
{
	my $self = shift;
	my ($o_gene) = @_;

	my $chr          = $o_gene -> GetChromosomeLabel();
	my $uid          = $o_gene -> GetNarcisseId();
	my $locus_tag    = $o_gene -> GetLocusTag();
	my $gene_name    = $o_gene -> GetGeneName();
	$gene_name       = $locus_tag if ( $gene_name eq "");
	$locus_tag       = $gene_name if ( $locus_tag eq "");
	my ($start,$end) = $o_gene -> GetEnds();
	my $strand       = $o_gene -> GetStrand();
	return ( $chr, $uid, $locus_tag, $gene_name, $start, $end, $strand );
}

=head2 procedure __GetMrna

 Title    : __GetMrna
 Usage    : my @a_data = $self -> __GetMrnas( $o_mrna )
 Function : Retrieve data relative to the mrna
 
 Args     : $o_mrna An mrna object
            $prot_flg = If &TRUE, return data for the protein file
            $protein_id How to set the protein id ? From CDS or from MRNA - PAS UTILISE !!!
 Return   : An array (the columns)
 
=cut
sub __GetMrna
{
	my $self     = shift;
	my ($o_mrna,$prot_flg) = @_;
	$prot_flg = &FALSE unless ( defined $prot_flg );
	
	my $o_gene = $o_mrna -> GetParent();
	my $uid = $o_gene -> GetNarcisseId();
	my $chr = $o_mrna -> GetChromosomeLabel();

	my $locus_id = $o_mrna -> GetLocusTagTranscriptNb();
	my $o_cds    = $o_mrna -> GetCDS();
	if ( ! defined($o_cds) )
	{
		warn "WARNING - dump Mrnas: $locus_id, Inconsistency between mRNA and CDS detected, removed...\n";
		return '';
	}
	
	my $o_annot = $self -> { _o_annot };
	my $debug   = $o_annot -> GetDebug();
	my $sequence_loaded = ($o_annot -> GetLength()!=0);
	
	# The translation should be done (if possible) in order to know the 'C/F/I' status of the transcript
	$o_cds -> GetTranslation ( $debug ) if ( $sequence_loaded );

	# The protein id should be the cds accession, unless we want the mrna accession
	#my $prot_id = ($protein_id eq 'MRNA') ? $o_mrna->GetAccession() : $o_cds -> GetAccession();
	my $prot_id = $o_cds -> GetProteinId();
	#$prot_id = $o_cds  -> GetAccession()    if ( $prot_id eq '');
	$prot_id = $o_mrna -> GetTranscriptId() if ( $prot_id eq '');
	#$prot_id = $o_mrna -> GetAccession()    if ( $prot_id eq "" );
	$prot_id = $o_mrna -> GetLocusTag()     if ( $prot_id eq "" );
	$prot_id = $o_gene -> GetLocusTag()     if ( $prot_id eq "" );
	$prot_id = $o_gene -> GetGeneName()     if ( $prot_id eq "" );

	# The accession is the mrna accession if possible, $prot_id if not possible
	my $accession = $o_mrna->GetTranscriptId();
	#$accession = $o_mrna -> GetAccession() if ( $accession eq '' );
	$accession = $prot_id  if ( $accession eq '' );
	
	# The gene name is the mrna gene name
	my $gene_name = $o_mrna -> GetGeneName(&TRUE);  # The gene name, or nothing
	
	# The transcript and protein id
	my $tuid        = $o_mrna -> GetNarcisseId();
	my $puid        = $tuid;
	$puid =~ s/t(\d+)$/p$1/;

	my ($start,$end) = $o_mrna -> GetEnds();
	my $strand       = $o_mrna -> GetStrand();

	if ( ! $prot_flg )
	{
		# my ($start_trans,$stop_trans,$nb_exons)=();
		
		# ACCES A UN MEMBRE PRIVE DE Locus::Mrna PAS BEAU - MANU
		# MAIS C'EST PLUS LA PEINE !!!
		# JG/API currently only defined in Ensembl.pm
		#if ( defined($o_mrna) && defined($o_mrna->{'__cds_start'}) ) 
		#{
		#	($start_trans,$stop_trans,$nb_exons) = ( $o_mrna->{'__cds_start'}, $o_mrna->{'__cds_end'}, $o_mrna->{'__nb_exons'});
		#}
	
		# that is why I must keep the older way to get the cds_start
		#else 
		#{
		#	($start_trans,$stop_trans,$nb_exons) = $o_mrna -> GetTranslationStartStop();
		#}
		my ($start_trans,$end_trans) = $o_mrna -> GetTranslationEnds();
		
		# Warning GetTranslationEnds may change the consistency, thus consistency MUST be checked here
		my $consistency = $o_cds  -> GetConsistency();
		$consistency = 'F' if ( $consistency =~ /F/ );  # FS => F, FE => F, FSFE=> F
		my $o_exon_set= $o_mrna -> GetExonSet();
		if ( ! defined $o_exon_set )
		{
			my ($s,$e) = $self -> GetEnds();
			my $locus_tag = $self -> GetLocusTag();
			throw Error::Simple ( "ERROR - No exons for this mrna: $locus_tag [$s..$e]" );
		}
		my $regions     = $o_exon_set -> GetRegions();
		my $nb_exons    = ($start_trans==-1)? -1: $o_exon_set -> GetNbRegions();
		return ($chr,$tuid,$uid,$locus_id,$gene_name,$accession,$start,$end,$strand,$nb_exons,$start_trans,$end_trans,$prot_id,$regions,$consistency);
	}
	else
	{
		my $len     = $o_cds -> GetProteinSequenceLength();
		my $note   = $o_mrna -> GetProduct();
		$note     .= $o_mrna -> GetNote(&TRUE) if ( $note eq '');
		my $consistency = $o_cds -> GetConsistency();
		$consistency = 'F' if ( $consistency =~ /F/ );  # FS => F, FE => F, FSFE=> F
		return ($chr,$puid,$locus_id,$gene_name,$prot_id,$len,$strand,$consistency,$note,'');
	}
}

=head2 procedure __GetExon

 Title    : __GetExon
 Usage    : my @a_data = $self -> __GetExon( $o_mrna )
 Function : Retrieve data relative to the exons
 Args     : $o_exon A Locus::Exon object
 Return   : An array (the columns)

=cut
sub __GetExon
{
	my $self            = shift;
	my ($o_exon) = @_;

	my $chr    = $o_exon -> GetChromosomeLabel();
	
	# PAS JOLI-JOLI dis donc !
	#my $o_mrna = $o_exon -> GetParent();
	#$o_mrna = $o_mrna -> GetParent() if ( ! $o_mrna -> isa ( 'Locus::Mrna' ));
	
	my $o_mrna = $o_exon -> GetParent() -> GetParent();
	my $tuid   = $o_mrna -> GetNarcisseId();
	my $uid    = $o_exon -> GetNarcisseId();
	my ($exon_nb) = ($uid=~/e(\d+)$/);
	my $locus_id = $o_mrna -> GetLocusTagTranscriptNb();
	my $locuscds_id = $locus_id . '_' . $exon_nb;
	my ( $start, $end ) = $o_exon -> GetEnds();
	my $strand          = $o_exon -> GetStrand();
	
	return ( $chr, $uid, $tuid, $locuscds_id, $start, $end, $strand );
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;


