#
# $Id: UsingBioPerl.pm 1354 2014-01-30 09:09:01Z sallet $
#

package AnnotatedSeqFileWriter::UsingBioPerl;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=head1 NAME

AnnotatedSeqFileWriter::UsinbioPerl - Dumping AnnotatedSeqFile objects in some file, using the BioPerl API

=head1 SYNOPSIS

=head1 DESCRIPTION

 this class extends the AnnotatedSeqFileWriter, it dumps data from AnnotatedSeq to genbank/embl/... files

=cut

use strict;
use warnings;
use Carp;

use Error qw(:try);
use LipmError::IOException;

use Bio::SeqIO;
use Bio::Seq;
use Bio::SeqFeature::Generic;
use Bio::Location::Simple;
use Bio::Location::Split;
use Bio::Seq::RichSeq;
use Data::Dumper;
use General;
use GenbankTags;

use Locus;
use base qw( AnnotatedSeqFileWriter );

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1354 $ =~ /\d+/g); $r[0]};
}

=head2 procedure _Init

 Title        : _Init
 Usage        : my $o_gbk = New AnnotatedSeqFile::Embl(...);
 Function     : Constructor
 Args         : $o_annot The AnnotatedSeq object inside which we live
                $format  The file format (currently 'embl', 'genbank')

=cut

sub _Init
{
	my $self = shift;
	my ($o_annot,$format) = @_;

	$self -> SUPER::_Init($o_annot);
	$self -> { '_format' } = $format;
	my $o_genbank_tags = New GenbankTags;
	$self -> { __o_genbank_tags } = $o_genbank_tags;
}

sub _NewSeqIO
{
	my $self = shift;
	my ($format,$fh_out) = @_;
	return new Bio::SeqIO(-format => $format, -fh => $fh_out);
}

=head2 procedure Dump

 Title     : Dump
 Usage     : $o_annot_writer -> Dump( );
 Procedure : Dump data to gbk/embl files
 Args      : -append 1/0 (optional, def 0)

=cut
sub Dump
{
	my $self = shift;
	my (%h_params) = (-append => 0, @_);
	
	my $open_mode = ($h_params{ -append } == 0) ? 'w' : 'a';
	my $o_annot = $self -> { _o_annot };
	my $is_sequence_loaded = ($o_annot -> GetLength()!=0);
	$self -> { __is_sequence_loaded } = $is_sequence_loaded;

	# Open output file
	my $f_out  = $o_annot -> GetOutputFile();
	my $o_comp   = New FileCompressedAppend ( $f_out, $open_mode );
	$f_out       = $o_comp->GetFileName();
	unlink $f_out if ( $open_mode eq 'w' );
	my $fh_out   = new IO::File( $f_out, $open_mode ) || throw LipmError::IOException($f_out,'f');
	my $format   = $self -> { '_format' };
	#my $seqstream = $self -> _NewSeqIO($format,$fh_out);
	my $seqstream = $self -> _NewSeqIO($format,$fh_out);

	# Dump the information from the header
	my $rh_headers = $o_annot -> GetHeaders();
	
	# NOTE - Those annotations must be declared when constructing the object !
	# See __DumpHeader for other top annotations
	$h_params{ '-is_circular' } = ($rh_headers->{ topology } eq 'circular') if ( defined $rh_headers->{ topology });
	$h_params{ '-molecule'    } = $rh_headers->{ molecule } if ( defined $rh_headers->{ molecule });
	$h_params{ '-division'    } = $rh_headers->{ division } if ( defined $rh_headers->{ division });
	my $seq = new Bio::Seq::RichSeq(%h_params);
	
	my $sv = ( defined $rh_headers->{ version } ) ? $rh_headers->{ version } : 1;
	$seq->version( $sv );

	$self -> { __seq } = $seq;

	$self -> __DumpSource ();
	
	# Sort the biological objects in increasing order
	my $ra_sorted_loci = $o_annot -> GetSortedLoci();
	#my $create_cds = $o_annot -> GetCreateCds();
	foreach my $o_locus ( @$ra_sorted_loci )
	{
		# Errors are catched and ignored
		try
		{
			if ( $o_locus -> isa('Locus::Gene') )
			{
				$self -> __DumpGene($o_locus);
				
				# If a coding gene...
				if ( $o_locus -> GetSubType() eq 'protein_coding_gene' )
				{
					#... print the mrnas, with their exon decomposition
					my $ra_mrnas = $o_locus -> GetMrnas();
					foreach my $o_mrna ( @$ra_mrnas )
					{
						$self -> __DumpMrna ( $o_mrna );
					}
				}
				
				# dump the NcRna, whether coding gene or not
				my $ra_ncrnas = $o_locus -> GetNcrnas();
				foreach my $o_ncrna ( @$ra_ncrnas )
				{
					$self -> __DumpNcrna($o_ncrna);
				}
				
				# EXPERIMENTAL
				# Dump the "invisible" RNA, ie the exons defined alone in the gene
				my $o_exonset= $o_locus -> GetExonSet( );
				if (defined($o_exonset))
				{
					$self -> __DumpExonSet( $o_exonset, 0 );
				}

				my $ra_regions = $o_locus -> GetMiscRegions();
				foreach my $o_region (@$ra_regions)
				{
					$self -> __DumpLocus($o_region);
				}
			}
			elsif ( $o_locus -> isa('Locus::Ncrna') )
			{
				$self -> __DumpNcrna($o_locus);
			}
			else
			{
				$self -> __DumpLocus( $o_locus );
			}
		}
		catch Error with
		{
			my $error = shift;
		};
	}
	$self -> __DumpHeader();
	$seqstream->write_seq($seq);
}

=head2 procedure __DumpHeader

 Title     : __DumpHeader
 Usage     : $o_annot_writer -> __DumpHeader( $seq, $o_annot, $is_sequence_loaded );
 Procedure : Dump header data, called by Dump
             Dump a minimal set of header data, and if any, dump the data found in the header hash maintained by $o_annot

=cut
sub __DumpHeader
{
	my $self = shift;
	my ($seq, $o_annot, $is_sequence_loaded) = @$self { qw( __seq _o_annot __is_sequence_loaded ) };
	
	# Dump file header
	my $accession = $o_annot -> GetChromosomeLabel();
	if ( $accession eq '' ) 
	{
		$accession = $o_annot -> GetContigLabel();
	}
	else
	{
		$seq->accession_number($accession);
	}
	$seq->display_name($accession);
	if ( $is_sequence_loaded == &TRUE )
	{
		my $r_seq = $o_annot -> GetGenomicSequence();
		$seq -> seq ( $$r_seq, 'DNA' );
	}

	# Dump the information from the header
	my $rh_headers = $o_annot -> GetHeaders();
	
	# If nothing in headers, return
	my @a_header_keys = keys %$rh_headers;
	return if ( @a_header_keys == 0);

	# Line DE
	my $de = ( defined $rh_headers->{ desc } ) ? $rh_headers->{ desc } : "File generated by BioFileConverter";
	$seq->description( $de );
	
	# Lines OC, OS (taxonomy and species info)
	my $taxonomy = $rh_headers  -> {taxonomy};
	my $organism = $rh_headers  -> {organism};
	my @a_taxonomy_elements;
	@a_taxonomy_elements = split (";",$taxonomy) if (defined($taxonomy)); #OC line
    my @a_classification = ($organism, @a_taxonomy_elements); #OS, OC lines
    my $o_species = Bio::Species->new(-classification => \@a_classification );
	$seq->species($o_species);

	my $o_annot_coll = Bio::Annotation::Collection->new();
	
	my $cpt=1;
	while ( exists $rh_headers->{ "ref_${cpt}_location" } )
	{
		my %h_params;
		foreach my $k ( @a_header_keys )
		{
			next unless ($k =~ /^ref_${cpt}/);
			my ($k1) = ($k =~ /ref_${cpt}_(.+)/);
			$h_params{'-'.$k1}  = $rh_headers -> { $k };
		}
		my $o_annot_ref_publi = Bio::Annotation::Reference->new( %h_params );
		$o_annot_ref_publi->start(1); #RP line
		$o_annot_ref_publi->end($o_annot -> GetLength()); #RP line
		$o_annot_coll->add_Annotation('reference',$o_annot_ref_publi); 
		$cpt++;
	}
	
	# ne marche pas !
	#if ( defined $rh_headers->{ "project" } )
	#{
	#	my $o_project = new Bio::Annotation::SimpleValue->new(-value => "Project:".$rh_headers->{ "project" });
	#	#$o_annot_coll->add_Annotation('project',$o_project);
	#	$o_annot_coll->add_Annotation($o_project);
	#}
	
	$seq->annotation($o_annot_coll);
}

sub __DumpSource
{
	my $self = shift;
	my ($seq, $o_annot) = @$self { qw( __seq _o_annot ) };
	
	my $sstrt = $o_annot -> GetSeqRegionStart();
	my $send  = $o_annot -> GetSeqRegionEnd();
	
	if ( $sstrt >= $send )
	{
		$sstrt = 1;
		$send  = length ($seq->seq);
	}
	
	my %h_params;
	my $rh_headers = $o_annot -> GetHeaders();
	foreach my $k (%$rh_headers)
	{
		next unless $k =~ /^source_/;
		my ($k1) = ($k =~ /^source_(.+)/);
		$h_params{$k1}  = $rh_headers -> { $k };
	}
	
	my $contig = $o_annot -> GetChromosomeLabel();
	$h_params{'note'} = $contig;
	$seq->add_SeqFeature (new Bio::SeqFeature::Generic (-start => $sstrt, -end => $send, -primary_tag => 'source', -tag => \%h_params) ); 
}
	
sub __DumpGene
{
	my $self = shift;
	my ( $o_gene ) = @_;
	return unless ($o_gene->IsSelected());
	
	my $seq  = $self -> { __seq };
	$seq -> add_SeqFeature ( $self -> __NewFeatureAndQualifiers ( 'gene', $o_gene ) );
}

sub __DumpMrna
{
	my $self = shift;
	my ( $o_mrna ) = @_;
	
	my $seq        = $self -> { __seq };
	my $o_annot    = $self -> { _o_annot };
	#my $create_cds = $o_annot -> GetCreateCds();

	if ($o_mrna->IsSelected())
	{ 
		my $f = $self->__NewFeatureAndQualifiers ( 'mRNA', $o_mrna );
		# Print the exons, only if mRNA selected !
		my $o_exonset = $o_mrna -> GetExonSet();
		if (defined $o_exonset && $o_exonset->IsSelected())
		{
			$self -> __GetRegions ( $f, $o_exonset );
			$seq -> add_SeqFeature ( $f );
		}
		elsif (defined $o_exonset)
		{
			warn $self -> GetWarningHeader() . "No exon for " . $o_mrna->GetpkExtended() . "\n";
		}
	}
	
	# If relevant, print the UTR
	my $sub_type1 = ($o_mrna -> GetStrand() > 0) ? 'five_prime_UTR': 'three_prime_UTR';
	my $sub_type2 = ($o_mrna -> GetStrand() > 0) ? 'three_prime_UTR': 'five_prime_UTR';
	my $o_utr1   = $o_mrna -> GetUtr( $sub_type1 );
	my $o_utr2   = $o_mrna -> GetUtr( $sub_type2 );
	
	# print the first UTR
	$self -> __DumpUtr($o_utr1) if ( defined $o_utr1 );
	
	# Print the CDS
	my $o_cds = $o_mrna -> GetCDS();
	if ( !defined ( $o_cds ) )
	{
		warn $self -> GetWarningHeader() . "No cds for " . $o_mrna->GetpkExtended() . "\n";
	}
	else
	{
		$self -> __DumpCds( $o_cds );
	}
	
	# Print the second UTR
	$self -> __DumpUtr($o_utr2) if ( defined $o_utr2 );
}

sub __DumpNcrna
{
	my $self = shift;
	my ( $o_ncrna ) = @_;
	return unless ($o_ncrna->IsSelected());
	
	my $seq  = $self -> { __seq };
	
	my $f = $self -> __NewFeatureAndQualifiers ( $o_ncrna -> GetSubType(), $o_ncrna );
	my $o_exonset = $o_ncrna -> GetExonSet();
	if (defined $o_exonset)
	{
		$self -> __GetRegions ( $f, $o_exonset );
	}
	$seq -> add_SeqFeature ( $f );
}

# EXPERIMENTAL
# Dump the exons, useful ONLY when a Locus::ExonSet is directly integrated to the Gene
# (features Exon in gbk files)

sub __DumpExonSet
{
	my $self = shift;
	my ( $o_exonset, $min_region ) = @_;
	return unless ($o_exonset->IsSelected());
	
	# The default is NOT to dump the exons if there is only 1 region in the ExonSet
	# But this can be changed with the parameter $min_region
	$min_region = 1 unless ( defined $min_region);
	my @a_exons = $o_exonset->GetRegions();
	my $seq  = $self -> { __seq };
	if (@a_exons > $min_region)
	{
	    foreach my $o_region (@a_exons)
		{
			$seq -> add_SeqFeature ( $self -> __NewFeatureAndQualifiers ( 'exon', $o_region ) );
		}
	}

}

sub __DumpCds
{
	my $self = shift;
	my ($o_cds) = @_;
	return unless ($o_cds->IsSelected());

	my $seq  = $self -> { __seq };
	my $is_sequence_loaded = $self -> { __is_sequence_loaded };

	my @a_regions = $o_cds->GetRegions();
	my $f = $self -> __NewFeatureAndQualifiers ( 'CDS', $o_cds );

	if ( $is_sequence_loaded == &TRUE )
	{
		# GetTranslation may change the consistency status of the cds
		# (should not now because it is called earlier, by AnnotatedSeq::Load)
		# DO NOT write translation if we are in a pseudogene (ie NOT 'C')
		my $seq_prot = $o_cds -> GetTranslation();
		if ( $o_cds->Get( 'pseudogene' ) eq '' )
		{
			$f -> add_tag_value ( 'translation', $seq_prot );
		}
	}
	
	# The consistency status of the cds drives __GetRegions !
	$self -> __GetRegions( $f, $o_cds );
	$seq -> add_SeqFeature ( $f );
}

sub __DumpUtr
{
	my $self = shift;
	my ($o_utr) = @_;
	return unless ($o_utr->IsSelected());

	my $seq  = $self -> { __seq };
   	my $o_genbank_tags = $self -> { __o_genbank_tags };

	my $feature = $o_genbank_tags -> So2Feature ( $o_utr -> GetSubType() );
	my @a_regions = $o_utr -> GetRegions();
	my $f = $self -> __NewFeatureAndQualifiers ( $feature, $a_regions[0] );
	$self -> __GetRegions   ( $f, $o_utr );
	$seq -> add_SeqFeature ( $f );
}

sub __DumpLocus{
	my $self = shift;
	my ($o_locus) = @_;
	return unless ($o_locus->IsSelected());

   	my $o_genbank_tags = $self -> { __o_genbank_tags };
	my $seq     = $self -> { __seq };
	
	my $feature = $o_genbank_tags -> So2Feature ( $o_locus -> GetSubType() );
	my $f        = $self -> __NewFeatureAndQualifiers ( $feature, $o_locus );
	$self -> __GetRegions ( $f, $o_locus );
	$seq  -> add_SeqFeature ( $f );
}
=head2 procedure __NewFeatureAndQualifiers

 Title    : __NewFeatureAndQualifiers
 Procedure: Create and return new Bio::SeqFeature::Generic object, add several qualifiers to this object
 Usage    : $self -> __GetQualifiers ( $f, $o_locus )
 Args     : $f A feature name (if '', only create a Bio::SeqFeature::generic without any qualifier)
            $o_locus The considered locus
 Return   : A Bio::SeqFeature::Generic object
 Error    : Throw an Error::Simple if $feature is not recognized

=cut
sub __NewFeatureAndQualifiers
{
	my $self = shift;
	my ($feature, $o_locus) = @_;
	return new Bio::SeqFeature::Generic () if ( $feature eq '' );
	return $self->__NewFeatureAndQualifiersCDS ( @_ ) if ( $feature eq 'CDS' );
	return $self->__NewFeatureAndQualifiersMrna ( @_ ) if ( $feature eq 'mRNA' );
	return $self->__NewFeatureAndQualifiersDefault ( @_ );
}
sub __NewFeatureAndQualifiersDefault
{
	my $self = shift;
	my ($feature, $o_locus) = @_;
	my $sub_type = $o_locus -> GetSubType();
	my $submit = $self -> { _o_annot } -> GetSubmit();

	# SPECIAL TREATMENT FOR THE mobile_element feature
	my $mobile_element_type = '';
	if ( $feature =~ /^mobile_element_(.+)/ )
	{
		$feature = 'mobile_element';
		$mobile_element_type = $1;
		my $mobile_element_name = $o_locus -> Get ( 'mobile_element_name' );
		if ( $mobile_element_name ne '' )
		{
			$mobile_element_type .= ':' . $mobile_element_name;
		}
	}
		
   	my $o_genbank_tags = $self -> { __o_genbank_tags };

	my $f = undef;
	if ( $o_genbank_tags->IsAGenbankFeature( $feature ) )
	{
		$f = new Bio::SeqFeature::Generic ();
		$f -> location ($self->__Locus2Location($o_locus));
		$f -> primary_tag ( $feature );

		$self -> __AddTagValues( $f, 'mobile_element_type', $mobile_element_type ) if ( $mobile_element_type ne '' );
		
		my $gene = $submit ? $o_locus -> GetGeneNameSimple() : $o_locus -> GetGeneName(0);
		$self -> __AddTagValues( $f,'gene',$gene)      if ( $gene ne '' );
	
		my $locus_tag = $o_locus -> GetLocusTag();
		if ( $submit == &FALSE && $locus_tag eq '' )
		{
			$locus_tag = $o_locus -> GetParentLocusTag();
		}
		$self -> __AddTagValues( $f,'locus_tag',$locus_tag)      if ( $locus_tag ne '' );
	
		# In Submit mode, all the notes will be concatenate in one feature
		my $note = ($submit == &TRUE ? $o_locus -> GetNote(1) : $o_locus -> GetNote(0));
		
		$self -> __AddTagValues( $f,'note',$note)      if ( $note ne '' );

		my $ra_inference = $o_locus -> GetAsArrayConsolidated('inference');
		$self -> __AddTagValues( $f,'inference',$ra_inference) if ( @$ra_inference != 0 );
		
		# Add a note to keep the memory of the type, if the sub_type is not a recorded feature
		if ( $o_genbank_tags -> IsAType( $feature, $sub_type ) )
		{
			$self -> __AddTagValues( $f, 'note', "sub_type=$sub_type" ) unless $sub_type eq 'region';
		}
		
		my $product = $o_locus -> GetProduct();
		$self -> __AddTagValues( $f,'product',$product)      if ( $product ne '' );
	
		#my $dbxref = $o_locus -> GetDbXref();
		#$self -> __AddTagValues( $f,'db_xref',$dbxref) if ( $dbxref ne "" );
		my $ra_dbxref = $o_locus -> GetAsArrayConsolidated('Dbxref','ontology_term');
		$self -> __AddTagValues( $f,'db_xref',$ra_dbxref) if ( @$ra_dbxref != 0 );

		my $alias = $o_locus -> GetAlias();
		$self -> __AddTagValues( $f,'gene_synonym',$alias) if ( $alias ne "" );
	
		if ( $o_locus -> can ( 'GetTranscriptId' ) )
		{
			# get the transcript id as a string
			my $transcript_id = $o_locus -> GetTranscriptId(&TRUE);
			$self -> __AddTagValues( $f,'transcript_id',$transcript_id) if ( $transcript_id ne "" );
		}
	
		my $ra_tags = $o_locus -> GetOtherKeys( { inference => 1 } );

		foreach my $tag ( @$ra_tags )
		{
			next if ($tag eq 'partial');	# This tag is written to gff3 files, but NOT READ
			next if ($tag eq 'mobile_element_name'); # This tag is integrated to the mobile_element_type
			my $tag1 = "\u$tag";
			
			# If the tag is a qualifier, ok
			if ( $o_genbank_tags->IsTagAGenbankQualifier ( $feature, $tag ) )
			{
				$self -> __AddTagValues( $f,$tag,$o_locus -> Get ( $tag ));
			}
			
			# if the Capitalized tag is a qualifier, ok (we probably come from gff3)
			elsif ( $o_genbank_tags->IsTagAGenbankQualifier ( $feature, $tag1 ) )
			{
				$self -> __AddTagValues( $f,$tag1,$o_locus -> Get ( $tag ));
			}
			
			# All other cases: wrap this unknown tag inside a note, or skip it in submit mode
			elsif ( $submit == &FALSE )
			{
				my $note = $tag . '=' . $o_locus -> GetAsString ( $tag );
				$self -> __AddTagValues( $f,'note',$note);
				warn ( "WARNING $feature-$tag $tag is NOT a recorded qualifier for $feature - This tag is written to a note !\n" );
			}
			else
			{
				warn ( "WARNING $feature-$tag $tag is NOT a recorded qualifier for $feature and working in submit mode - This tag is SKIPPED !\n" );
			}
		}
	}
	else
	{
		warn ( "WARNING $feature $feature is NOT a recorded feature - This paragraphe is skipped !\n" );
		throw Error::Simple ( "$feature is NOT a recorded feature" );
	}
	return $f;
}

=head2 procedure __NewFeatureAndQualifiersCDS

 Title    : __NewFeatureAndQualifiersCDS
 Procedure: Same as __NewFeatureAndQualifiersDefault, but specialized for CDS: Take most qualifiers from the mRNA

=cut
sub __NewFeatureAndQualifiersCDS
{
	my $self = shift;
	my ($feature, $o_locus) = @_;
	my $o_parent = $o_locus -> GetParent();
	my $o_region0= $o_locus -> GetRegion0();
	my $submit = $self -> { _o_annot } -> GetSubmit();

   	my $o_genbank_tags = $self -> { __o_genbank_tags };

	my $f = undef;
	$f = new Bio::SeqFeature::Generic ();
	$f -> location ($self->__Locus2Location($o_locus));
	$f -> primary_tag ( $feature );

	my $gene = $submit ? $o_locus -> GetGeneNameSimple() : $o_locus -> GetGeneName();
	$self -> __AddTagValues( $f,'gene',$gene)      if ( $gene ne '' );
	
	my $locus_tag = $o_locus -> GetLocusTag();
	
	# If no locus tag ask for the locus tag on first cds region
	if ( $locus_tag eq '' ) 
	{
		$locus_tag = $o_locus -> GetRegion0() -> GetLocusTag();
	}
	if ( $submit == &FALSE && $locus_tag eq '' )
	{
		$locus_tag = $o_locus -> GetParentLocusTag();
	}
	$self -> __AddTagValues( $f,'locus_tag',$locus_tag)      if ( $locus_tag ne '' );
	
	if ($submit == &TRUE)
	{
		my $note = $o_locus -> GetAsStringConsolidated('note',$o_region0,$o_parent);
		$self -> __AddTagValues( $f,'note',$note)      if ( $note ne '' );
	}
	else
	{
		my $ra_note = $o_locus -> GetAsArrayConsolidated('note',$o_region0,$o_parent);
		foreach my $n (@$ra_note)
		{
			$self -> __AddTagValues( $f,'note',$n );
		}
	}

	my $ra_inference = $o_locus -> GetAsArrayConsolidated('inference',$o_region0,$o_parent);
		$self -> __AddTagValues( $f,'inference',$ra_inference) if ( @$ra_inference != 0 );

	my $function = $o_locus -> GetAsStringConsolidated('function',$o_region0,$o_parent);
	$self -> __AddTagValues( $f,'function',$function)      if ( $function ne '' );
		
	my $product = $o_locus -> GetAsStringConsolidated('product',$o_region0,$o_parent);
	$self -> __AddTagValues( $f,'product',$product)      if ( $product ne '' );
	
	my $exception = $o_locus -> Get ( 'exception', $o_region0 );
	$self -> __AddTagValues( $f,'exception',$exception) if ( $exception ne '' );
	
	my $ra_dbxref = $o_locus -> GetAsArrayConsolidated('Dbxref','ontology_term',$o_region0,$o_parent);
	$self -> __AddTagValues( $f,'db_xref',$ra_dbxref) if ( @$ra_dbxref != 0 );
		
	my $alias = $o_locus -> GetAlias();
	$self -> __AddTagValues( $f,'gene_synonym',$alias) if ( $alias ne "" );
	
	my $protein_id = $o_locus -> GetProteinId(&TRUE);
	$self -> __AddTagValues( $f,'protein_id',$protein_id) if ( $protein_id ne "" );

	my $ra_tags = $o_locus -> GetConsolidatedOtherKeys({inference=>1,function=>1, exception=>1},$o_region0,$o_parent);   # Get keys from the CDS and from its parent
	foreach my $tag ( @$ra_tags )
	{
		my $tag1 = "\u$tag";

		my $val = $o_locus->GetAsArrayConsolidated($tag,$o_region0,$o_parent);

		# If the tag is a qualifier, ok
		if ( $o_genbank_tags->IsTagAGenbankQualifier ( $feature, $tag ) )
		{
			$self -> __AddTagValues( $f,$tag,$val);
		}
		
		# if the Capitalized tag is a qualifier, ok (we come from gff3)
		elsif ( $o_genbank_tags->IsTagAGenbankQualifier ( $feature, $tag1 ) )
		{
			$self -> __AddTagValues( $f,$tag1,$val);
		}

		# All other cases: wrap this unknown tag inside a note, or skip it in submit mode
		elsif ( $submit == &FALSE )
		{
			next if ($tag eq 'partial');	# This tag is written to gff3 files, but NOT READ
			my $note = $tag . '=' . join(' ',@$val);
			$self -> __AddTagValues( $f,'note',$note);
			warn ( "WARNING $feature-$tag $tag is NOT a recorded qualifier for $feature - This tag is written to a note !\n" );
		}
		else
		{
			warn ( "WARNING $feature-$tag $tag is NOT a recorded qualifier for $feature and working in submit mode - This tag is SKIPPED !\n" );
		}
	}
	return $f;
}

=head2 procedure __NewFeatureAndQualifiersMrna

 Title    : __NewFeatureAndQualifiersMrna
 Procedure: Same as __NewFeatureAndQualifiersDefault, but specialized for mRNA: Take most qualifiers from the mRNA
            Only a few attributes are written

=cut
sub __NewFeatureAndQualifiersMrna
{
	my $self = shift;
	my ($feature, $o_locus) = @_;
	my $sub_type = $o_locus -> GetSubType();
	my $submit = $self -> { _o_annot } -> GetSubmit();

   	my $o_genbank_tags = $self -> { __o_genbank_tags };

	my $f = undef;

	$f = new Bio::SeqFeature::Generic ();
	$f -> location ($self->__Locus2Location($o_locus));
	$f -> primary_tag ( $feature );
	
	my $gene = $submit ? $o_locus -> GetGeneNameSimple() : $o_locus -> GetGeneName();
	$self -> __AddTagValues( $f,'gene',$gene)      if ( $gene ne '' );

	my $locus_tag = $o_locus -> GetLocusTag();
	if ( $submit == &FALSE && $locus_tag eq '' )
	{
		$locus_tag = $o_locus -> GetParentLocusTag();
	}
	$self -> __AddTagValues( $f,'locus_tag',$locus_tag)      if ( $locus_tag ne '' );

	#my $note = $o_locus -> GetNote();
	#$self -> __AddTagValues( $f,'note',$note)      if ( $note ne '' );
	
	#my $product = $o_locus -> GetProduct();
	#$self -> __AddTagValues( $f,'product',$product)      if ( $product ne '' );

	#my $dbxref = $o_locus -> GetDbXref();
	#$self -> __AddTagValues( $f,'db_xref',$dbxref) if ( $dbxref ne "" );
	
	my $alias = $o_locus -> GetAlias();
	$self -> __AddTagValues( $f,'gene_synonym',$alias) if ( $alias ne "" );

	my $transcript_id = $o_locus -> GetTranscriptId(&TRUE);
	$self -> __AddTagValues( $f,'transcript_id',$transcript_id) if ( $transcript_id ne "" );

	my $exception = $o_locus -> Get ( 'exception', $o_locus );
	$self -> __AddTagValues( $f,'exception',$exception) if ( $exception ne '' );

	return $f;
}

=head2 procedure __AddTagValues

 Title    : __AddTagValues
 Procedure: Call add_tag_values, adding one or several tag values to a feature
 Usage    : my $l = $self -> __AddTagValues($f,$tag,$value)
 Args     : $f  : The feature - MODIFIED
            $tag: The tag 
            $value: A scalar or a ref to an array

=cut
sub __AddTagValues
{
	my $self = shift;
	my ($f, $tag, $value) = @_;
	if ( ref($value) eq 'ARRAY' )
	{
		foreach my $v ( @$value )
		{
			$f -> add_tag_value ( $tag, $v );
		}
	}
	else
	{
		$f -> add_tag_value ( $tag, $value );
	}
}

=head2 function __GetRegions

 Title    : __GetRegions
 Procedure: Dump the regions of the locus, with fuzzy locations if necessary
            
 Usage    : $self -> __GetRegions( $f, $o_locus )

=cut
sub __GetRegions
{
	my $self = shift;
	my ( $f, $o_locus ) = @_;
	my $l           = new Bio::Location::Split;
	my $consistency = $o_locus -> GetConsistency();
	my @a_regions   = $o_locus -> GetRegions();
	
	# The consistency status of $o_locus may be transferred to the FIRST and LAST regions only if it is FS or FE
	if ( $a_regions[0] -> GetStrand() > 0 )
	{
		for ( my $i = 0; $i <= $#a_regions; $i++ )
		{
			my $o_region = $a_regions[$i];
			$l -> add_sub_Location ( $self -> __Locus2Location($o_region));
		}
	}
	else
	{
		for ( my $i = $#a_regions; $i >= 0; $i-- )
		{
			my $o_region = $a_regions[$i];
			$l -> add_sub_Location ( $self -> __Locus2Location($o_region));
		}
	}
	$l -> splittype ( $o_locus->_GetSplitType() );
	$f -> location ( $l );
}

=head2 procedure __Locus2Location

 Title    : __Locus2Location
 Procedure: Take a locus as entry and returns the corresponding location, may be a fuzzy location
 Args     : $o_locus  : The locus
            $consistency : The consistency flag, default = $o_locus -> GetConsistency()

=cut
sub __Locus2Location
{
	my $self = shift;
	my ( $o_locus, $consistency ) = @_;
	$consistency = $o_locus -> GetConsistency() unless defined ($consistency);
	
	my %h_prms = (-start  => $o_locus->GetStart(),-end => $o_locus->GetEnd(),-strand => $o_locus->GetStrand());
	if ( $consistency eq 'C' )
	{
		return Bio::Location::Simple->new(%h_prms);
	}
	else
	{
		$h_prms{-start_fuz}='BEFORE' if ($o_locus->GetStrand()>0 && $consistency=~/FS/);
		$h_prms{-start_fuz}='BEFORE' if ($o_locus->GetStrand()<0 && $consistency=~/FE/);
		$h_prms{-end_fuz}  ='AFTER'  if ($o_locus->GetStrand()>0 && $consistency=~/FE/);
		$h_prms{-end_fuz}  ='AFTER'  if ($o_locus->GetStrand()<0 && $consistency=~/FS/);
		return Bio::Location::Fuzzy->new(%h_prms);
	}
}

1;

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;


