#
# $Id$
#

package Assembly::Agp;
use base qw( Assembly );

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

Assembly - Loads an agp or contig.md assembly file
           Corrects the positions using the assembly
           Create a single fasta file, using a multifasta as entry

=head1 SYNOPSIS

 my $o_assembly = New Assembly($assembly_file_name, $grouplabel, $chromosome);

 

=head1 DESCRIPTION

 
=cut

use strict;
use warnings;
use Data::Dumper;
use Error::Simple;
use Carp;
use LipmError::ParameterException;
use LipmObject;
use General;
use base qw( LipmObject );
use DataReader::Whole;
use AssemblyError;

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1009 $ =~ /\d+/g); $r[0]};
}

=head2 procedure _Init

 Title        : Constructor
 Usage        : my $o_assembly = New Assembly($assembly_filename, $grouplabel, $chromosome, $random);
 Function     : Initialize the object, reading the assembly file
 Args         : $assembly_filename The name of the assembly file (agp or config.md file)
                $grouplabel        identifier of the required assembly 
                                   (for example human, can be Celera or Primary_Assembly or HuRef)
                $chr               The chromosome to assemble
                $random (0/1)      Consider the chromosome as random - or not              

=cut

sub _Init
{
	my $self = shift;
	my ($assembly_filename, $grouplabel, $chromosome, $random) = @_;
	
	$self->SUPER::_Init( @_ );

	$self->__LoadAgpFile();
	$self->SetCurrentChromosome($chromosome);
	#$self->SetFirstContig();

	#use Data::Dumper;
	#print Dumper ($self -> { __rh_contigs });
	#exit 0;
}

=head2 GetContigNames
 Title     : GetContigNames
 Usage     : my $ra_contigs = $o_assembly -> GetContigNames();
 Function  : Return the ordered list of contig names, removing the 'start' and 'end' contigs
             If the current chromosome is not selected, throw an error

=cut
sub GetContigNames
{
	my $self  = shift;
	$self->__ErrorNoCurrentChromosome();

	my @a_names;
	my $ra_data = $self -> { __rh_contigs } -> { data };
	
	# Loop in the data, avoiding the 'start' and 'end' contigs
	for (my $i=0; $i<=$#$ra_data; $i++)
	{
		next if ($ra_data->[$i]->{feature_name} eq 'start');
		next if ($ra_data->[$i]->{feature_name} eq 'end');
		next if ($ra_data->[$i]->{feature_type} =~ /^[NU]$/);
		push @a_names,$ra_data->[$i]->{feature_name};
	}
	return \@a_names;
}

=head2 GetAssemblyGaps
 Title     : GetAssemblyGaps
 Usage     : my $ra_assembly_gaps = $self -> GetAssemblyGaps()
 Procedure : Retrieve the assembly gaps in an array of hashes
			 The field names of the hashes is the column names of an agp files, for the U/N records
 Args      : None

=cut
sub GetAssemblyGaps
{
	my $self = shift;
	$self->__ErrorNoCurrentChromosome();
	#my @a_header = qw( chromosome chr_start chr_stop feature_id feature_type feature_name feature_start feature_end orientation );
	my @a_assembly_gaps;
	my $ra_components = $self -> { __rh_contigs } -> { data };
	foreach my $rh_line ( @$ra_components )
	{
		next if ( $rh_line -> { feature_type } !~ /^[NU]$/ );
		my %h_component;
		my %h_line = %$rh_line;
		@h_component{ qw/ object object_beg object_end component_type gap_length gap_type linkage linkage_evidence / } =
			@h_line{ qw/ chromosome chr_start chr_stop feature_type feature_name feature_start feature_end orientation / };
		push @a_assembly_gaps,\%h_component;
	}
	return \@a_assembly_gaps;
}
	
=head2 __IterContig
 Title     : __IterContig
 Usage     : $self -> __Itercontig ($step)
 Function  : Starting from the current contig, sets the current contig $step contigs forward (if $step>0) or backward (if $step<0)
 Args      : $step integer, may be > 0 or < 0
 Return    : The name of the contig
 Access    : Private
 Error     : Throw an error instead of setting the current contig before the first or after the last one

=cut

sub __IterContig
{
	my $self   = shift;
	my ($step) = @_;

	$self->__ErrorNoCurrentChromosome();

	if ( $step != 0)
	{
		my $index   = $self -> { __current_contig_index };
		my $ra_data = $self -> { __rh_contigs } -> { data };
		my $sign = $step/abs($step);
		$step = abs($step);
		
		for ( my $i=0; $i<$step; $i++)
		{
			do
			{
				$index += $sign;
				last if ($index>$#$ra_data || $index<0);
			}
			while($ra_data->[$index]->{feature_type} =~ /^[NU]$/ );
		}
		
		if ( $index > $#$ra_data )
		{
			throw Error::Simple ( 'Already at end' );
		};
		if ( $index < 0 )
		{
			throw Error::Simple ( 'Already at start' );
		};
		$self -> { __current_contig_index } = $index;
	}
	return $self->GetCurrentContigName();
}
#sub __IterContig
#{
	#my $self   = shift;
	#my ($step) = @_;
	
	#my $index   = $self -> { __current_contig_index };
	#my $ra_data = $self -> { __rh_contigs } -> { data };
	
	
	
	#if ( $step>0 && $index+$step>$#$ra_data )
	#{
		#throw Error::Simple ( 'Already at end' );
	#};
	#if ( $step<0 && $index+$step<0 )
	#{
		#throw Error::Simple ( 'Already at start' );
	#};
	#$self -> { __current_contig_index } = $index + $step;
	#return $self->GetCurrentContigName();
#}
=head2 SetFirstContig
 Title     : SetFirstContig
 Usage     : $self -> SetFirstContig ()
 Procedure : Position the object to the first contig
 Args      : None

=cut		
sub SetFirstContig
{
	my $self = shift;
	$self->__ErrorNoCurrentChromosome();

	my $ra_data = $self -> { __rh_contigs } -> { data };
	my $index = 0;
	while ($ra_data->[$index]->{feature_type} =~ /^NU$/ && $index<$#$ra_data)
	{
		$index++;
	}
	$self -> { __current_contig_index } = $index;
}

=head2 SetCurrentChromosome
 Title     : SetCurrentChromosome
 Usage     : $self -> SetCurrentChromosome ()
 Procedure : Position the Assembly::Agp to the wanted chromosome, if possible
 Args      : The chromosome label to set. You can pass "" to unset any chromosome

=cut		
sub SetCurrentChromosome
{
	my $self = shift;
	my ($chromosome) = @_;
	$chromosome="" unless(defined $chromosome);
	
	if ( $chromosome eq '' )
	{
		$self -> { __chromosome } = $chromosome;
		$self -> { __rh_contigs } = undef;
		$self -> { __current_contig_index } = undef;
	}
	else
	{
		my $rh_chromosomes = $self -> { __rh_chromosomes };
	
		if ( not defined $rh_chromosomes -> { $chromosome } )
		{
			throw Error::Simple ( "$chromosome is unknown" );
		}
	
		$self->{ __chromosome } = $chromosome;
		$self->{ __rh_contigs } = $rh_chromosomes->{$chromosome};
		$self->SetFirstContig();
	}
}

=head2 GetCurrentContigName, GetStart, Getend, GetStrand
 Title    : GetCurrentContigName, GetStart, Getend, GetStrand, GetFeatureStart, GetFeatureEnd
 Usage    : my $contig = $self -> GetCurrentContigName ()
 Function : Get the current contig name, the start/end position of the current contig, or the strand
 Args     : None

=cut

sub GetCurrentContigName
{
	my $self = shift;
	$self->__ErrorNoCurrentChromosome();
	
	my $index   = $self -> { __current_contig_index };
	my $ra_data = $self -> { __rh_contigs } -> { data };
	return $ra_data->[$index]->{feature_name};
}

sub GetStrand
{
	my $self = shift;
	$self->__ErrorNoCurrentChromosome();

	my $index   = $self -> { __current_contig_index };
	my $ra_data = $self -> { __rh_contigs } -> { data };
	my $orientation = $ra_data->[$index] -> { orientation };

	# the DEFAULT is returning '+' (cf the agp specifications)
	return ( $orientation eq '-' ) ? '-' : '+';
}

sub GetStart
{
	my $self = shift;
	$self->__ErrorNoCurrentChromosome();

	my $index   = $self -> { __current_contig_index };
	my $ra_data = $self -> { __rh_contigs } -> { data };
	return $ra_data->[$index] -> { chr_start };
}

sub GetEnd
{
	my $self = shift;
	$self->__ErrorNoCurrentChromosome();

	my $index   = $self -> { __current_contig_index };
	my $ra_data = $self -> { __rh_contigs } -> { data };
	return $ra_data->[$index] -> { chr_stop };
}

sub GetFeatureStart
{
	my $self = shift;
	$self->__ErrorNoCurrentChromosome();

	my $index   = $self -> { __current_contig_index };
	my $ra_data = $self -> { __rh_contigs } -> { data };
	return $ra_data->[$index] -> { feature_start };
}

sub GetFeatureEnd
{
	my $self = shift;
	$self->__ErrorNoCurrentChromosome();

	my $index   = $self -> { __current_contig_index };
	my $ra_data = $self -> { __rh_contigs } -> { data };
	return $ra_data->[$index] -> { feature_end };
}

sub GetChromosomeLabels
{
	my $self = shift;
	my $rh_chromosomes = $self -> { __rh_chromosomes };
	my @a_chroms = sort keys %$rh_chromosomes;
	return \@a_chroms;
}

sub __ErrorNoCurrentChromosome
{
	my $self = shift;
	if ( $self -> {__chromosome} eq '' )
	{
		throw Error::Simple ( 'ERROR - No current chromosome defined yet - Please call SetCurrentChromosome first' );
	}
}
=head2 procedure __LoadAgpFile

 Title     : __LoadAgpFile
 Usage     : $self->__LoadAgpFile();
 Procedure : Load the tiling/golden path from one or several agp file

 __rh_contigs description:
                key = $chr_start OR $chr_end OR data
                val = 
                      1/ If data, the val is an ARRAY val = the line read from seq_contig.md, with some corrections in the case of a random chromosome 
                         (RANDOM NOT IMPLEMENTED !)
                         keys = chromosome (part number in agp specification),
                                chr_start,chr_stop,orientation (object beg and end, orientation in agp specification)
	                            feature_type (component type in agp specification),
	                            feature_start,feature_end (component beg and end in agp specification)
	                            chromosome (irrelevant in agp specification)
	                            feature_name (component_id in agp specification)
	                  2/ If $chr_start or $chr_end, the val is an INTEGER

=cut

sub __LoadAgpFile
{
	my $self = shift;
	my @a_f_assembly = split(',',$self -> { __assembly_filename });
		
	# Agp file are organized according to http://www.ncbi.nlm.nih.gov/projects/genome/assembly/agp/AGP_Specification.shtml
	# columns 0-4 : object  object_begin  object_end  part_number feature_type 
	# feature_type != N   columns 5-8: feature_id feature_beg  feature_end orientation
	# feature_type == N   columns 5-8: gap_length  gap_type  linkage empty
	
	my @a_header = qw( chromosome chr_start chr_stop feature_id feature_type feature_name feature_start feature_end orientation );
	my $o_reader = New DataReader::Whole( -force_headers => \@a_header );
	
	my @a_contigs;
	foreach my $f_assembly(@a_f_assembly)
	{
		$o_reader   -> Prepare ( -from => $f_assembly, -select => [ 'chromosome', 'chr_start', 'chr_stop', 'feature_name', 'feature_start', 'feature_end', 'orientation', 'feature_type' ]);
		my $ra_contigs  =  $o_reader -> FetchallArrayref( -line_type => 'HASH' );
		if ( @$ra_contigs == 0 )
		{
			throw Error::Simple " ERROR - nothing read from file $f_assembly";
		}
		push @a_contigs, @$ra_contigs;
	}
	
	$self -> __FinalizeData( \@a_contigs );
}

=head2 procedure __FinalizeData

 Title     : __FinalizeData (was __FromAgpToContigMd)
 Procedure : Adding extra information for compatibility reasons
 Args      : $ra_contigs the data read by __LoadAgpFile

=cut

sub __FinalizeData
{
	my $self = shift;
	my ($ra_contigs) = @_;
	
	my $chrom      = $self -> { __chromosome };
	
	# %h_chromosomes = contains data for each chromosome (object) - key = chromosome label, val = an array
	my %h_chromosomes;
	foreach my $rh_c (@$ra_contigs)
	{
		my $chrom = $rh_c->{chromosome};
		next if ($chrom =~ /^#/);
		
		if ( !exists $h_chromosomes{$chrom} )
		{
			my @a_data;
			my %h_start;
			$h_start { chromosome }   = $chrom;
			$h_start { chr_start  }   = 1;
			$h_start { chr_stop   }   = 1;
			$h_start { feature_name } = 'start';
			push @a_data,\%h_start;
			$h_chromosomes{$chrom} = \@a_data;
		}
		my $ra_data = $h_chromosomes{$chrom};
		#push @$ra_data,$rh_c if ($rh_c->{feature_type} !~ /^[NU]$/);
		push @$ra_data,$rh_c;
	}
	
	foreach my $chrom (sort keys %h_chromosomes)
	{
		my $ra_data = $h_chromosomes{$chrom};
		my $rh_start= $ra_data->[0];

		my %h_end;
		my $chr_start = $ra_data->[1]->{ chr_start }; # Using 1 as index because the 'start' contig is not counted here
		my $chr_end = $ra_data->[-1]->{chr_stop} + 1 ;

		#
		# NOTE - The 'start' pseudo-contig is at position 1, ie the FIRST nucleotide of the sequence (see above)
		#        The 'end' pseudo-contig starts PAST LAST NUCLEOTIDE
		#        Both pseudo-contigs are 1 nucleotide long

		$h_end { chromosome }   = $chrom;
		$h_end { chr_start  }   = $chr_end;
		$h_end { chr_stop   }   = $chr_end;
		$h_end { feature_name } = 'end';
		push @$ra_data,\%h_end;

		my $key_start = $chrom."_start";
		my $key_end   = $chrom."_end";
		my $key_chrom = $chrom;

		my %h_converted_contigs = ();	
		$h_converted_contigs{ $key_start } = $chr_start;
		$h_converted_contigs{ $key_end }   = $chr_end;
		$h_converted_contigs{ data }        = $ra_data;

		$h_chromosomes{$chrom} = \%h_converted_contigs;
	}
	$self -> { __rh_chromosomes } = \%h_chromosomes;
}




#sub __FinalizeData
#{
	#my $self = shift;
	#my ($ra_contigs) = @_;
	
	#my $chrom      = $self -> { __chromosome };
	
	#my @a_data;
	#my %h_start;
	#$h_start { chromosome }   = $chrom;
	#$h_start { chr_start  }   = 1;
	#$h_start { chr_stop   }   = 1;
	#$h_start { feature_name } = 'start';
	#push @a_data,\%h_start;
	
	## Skip the gaps and the lines related to other chromosomes, BUT keep every chromosome label
	#my %h_chroms;
	#foreach my $rh_c (@$ra_contigs)
	#{
		#my $chr = $rh_c->{chromosome};
		#next if ($chr =~ /^#/);
		#$h_chroms{$chr}=1;
		#next unless ($chr eq $chrom);
		#push @a_data,$rh_c if ($rh_c->{feature_type} !~ /^[NU]$/);
	#}
	
	#my @a_chroms = sort keys %h_chroms;
	#$self -> { __ra_chromosomes } = \@a_chroms;
	
	#my %h_end;
	#my $chr_start = $a_data[1]->{ chr_start }; # Using 1 as index because the 'start' contig is not counted here

	#my $chr_end;
	## Is there a gap at end of @$ra_contigs ?
	#if ( $ra_contigs->[-1]->{feature_type} =~ /^[NU]$/)
	#{
		#$chr_end = $ra_contigs->[-1]->{chr_stop} + 1 ;
	#}
	#else
	#{
		#$chr_end   = $a_data[-1]->{ chr_stop } + 1; # Using -1 as index because the 'end' contig is NOT YET in the array
	#}
	
	## NOTE - The 'start' pseudo-contig is at position 1, ie the FIRST nucleotide of the sequence
	##        The 'end' pseudo-contig starts PAST LAST NUCLEOTIDE
	
	#$h_end { chromosome }   = $chrom;
	#$h_end { chr_start  }   = $chr_end;
	#$h_end { chr_stop   }	= $chr_end;
	#$h_end { feature_name } = 'end';
	#push @a_data,\%h_end;

	#my $key_start = $chrom."_start";
    #my $key_end   = $chrom."_end";
	#my $key_chrom = $chrom;

	#my %h_converted_contigs = ();	
	#$h_converted_contigs{ $key_start } = $chr_start;
	#$h_converted_contigs{ $key_end }   = $chr_end;
	#$h_converted_contigs{ data }        = \@a_data;

	#$self -> { __rh_contigs } = \%h_converted_contigs;
#}


=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;

