#
# $Id: CdsRegion.pm 1288 2013-10-21 08:39:57Z manu $
#

package Locus::CdsRegion;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=head1 NAME

Locus::CdsRegion - describes a region of a Cds

=head1 SYNOPSIS

 my $o_cds_r = New Locus::CdsRegion('start'=>integer,'end'=>integer,'strand'=>'-1|1|+|-','locus_tag'=>string,parent=>$o_gene);

=head1 DESCRIPTION

 A CdsRegion extends the Locus class. Its parent should be a Cds object.
 A CdsRegion may have a locus_tag, and has a narcisseId

=cut


use strict;
use warnings;
use General;
use Carp;
use base qw( Locus );
sub GetDefaultSubType { return 'CDS' };

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1288 $ =~ /\d+/g); $r[0]};
}

=head2 function IsSelected

 Prerequisite: SetParent should have already be called on this object
 See Locus.pm for the generic documentation
 We call the Parent's IsSelected function, because the status of this object is THE SAME as the status of its parent

=cut
sub IsSelected
{
	my $self = shift;
	return $self -> GetParent() -> IsSelected();
}

=head2 procedure SetParent

 Title        : SetParent
 Usage        : $o_cds->setParent( $o_cds );
 Procedure    : Set the parent
 Prerequisite : The parent should not already set, $o_cds should be a Locus::Cds object
 
 Args         : $o_cds : The parent Cds
 Globals      : none

=cut
sub SetParent
{
	my $self = shift;
	my ($o_cds) = @_;
	
	if ( ref($o_cds) ne 'Locus::Cds' )
	{
		throw LipmError::ParameterException( 'o_parent', 'type', -text => 'should be a Locus::Cds object' );
	}
	$self->SUPER::SetParent( $o_cds );
	
	my $nc_id  = $o_cds -> GetNarcisseId();
	$nc_id    .= 'r' . $self -> __GetRank();
	$self -> { __narcisse_id } = $nc_id;
}

=head2 function GetParentId

 Title     : GetParentId
 Function  : Returns the Id of the grandparent, because it is considered that the parent, which is a Locus::Cds object,
             is only a technical object useful for the implementation, it does not have any biological meaning, so that it will not be written in the output files.
             Thus, it is more convenient ignoring its NarcisseId
 Returns   : the parent's id in a biological sense

=cut
sub GetParentId	
{
	my $self = shift;
	return $self->GetParent()->GetParentId();
}
=head2 function GetProteinId

 Title : GetProteinId
 Usage : my $id = $o_cds -> GetProteinId();
=cut
sub GetProteinId
{
	my $self = shift;	
	my $as_string = @_;
	return ($as_string) ? $self -> GetAsString ( 'protein_id' ) : $self -> Get ( 'protein_id' );
}

=head2 functions GetScore, GetPhase

=cut
sub GetScore
{
	my $self = shift;
	my $rvl  = $self -> Get ( 'score' );
	return $rvl eq '' ? '.' : $rvl;
}
sub GetPhase
{
	my $self = shift;
	my $rvl  = $self -> Get ( 'phase' );
	return $rvl eq '' ? '.' : $rvl;
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;


