#
# $Id: MiscRegion.pm 1342 2013-12-17 12:52:36Z manu $
#

package Locus::MiscRegion;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=head1 NAME

Locus::MiscRegion - describes any region not already described by the other Locus:: objects

=head1 SYNOPSIS

=head1 DESCRIPTION

 A MiscRegion extends the Locus class, it may splitted in MiscRegions (like an rna).
 Any type is allowed.
 For bioperl representation, the split type is 'order'

=cut

use strict;
use warnings;
use Carp;
use Bio::Seq;
use Data::Dumper;

use General;
use Locus;

use base qw( Locus );

sub GetDefaultSubType { return 'region' };


# used by Locus::_GetBioPerlRegions
sub _GetSplitType() { return 'order' };

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1342 $ =~ /\d+/g); $r[0]};
}

=head2 function _GetBioPerlFeature

 Title    : _GetBioPerlFeature
 Function : Call the father's method, set the primary_tag to Mrna
 Usage    : my $f = $self -> GetBioPerlFeature();
 Args     : none
 Return   : A Bio::SeqFeature::Generic object

=cut
sub _GetBioPerlFeature
{
	my $self = shift;
	my $f = $self->SUPER::_GetBioPerlFeature ( @_ );
	$f -> primary_tag ( $self -> GetSubType() );
	$f -> location ( $self -> _GetBioPerlRegions() );
	return $f;
}

=head2 procedure SetParent

 Title        : SetParent
 Usage        : $o_gene->SetParent( $o_annot );
 Procedure    : Set the parent
 Prerequisite : The parent should not already set
 
 Args         : $o_annot : The parent, an AnnotatedSeqfile.pm object OR another MiscRegion
 Globals      : none

=cut
sub SetParent
{
	my $self = shift;
	my ($o_annot) = @_;
	
	if ( !$o_annot->isa('AnnotatedSeq') && !$o_annot->isa('Locus::MiscRegion') && !$o_annot->isa('Locus::Gene'))
	{
		throw LipmError::ParameterException( 'o_parent', 'type', -text => 'should be an AnnotatedSeq or Locus::MiscRegion or Locus::Gene object, not a '. ref($o_annot) );
	}
	
	# Call Locus::SetParent, but do not adjust the parent boundaries
	$self->SUPER::SetParent( $o_annot, &FALSE );
	
	my $nc_id = $o_annot -> GetNarcisseId();
	$nc_id .= 'm' . $self -> GetCount();
	$self -> { __narcisse_id } = $nc_id;
}

=head2 procedure AddTo

=cut
sub AddTo { 
	my $self = shift;
	my ($o_annot) = @_;
	$o_annot->AddMiscRegion($self);
}

=head2 procedures Select,Deselect

 See Locus.pm for the generic documentation
 we select/deselect the Locus::Cds, the Locus::ExonSet and the Locus::Utr objects

=cut
sub Select
{
	my $self = shift;
	my ( $inside ) = @_;
	$inside = &TRUE unless (defined $inside);
	if ( $inside )
	{
		my @a_regions = $self -> GetRegions();
		if (@a_regions>2)
		{
			foreach my $o_region (@a_regions)
			{
				$o_region -> Select($inside);
			}
		}
	}
	delete $self -> { __deselected };
}

sub Deselect
{
	my $self   = shift;
	my ( $inside ) = @_;
	$inside = &TRUE unless (defined $inside);
	if ( $inside )
	{
		my @a_regions = $self -> GetRegions();
		if (@a_regions>2)
		{
			foreach my $o_region (@a_regions)
			{
				$o_region -> Deselect($inside);
			}
		}
	}
	$self -> { __deselected } = &TRUE;
}

=head2 function GetIntrons

 Title        : GetIntrons
 Usage        : my @a_introns = $o_region->GetIntrons();
 Function     : Return an array of Locus, each locus representing an "intron", ie the nt between several regions
 
 Args         : None
 Return       : The Array of introns
 Globals      : none
 
=cut
sub GetIntrons
{
	my $self = shift;
	if ( !defined $self -> { __ra_introns } )
	{
		my @a_introns;
		my @a_regions = $self -> GetRegions();
		my $strand = $self->GetStrand();

		if ( @a_regions > 1 )
		{
			for (my $i=1; $i<@a_regions;$i++)
			{
				my $s = $a_regions[$i-1]->GetEnd()+1;
				my $e = $a_regions[$i]->GetStart()-1;
				next if ($s>$e);

				my $o_intron = New Locus::MiscRegion ( start=>$s,end=>$e,strand=>$strand );
				$o_intron->SetParent($self);
				push @a_introns, $o_intron;
			}
		}
		$self -> { __ra_introns } = \@a_introns;
	}
	return @{$self -> { __ra_introns }};
}
=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;


