#
# $Id$
#
 
package Submit::ToNcbi;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME



=head1 SYNOPSIS


=head1 DESCRIPTION

=cut

use strict;
use warnings;
use Carp;

use Error qw(:try);
use General;
use base qw( Submit );

BEGIN
{
    our $VERSION = do {my @r = (q$Rev$ =~ /\d+/g); $r[0]};
}


=head2 procedure _ApplyRules
 Title    : _ApplyRules
 Function : Apply rules useful for Ncbi submissionShould be called by the derived class ApplyRules BEFORE or AFTER the code of ApplyRules
            This function MODIFIES the objects managed by $o_ann_seq
            It should first define some variables, then call the Pre script, then implements rules, the calls the Post script:
               eval $self->_ApplyPre(); die $@ if $@;
               ...
               eval $self->_ApplyPost(); die $@ if $@;
 Arg      : the AnnotatedSeq object
 Access   : Protected

=cut
sub _ApplyRules
{
	my $self = shift;
	my ($o_ann_seq) = @_;
	#$self->SUPER::ApplyRules();
	
	# building the environment, you may use it in the pre and post scripts
	my $ra_objects= $o_ann_seq->GetAllLoci();
	my $ra_genes  = $o_ann_seq->GetGenes();
	my $ra_ncrnas = $o_ann_seq->GetNcrnas();
	my $ra_repeats= $o_ann_seq->GetRepeats();
	my $ra_regions= $o_ann_seq->GetMiscRegions();
	
	eval $self->_ApplyPre(); throw Error::Simple($@) if $@;
	
	
	#SC: soumission des annotations, pas necessaire de soumettre les gap
	#suppression feature gap
	my @a_gapfeatures = ('assembly_gap', 'gap');
	foreach my $o_locus (@$ra_objects)
	{
		my $sub_type = $o_locus->GetSubType();
		if (($o_locus->isa('Locus::MiscRegion') && grep (/^$sub_type$/, @a_gapfeatures)))
		{
			$o_locus->Deselect();
		}
	}
	
	
	# 2/ If we are in kingdom Bacteria deselect the mRNA, but first retreive their product attribute !
	if ( $o_ann_seq->GetKingdom() eq 'bacteria' )
	{
		print STDERR "Working with bacteria, removing mRNA objects\n";
		foreach my $o_gene ( @$ra_genes )
		{
			next if ( $o_gene->GetSubType() eq 'ncRNA_gene' );
			my $ra_mrnas = $o_gene->GetMrnas();
			foreach my $o_mrna ( @$ra_mrnas )
			{
				my $gene_product = $o_gene->GetProduct(&TRUE);
				my $mrna_product = $o_mrna->GetProduct(&TRUE);
				my $o_cds = $o_mrna -> GetCDS;
				
				my $cds_product  = $o_cds->GetProduct(&TRUE);
				if ( $cds_product eq '' )
				{
					if ( $mrna_product eq '' )
					{
						$mrna_product = 'Hypothetical protein';
					}
					#~ $o_gene -> Set ( 'product', $mrna_product );
					$o_cds->Set ( 'product', $mrna_product );
				}
				$o_mrna->Deselect(&FALSE);
			}
		}
	}
	
	

	# 3/ Deselect redundant CDS/UTR, keeping the locus_tag of the FIRST CDS with a given pk
	#    Add the translation table to ALL CDS
	print STDERR "Removing redundant CDS or 3'UTR\n";
	my %h_cds=();
	my %h_utr=();
	my $trans_table = $o_ann_seq -> GetCodonTable();
	foreach my $o_gene ( @$ra_genes )
	{
		next if ( $o_gene->GetSubType() eq 'ncRNA_gene' );
		my $ra_mrnas = $o_gene->GetMrnas();
		foreach my $o_mrna ( @$ra_mrnas )
		{
			$o_mrna->Deselect(&FALSE);
			my $o_cds = $o_mrna->GetCDS();
			my $pk = $o_cds->Getpk();
			if ( exists $h_cds{ $pk } )
			{
				print STDERR "REDUNDANT_CDS - " . $o_cds->GetpkExtended() . " Deselected\n";
				$o_cds->Deselect();
			}
			else
			{
				$h_cds{$pk} = 1;
				$o_cds -> Set ( 'transl_table', $trans_table );
			}

			#SC: idem pour les 3'UTR
			my $o_utr = $o_mrna->GetUtr('three_prime_UTR');
			my $pk = $o_cds->Getpk();
			if ( defined $o_utr )
			{
				if ( exists $h_utr{ $pk } )
				{
					print STDERR "REDUNDANT_UTR - " . $o_utr->GetpkExtended() . " Deselected\n";
					$o_utr->Deselect();
				}
				else
				{
					$h_utr{$pk} = 1;
				}
			}
		}
	}
	
	# 4a/ removing the note, replace it with product... or removing the product, replace it with a note !
	#SC: les mobile_element|repeat_region|rep_origin ne peuvent pas avoir de champ product
	#SC: donc on reverse dans des notes
	#SC: dans le GFF3 de melilo, le champ product etait dans le champ Note
	#SC: il faudrait voir avec Erika qui avait fait cela, il me semble que le champs "product=xxxx" de la colonne 9 correspondrait plus.
	# 
	print STDERR "Replacing product with note for some elements, the reverse for others, and adding product for CDS/gene if not already done\n";
	foreach my $o_locus (@$ra_objects)
	{
		
		next unless $o_locus->IsSelected();
		my $sub_type = $o_locus->GetSubType();
		if (($o_locus->isa('Locus::MiscRegion') && $sub_type ne 'mobile_element' && $sub_type ne 'rep_origin') || $o_locus->isa('Locus::Repeat'))
		{
			my $product = $o_locus -> GetProduct();
			if ( $product ne "" )
			{
				$o_locus -> Delete ('product');
				$o_locus -> AddValue ( 'note', $product );
			}
		}
	}


	
	# 8/ Computing product for tRNA, rRNA and misc_feature
	#    product is mandatory for rRNA and misc_feature
	#    Concerning gene, and only if bacteria, I take the CDS's product if none specified
	#SC: le champ product des tRNA est contraint pour la soumission
	#SC: tRNA-Aam
	#SC: cf. avec Erika si le format anti_codon=xxx_xxx est stable en sortie de Eugene
	#EC: DEUX FORMATS sont geres (!)
	#    1/ anti_codon=Met_CAT      ==> product=tRNA-Met
	#    2/ product=tRNA-Met        ==> product=tRNA-Met
	# Lorsqu'on aura un retour de embl on verra...
	
	print STDERR "Computing product for tRNA\n";
	foreach my $o_locus (@$ra_objects)
	{
		if ($o_locus->isa('Locus::Rna') && $o_locus -> GetSubType() eq 'tRNA' )
		{
			my $product = $o_locus -> GetProduct();
			if ( $product !~ /^tRNA-[A-Z][a-z]{2}$/ )
			{
				if ( $product =~ /^([A-Z][a-z]{2}) tRNA$/ )
				{
					$product = 'tRNA-' . $1;
				}
				else
				{
					$product = '';
				}
			}
			if ($product eq '')
			{
				my $anti_codon = $o_locus->Get ('anti_codon');
				if ( $anti_codon =~ /^(.+)_/ )
				{
					$product = 'tRNA-'.$1;
				}
				else
				{
					warn $self -> GetWarningHeader() . "The anti_codon attribute does not have a normalized format: ".$anti_codon." for tRNA ".$o_locus->GetpkExtended() . "\n";
				}
			}
			
			if ( $product eq '' )
			{
				my $name = $o_locus -> GetGeneName();
				#print STDERR "KOUKOU $name\n";
				#print STDERR $o_locus->GetParent()->SafeDumper;
				#print STDERR $o_locus->SafeDumper;
				if ( $name =~ /^tRNA-[A-Z][a-z]{2}$/ )
				{
					$product = $name;
				}
				if ( $name =~ /_([A-Z][a-z]{2})_[A-Z]{3}$/ )
				{
					#print STDERR "KOUKOU2 $name\n";
					$product = 'tRNA-'.$1;
				}
			}
			
			if ( $product eq '' )
			{
				$product = 'tRNA-Xxx';
			}

			if ($product  !~ /^tRNA-[A-Z][a-z]{2}$/ )
			{
				warn $self -> GetWarningHeader() . "The computed product does not have a normalized format: ".$product." for tRNA ".$o_locus->GetpkExtended() . "\n";
			}
			$o_locus->Delete('product');
			$o_locus->Set('product',$product);
		}
		elsif ( ($o_locus->isa('Locus::Rna') && $o_locus -> GetSubType() eq 'rRNA') ||
				 ($o_locus->isa('Locus::MiscRegion') && $o_locus -> GetSubType() eq 'misc_feature') )
		{
			my $product = $o_locus -> GetProduct(&TRUE);
			if (! defined ( $product ) )
			{
				$o_locus -> Set ( 'product', 'hypothetical protein' );
			}
		}
	}
	
	# 4b/ gene etc
#SC: les genes synonymes ne peuvent pas etre mis ailleurs que dans des notes
#SC: nous avons fait le choix de structure ce champ note avec le mot cle "Synonym:"
#SC: les synonymes sont par exemple des locus_tag de genes de reference sur lequel on a recupere l'annotation par transfert
#SC: ex melilo: locus_tag=SM2011_a0001, synonym=SMa0001

	print STDERR "replacing Name attributes by Synonym in a note\n";
	print STDERR "Replacing gn attributes by adding gene attributes for Locus::Gene and putting them in a note for Locus::Repeat\n";
	print STDERR @$ra_objects . " objects to examine\n";
	# Delete those attributes even on deselected mRNA, because of the note consolidation process
	my $contig = $o_ann_seq -> GetContigLabel();
	foreach my $o_locus (@$ra_objects)
	{

		my $sub_type = $o_locus->GetSubType();
		my $gene = $o_locus->Get('gene');
		
		my $lt   = $o_locus -> GetLocusTag();
		if ( $gene =~ /$lt/i || $gene =~ /$contig/i )
		{
			$o_locus -> Delete ( 'gene' );
		}
		
		# Keep gene_synonym to a note
		my $gs = $o_locus -> GetAlias ( &TRUE );
		if ( $gs ne '' && $gs ne $gene )
		{
			$o_locus -> AddValue ( 'note', "corresponds to $gs" );
			#$o_locus -> AddValue ( 'note', "Synonym:$gs" );
		}
		# See Locus::GetAlias
		$o_locus -> Delete ( 'gene_synonym' );

		#SC: les mobile_element|repeat_region|rep_origin ne peuvent pas avoir de champ gene (nom de gene)
		if (($o_locus->isa('Locus::MiscRegion') && ($sub_type eq 'insertion_sequence' || $sub_type eq 'rep_origin')) || $o_locus->isa('Locus::Repeat'))
		{
			if ( $gene ne '' )
			{
				$o_locus->AddValue('note',$gene);
			}
		}
	}

	# 5/ adding a note about RNASeq to the UTR, only if bacteria
#SC: dans le cas de melilo, le RNAseq avait permis de definir les UTR
#SC: chez les bacteries, si on a des UTR c'est surement lie au fait que l'on avait du RNAseq oriente pour les detecter
	if ( $o_ann_seq->GetKingdom() eq 'bacteria' )
	{
		print STDERR "Adding note relative to RNASeq for UTR and removing locus_tag, product, previous notes (bacteria only)\n";
		foreach my $o_locus (@$ra_objects)
		{
			if ($o_locus->isa('Locus::Utr') || $o_locus->isa('Locus::UtrRegion'))
			{

				# D'apres Stephanie il faut un locus_tag sur les UTR
				# Donc au contraire s'il n'existe pas je le remets
				my $o_lt = $o_locus->GetLocusTag();
				if ($o_lt eq '' )
				{
					$o_lt = $o_locus->GetParent()->GetParent()->GetLocusTag();
					if ( $o_lt ne '' )
					{
						$o_locus -> Set ('locus_tag',$o_lt);
					}
				}
				#$o_locus->Delete('locus_tag');
				$o_locus->Delete('gene');
				$o_locus->Delete('product');
				$o_locus->Delete('note');
				$o_locus->Set('note','based on oriented RNAseq data');
			}
		}
	}
	
	# 6/ Special treatment for the ncRNA
	#SC: pour les ncRNA, le champ product est obligatoire ainsi que la classe
	print STDERR "Special treatment for the ncrnas\n";
	foreach my $o_locus (@$ra_objects)
	{
		next unless $o_locus->IsSelected();
		next unless ( $o_locus->isa('Locus::Ncrna') && $o_locus->GetSubType() eq 'ncRNA' );
		$o_locus->Set('product','novel ncRNA; function unknown') unless ($o_locus->Get('product') ne '');
		$o_locus->Set('ncRNA_class','other');
		$o_locus->AddValue('note','based on oriented RNAseq data');
	}
	
	# 7/ Computing protein id
	#SC: le protein_id est obligatoire
	#SC: le prefixe par defaut est: gnl|BFC|$locus_tag mais BFC peut etre remplace par le switch --protein-namespace
	#SC: ce protein_id devant etre unique, cela colle
	my $protein_namespace = $self -> _Get ( '_protein-namespace' );
	$protein_namespace = 'BFC' if ($protein_namespace eq '' );
	print STDERR "Computing protein_id for the CDS\n";
	foreach my $o_locus (@$ra_objects)
	{
		if ($o_locus->isa('Locus::Cds'))
		{
			my $locus_tag = $o_locus->GetLocusTag();
			if ( $locus_tag eq '')
			{
				$locus_tag = $o_locus->GetParent()->GetParent()->GetLocusTag();
				$o_locus -> Set('locus_tag',$locus_tag);
			}
			if ( $locus_tag ne "" )
			{
				$o_locus->GetRegion0()->Set('protein_id',"gnl|$protein_namespace|$locus_tag");
			}
			else
			{
				warn "ERROR - NO LOCUS_TAG NO locus tag FOR THIS CDS ".$o_locus->GetpkExtended() . "\n";
			}
		}
	}
	
	# 9/ Last cleaning
	print STDERR "removing other attributes and cleaning product (removing trailing .)\n";
	foreach my $o_locus (@$ra_objects)
	{
		#SC: si le /function est defini, gardons le
		$o_locus -> DeleteExcept( 'EC_number','eC_number','ncRNA_class','pseudogene','function','transl_table' );
		$o_locus->Delete('Dbxref');
		my $product = $o_locus->GetProduct(1);
		if ( $product =~ /\.$/ )
		{
			$product =~ s/\.$//;
			$o_locus -> Delete ( 'product' );
			$o_locus -> Set ('product',$product);
		}
	}
	eval $self->_ApplyPost(); throw Error::Simple($@) if $@;
}

=head2 procedure _Checkrules
 Title    : _CheckRules
 Function : Checkrules are used to CHECK that the data are correct for submission.
            It is called by Applyrules AFTER _Applyrules, thus even after the post script
            If something is wrong with the data, an Error::Simple exception is thrown
 Arg      : the AnnotatedSeq object
 Access   : Protected

=cut
sub _CheckRules
{
	my $self = shift;
	my ($o_ann_seq) = @_;
	
	my $ra_genes = $o_ann_seq->GetGenes();
	my $no_lt = 0;
	foreach my $o_gene (@$ra_genes)
	{
		$no_lt++ if ( $o_gene -> GetLocusTag() eq '' );
	}
	if ( $no_lt != 0 )
	{
		throw Error::Simple("ERROR - $no_lt genes DO NOT HAVE ANY locus_tag");
	}
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
