#
# 	sebastien.letort@toulouse.inra.fr
#	Created: July 31, 2008
#   revised by emmanuel.courcelle@toulouse.inra.fr - Oct 2008
#   modified by thomas.faraut@toulouse.inra.fr - march 2009
#   stolen to the Narcisse project by manu - Aug 2011
#	$Id: DataReader.pm 880 2012-04-10 15:52:00Z manu $
#
# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=pod

=head1 NAME

	DataReader - Reading tabulated data files (.tsv)

=head1  SYNOPSIS

    This is an abstract class - DO NOT USE directly, please create a subclass of DataReader
	my $o_reader = New DataReader::some_subclass();

=head1 DESCRIPTION

	This class is used to access tabular datafiles, it is designed for performance. 
	You should read the documentation about the subclasses, because you will have to use one of another, depending on the optimisation wanted
	(memory or cpu) or on the precise request (where conditions or not)

=head1 FILE FORMAT

    The files read by this hierarchy of classes are tabulated files (the columns are separated by TAB (\t), tsv-like files):
		- You may insert COMMENTS (1st character = ;), EMPTY LINES (ONLY \n)
		- If the FIRST LINE is /path/to/some/file (1st char is a /) it is ignored
		- If some other line is /path/to/some/file it is an exit condition (see later, CONCATENATING FILES)
		- If the FIRST LINE (except for /path, comments or empty lines) starts with @ it is recognized as an "indexation line" by DataReader::Region, 
		  it is ignored by all other subclasses
		- The FIRST LINE (ignoring comments, empty lines or index line) SHOULD be a Header lines
		- The OTHER LINES are comments, empty lines, or data
	
	Two columns may be identified as 'start' or 'end' segments, default labelled 'start' and 'end' (but this can be changed)
	The file MUST BE sorted in numerical increasing order, first key start column, second key end column, to be able to specify
	range requests
		
    Here is an example:

; Starting with a comment
@... (see Region)
element	chrom	start	end	sign	s_chrom	s_start	s_end	s_order	matching	q_chain
1	1	1068612	1225685	+	1	961719	1077048	2	63204	2:4:

=head1 HOW TO USE

   3 stages are necessary to use those objects:
   
   1/ Create a new object
   2/ Call Prepare, specifying a FROM parameters, and may be specifying SELECT and WHERE conditions. 
      FROM parameter may be:
           - A file name, in this case the file will be opened by the object
           - A file handle, ie an already (and may be seeked) file
      
      SELECT conditions are only vailable with FetchAllHashref and are:
           - A ref to an array, specifying the columns you are interested with
   
      WHERE conditions may be:
           - RANGE specifications (start..end)
           - VALUE specifications (only EQUALITY is supported, with case sensitivity or not)
           
   3/ Call FetchAllArrayref or FetchAllHashref
   
   The results are deposited in one of 4 data structures:
       -An array of ...  (FetchAllArrayref):
            -arrays
            -hashes (keys = column label)
       -A hash of ... (FetchAllHashref, you should specify which column label is used as key):
            -arrays
            -hashes

=head1 CONCATENATING FILES

   It is possible to concatenate several file. The following is the concatenation of 2 files:

/path/to/some/file
element	start	end	color
1	10	20	red
; some comment here
/path/to/some/other/file
element	start	end	size
1	10	20	big
2	30	40	small
; some comment here

   If an opened file handle is passed to Prepare, OR the file name, only the 4 first lines will be read
   If an opened file handle is passed to Prepare AND the file is positioned at the line 5 (/path/to/some/other/file), 
   only the SECOND part of the file will be read. A part from this, youhave THE SAME behaviour as reading two different files.

=head1 LIMITATIONS

   1/ The label of the FIRST COLUMN should not start with a # character, as this character will be removed to be compatible 
   with old NArcisse specifications. Thus in the previous example, the label line could be written as:
 #element	chrom	start etc.
 
   2/ A data line should NOT start with a ; as this is interpreted as a comment, nor with a / as this is an exit condition
   
=cut

package DataReader;

use strict;
use warnings;

use base qw( LipmObject );

use POSIX;  # for the ceil function
use Data::Dumper;
use Carp;
use IO::File;


use LipmError::IOException;
use LipmError::ParameterException;
use LipmError::ObjectException;

use General;

our $USE_INLINE_C = (__EnableInlineCIfPossible() eq '');
use constant MAX_CHROM_SIZE => 4000000000;

# Column separator
our $SEP = "\t";

=head2 method _Init

 Usage        : my $o_x = New DataReader::Whole();
                my $o_x = New DataReader::Region();
 Fonction     : Constructor
 Args         : -force_headers => undef, If specified (default undef), the headers are NOT read from the file
                -indexed       => &FALSE, If specified (default &FALSE), the first line is an "indexation" line. Useful ONLY with -force_headers 
none
 Attibuts     : The attributes are protected because they are needed by DataReader derived classes
  		_file       		: filename from which we want data
  		_ra_data    		: ref list on row ref **row ref can be HASH or ARRAY, in this class HASH**
  		_rh_headers 		: keys are column names, values are column index
  		_ra_headers 		: list of columns names
  		_ra_wanted_headers 	: list of column names to retrieve 
  		_ra_sorted_keys 	: when data are retrieved in a hash, we need to give a way to retrieve the order of rows
  		_ra_condition_values : value constraint associated to the _ra_column_names
  		_ra_column_names 	: list of column names which have a constraint
 
  		_return_type  		: indicates the return type for the data: as an array or as a hash (default : HASH)

=cut
sub _Init
{
	my $self = shift;
	my %h_args = (
        -force_headers => undef,
        -indexed       => &FALSE,
        @_
         );
         
	$self->SUPER::_Init( @_ );
	$self->_Set( _return_type  => 'HASH',
				 _rh_return    => {},
				 _ra_return    => [],
				 _file         => '',
				 _fh_file      => undef,		# Will be used by DataReader::Region::Tell
				 _rh_headers   => {},
				 _ra_headers   => [],
				 _ra_data      => undef,
				 _key_name     => 'element',
				 _ra_wanted_headers => [],
				 _ra_sorted_keys    => [],
				 _case_sensitive    => &TRUE,
				 _force_headers     => $h_args{-force_headers},
				 _indexed           => $h_args{-indexed}
			);

	$self->__InitConditionsToDefault();

	$self->_Set(
				_ra_condition_values => [],
				_b_condition_value  => &FALSE,
				_ra_column_names     => []
		);
	return;
}
	
=head2 Static procedures DisableInlineC, EnableInlineC

 Usage        : DataReader::DisableInlineC();
                DataReader::EnableInlineC();
 Fonction     : Disable/Enable the C version of the read function.
                EnableInlineC calls __EnableInlineCIfPossible and croaks if &FALSE is returned
                The C function is enabled/disabled for ALL INSTANCES of the object, EVEN FOR DERIVED OBJECTS
 Args         : none

=cut
sub DisableInlineC { $USE_INLINE_C = &FALSE; };
sub EnableInlineC  
{ 
	return if &IsInlineCEnabled();
	
	my $msg = &__EnableInlineCIfPossible();
	if ( $msg eq "" )
	{
		$USE_INLINE_C = &TRUE;
	}
	else
	{
		croak ( "ERROR - InlineC could NOT be enabled - Here is the message: $msg" );
	}
}

=head2 Static function __EnableInlineCIfPossible

 Usage     : my $flg = __EnableInlineCIfPossible
 Fonction  : Enable the C version of the read function, only if possible (modules found etc.)
             This function is called when the module is loaded
 Args      : none
 Return    : "" if enabled, a message if not
 Access    : private
=cut
sub __EnableInlineCIfPossible
{ 
	eval "use DataReader_inlineC";
	if ( $@ ne "" ) 
	{
		return $@;
	}
	return ""; 
};

=head2 Static function IsInlineCEnabled
=cut
sub IsInlineCEnabled { return $USE_INLINE_C; };


=head2 Functions GetHeaders,GetHeadersColumns

 Usage        : my $ra_headers = $o_obj -> GetHeaders();
                my $rh_headers = $o_obj -> GetHeadersColumns();
 Functions    : return the list of headers in the correct order, 
                OR a hash, the keys are the headers and the values are the corresponding col numbers

=cut
sub GetHeaders
{
	my $self = shift;
	return $self->_Get('_ra_headers');
}
sub GetHeadersColumns
{
	my $self = shift;
	return $self->_Get('_rh_headers');
}
sub GetSortedKeys
{
	my $self = shift;
	return $self->_Get('_ra_sorted_keys');
}
	
#---------------------------------
#    Reading the file
#---------------------------------

=head2 Procedure Prepare

 Usage        : $o_obj->Prepare( %h_args );
 Procedure    : initialize attributes (in particular reset _ra_return _rh_return and (REMOVED) _ra_sorted_keys)
                (see NarcisseDataReader::prepare, the difference here is that there is no cache and no VALUE condition)
 Args         : NAMED ARGUMENTS:
               	-from      => $file     The file name, OR a file handler
               	-from_key  => Defaults to $file is $file is a filename, required if $file is a file handler
               	              Only used by DataReader::Whole to manage the cache
               	-where     => $rh_cond  A condition, expressed as a hash:
                       { VALUE => { col1_name => val1, col2_name => val2, ... },
               		     RANGE => { start => x, end => x, start_label => 'q_start', end_label => 'q_end' }
               		   }
               	-case_sensitive => $bool (default &TRUE), set to &FALSE if you want the API to be insensitive.
               		                      Only used with -where => { VALUE } condition.
                -select => [ 'col3_name', 'col4_name' ]   Select only some fields to return - WORKS ONLY WITH -line_type = HASH !
 Returns	  : none

=cut
sub Prepare
{
	my $self = shift;
	my %h_args = (
        -from   => '-from option',	# dummy name, useful for error message generation
        -where  => {},
        -select => [],  
        -force_reading  => 0,
        -case_sensitive => &TRUE,
        @_
         );
        	
	my $file = $h_args{ -from };

	$self->_Set( _file => $file,
				 _case_sensitive => $h_args{ -case_sensitive },
			 );
				 
	# Clearing the return structures we are preparing the object for new data to retrieve
	$self->_Set( '_ra_return' 		=> [],
	             '_rh_return' 		=> {}
	             #(REMOVED)'_ra_sorted_keys'  => []
	            );
	$self->__InitConditionsToDefault();

	my $rh_where = $h_args{ -where };

	# VALUE condition
	if( defined( $$rh_where{ VALUE } ) )
	{
		my $rh_conditions = $$rh_where{ VALUE };
		my @a_col_names   = keys %$rh_conditions;
		my @a_col_values  = map( "\Q$_\E", values %$rh_conditions );
		$self->_Set( '_b_condition_value' => &TRUE,
					 _ra_column_names 	  => \@a_col_names,
					 _ra_condition_values => \@a_col_values,
					 _rh_value_condition  =>  $rh_conditions
					 );
		$self->{ _rh_where_condition }{ value } = &TRUE;
    }

	# RANGE condition
	my $rh_cond = $self->_Get( '_rh_range_condition' );
	if ( defined( $$rh_where{ RANGE } ) )  # Modify default values according to the arguments
	{
		my $rh_range = $$rh_where{ RANGE };

		# Setting the start and end conditions, start and end have to be non-zero values (if no default values are used)
		$$rh_cond{ start }{ value } = $$rh_range{ start } || 0;
		$$rh_cond{ end   }{ value } = $$rh_range{ end   } || &MAX_CHROM_SIZE;

		$$rh_cond{ start }{ label } = $$rh_range{ start_label } || 'start';
		$$rh_cond{ end   }{ label } = $$rh_range{ end_label   } || 'end';

		$self->{ _rh_where_condition }{ range } = &TRUE;
	}

	# Invalid start-end
	if( $$rh_cond{ start }{ value } > $$rh_cond{ end }{ value } )
	{
		my $msg = "ERROR IN REQUESTED RANGE - ( start < end ) is not respected reading  $file \n";
		$msg   .= "start=$$rh_cond{ start }{ value },";
		$msg   .= " end=$$rh_cond{ end }{ value } using default values instead";
		warn($msg);
		$self->_Trace( $msg, '.', 'ID' );

		#We are using default values instead
		$self->{ _rh_where_condition }{ range } = &FALSE;
		$$rh_cond{ start }{ value } = 0;
		$$rh_cond{ end }{ value }   = &MAX_CHROM_SIZE;
	}

	# Selecting specific fields to retrieve
	my $ra_select = $h_args{ -select };
	$self->_Set( _ra_select => $ra_select );

	return;
}


=head2 Procedure __InitConditionsToDefault

 Usage        : $o_obj->__InitConditionsToDefault();
 Procedure    : Set _rh_value_condition, _rh_range_condition, and _rh_where_condition
                with their default values.
 Args         : none
 Returns 	  : none

=cut

sub __InitConditionsToDefault
{
	my $self = shift;

	my $rh_range_condition = {
			start => { value => 0,               label => 'start' },
			end   => { value => &MAX_CHROM_SIZE, label => 'end'},
		};

	$self->_Set(
			_b_condition_value  => &FALSE,
			_rh_value_condition => {},
			_rh_range_condition => $rh_range_condition,
			_rh_where_condition => { range => &FALSE, value => &FALSE }
		);

	return;
}

=head2 Procedure _SetHeaderField

 Usage        : $o_obj->_SetHeaderField( $header_line );
 Procedure    : Retrieve the header fields from the first line, and store them in _ra_headers and _rh_headers
 Args         : The header line as an array
 Returns 	  : none
=cut

sub _SetHeaderField
{
	my $self = shift;
	my ( $ra_headers ) = @_;

	my @a_headers;
	my %h_headers;
	if (defined $ra_headers )
	{
		@a_headers = @$ra_headers;
		if (@a_headers > 0 )
		{
			$a_headers[0]  =~ s/^#//;
			@h_headers{ @a_headers } = (0..$#a_headers);
		}
		$self->_Set( _ra_headers => \@a_headers,
					 _rh_headers => \%h_headers );
	 }
	 else
	 {
		$self->_Set( _ra_headers => undef,
					 _rh_headers => undef );
	 };

	return;
}

=head2 Procedure _AssertLineType

 Usage     : $self -> _AssertLineType();
 Procedure : Croak if line_type is not HASH or ARRAY or if LINE_TYPE is set to ARRAY together with a -select argument
=cut

sub _AssertLineType
{
	my $self = shift;
	my $line_type = $self -> _Get ( '_line_type' );
	if ( $line_type ne 'ARRAY' && $line_type ne 'HASH' )
	{
		croak "ERROR - Illegal value for line_type: $line_type ";
	}
	
	my $ra_select = $self -> _Get ( '_ra_select' );
	if ( scalar (@$ra_select) > 0 && $line_type eq 'ARRAY' )
	{
		croak "ERROR - You MAY NOT use the -select switch with -line_type=ARRAY !";
	}
}

#----------------------------------
#    Retrieving data
#---------------------------------

=head2 function FetchallHashref

 Usage        : $rh_data = $o_obj->FetchallHashref( 'key' );
 Function     : Return the data read in a hash, the key is the value found in column 'key'
 Returns      : a reference on the hash containing the data 
 Args         : key_name : header field used as a the hash key

=cut
sub FetchallHashref { my $self = shift; throw LipmError::ObjectException( $self, 'abstract method' ); };

=head2 function FetchallArrayref

 Usage        : $ra_data = $o_xx->FetchallArrayref();
 Function     : Return the data read in an array
 Returns      : a reference on the array containing the data 
 Args         : none

=cut
sub FetchallArrayref { my $self = shift; throw LipmError::ObjectException( $self, 'abstract method' ); } ;

=head2 Procedure _GetCondSpecificValues

 Usage        : $self->_GetCondSpecificValues
 Procedure    : returns the necessary information for testing the VALUE condition specified by the -where argument
                a reference on an array containing the column indexes of the tested fields
                a reference on an array containing the values requested for the aforementioned fields
 Args         : none
 Returns      : two references on arrays 
=cut
sub _GetCondSpecificValues
{
	my $self = shift;
	
	my @a_attr = qw( _ra_column_names _ra_condition_values _rh_headers );
	my ( $ra_col_names, $ra_vals, $rh_headers ) = $self->_Get( @a_attr );
	my @col_names_number = @{ $rh_headers } { @$ra_col_names };
		
	return ( \@col_names_number, $ra_vals );
}


=head2 Procedure _ValidateValueConditions

 Usage        : $self -> _ValidateValueConditions( )
 Procedure    : iterates among all the condition values to check them, returns 1 if at least one is verified
 Args         :
     $ra_row         => a reference on the line (an array)
     $ra_value_col   => a reference on the column numbers of the tested fields
     $ra_value_val   =>  a reference on the requested values
     $case_sensitive => boolean, if false, will use 'i' flag on regexp.
 Returns      : &FALSE if the condition is not fullfilled, &TRUE if it is


=cut
sub _ValidateValueConditions
{
	my $self = shift;
	my( $ra_row, $ra_value_col, $ra_value_val, $case_sensitive ) = @_;

	my $flag = &FALSE;
	if( $case_sensitive == &FALSE )
	{
		for( my $j=0; $j < @$ra_value_col; ++$j )
		{
			my $val = $ra_value_val -> [ $j ];
			if( $$ra_row[ $ra_value_col -> [ $j ] ] =~ /^$val$/i )
			{
				++$flag;
				last;   # exit from for loop
			}
		}
	}
	else
	{
		for( my $j=0; $j < @$ra_value_col; ++$j )
		{			
			my $val = $ra_value_val -> [ $j ];
			# Please do not replace with an 'eq', because some special chars were escaped in prepare
			# and this will work ONLY with reexp matches !
			if( $$ra_row[ $ra_value_col -> [ $j ] ] =~ /^$val$/ )
			{
				++$flag;
				last;   # exit from for loop
			}
		}
	}

	return $flag;
}

#(REMOVED)=head2 Procedure _GetHashSpecificValues
#
# Usage        : $self->_GetHashSpecificValues( )
# Procedure    : Constructs the hash specific values              
# Args         : none
# Returns      : $key_name_col => the key_name index
#                $ra_sorted_keys => a reference on the array that will store the keys according to the already sorted file order 
#=cut
#
#sub _GetHashSpecificValues
#{
#	my $self = shift;
#	
#	my @a_attr = qw( _rh_headers _key_name );
#	my ( $rh_headers, $key_name ) = $self->_Get( @a_attr );
#	my $ra_sorted_keys = $self->_Get( '_ra_sorted_keys' );
#	if ( !defined $rh_headers->{$key_name} )
#	{
#		croak "ERROR - $key_name is NOT the name of a column";
#	}		
#	return  ( $$rh_headers{ $key_name }, $ra_sorted_keys ) ;
#}# _GetHashSpecificValues


=head2 Function _GetHeaderColumn

 Usage        : my $col       = $self -> _GetHeaderColumn ( $ra_header, $col_name );
                my @a_columns = $self -> _GetHeaderColumn ( $ra_header, @a_col_names );
 Function     : Returns the column position, given a header name
 Args         : $ra_header A ref to the array representing the header
                $key       The key  to search for
 Returns 	  : In array context:  The positions of the keys in the headers
                In scalar context: Only the FIRST key is considered
 
=cut
sub _GetHeaderColumn
{
	my $self = shift;
	my ($ra_headers,@a_col_names) = @_;
	
	my @a_indexes;
	foreach my $col_name (@a_col_names)
	{
		my $i = 0;
		while ( $i <= $#$ra_headers && $ra_headers -> [$i] ne $col_name ) { $i++; };
		croak ("ERROR - key name $col_name absent from header") if ( $i > $#$ra_headers);
		push @a_indexes,$i;
	}
	
	return wantarray ? @a_indexes : $a_indexes[0];
}

=head2 Procedure Finish

 Title        : Finish
 Usage        : $o_xx->Finish( none );
 Procedure    : Empty the memory cache
 Args         : none

=cut
sub Finish {};

1;
