#
# $Id: FileFormats.pm 1081 2013-03-05 13:47:04Z manu $
#

package FileFormats;

use strict;
use warnings;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

FileFormats - Help managing some BioInformatics file formats

=head1 SYNOPSIS
 my $o_ffmt = New FileFormats();
 my $format = $o_ffmt->GuessFormat('data.gff3.gz');
 my $ext    = $o_ffmt->GetExtension( $format );
 my $rh_info= $o_ffmt->GetFormatInfo( 'FASTA' );
 print $o_ffmt->GetFormats();

=head1 DESCRIPTION

 Thanks to this object, you can guess a file format from its file name, or select a correct file extension, knowing the format.
 This object can also suggest a wise name for your format-specific objects, etc.
 You may pass a compressed file name to GuessFormat.
 
=cut

use General;
use Error qw(:try);
use base qw( LipmObject );
use LipmError::ParameterException;
use LipmError::ObjectException;

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1081 $ =~ /\d+/g); $r[0]};
}

our %H_FORMATS = (
	GFF3    => { extensions    => [ 'gff3'],
	             preferred_ext => 'gff3',
		         class         => 'GFF3'
		       },
	GTF     => { extensions    => [ 'gtf'],
	             preferred_ext => 'gtf',
		         class         => 'GTF'
		       },
	ENSEMBL => { extensions    => [ 'ensembl' ],
				 preferred_ext => 'ensembl',
				 class         => 'Ensembl'
			   },
 	GENBANK => { extensions    => [ 'genbank', 'gbk', 'gb'],
	             preferred_ext => 'gbk',
		         class         => 'Genbank'
		       },
	EMBL    => { extensions    => [ 'embl' ],
				 preferred_ext => 'embl',
				 class         => 'Embl'
			   },
	FASTA   => { extensions    => [ 'fa', 'fasta' ],
	             preferred_ext => 'fa',
	             class         => 'Fasta' 
	           },
	NARCISSE => { extensions => [ 'ncs' ],
	              preferred_ext => 'ncs',
	              class         => 'Narcisse'
	            },
	NARCISSE_GENETIC => { extensions => [ 'ncsg' ],
	                      preferred_ext => 'ncsg',
	                      class         => 'Genetic'
					    }
	 );

# keys = The file extensions,values = The corresponding format
# This hash will be computed from %H_FORMATS by the constructor
our $RH_EXTENSIONS_FMT;

sub _Init
{
	my $self = shift;
	if ( ! defined $FileFormats::RH_EXTENSIONS_FMT )
	{
		my %h_extensions_fmt;
		foreach my $fmt (keys %FileFormats::H_FORMATS)
		{
			my $ra_ext = $FileFormats::H_FORMATS{$fmt}->{extensions};
			foreach my $ext (@$ra_ext)
			{
				if (exists $h_extensions_fmt{$ext})
				{
					throw LipmError::ParameterException('%H_FORMATS','value', -text => "Duplicate extension in %H_FORMATS");
				}
				$h_extensions_fmt{$ext} = $fmt;
			}
		}
		$FileFormats::RH_EXTENSIONS_FMT = \%h_extensions_fmt;
	}
}
		
=head2 procedure GuessFormat

 Title    : GuessFormat
 Usage    : 1/ my $fmt    = $o_ffmt -> GuessFormat ( $file_name );
            2/ my@a_ffmts = $o_ffmt -> GuessFormat ( $file_name );
 Function : Guess and return the file format from its name
 Args     : $name some file name
 Return   : 1/ The guessed format, or "" if the format could not be guessed
            2/ An array containing:
                  - The guessed format
                  - The file name (ie without any extension)
                  - The file extension
                  - The compression path, if any
               The array is empty if the format is unknown
 
 Global   : $RH_EXTENSIONS_FMT
 
=cut
sub GuessFormat
{
	my $self   = shift;	
	my ($file_name) = @_;
	
	my ($name,$ext) = ("","");
	my $compression_path = &GetCompressPath( $file_name );
	
	# file.gff3
	if ( $compression_path eq "" )
	{
		($name,$ext) = ($file_name =~ /^(.+)\.(\w+)$/);
	}
	
	# file.gff3.gz
	else
	{
		($name,$ext) = ($file_name =~ /^(.+)\.(\w+)\.\w+$/);
	}
	
	if ( exists $FileFormats::RH_EXTENSIONS_FMT -> { $ext } )
	{
		my $fmt = $FileFormats::RH_EXTENSIONS_FMT -> { $ext };
		return wantarray?($fmt,$name,$ext,$compression_path):$fmt;
	}
	else
	{
		return (wantarray)?():"";
	}
}

=head2 procedure GetExtension

 Title    : GetExtension
 Usage    : my $ext    = $o_ffmt->GetExtension( $format );
 Function : Return the extension you should use when you give a name to a file, from the file format.
 Args     : $format some file format
 Return   : The extension you should use
 Error    : throw a ParameterExtension if the format is unknown
 
 Global   : %H_FORMATS
 
=cut
sub GetExtension
{
	my $self = shift;
	my ($fmt)= @_;
	if (exists $H_FORMATS{$fmt})
	{
		return $H_FORMATS{$fmt} -> { preferred_ext };
	}
	else
	{
		throw LipmError::ParameterException('fmt','value', -text => "Format unknown: $fmt");
	}
}
=head2 procedure GetClassName

 Title    : GetClassName
 Usage    : my $ext    = $o_ffmt->GetClassName( $format );
 Function : Return the class name you should use when dealing with the format
 Args     : $format some file format
 Return   : The class name you should use
 Error    : throw a ParameterExtension if the format is unknown
 
 Global   : %H_FORMATS
 
=cut
sub GetClassName
{
	my $self = shift;
	my ($fmt)= @_;
	if (exists $H_FORMATS{$fmt})
	{
		return $H_FORMATS{$fmt} -> { class };
	}
	else
	{
		throw LipmError::ParameterException('fmt','value', -text => "Format unknown: $fmt");
	}
}

=head2 procedure GetFormatInfo

 Title    : GetFormatInfo
 Usage    : my %h_info = $o_fmt -> GetFormatInfo( 'format' );
 Function : Return all information about a format
 Args     : $format some file format
 Return   : A ref to a hash with keys: extension, preferred_ext, class
 Error    : throw a ParameterExtension if the format is unknown
 
 Global   : %H_FORMATS
 
=cut
sub GetFormatInfo
{
	my $self = shift;
	my ($fmt) = @_;
	if (exists $H_FORMATS{$fmt})
	{
		return $H_FORMATS{$fmt};
	}
	else
	{
		throw LipmError::ParameterException('fmt','value', -text => "Format unknown: $fmt");
	}
}

=head2 procedure GetFormattedInfo

 Title    : GetFormattedInfo
 Usage    : print $o_fmts->Getformattedinfo();
 Function : Return the list of allowed formats, pretty-formatted to be printed
 Args     : none
 Return   : A string to print
 
 Global   : $RH_EXTENSIONS_FMT
 
=cut
sub GetFormattedInfo
{
	my $self = shift;
	my $blabla = "\n";
	$blabla .= "!---------------------------------!\n";
	$blabla .= "! AUTOMATIC FORMAT DETERMINATION  !\n";
	$blabla .= "!---------------------------------!\n";
	
	$blabla .= sprintf ( "! File extension ! Guessed format !\n" );
	$blabla .= sprintf ( "!----------------!----------------!\n" );
	foreach my $key (sort { $FileFormats::RH_EXTENSIONS_FMT->{$a} lt $FileFormats::RH_EXTENSIONS_FMT->{$b} } keys %$FileFormats::RH_EXTENSIONS_FMT)
	{
		$blabla .= sprintf ( "!  %10s    !     %10s !\n",$key,$FileFormats::RH_EXTENSIONS_FMT->{$key} );
	};
	$blabla .= sprintf ( "!----------------!----------------!\n" );
	
	return $blabla;
}
	
=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
