#
# $Id: Splitter.pm 945 2012-06-27 07:51:23Z manu $
#

package Splitter;

use strict;
use warnings;

use base qw( LipmObject );

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

=head1 NAME

Splitter - Split an annotation file

=head1 SYNOPSIS
 my $o_split = New Splitter::Fasta( $src_filename);
 my $ra_contigs = $o_split -> GetContigs( $regexp );
 $o_split -> CopyContigs ( $dst_filename, $regexp );

=head1 DESCRIPTION

 A contig is a sequence (fasta files), or the annotation related to a sequence (gff3, genbank etc. files)
 Splitter is an abstract class, useful things are implemented in the derived classes.
 The file is read and indexed (the position of each sequence, is kept in memory), and it is possible to 
 copy a file containing only a limited number of contigs
 
=cut

use General;
use Error qw(:try);
use LipmError::ParameterException;
use LipmError::ObjectException;

use File::Temp qw/ :seekable /;

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 945 $ =~ /\d+/g); $r[0]};
}

=head2 procedure _Init

 Title        : Constructor
 Usage        : Use child classes (AnnotatedSeqFile::GenBank, AnnotatedSeqFile::GFF3, etc.)
 Function     : Initialize the object: open the file, copy it if compressed, then call the _Index() procedure
 Args         : $src_filename The source file (may be compressed, see GetStreamIn)
 Globals      : none

=cut

sub _Init
{
	my $self = shift;
	my ($src_filename) = @_;
	$self->SUPER::_Init(@_);
	
	$self -> _Set ( '__filename', $src_filename );
	
	# If the input file is compressed (or stdin, see GetStreamIn)), we first copy the file to a temporary
	if ($src_filename eq '' || $src_filename eq '-' || $src_filename eq 'stdin' || &GetCompressPath( $src_filename ) ne '' )
	{
		my $fh_out = new File::Temp();
		my $fh_in  = &GetStreamIn ( $src_filename );
		while ( my $line = <$fh_in> )
		{
			print $fh_out $line;
		}
		$self -> _Set('_fh_contigs',$fh_out);
	}
	else
	{
		my $fh_in  = &GetStreamIn ( $src_filename );
		$self -> _Set('_fh_contigs',$fh_in);
	}
	$self -> _Set ('_rh_contigs',{});
	$self -> _Set ('_rh_contig_size',{});
	$self -> _Index();
	
}

=head function GetFileHeader
 return a fileheader, may be overloaded in derived classes (only used in GFF3 files now)

=cut
sub GetFileHeader
{
	return '';
}

=head function GetContigSize
 Title    : GetContigSize
 Usage    : my $size = $o_splitter -> GetContigSize('some_contig');
 Function : Return the "size" of the contig, the unit depends on the derived class
            Return 0 if the contig does not exist

=cut
  
sub GetContigSize
{
	my $self = shift;
	my ($ctg) = @_;
	my $rh_contig_size = $self -> _Get ( '_rh_contig_size' );
	return (defined $rh_contig_size -> { $ctg })?$rh_contig_size -> { $ctg }:0;
}

=head2 procedure _Index

 Title        : _Index
 Usage        : $self -> _Index()
 Function     : Initializes the _h_contigs hash (key = contig name, val = offset in the file)
 Args         : None
 Globals      : none

=cut
sub _Index
{
	throw LipmError::ObjectException( 'Splitter', 'abstract method' );
}

=head2 function GetContigNames

 Title    : GetContigNames
 Usage    : my $ra_contigs = $o_splitter->GetContigNames( $regexp );
            my $ra_contigs = $o_splitter->GetContigNames( '\Qcontig1.2\E' );
 Function : return all contig Ids matching the $regexp.
 Returns  : $ra_ids, a sorted list ref to all ids.
 Args     : $regexp, a regular expression.
            will be used as ($id =~ /$regexp/);
            if $regexp is undef, all ids are returned.
 Note     : It can be a good idea to excape special characters like '.' with '\Q'.$id.'\E'

=cut
sub GetContigNames
{
	my $self     = shift;
	my ($regexp) = @_;

	my $rh_contigs = $self -> _Get ( '_rh_contigs' );
	my @a_ids = sort keys %$rh_contigs;
	return \@a_ids if( !defined $regexp );

	# $regexp is defined, search through the list
	my $ra_filtered_ids = [];
	foreach my $id ( @a_ids )
	{
		push( @$ra_filtered_ids, $id) if( $id =~ /$regexp/ );
	}
	return $ra_filtered_ids;
}

=head2 function GetContigs

 Title    : GetContigs
 Usage    : my $r_contigs = $o_splitter->GetContigs( \@a_contigs );
 Function : For each id in the array, return the corresponding contig
            If some id is not found, throw an exception
            If the array is undef, throw an exception
            It the array is empty, return ''
            NOTE - If you want to retrieve some contigs, this is much more performant than using the Regex version,
            as we do not have to scan all the indexed keys !
 Returns  : $r_contigs, a ref to string with the extracted contigs, formatted as in the original file
 Args     : $ra_contigs, ... A list of contigs to retrieve

=cut

sub GetContigs
{
	my $self         = shift;;
	my ($ra_contigs) = @_;
	throw LipmError::ParameterException('ra_contigs','type') unless (&SafeIsArray ($ra_contigs));
	$self->AssertAllFound($ra_contigs);
	if ( @$ra_contigs == 0 )
	{
		my $seq = "";
		return \$seq;
	}
	else
	{
		return $self -> _GetContigs($ra_contigs);
	}
}
sub _GetContigs{throw LipmError::ObjectException( 'AnnotatedSeqFileReader', 'abstract method' );}; 
=head2 function GetContigsRegex

 Title    : GetContigsRegex
 Usage    : my $r_contigs = $o_splitter->GetContigsRegex( $regexp );
            my $r_contig  = $o_splitter->GetContigsRegex( '\Qcontig1.2\E' );
 Function : return all contigs whose Ids match the $regexp.
 Returns  : $r_contigs, a ref to string with the extracted contigs, formatted as in the original file
 Args     : $regexp, a regular expression, see GetContigNames

=cut

sub GetContigsRegex
{
	my $self     = shift;
	my ($regexp) = @_;
	my $ra_contigs = $self -> GetContigNames( $regexp );
	return $self -> GetContigs( $ra_contigs );
}

=head2 function GetContig

 Title    : GetContig
 Usage    : my $r_contigs = $o_splitter->GetContig( 'contig' );
 Function : return the contig corresponding to the id passed by parameter, or through an exception
            NOTE - If you want to retrieve a given contig, this is much more performant than using the Regex version,
            as we do not have to scan all the indexed keys !
 Returns  : $r_contigs, a ref to string with the extracted contigs, formatted as in the original file
 Args     : 'contig', the contig to retrieve
 Note     : GetContig calls GetContigs (could be the opposite) for performance reason

=cut

sub GetContig
{
	my $self     = shift;
	my ($contig) = @_;
	if ( ! defined ( $contig) ) 
	{
		throw LipmError::ParameterException( 'contig', 'missing' );
	}
	return $self->GetContigs( [ $contig ] );
}

=head2 procedure CopyContigsRegex, CopyContigs, Copycontig

 Title    : CopyContigs
 Usage    : $o_splitter->CopyContigsRegex( $dst_filename, $regexp );
            $o_splitter->CopyContigs( $dst_filename, \@a_contigs );
            $o_splitter->GetContig( $dst_filename, 'contig' );
 Function : Copy each contig to dst_filename (should be writable)
            For the meaning of the second parameter, see the corresponding Get functions...
            ...But please note that the method GetContigs is NOT used, to get a lower memory print.
 Args     : $dst_filename The destination filename (should be writeable)
            $regexp, a regular expression, see GetContigNames

=cut
sub CopyContig
{
	my $self     = shift;
	my ($dst_filename,$contig) = @_;
	if ( ! defined ( $contig) ) 
	{
		throw LipmError::ParameterException( 'contig', 'missing' );
	}
	return $self -> CopyContigs( $dst_filename, [ $contig ] );
}

sub CopyContigsRegex
{
	my $self     = shift;
	my ($dst_filename, $regexp) = @_;
	my $ra_contigs = $self -> GetContigNames( $regexp );
	$self -> CopyContigs( $dst_filename, $ra_contigs );
}

sub CopyContigs
{
	my $self         = shift;
	my ($dst_filename, $ra_contigs) = @_;
	throw LipmError::ParameterException('ra_contigs','type') unless (&SafeIsArray ($ra_contigs));
	$self->AssertAllFound($ra_contigs);
			
	if ( @$ra_contigs == 0 )
	{
		return;
	}
	else
	{
		$self -> _CopyContigs($dst_filename, $ra_contigs);
	}
}
sub _CopyContigs{throw LipmError::ObjectException( 'AnnotatedSeqFileReader', 'abstract method' );};

=head2 procedure CopyContigsRegex, CopyContigs, Copycontig

 Title    : AssertAllFound
 Usage    : $self->AssertAllFound($ra_contigs);
 Function : If at least 1 $id_config is not found in the list of contigs, throw an exception
 Args     : $ra_contigs: ref to an array of contig ids

=cut
sub AssertAllFound
{
	my $self         = shift;
	my ($ra_contigs) = @_;

	my @a_not_found;
	my $rh_contigs = $self -> _Get ( '_rh_contigs' );
	foreach my $id_contig (@$ra_contigs)
	{
		if ( ! exists $rh_contigs -> { $id_contig } )
		{
			push @a_not_found, $id_contig;
		}
	}
	if ( @a_not_found != 0 )
	{
		my $filename = $self -> _Get ( '__filename' );
		my $msg = "file $filename : Some contigs are not found: " . join ( ' ', @a_not_found);
		throw LipmError::ParameterException('ra_contigs','value',-text => "$msg\n" );
	}
}
	
=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;
