#!/usr/bin/perl -w
#
# $Id: Ensembl.pm 1081 2013-03-05 13:47:04Z manu $
#

package AnnotatedSeqFileReader::Ensembl;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=head1 NAME

 AnnotatedSeqFileReader::Ensembl - Query the Ensembl database to load an AnnotatedSeq object


=head1 SYNOPSIS  
    (1) BioFileConverter.pl --input-ann-file all.ensembl
    (2) BioFileConverter.pl --input-ann-file gorilla_gorilla.all.ensembl
    (3) BioFileConverter.pl --input-ann-file gorilla_gorilla.1.ensembl --output-ann-file gorilla.gff3
    (4) BioFileConverter.pl --input-ann-file gorilla_gorilla.1.ensembl --output-ann-file gorilla.gff3 --output-seq-file gorilla_seq.fa
    
    (1) Retrieve the list of species available
    (2) Retrieve the list of chromosomes available for the species gorilla_gorilla
    (3) Retrive annotation data for chromosome 1 of gorilla_gorilla and store them to a gff3 file
    (4) Same as (3), but store also the sequence in a fasta file
	See AnnotatedSeq

=head1 DESCRIPTION

=cut 

use strict;
use warnings;
use Carp;

use Error qw(:try);
use Bio::SeqIO;
use Bio::Seq;
use General;
use AnnotatedSeq;
use Data::Dumper;
use Bio::EnsEMBL::Registry;

use base qw(AnnotatedSeqFileReader);

sub Load
{
	my $self       = shift;
	my $o_annot    = $self -> { _o_annot };
	my $input_file = $o_annot->{__in_ann_file};

	my $registry = 'Bio::EnsEMBL::Registry';
	if ( exists $ENV{'ENSEMBL_REGISTRY'} )
	{
		$registry -> load_all();
	}
	else
	{
		$registry -> load_registry_from_db(
		-host => 'ensembldb.ensembl.org',
		-user => 'anonymous'
		);
	}

	# Asking for the list of species
	if ( $input_file =~ /^all\./ )
	{
		$self -> __LoadSpeciesList($registry);
		return;
	};

	# Asking for the list of chromosomes
	my ( $species, $chromosome ) = split ( '\.', $input_file );
	if ( $chromosome eq 'all' )
	{
		$self -> __LoadChromosomesList($registry, $species);
		return;
	};
	
	# downloading data for a given chromosome/species
	#my $db_adaptor = $registry->get_DBAdaptor($species, 'core');
	#print STDERR "db_adaptor $db_adaptor\n";
	
	my $slice_adaptor = $registry->get_adaptor($species, 'core', 'slice');
	# MANU - Pas sur que ca marche quand on a specifie un fichier de configuration !!!
	if ( !defined ($slice_adaptor) )
	{
		$slice_adaptor = GetSliceAdaptor($registry);
	}
	
	#print STDERR "slice_adaptor $slice_adaptor\n";
	
	my $slice = $slice_adaptor->fetch_by_region( 'chromosome', $chromosome );
	
	my $coord_sys  = $slice->coord_system()->name();
	my $seq_region = $slice->seq_region_name();
	my $start      = $slice->start();
	my $end        = $slice->end();
	my $strand     = $slice->strand();

	print "species $species: $coord_sys chromosome $seq_region $start-$end (strand $strand)\n";
	
	my $ra_genes = $slice->get_all_Genes();
	my $nb_genes = @$ra_genes;
	$| = 1;
	my $i = 0;
	while (my $o_egene = shift @$ra_genes)
	{
		$i++;
#last if ( $i>=100);  # (for debug)
		my $biotype = $o_egene -> biotype();
		
		# skip pseudogenes
		next if ( $biotype =~ /pseudogene/ );
		my $coding = ($biotype =~ /protein_coding/) ? 1 : 0;
		
		print "\rGene $i" . "/" . "$nb_genes (".$o_egene -> display_id().")". 20x' '."";
		
		my %h_param = ( start     => $o_egene -> start(),
						end       => $o_egene -> end(),
						strand    => $o_egene -> strand(),
						locus_tag => $o_egene -> stable_id(),
						ID        => $o_egene -> stable_id());
						
		# Nom du gene
		# Si EntrezGene est renseigne on l'utilise en nom de gene.
		$h_param{Name} = $self->__GetXref($o_egene, 'EntrezGene');
		$h_param{Name} = $o_egene->display_id() if ($h_param{Name} eq '');
		$h_param{note} = $self->__GetEntrezNote($o_egene);
		
		# If not coding, we add an ncRNA, not a gene
		my $o_gene;		
		if ( $coding == &FALSE )
		{
			my $o_ncrna = New Locus::Ncrna ( %h_param );
			$o_annot -> AddNcrna ( $o_ncrna );
			next;
		}
		
		# coding, we add a gene, then later a transcript
		else
		{
			$o_gene = New Locus::Gene ( %h_param );
			$o_annot -> AddGene ( $o_gene );
		}

		my $ra_emrna = $o_egene -> get_all_Transcripts();
		while (my $o_erna = shift @$ra_emrna)
		{
			my %h_param = ( start     => $o_erna -> start(),
							end       => $o_erna -> end(),
							strand    => $o_erna -> strand(),
							locus_tag => $o_erna -> stable_id(),
							ID        => $o_erna -> stable_id() );

			# transcript_id => RefSeq mRNA or RefSeq mRNA predicted
			# protein_id    => RefSeq peptide or RefSeq peptide predicted
			# If none exist, the EnsEMBL id is reused
			$h_param{transcript_id} = $self->__GetXref($o_erna, 'RefSeq mRNA');
			$h_param{transcript_id} = $self->__GetXref($o_erna, 'RefSeq mRNA predicted') if ($h_param{transcript_id} eq '');
			$h_param{transcript_id} = $o_erna->display_id()                              if ($h_param{transcript_id} eq '');
			$h_param{note}          = $self->__GetEntrezNote($o_erna);
			
			# Some transcripts are NOT protein_coding, we declare them as ncRNA
			my $t_biotype = $o_erna -> biotype();
			my $t_coding  = ($t_biotype =~ /protein_coding/) ? 1 : 0;
			
			my $o_rna;
			if ($t_coding==&TRUE)
			{
				$o_rna = New Locus::Mrna ( %h_param ); 
				$o_gene -> AddmRNA ( $o_rna );
			}
			else
			{
				$o_rna = New Locus::Ncrna ( %h_param );
				$o_annot -> AddNcrna ( $o_rna );
			}
			
			my $ra_exons = $o_erna -> get_all_Exons();
			while (my $o_eexon = shift @$ra_exons )
			{
				my %h_param = ( start     => $o_eexon -> start(),
								end       => $o_eexon -> end(),
								strand    => $o_eexon -> strand(),
								locus_tag => $o_eexon -> stable_id(),
								ID        => $o_eexon -> stable_id());
				#Modified by Thomas 28/11/2007
				#removed by manu 20/3/2012 => Prefer using Locus::Mrna::GetTranslationEnds()
				#$h_transcript_param{cds_start} = $o_transcript->cdna_coding_start;
				#$h_transcript_param{cds_end} = $o_transcript->cdna_coding_end;
				#$h_transcript_param{product} = $o_gene->description if (defined $o_gene->description);
				
				my $o_exon = New Locus::Exon ( %h_param );
				$o_rna -> AddExon ( $o_exon );
			}
			if ( $t_coding == &TRUE )
			{
				# Trying the same stuff as for gene and transcript !
				my $protein_id = $self->__GetXref($o_erna, 'RefSeq peptide');
				$protein_id = $self->__GetXref($o_erna, 'RefSeq peptide predicted') if ($protein_id eq '');
				$protein_id = $o_erna->display_id()                                 if ($protein_id eq '');

				my $o_etranslation = $o_erna -> translation();
				if (defined $o_etranslation)
				{
					my $ra_etexons = $o_erna -> get_all_translateable_Exons();
					while (my $o_eexon = shift @$ra_etexons )
					{
						my %h_param = ( start     => $o_eexon -> start(),
										end       => $o_eexon -> end(),
										strand    => $o_eexon -> strand(),
										locus_tag => $o_eexon -> stable_id(),
										protein_id=> $protein_id );
										
						my $o_cds = New Locus::CdsRegion ( %h_param );
						$o_rna -> AddCDS ( $o_cds );
					} # while $o_eexon
				} # if defined $o_etranslation
			} # if $t_coding
		} # while $o_erna
	} # while $o_egene
	
	print "\n Done with the genes\n";

	my $ra_repeats = $slice->get_all_RepeatFeatures();
	my $nb_repeats = @$ra_repeats;
	$i = 0;
	while (my $o_erepeat = shift @$ra_repeats)
	{
		$i++;
#last if ( $i>=10);
		print "\rRepeat $i" . "/" . "$nb_repeats (".$o_erepeat -> display_id().")". 40x' '."";

		my %h_param = ( start     => $o_erepeat -> start(),
						end       => $o_erepeat -> end(),
						#strand    => $o_erepeat -> strand(),
						strand    => 1,
						locus_tag => $o_erepeat -> display_id());

		my $o_repeat = New Locus::Repeat(%h_param);
		
		$o_annot->AddRepeat($o_repeat);
	}
	print "\n Done with the repeats\n";

} # Load


# this is a PATCH, j'espere provisoire (manu)
sub GetSliceAdaptor
{	
	return undef;
	my ($registry) = @_;
	my $slice_adaptor;		
	my $ra_adaptors = $registry->get_all_adaptors();
	foreach my $o_adaptor (@$ra_adaptors)
	{
		if ( $o_adaptor->isa('Bio::EnsEMBL::DBSQL::SliceAdaptor') )
		{
			$slice_adaptor = $o_adaptor;
			last;
		}
	}
	return $slice_adaptor;
}

=head2 procedure __LoadSpeciesList

 Title        : __LoadSpeciesList
 Usage        : $self->__LoadSpeciesList( $registry );
 Function     : Prints the list of available species 
 Args         : $registry An Ensembl $registry

=cut
sub __LoadSpeciesList
{
	my $self = shift;
	my ($registry) = @_;
	print "Retrieving the list of species from the database...\n\n";
	my $ra_db_adaptors = $registry -> get_all_DBAdaptors();
	foreach my $db_adaptor (@$ra_db_adaptors)
	{
		my $db_connection = $db_adaptor -> dbc();
		
		printf(
			"species/group\t%s/%s\ndatabase\t%s\nhost:port\t%s:%s\n\n",
			$db_adaptor->species(),
			$db_adaptor->group(),
			$db_connection->dbname(),
			$db_connection->host(),
			$db_connection->port()
		);
	}
	return;
}

=head2 procedure __LoadChromosomesList

 Title        : __LoadChromosomesList
 Usage        : $self->__LoadChromosomesList( $registry );
 Function     : Prints the list of available chromosomes for a given species
 Args         : $registry An Ensembl $registry
				$species  The species to consider

=cut
sub __LoadChromosomesList
{
	my $self = shift;
	my ($registry,$species) = @_;
	
	my $slice_adaptor = $registry->get_adaptor($species,'core','slice');
	
	# MANU - Pas sur que ca marche quand on a specifie un fichier de configuration !!!
	if ( !defined ($slice_adaptor) )
	{
		$slice_adaptor = GetSliceAdaptor($registry);
	}
		
	my $ra_slices     = $slice_adaptor -> fetch_all('chromosome');
	my (@a_chr);
	foreach my $slice (@$ra_slices)
	{
		push @a_chr, { name => $slice->seq_region_name(), length => $slice->seq_region_length() };
	}
	
	print "Retrieving the list of chromosomes for the species $species (from biggest to smallest)...\n";
	foreach my $c ( sort { $b->{length} <=> $a->{length} } @a_chr )
	{
		print $c -> { name }.":\t" . $c -> { length }." bp\n";
	}
	return;
}

=head2 procedure __GetXref

 Title        : __GetXref
 Usage        : $self->__GetXref($o_egene, $db_name);
 Function     : returns the id of the given gene in $db_name. If there is no id for this database, the function returns "".
 Returns      : identifier of given gene or "" if the gene does not have a known identifier for this database
 Args         : $o_egene A Bio::EnsEMBL::Gene object
 				$db_name Name of the external database.

=cut

sub __GetXref{
	my $self = shift;
	my ($o_gene,$db_name) = @_;
#print STDERR "KOUKOU " . $o_gene->display_id . "\n";
	my $ra_dbentries = $o_gene->get_all_DBLinks;
	foreach my $db_entry (@$ra_dbentries)
	{
#print STDERR $db_entry->db_display_name() . " => " . $db_entry->display_id . "\n";
		if ( $db_entry->db_display_name() eq $db_name )
		{
			return $db_entry->display_id;
		}
	}
	return "";
}
=head2 procedure __GetEntrezNote

 Title        : __GetEntreznote
 Usage        : $self->__GetEntreNote($o_egene);
 Function     : returns note field from the Entrez external database.
 Returns      : identifier of given gene or "" if the gene does not have a known identifier for this database
 Args         : $o_egene A Bio::EnsEMBL::Gene object

=cut
sub __GetEntrezNote
{
	my $self = shift;
	my ($o_ensembl_obj) = @_;
	
	my @a_dbe = $o_ensembl_obj -> get_all_DBEntries('EntrezGene');
	my $ra_dbe   = $a_dbe[0];
	if ( defined ($ra_dbe) && @$ra_dbe>0 )
	{
		return $ra_dbe->[0]->description();
	}
	else
	{
		return "";
	}
}


sub __GetEntrezData_OBSOLETE
{
	my $self = shift;
	my ($o_ensembl_obj,$rh_params) = @_;
	
	my @a_dbe = $o_ensembl_obj -> get_all_DBEntries('EntrezGene');
	my $ra_dbe   = $a_dbe[0];
	return unless ( defined ($ra_dbe) && @$ra_dbe>0 );
	my $dbe = $ra_dbe->[0];
	
	my $alias = $dbe->display_id();
	if ( $o_ensembl_obj->isa('Bio::EnsEMBL::Gene') )
	{
		$rh_params->{Alias} = $alias unless ( exists $rh_params->{Name} && $rh_params->{Name} eq $alias );
	}
	elsif ( $o_ensembl_obj->isa('Bio::EnsEMBL::Transcript') )
	{
		$rh_params->{Alias} = $alias unless ( exists $rh_params->{transcript_id} && $rh_params->{transcript_id} eq $alias );
	}
	$rh_params->{note} = $dbe->description();
}
	
	
	
sub __GetDbEntries_OBSOLETE
{
	my $self = shift;
	my ($ra_db_entries) = @_;
	
	my %h_results = ();
	#foreach my $dbe (@$ra_db_entries)
	#{
	#	$h_results{$dbe->dbname()} = $dbe->display_id();
	#}
	my $dbe = $ra_db_entries->[0];
	if (defined $dbe)
	{
		$h_results{ 'Alias' } = $dbe->display_id();
		$h_results{ 'Note'  } = $dbe->description();
	}
	return %h_results;
}

=head2 procedure LoadGenomicSequence

 Title        : LoadGenomicSequence
 Usage        : my $r_seq = $o_gbk->LoadGenomicSequence();
 Function     : Load the genomic sequence from the Ensembl database
 Returns      : The genomic sequence
 Args         : none
 Globals      : none

=cut
#sub LoadGenomicSequence {};
sub LoadGenomicSequence
{
	my $self      = shift;
	my $o_annot   = $self -> { _o_annot };
	my $input_file= $o_annot->{__in_ann_file};

	if ( $input_file =~ /^all\./ )
	{
		my $seq = "";
		return \$seq;
	}
	else
	{
		my ( $species, $chromosome ) = split ( '\.', $input_file );
		if ( $chromosome eq 'all' )
		{
			my $seq = "";
			return \$seq;
		}
		else
		{
			my $registry = 'Bio::EnsEMBL::Registry';
			if ( exists $ENV{'ENSEMBL_REGISTRY'} )
			{
				$registry -> load_all();
			}
			else
			{
				$registry -> load_registry_from_db(
				-host => 'ensembldb.ensembl.org',
				-user => 'anonymous'
				);
			}

			my $slice_adaptor  = $registry->get_adaptor($species,'core','slice');
			my $slice          = $slice_adaptor->fetch_by_region( 'chromosome', $chromosome );
			my $length         = $slice -> seq_region_length ();
			my $masked_slice   = $slice -> get_repeatmasked_seq( undef, 1 ); # Get the sequence, mask everything, but softmasking
			my $genomic_seq    = "";
			
			my $already_stored = 0;
			my ($chunk_start,$chunk_end,$chunk_size) = (1,1,100000);
			my $nb_iterations;
			if ( $length % $chunk_size > 0 )
			{
				$nb_iterations = int ( $length / $chunk_size ) + 1;
			}
			else
			{
				$nb_iterations = $length / $chunk_size;
			}
			
			for ( my $i=0; $i<$nb_iterations; $i++ )
			{
				my $chunk_start     = 1 + $chunk_size * $i;
				my $chunk_real_size =($i==$nb_iterations-1)?$length-$chunk_start+1:$chunk_size;
				my $chunk_end       = $chunk_start + $chunk_real_size - 1;
				
				# Try 3 times before giving up
				my $error = &FALSE;
				for (my $j=0; $j<=2; $j++)
				{
					my $seq = $masked_slice -> subseq($chunk_start,$chunk_end);
					if ( length ($seq) < $chunk_real_size )
					{
						print STDERR "\nreceived: " . length ($seq) . " wanted: $chunk_real_size\n";
						$error = &TRUE;
					}
					else
					{
						$genomic_seq .= $seq;
						print "\rreceived: $chunk_start..$chunk_end total length = " . length ($genomic_seq) . 40x" ";
						last;
					}
				}
				if ($error == &TRUE)
				{
					throw Error::Simple("ERROR - Could not retrieve the sequence !");
				}
			}
			print "\n Done with the sequence\n";

			if ( length($genomic_seq) == 0 )
			{
				throw Error::Simple("ERROR - There is not genomic sequence embedded in this ensembl database");
			}
			else
			{
				return \$genomic_seq;
			}
		}
	}
}


sub DetectContig
{	
	my $self       = shift;
	my $o_annot    = $self    -> { _o_annot };
	my $input_file = $o_annot -> { __in_ann_file };
	my ($contig);
	if ($input_file=~/^(all)\.?/)
	{
		$contig = 'all';
	}
	else
	{
		my @a_file = split('\.',$input_file);
		$contig = $a_file[1];
	}
	$o_annot -> { __contig } = $contig;
}











1;


=head2 procedure DownloadEnsemblData

 Title        : DownloadEnsemblData
 Usage        : DownloadEnsemblData($o_sp,$o_param)
 Fonction     : Download masked fasta sequences associated to $org et creates the release file.
 Args         : $o_sp (the species observer)
                $o_param IS MODIFIED (the parameters FILE_SEQ_* are added, where * stands for the chromosome names)

=cut


#sub DownloadEnsemblData
#{
	#my ($o_sp, $o_param) = @_;

	#my $org = ( $o_param -> IsDefined ( 'ORG_ENSEMBL' ) ) ? $o_param -> Get( 'ORG_ENSEMBL' ) : $o_sp -> GetLongName();
	#my $registry = 'Bio::EnsEMBL::Registry';

	## Connecting to Ensembl
	#my $o_io = $o_sp -> GetInstanceObserver();
	#my $ensembl_host = $o_io -> GetCfg ( 'ENSEMBL_HOST');
	#my $ensembl_user = $o_io -> GetCfg ( 'ENSEMBL_USER');
	
	#$registry->load_registry_from_db(
    	#-host => $ensembl_host,
    	#-user => $ensembl_user
	#);

	## Release
	#my $db_adaptor = $registry->get_DBAdaptor($org, 'core');
	#my $dbc = $db_adaptor->dbc;
	
	## Recuperation de la date de creation
	#my $created = &__get_dbCreationDate($dbc, $dbc->dbname);
	
	## Recuperation de la version de l'assemblage
	#my $assembly = $db_adaptor->get_CoordSystemAdaptor->fetch_by_name('chromosome')->version;

	#$assembly = "" if (! defined $assembly);

	#my $file_ens = $o_sp -> GetReleaseFile ( -check => &FALSE );
	#my $f_ens = new IO::File($file_ens,"w") or Carp::croak("Error: unable to create >$file_ens<\n");
	#print $f_ens "#release\tdate\tsite\tassembly\n";
	#print $f_ens $db_adaptor->dbc->dbname."\t$created\tensembl\t$assembly";
	#$f_ens->close;

	## Sequences masquees des chromosomes
	#my $slice_adaptor = $registry->get_adaptor($org,'core','Slice');
	#my $ra_chr = $slice_adaptor->fetch_all('chromosome');
	
	## If the parameter CHROMOSOMES is specified, it is used to restrict the choice
	#my %h_restrict_chr = ();
	#if ( $o_param -> IsDefined ( 'CHROMOSOMES' ) )
	#{
		#foreach my $chr ( split ( ',', $o_param -> Get ('CHROMOSOMES')) )
		#{
			#$h_restrict_chr { $chr } = 1;
		#}
	#}
	#else
	#{
		#$h_restrict_chr { 'all_chromosomes' } = 1;
	#}
	
	#my $seq_file_key;
	#my %h_chr_name=();
	#foreach my $chr (@$ra_chr)
	#{
		#my $chr_name = $chr->seq_region_name;
		#if ( exists $h_restrict_chr { 'all_chromosomes' } || exists $h_restrict_chr { $chr_name } )
		#{
			#$h_chr_name{$chr_name}=1;
		#}
	#}
	#foreach my $chr_name (sort keys(%h_chr_name))
	#{
		#my $slice_adaptor_chr = $registry->get_adaptor($org,'core','Slice');
		#my $chr = $slice_adaptor_chr->fetch_by_region('chromosome',$chr_name);
		##next if ( $chr_name !~ /MT/ ); warn "############### JG Attention: mode debug ##############\n";
		#next if ( ! $o_param->Get('INCL_RANDOM')  && $chr_name =~ /_random$/ );
		#if ( $chr_name =~ /_random$/ )
		#{
			#$chr_name =~ s/_random$/_r/;
		#}

##		if ($chr_name eq '11' || $chr_name eq '17' || $chr_name eq '21' ||
##		$chr_name eq '7' || $chr_name eq 'Y' || $chr_name eq '13' || $chr_name eq '16' ||
##		$chr_name eq '18' || $chr_name eq '1' || $chr_name eq '22' || $chr_name eq '2'){ next; }

		#print "Starting downloading chromosome sequence $chr_name\n";

		#$seq_file_key = "FILE_SEQ_$chr_name";
		#my $file = File::Spec->catfile( $o_sp -> GetDownloadPath(), $seq_file_key );
		##my $file = $rep_download."/".$org.$chr_name.".fasta";

		#$o_param -> Set( $seq_file_key, $file);

		## Download the file only if not already done
		## (can be dangerous if the file is not complete...)
		#if ( -e $file && ! -z $file )
		#{
			#print "Skipping $file: the file already exists!\n";
			#next;
		#}

		#my $f_chr = new IO::File("$file","w") or Carp::croak("Error: unable to create >$file<\n");
		#my $seq = undef;
		#my $ctries=1;
		#while ( !defined($seq) && $ctries <= 3 )
		#{
			#if ( $ctries > 1 )
			#{
				#printf STDERR "Try number #%d\n",$ctries;
			#}
			#eval
			#{
				##print Dumper($chr);
				#$seq = $chr->seq_region_Slice->get_repeatmasked_seq(undef, 1)->seq;
				#if ( length($seq) != $chr->{seq_region_length} )
				#{
					#warn "Sequence length does not match, I try again\n";
					#$seq = undef;
				#}
			#};
			#$ctries++;
		#}
		#print $f_chr ">$org#$chr_name\n$seq\n";
		#$f_chr->close;

		#delete $chr->{'seq'};
		#undef $slice_adaptor_chr;
		#undef $chr;
	#}
#}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut
