#
# $Id: Mrna.pm 1342 2013-12-17 12:52:36Z manu $
#

package Locus::Mrna;

# Copyright INRA/CNRS

# Emmanuel.Courcelle@toulouse.inra.fr
# Jerome.Gouzy@toulouse.inra.fr
# Thomas.Faraut@toulouse.inra.fr

# This software is a computer program whose purpose is to provide a
# web-based interface for analyzing the different levels of genome
# conservations.

# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".

# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.

# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.

# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.


=head1 NAME

Locus::Mrna - describes a mRNA

=head1 SYNOPSIS

 my $o_gene = New Locus::Mrna('start'=>integer,'end'=>integer,'strand'=>'-1|1|+|-','locus_tag'=>string,parent=>$o_gene);

=head1 DESCRIPTION

 A Mrna extends the Locus::Rna class, including a Locus::CDS and two Locus::Utr objects
 The parent MUST be a Locus::Gene object

=cut

use strict;
use warnings;
use Carp;
use Bio::Seq;
use Data::Dumper;

use General;
use Locus;

use base qw( Locus::Rna );
sub GetDefaultSubType { return 'mRNA' };

BEGIN
{
    our $VERSION = do {my @r = (q$Rev: 1342 $ =~ /\d+/g); $r[0]};
}

=head2 procedure _Init

 Title        : _Init
 Function     : Constructor, see Locus::_Init for the details
 Usage        : my $o_rmna = New Locus::Mrna ( %h_param );
                my $o_rmna = New Locus::Mrna ( %h_param, $o_locus );

=cut

sub _Init
{
	my $self = shift;
	$self->SUPER::_Init(@_);

	# A Locus::CDS object living inside this Mrna
	$$self{'__o_cds'} = undef;
	
	# Two Locus::Utr objects living inside this Mrna
	$$self{'__o_five_prime_UTR'}  = undef;
	$$self{'__o_three_prime_UTR'} = undef;
	
	# See SetParent and GetLocusNarcisse Id 
	$$self{'__narcisse_index' } = 0;
}

=head2 procedures Select,Deselect

 See Locus.pm for the generic documentation
 we select/deselect the Locus::Cds, the Locus::ExonSet and the Locus::Utr objects

=cut
sub Select
{
	my $self = shift;
	my ( $inside ) = @_;
	$inside = &TRUE unless (defined $inside);
	if ($inside)
	{
		$self->GetCDS()    -> Select() if ( defined ( $self-> { __o_cds } ) );
		$self->GetExonSet()-> Select() if ( defined ( $self-> { __o_exonset } ) );
		$self->GetUtr( 'five_prime_UTR' )  -> Select() if ( defined ( $self-> { __o_five_prime_UTR } ) );
		$self->GetUtr( 'three_prime_UTR' ) -> Select() if ( defined ( $self-> { __o_three_prime_UTR } ) );
	}
	delete $self -> { __deselected };
}

sub Deselect
{
	my $self   = shift;
	my ( $inside ) = @_;
	$inside = &TRUE unless (defined $inside);
	if ($inside)
	{
		$self->GetCDS()    -> Deselect() if ( defined ( $self-> { __o_cds } ) );
		$self->GetExonSet()-> Deselect() if ( defined ( $self-> { __o_exonset } ) );
		$self->GetUtr( 'five_prime_UTR' )  -> Deselect() if ( defined ( $self-> { __o_five_prime_UTR } ) );
		$self->GetUtr( 'three_prime_UTR' ) -> Deselect() if ( defined ( $self-> { __o_three_prime_UTR } ) );
	}
	$self -> { __deselected } = &TRUE;
}

=head2 procedure SetParent

 Title        : SetParent
 Usage        : $o_locus->setParent( $o_gene );
 Procedure    : Set the parent
 Prerequisite : The parent should not already set, $o_gene should be a Locus::Gene object
 
 Args         : $o_gene : The parent gene
 Globals      : none

=cut
sub SetParent
{
	my $self = shift;
	my ($o_gene) = @_;
	
	if ( ref($o_gene) ne 'Locus::Gene' )
	{
		throw LipmError::ParameterException( 'o_parent', 'type', -text => 'should be a Locus::Gene object' );
	}
	$self->SUPER::SetParent( $o_gene );
	
	my $nc_id = $o_gene -> GetNarcisseId();
	
	# GetNbMrnas() returns the number of mrnas already added to the gene
	# If this is the first mrna added, we'll retrieve the nb 0
	$nc_id .= 't' . $o_gene->GetNbMrnas();
	$self -> { __narcisse_id } = $nc_id;
	$self -> { __narcisse_index } = $o_gene->GetNbMrnas();
	
}

=head2 procedure GetLocusTagTranscriptNb

 Title     : GetLocusTagTranscriptNb (was GetNarcisseLocusId)
 Usage     : my $id = $o_locus->GetLocusTagTranscriptNb()
 Function  : If the gene locus tag/gene name is different from the Mrna locus tag/gene name, return the Mrna locus tag
             If they are equal (or if there is no locus_tag) and there is only ONE mrna, return the locus tag
             If they are equal BUT there are several mrnas return a Narcisse locus tag: gene_locus_tag#Nb
 
 Args      : none
 Return    : A built locus id
 Globals   : none

=cut
sub GetLocusTagTranscriptNb
{
	my $self = shift;
	
	my $o_parent         = $self -> GetParent();
	my $parent_locus_tag = $o_parent -> GetLocusTag();
	$parent_locus_tag    = $o_parent -> GetGeneName() if ( $parent_locus_tag eq '' );
	my $locus_tag        = $self -> GetLocusTag();
	$locus_tag           = $parent_locus_tag if ( $locus_tag eq '' );
	
	if ( $locus_tag eq $parent_locus_tag && $o_parent -> GetNbMrnas() > 1 )
	{
		return $locus_tag . '#' . $self -> { __narcisse_index };
	}
	else
	{
		return $locus_tag;
	}
}

# Call Locus::GetAsStringConsolidated with the FIRST exon and the FIRST CDS
sub GetAsStringConsolidated
{
	my $self = shift;
	my ($tag) = @_;
	my @a_loci;
	my $o_exon_set = $self -> GetExonSet();
	if ( defined $o_exon_set )
	{
		push @a_loci, $o_exon_set->GetRegion0();
	}

	my $o_cds = $self -> GetCDS();
	if ( defined $o_cds )
	{
		push @a_loci, $o_cds->GetRegion0();
	}
	return $self->SUPER::GetAsStringConsolidated($tag,@a_loci);
}
sub GetConsolidatedNote
{
	my $self = shift;
	return $self -> GetAsStringConsolidated( 'note' );
}
sub GetConsolidatedProduct
{
	my $self = shift;
	return $self -> GetAsStringConsolidated ( 'product' );
}

# Call Locus::GetconsolidatedOtherKeys with th eFIRST exon and the FIRST CDS
sub GetConsolidatedOtherKeys
{
	my $self = shift;
	my ($rh_ignored) = @_;

	my @a_loci;
	my $o_exon_set = $self -> GetExonSet();
	if ( defined $o_exon_set )
	{
		push @a_loci, $o_exon_set->GetRegion0();
	}

	my $o_cds = $self -> GetCDS();
	if ( defined $o_cds )
	{
		push @a_loci, $o_cds->GetRegion0();
	}
	return $self->SUPER::GetConsolidatedOtherKeys ( $rh_ignored, @a_loci );
}

=head2 procedure GetConsolidatedOtherKeys_DELETED

 Title     : GetConsolidatedOtherKeys
 Usage     : my $ra_other_keys = $o_locus->GetConsolidatedOtherKeys()
 Function  : Consolidate the 'other' keys of mRNA, CDS and ExonSet
 Args      : $rh_ignored, see GetOtherKeys
 Return    : The keys
 Globals   : none

=cut
sub GetConsolidatedOtherKeys_DELETED
{
	my $self = shift;
	my ($rh_ignored) = @_;
	
	my @a_other_keys = ();
	my $o_exon_set = $self -> GetExonSet();
	if ( defined $o_exon_set )
	{
		my $ra_keys = $o_exon_set -> GetOtherKeys( $rh_ignored );
		push @a_other_keys, @$ra_keys;
	}
	
	my $o_cds = $self -> GetCDS();
	if ( defined $o_cds )
	{
		my $ra_keys = $o_cds -> GetOtherKeys( $rh_ignored );
		push @a_other_keys, @$ra_keys;
	}
	
	my $ra_keys = $self -> GetOtherKeys( $rh_ignored );
	push @a_other_keys, @$ra_keys;
	
	my %h_other_keys;
	foreach my $k (@a_other_keys) 
	{
		$h_other_keys{$k}=1;
	}
	@a_other_keys = keys %h_other_keys;
	return \@a_other_keys;
}



=head2 procedure AddCDS

 Title        : AddCDS
 Usage        : (1) $o_mrna->AddCDS( $o_cds_region );
                (2) $o_mrna->AddCDS( $o_cds );
 Procedure    : (1) Add a CdsRegion to the CDS object inside the Mrna, may be creating the Cds
                (2) Add the Cds to the Mrna, throw an error if it exists already
 Args         : $o_cds_region or $o_cds: The region to add to the CDS, OR the CDS to add
                NOTE - $o_cds_region MAY be modified by this procedure
 Globals      : none

=cut
sub AddCDS
{
	my $self = shift;
	my ($o_cds_region) = @_;
	if ( ref($o_cds_region) eq 'Locus::CdsRegion' )
	{
		if ( ! defined 	$self -> { __o_cds } )
		{
			my $o_cds = New Locus::Cds ( $o_cds_region );
			$o_cds -> SetParent ( $self );
			$self -> { __o_cds } = $o_cds;
		}
	
		my $o_cds = $self -> { __o_cds };
		$o_cds -> AddRegion ( $o_cds_region );
	}
	elsif ( ref($o_cds_region) eq 'Locus::Cds' )	
	{
		if ( defined $self -> { __o_cds } )
		{
			my $locus_tag = $self -> GetLocusTag();
			my ($s,$e)    = $self -> GetEnds();
			throw Error::Simple( "ERROR - The CDS is already added for this mRNA $locus_tag [$s..$e]" );
		}
		else
		{
			$o_cds_region -> SetParent ( $self );
			$self -> { __o_cds } = $o_cds_region;
		}
	}
	else 
	{
		throw LipmError::ParameterException( 'o_cds_region', 'type', -text => 'should be a Locus::CdsRegion or Locus::Cds object, not a '.ref($o_cds_region) );
	}

}

=head2 procedure AddExon

 Title        : AddExon
 Usage        : (1) $o_mrna->AddExon( $o_exon );
                (2) $o_mrna->AddExon( $o_exon_set );
 Procedure    : (1) Add a Exon to the ExonSet object inside the Mrna, may be creating the ExonSet
                (2) Add the ExonSet to the Mrna, throw an error if it exists already
 Args         : $o_exon or $o_exon_set: The region to add to the ExonSet, OR the ExonSet to add
                NOTE - $o_exon_region MAY be modified by this procedure
                NOTE - JAMAIS TESTE - JAMAIS UTILISE - PEUT-ETRE TOTALEMENT INUTILE !!!
 Globals      : none
 

=cut
#sub AddExon
#{
	#my $self = shift;
	#my ($o_exon) = @_;
	#if ( ref($o_exon) eq 'Locus::Exon' )
	#{
		#my $o_exon_set = $self -> GetExonSet();
	 	#if ( !defined $o_exon_set )
	 	#{
			#$o_exon_set = New Locus::ExonSet( $self );
			#$self -> AddExon ( $o_exon_set );
		#}		
		#$o_exon_set -> AddRegion ( $o_exon );
	#}
	#elsif ( ref($o_exon) eq 'Locus::ExonSet' )	
	#{
		#if ( defined $self -> GetExonSet() )
		#{
			#my $locus_tag = $self -> GetLocusTag();
			#my ($s,$e)    = $self -> Getends();
			#throw LipmError::Simple( "ERROR - The ExonSet is already added for this mRNA $locus_tag [$s..$e]" );
		#}
		#else
		#{
			#$o_exon -> SetParent ( $self );
			#$self -> { __o_exonset } = $o_exon;
		#}
	#}
	#else 
	#{
		#throw LipmError::ParameterException( 'o_exon', 'type', -text => 'should be a Locus::Exon or Locus::ExonSet object, not a '.ref($o_exon) );
	#}
#}

=head2 procedure SetTranscriptId

 Title : SetTranscriptId
 Usage : $o_mrna -> SetTranscriptId();
 TODO  : UTILISE UNIQUEMENT POUR GTF.pm, PAS FINALISE ==> NOT USED YET
=cut
sub SetTranscriptId
{
	my $self =shift;
	my ($transcript_id) = @_;
	if ( defined $self -> { _d_rh_core } -> { transcript_id } )
	{
		my ($s,$e) = $self -> GetEnds();
		throw Error::Simple( "ERROR - The transciprt_id is already set for this mRNA [$s..$e]" );
	}
	else
	{
		$self -> { _d_rh_core } -> { transcript_id } = $transcript_id;
	}
}

=head2 function GetAllChildren

 Title    : GetAllChildren
 Function : Return the children loci, ie the loci embedded by this locus
            This is a recursive function
 Usage    : my $ra_children = $o_locus -> GetAllChildren();
 Args     : none
 Return   : A ref to an array containing the children of the locus, as well as the children of the children, etc.
 Globals  : none

=cut
sub GetAllChildren
{
	my $self = shift;
	my @a_children = @{$self->SUPER::GetAllChildren()};
	
	my $o_cds = $self -> GetCDS();
	push @a_children, $o_cds, @{$o_cds->GetAllChildren()} if ( defined $o_cds );
	
	my $o_utr_3 = $self -> GetUtr('three_prime_UTR');
	push @a_children, $o_utr_3, @{$o_utr_3->GetAllChildren()} if ( defined $o_utr_3 );
	
	my $o_utr_5 = $self -> GetUtr('five_prime_UTR');
	push @a_children, $o_utr_5, @{$o_utr_5->GetAllChildren()} if ( defined $o_utr_5 );
	
	return \@a_children;
}

=head2 function GetCDS

 Title    : GetCDS
 Usage    : my $o_cds  = $o_mrna->GetCDS();
 Function : Accessor to the CDS object
 Args     : $create_cds (default &FALSE) If &TRUE and if the CDS does not exist, generate the CDS from the exons and the UTR
 Globals  : none
 Return   : The Cds if found or juste created, or undef

=cut
sub GetCDS{
	my $self = shift;
	my ($create_cds) = @_;
	$create_cds = &FALSE unless ( defined $create_cds );
	
	if ( defined $self -> { __o_cds } )
	{
		return $self -> { __o_cds };
	}
	else
	{
		return ($create_cds) ? $self -> __CreateCDS() : undef;
	}
}

=head2 procedure AddUtr

 Title        : AddUtr
 Usage        : $o_mrna->AddUtrRegion( $o_utr );
 Procedure    : Add a region to one of the utrs, may be creating an Utr object
 Prerequisite : $o_utr should be a Locus::UtrRegion object with the correct type
 
 Args         : $o_utr_region : The utr region to add to the UTR
 Globals      : none

=cut
sub AddUtr
{
	my $self = shift;
	my ($o_utr_region) = @_;
	
	if ( ref($o_utr_region) eq 'Locus::UtrRegion' )
	{
		my $sub_type = $o_utr_region -> GetSubType();
		if ( !defined($self -> { "__o_$sub_type" }) )
		{
			my $locus_tag       = $o_utr_region -> GetLocusTag();
			my $o_utr           = New Locus::Utr ( locus_tag => $locus_tag, sub_type => $sub_type, $o_utr_region );
			$o_utr -> SetParent ( $self );
			$self -> { "__o_$sub_type" } = $o_utr;
		}
		my $o_utr = $self -> { "__o_$sub_type" };
		$o_utr -> AddRegion ( $o_utr_region );
	}
	elsif ( ref($o_utr_region) eq 'Locus::Utr' )
	{
		my $sub_type = $o_utr_region -> GetSubType();
		if ( defined($self -> { "__o_$sub_type" }) )
		{
			throw Error::Simple( "ERROR - The UTR is already added for this mRNA " . $self -> GetpkExtended() . "\n" );
		}
		$o_utr_region -> SetParent ( $self );
		$self -> { "__o_$sub_type" } = $o_utr_region;
	}
	else
	{
		throw LipmError::ParameterException( 'o_utr', 'type', -text => 'should be a Locus::UtrRegion or Locus::Utr object' );
	}
}
=head2 procedure RemoveUtr

 Title        : RemoveUtr
 Usage        : $o_mrna->RemoveUtr( $type )
 Procedure    : Remove the Utr 
 
 Args         : $type : three_prime_UTR or five_prime_UTR
 Globals      : none

=cut
sub RemoveUtr
{
	my $self = shift;
	my ($sub_type) = @_;
	
	$self -> { "__o_$sub_type" } = undef;
}

=head2 procedure CreateUtr

 Title        : CreateUtr
 Usage        : $o_mrna->CreateUtr();
 Procedure    : Create TWO utrs (3' and 5') if possible AND if not already created
 
 Args         : None
 Globals      : none
 
=cut
sub CreateUtr
{
	my $self = shift;

	my $o_cds = $self->GetCDS();
	return unless defined($o_cds);
	
	my $o_exon = $self->GetExonSet();
	return unless defined($o_exon);
	
	my ($cds_s, $cds_e) = $o_cds->GetEnds();
	my ($exon_s,$exon_e) = $o_exon->GetEnds();
	my $strand = $self -> GetStrand();

	foreach my $type ( qw/ three_prime_UTR five_prime_UTR / )
	{
		my $o_utr = $self -> GetUtr ( $type );
		
		# If the utr already exists, don't touch it !
		next if ( defined($o_utr) );
		
		my ($utr_s,$utr_e);
		if ( $strand==1&&$type eq 'five_prime_UTR' || $strand==-1&&$type eq 'three_prime_UTR' )
		{
			next if ($cds_s==$exon_s);
			$utr_s=$exon_s;
			$utr_e=$cds_s-1;
		}
		else
		{
			next if ($cds_e==$exon_e);
			$utr_s=$cds_e+1;
			$utr_e=$exon_e;
		}

		$o_utr = New Locus::UtrRegion( start => $utr_s, end => $utr_e, strand => $strand, sub_type => $type );
		$self -> AddUtr($o_utr);
		warn ( $self -> GetWarningHeader() . "Create utr " . $o_utr->GetpkExtended() ."\n" );
	}
}

=head2 function GetIntrons

 Title        : GetIntrons
 Usage        : my @a_introns = $o_mrna->GetIntrons();
 Function     : Using the Utr and the CDS, return an array of Locus, each locus representing an intron
 NOTE         : CE SERAIT MIEUX DE FAIRE UN OBJET IntronSet CONTENANT DES REGIONS, cf. ExonSet et Exon MAIS POUR L'iNSTANT ON RESTE COMME CA
 
 Args         : None
 Return       : The Array of introns
 Globals      : none
 
=cut
sub GetIntrons
{
	my $self = shift;
	if ( !defined $self -> { __ra_introns } )
	{
		my @a_introns;
		$self->CreateUtr();
		my $o_cds = $self -> GetCDS();
		return @a_introns if (! defined $o_cds );
		
		my (@a_utr_type) = ( 'five_prime_UTR', 'three_prime_UTR' );
		my $strand = $self->GetStrand();
		@a_utr_type = reverse (@a_utr_type) if ($strand < 0);
		my @a_utr = ($self->GetUtr($a_utr_type[0]),$self->GetUtr($a_utr_type[1]));

		my @a_regions;
		push @a_regions, $a_utr[0]->GetRegions() if ( defined $a_utr[0] );
		push @a_regions, $o_cds->GetRegions();
		push @a_regions, $a_utr[1]->GetRegions() if ( defined $a_utr[1] );
		
		for (my $i=1; $i<@a_regions;$i++)
		{
			my $s = $a_regions[$i-1]->GetEnd()+1;
			my $e = $a_regions[$i]->GetStart()-1;
			next if ($s>$e);
			my $o_intron = New Locus::MiscRegion ( start=>$s,end=>$e,strand=>$strand );
			$o_intron->SetParent($self->GetParent());
			push @a_introns, $o_intron;
		}
		$self -> { __ra_introns } = \@a_introns;
	}
	return @{$self -> { __ra_introns }};
}
 
=head2 function GetUtr

 Title    : GetUtr
 Usage    : $o_utr = $o_mrna->GetUtr( $sub_type );
 Function : Get the UTR, three_prime_UTR or five_prime_UTR. 
 
 Args     : $sub_type : The sub_type of the utr to get
 Return   : the utr if set, or undef
 Globals  : none

=cut
sub GetUtr
{
	my $self   = shift;
	my ($sub_type) = @_;
	
	if ( $sub_type =~ /_prime_UTR$/ )
	{
		if ( defined $self -> { "__o_$sub_type" } )
		{
			return $self -> { "__o_$sub_type" }
		}
		else
		{
			return undef;
		}
	}
	else
	{
		throw LipmError::ParameterException ('type','value', -text => "$sub_type is not an allowed sub_type for GetUtr\n");
	}
}

=head2 function __CreateCDS

 Title        : __CreateCDS
 Function     : Using the exons AND the Utr, create the CDS object
 Prerequisite : All information about exons and UTRs should be already completely loaded (but this is NOT checked)
 Usage        : my $o_cds = $self -> __CreateCDS()
 Returns      : The $o_cds created (or the $o_cds already existing if any), or undef if nothing created (no exons)
 Access       : Private
 Args         : none
 Globals      : none

=cut
sub __CreateCDS
{
	my $self = shift;
	
	if ( ! defined $self -> { __o_cds } )
	{
		my $strand     = $self -> GetStrand();
		my $o_exon_set = $self -> GetExonSet();
		my $o_sutr     = ($strand==1) ? $self -> GetUtr ( 'five_prime_UTR' ) : $self -> GetUtr ( 'three_prime_UTR' );
		my $o_eutr     = ($strand==1) ? $self -> GetUtr ( 'three_prime_UTR' ) : $self -> GetUtr ( 'five_prime_UTR' );
		
		if ( defined $o_exon_set )
		{
			my @a_exons = $o_exon_set -> GetRegions();
			my ( $start_cds, $end_cds ) = (0,0);
			if ( defined $o_sutr )
			{
				my @a_utr  = $o_sutr -> GetRegions();
				$start_cds = $a_utr[-1] -> GetEnd()+1;	# The CDS starts AFTER the starting utr
			}
			else
			{
				$start_cds = $a_exons[0] -> GetStart();  # No UTR defined -> The CDS starts just with the exon
			}
			if ( defined $o_eutr )
			{
				my @a_utr = $o_eutr -> GetRegions();
				$end_cds = $a_utr[0] -> GetStart() - 1; # The CDS ends BEFORE the ending CDS
			}
			else
			{
				$end_cds = $a_exons[-1] -> GetEnd();     # no UTR defined -> The CDS ends just with the exon
			}
			
			# Scan the exons, and call AddCDS accordingly
			for my $o_exon ( @a_exons )
			{
				
				my ($start,$end);
				
				# -------  *
				# * --------
				# exon BEFORE $start_cds OR exon AFTER $end_cds
				if ( $o_exon->GetEnd() < $start_cds || $o_exon->GetStart() > $end_cds)
				{
					next;
				}
				
				# ---*----
				# $start_cds INSIDE exon
				if ( $o_exon -> GetStart < $start_cds )
				{
					$start = $start_cds;
				}
				
				# * -----------
				# $start_cds BEFORE exon				
				else
				{
					$start = $o_exon -> GetStart();
				}
				
				# ---------    *
				# exon BEFORE $end_cds
				if ( $o_exon -> GetEnd() <= $end_cds )
				{
					$end = $o_exon -> GetEnd();
				}
				
				# ----*-----
				# $end_cds INSIDE exon
				elsif ( $o_exon -> GetStart() <= $end_cds )
				{
					$end = $end_cds;
				}
				$self -> AddCDS ( New Locus::CdsRegion ( start => $start, end => $end, strand => $strand, $self->GetAncestor() ) );
			}
		};
	}
	return $self -> GetCDS();
}

=head2 function GetTranslationEnds

 Title        : GetTranslationEnds
 Function     : Return, may be identifying it, the translational start on a mRNA
 Usage        : my ($min_trans,$max_trans,$nb_exons) = &GetTranslationStartStop()
 Prerequisite : none
 Returns      : $min_trans: lower boundary participating to the translation   (ie the cds start, relative to the Mrna start)
                $max_trans: higher boundary participating to the translation  (ie the cds stop, relative to the Mrna end)
                (-1,-1) is returned when there is an inconsistency (whether already set by Gene::DumpFastaNarcisse OR set by this function)
                WARNING - The consistency status of the cds object inside this mrna MAY BE MODIFIED !
                NOTE - All the proteins marked as Fragment or Inconsistent should be marked ALSO as Fragment or Inconsistent by GetTranslation
                       However, the reverse is not true, ie GetTranslation should detect MORE Inconsistent or Fragment proteins than this method
 Args         : none
 Globals      : none

=cut

sub GetTranslationEnds
{
	my $self = shift;
	
	if ( !defined ($self -> { _d_rh_core } -> {cds_start}) || !defined ($self -> { _d_rh_core } -> {cds_end}) )
	{
		my $o_cds    = $self -> GetCDS();
		if ( $o_cds -> GetConsistency() eq 'I' ) 
		{
			$self -> { _d_rh_core } -> {cds_start} = -1;
			$self -> { _d_rh_core } -> {cds_end}   = -1;
		}
		else
		{
			my $strand  = $self -> GetStrand();
			my $o_exonset  = $self -> GetExonSet();
			if ( ! defined ( $o_exonset ) )
			{
				my ($s,$e)    = $self -> GetEnds();
				my $locus_tag = $self -> GetLocusTag();
				throw Error::Simple ( "ERROR - mRNA $locus_tag [$s..$e] does not have any ExonSet" );
			}
			
			my @a_exons = $o_exonset -> GetRegions();
			my $nb_exons = @a_exons;
			my $len_mrna = $o_exonset -> GetRegionsLength();
		
			my $o_cds    = $self -> GetCDS();
			return (-1,-1,-1) if ($o_cds -> GetConsistency() eq 'I');
			
			my @a_cds_regions = $o_cds -> GetRegions();
			my $mincds = $a_cds_regions[0] -> GetStart();
			my $maxcds = $a_cds_regions[-1]-> GetEnd();
			my $mintrans=0;
			foreach my $o_exon (@a_exons)
			{
				my($start,$end) = $o_exon -> GetEnds();
				if ( $mincds > $end )
				{
					$mintrans += ($end-$start+1);
				}
				else
				{
					$mintrans +=  ($mincds-$start+1);
					last;
				}
			}
			
			my $maxtrans=0;
			foreach my $o_exon (reverse @a_exons)
			{
				my($start,$end) = $o_exon -> GetEnds();
				if ( $maxcds < $start )
				{
					$maxtrans += ($end-$start+1);
				}
				else
				{
					$maxtrans +=  ($end-$maxcds+1);
					last;
				}
			}
			if ( $mintrans < 1 || $maxtrans < 1 )
			{
				warn $self ->GetWarningHeader('') . "strong inconsistency between CDS and mRNA detected " . $self->GetpkExtended() ."\n";
				$o_cds -> SetInconsistent();
				$self -> { _d_rh_core } -> {cds_start} = -1;
				$self -> { _d_rh_core } -> {cds_end}   = -1;
			}
			else
			{
				$maxtrans = $len_mrna-$maxtrans+1;
				my $mod = ((abs($maxtrans-$mintrans)+1 ) % 3);
				if ( $mod != 0 )
				{
					$maxtrans -= $mod;
					warn $self ->GetWarningHeader('') . "not mult of 3 subsequence (length $len_mrna) " . $self->GetpkExtended() ."\n";
					$o_cds -> SetFragment('F');
				}
	
				if ( $strand < 0 )
				{
					my ($min,$max) = ($mintrans,$maxtrans);
					$mintrans = $len_mrna - $max + 1;
					$maxtrans = $len_mrna - $min + 1;
				}
				$self -> { _d_rh_core } -> {cds_start} = $mintrans;
				$self -> { _d_rh_core } -> {cds_end}   = $maxtrans;
			}
		}
	}
	return ($self -> { _d_rh_core } -> {cds_start},$self -> { _d_rh_core } -> {cds_end});
}

=pod

=head1 COPYRIGHT NOTICE

This software is governed by the CeCILL license - www.cecill.info

=cut

1;


